
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_monitor_framer.c                                    */
/*  DESCRIPTION:  Monitor framing                                            */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "khe_set.h"


/*****************************************************************************/
/*                                                                           */
/*  KHE_MONITOR_FRAMER_STATE                                                 */
/*                                                                           */
/*  State of a framer object.                                                */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_MONITOR_FRAMER_INIT,
  KHE_MONITOR_FRAMER_LOADING,
  KHE_MONITOR_FRAMER_COMPLETE
} KHE_MONITOR_FRAMER_STATE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_MONITOR_ON_DAY - one monitor on one day                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_mf_monitor_on_day_rec {
  KHE_SET			item_indexes;
} *KHE_MF_MONITOR_ON_DAY;

typedef HA_ARRAY(KHE_MF_MONITOR_ON_DAY) ARRAY_KHE_MF_MONITOR_ON_DAY;


/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_MONITOR - one monitor                                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_mf_monitor_rec {
  KHE_MONITOR			monitor;		/* the monitor       */
  int				monitor_index;		/* index of monitor  */
  int				first_day_index;	/* first day index   */
  int				last_day_index;		/* last day index    */
  ARRAY_KHE_MF_MONITOR_ON_DAY	days;			/* one for each day  */
  KHE_SET			item_indexes;		/* last calculated   */
  int				seq_num;		/* last seq num      */
} *KHE_MF_MONITOR;

typedef HA_ARRAY(KHE_MF_MONITOR) ARRAY_KHE_MF_MONITOR;


/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_RESOURCE_ON_DAY - one resource on one day                         */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_mf_resource_on_day_rec {
  KHE_SET			monitor_indexes;
} *KHE_MF_RESOURCE_ON_DAY;

typedef HA_ARRAY(KHE_MF_RESOURCE_ON_DAY) ARRAY_KHE_MF_RESOURCE_ON_DAY;


/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_RESOURCE                                                          */
/*                                                                           */
/*  Represents one resource within the monitor framer.                       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_mf_resource_rec {
  KHE_RESOURCE			resource;		/* the resource      */
  ARRAY_KHE_MF_MONITOR		monitors;		/* all its monitors  */
  ARRAY_KHE_MF_RESOURCE_ON_DAY	days;			/* one for each day  */
  KHE_SET			monitor_indexes;	/* last calculated   */
  int				seq_num;		/* last seq num      */
} *KHE_MF_RESOURCE;

typedef HA_ARRAY(KHE_MF_RESOURCE) ARRAY_KHE_MF_RESOURCE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_MONITOR_FRAMER - a monitor framer object                             */
/*                                                                           */
/*****************************************************************************/

struct khe_monitor_framer_rec {
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_FRAME			frame;
  KHE_SET			day_indexes;
  KHE_MONITOR_FRAMER_STATE	state;
  int				seq_num;	/* new for each set of days  */
  ARRAY_KHE_MF_RESOURCE		resources;	/* the resources             */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_MONITOR_FRAMER_STATE"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  char *KheMfStateShow(KHE_MONITOR_FRAMER_STATE mfs)                       */
/*                                                                           */
/*  Show mfs.                                                                */
/*                                                                           */
/*****************************************************************************/

static char *KheMfStateShow(KHE_MONITOR_FRAMER_STATE mfs)
{
  switch( mfs )
  {
    case KHE_MONITOR_FRAMER_INIT:	return "init";
    case KHE_MONITOR_FRAMER_LOADING:	return "loading";
    case KHE_MONITOR_FRAMER_COMPLETE:	return "complete";

    default:
      HnAbort("KheMfStateShow internal error (%d)", mfs);
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_MF_MONITOR_ON_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KHE_MF_MONITOR_ON_DAY KheMfMonitorOnDayMake(KHE_MONITOR_FRAMER mf)  */
/*                                                                           */
/*  Make a new monitor on day object.                                        */
/*                                                                           */
/*****************************************************************************/

static KHE_MF_MONITOR_ON_DAY KheMfMonitorOnDayMake(KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR_ON_DAY res;
  HaMake(res, mf->arena);
  KheSetInit(res->item_indexes, mf->arena);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfMonitorOnDayDebug(KHE_MF_MONITOR_ON_DAY mfmd, int day_index,   */
/*    KHE_MONITOR_FRAMER mf, int verbosity, int indent, FILE *fp)            */
/*                                                                           */
/*  Debug printf of mfmd onto fp with the given verbosity and indent.        */
/*                                                                           */
/*****************************************************************************/

static void KheMfMonitorOnDayDebug(KHE_MF_MONITOR_ON_DAY mfmd, int day_index,
  KHE_MONITOR_FRAMER mf, int verbosity, int indent, FILE *fp)
{
  KHE_TIME_GROUP tg;
  tg = KheFrameTimeGroup(mf->frame, day_index);
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "%s: %s", KheTimeGroupId(tg), KheSetShow(mfmd->item_indexes));
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_MF_MONITOR"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_MONITOR KheMfMonitorMake(KHE_MONITOR m, int monitor_index,        */
/*    KHE_MONITOR_FRAMER mf)                                                 */
/*                                                                           */
/*  Make a new mf monitor with these attributes (just the basic object).     */
/*                                                                           */
/*****************************************************************************/

static KHE_MF_MONITOR KheMfMonitorMake(KHE_MONITOR m, int monitor_index,
  KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR res;
  HaMake(res, mf->arena);
  res->monitor = m;
  res->monitor_index = monitor_index;
  res->first_day_index = res->last_day_index = -1;  /* placeholder */
  HaArrayInit(res->days, mf->arena);
  KheSetInit(res->item_indexes, mf->arena);
  res->seq_num = -1;  /* signals item_indexes out of date */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfMonitorSetDayIndexRange(KHE_MF_MONITOR mfm,                    */
/*    int first_day_index, int last_day_index, KHE_MONITOR_FRAMER mf)        */
/*                                                                           */
/*  Set the first_day_index and last_day_index attributes of mfm, and        */
/*  make one monitor on day object for each day in this range.               */
/*                                                                           */
/*****************************************************************************/

static void KheMfMonitorSetDayIndexRange(KHE_MF_MONITOR mfm,
  int first_day_index, int last_day_index, KHE_MONITOR_FRAMER mf)
{
  int i;
  mfm->first_day_index = first_day_index;
  mfm->last_day_index = last_day_index;
  for( i = first_day_index;  i <= last_day_index;  i++ )
    HaArrayAddLast(mfm->days, KheMfMonitorOnDayMake(mf));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfMonitorSetItemAffectedByDay(KHE_MF_MONITOR mfm,                */
/*    int item_index, int day_index)                                         */
/*                                                                           */
/*  Record the fact that the item whose index is item_index is affected by   */
/*  the day whose index is day_index.                                        */
/*                                                                           */
/*****************************************************************************/

static void KheMfMonitorSetItemAffectedByDay(KHE_MF_MONITOR mfm,
  int item_index, int day_index)
{
  KHE_MF_MONITOR_ON_DAY mfmd;
  mfmd = HaArray(mfm->days, day_index - mfm->first_day_index);
  KheSetInsert(mfmd->item_indexes, item_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfMonitorSetAffectedDays(KHE_MF_MONITOR mfm,                     */
/*    KHE_MONITOR_FRAMER mf)                                                 */
/*                                                                           */
/*  This is called after all calls to KheMfMonitorSetItemAffectedByDay are   */
/*  complete.  It adds mfm's index to each day the monitor is affected by.   */
/*                                                                           */
/*****************************************************************************/
static void KheMfResourceSetMonitorAffectedByDay(KHE_MF_RESOURCE mfr,
  int item_index, int day_index);

static void KheMfMonitorSetAffectedDays(KHE_MF_MONITOR mfm,
  KHE_MF_RESOURCE mfr, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR_ON_DAY mfmd;  int i, day_index;
  HaArrayForEach(mfm->days, mfmd, i)
    if( KheSetCount(mfmd->item_indexes) > 0 )
    {
      day_index = mfm->first_day_index + i;
      KheMfResourceSetMonitorAffectedByDay(mfr, mfm->monitor_index, day_index);
    }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfMonitorBringItemIndexesUpToDate(KHE_MF_MONITOR mfm,            */
/*    KHE_MONITOR_FRAMER mf)                                                 */
/*                                                                           */
/*  Bring mfm->item_indexes up to date.  It needs to contain the union,      */
/*  over the days of interest d, of mfm's item indexes on day d.             */
/*                                                                           */
/*****************************************************************************/

static void KheMfMonitorBringItemIndexesUpToDate(KHE_MF_MONITOR mfm,
  KHE_MONITOR_FRAMER mf)
{
  int day_index, i;  KHE_MF_MONITOR_ON_DAY mfmd;
  if( mfm->seq_num < mf->seq_num )
  {
    KheSetClear(mfm->item_indexes);
    KheSetForEach(mf->day_indexes, day_index, i)
      if( day_index >= mfm->first_day_index && day_index <= mfm->last_day_index)
    {
      mfmd = HaArray(mfm->days, day_index - mfm->first_day_index);
      KheSetUnion(mfm->item_indexes, mfmd->item_indexes);
    }
    mfm->seq_num = mf->seq_num;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfMonitorDebug(KHE_MF_MONITOR mfm, KHE_MONITOR_FRAMER mf,        */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of mfm onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

static void KheMfMonitorDebug(KHE_MF_MONITOR mfm, KHE_MONITOR_FRAMER mf,
  int verbosity, int indent, FILE *fp)
{
  KHE_MF_MONITOR_ON_DAY mfmd;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Monitor(%s)\n", indent, "", KheMonitorId(mfm->monitor));
    HaArrayForEach(mfm->days, mfmd, i)
      KheMfMonitorOnDayDebug(mfmd, mfm->first_day_index + i, mf,
	verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "%s", KheMonitorId(mfm->monitor));
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_MF_RESOURCE_ON_DAY"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_RESOURCE_ON_DAY KheResourceOnDayMake(KHE_MONITOR_FRAMER mf)       */
/*                                                                           */
/*  Make a new resource on day object.                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_MF_RESOURCE_ON_DAY KheResourceOnDayMake(KHE_MONITOR_FRAMER mf)
{
  KHE_MF_RESOURCE_ON_DAY res;
  HaMake(res, mf->arena);
  KheSetInit(res->monitor_indexes, mf->arena);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceOnDayDebug(KHE_MF_RESOURCE_ON_DAY mfrd,                */
/*    int day_index, KHE_MF_RESOURCE mfr, KHE_MONITOR_FRAMER mf,             */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of mfrd onto fp with the given verbosity and indent.        */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceOnDayDebug(KHE_MF_RESOURCE_ON_DAY mfrd,
  int day_index, KHE_MF_RESOURCE mfr, KHE_MONITOR_FRAMER mf,
  int verbosity, int indent, FILE *fp)
{
  KHE_TIME_GROUP tg;  int monitor_index, i;  KHE_MF_MONITOR mfm;
  tg = KheFrameTimeGroup(mf->frame, day_index);
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Day %s\n", indent, "", KheTimeGroupId(tg));
    KheSetForEach(mfrd->monitor_indexes, monitor_index, i)
    {
      mfm = HaArray(mfr->monitors, monitor_index);
      fprintf(fp, "%*s  %s\n", indent, "", KheMonitorId(mfm->monitor));
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "Day(%s)", KheTimeGroupId(tg));
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_MF_RESOURCE"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceSetMonitorAffectedByDay(KHE_MF_RESOURCE mfr,           */
/*    int monitor_index, int day_index)                                      */
/*                                                                           */
/*  Inform mfr that its monitor with index monitor_index is affected by      */
/*  what happens on the day with index day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceSetMonitorAffectedByDay(KHE_MF_RESOURCE mfr,
  int monitor_index, int day_index)
{
  KHE_MF_RESOURCE_ON_DAY mrd;
  mrd = HaArray(mfr->days, day_index);
  KheSetInsert(mrd->monitor_indexes, monitor_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceAddAvoidUnavailableTimesMonitor(KHE_MF_RESOURCE mfr,   */
/*    KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm, int m_index,                 */
/*    KHE_MONITOR_FRAMER mf)                                                 */
/*                                                                           */
/*  Update mfr's data structures to take account of avoid unavailable times  */
/*  monitor autm, whose index is m_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceAddAvoidUnavailableTimesMonitor(KHE_MF_RESOURCE mfr,
  KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm, int m_index, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR mfm;  KHE_TIME first_time, last_time, time;
  int first_day_index, last_day_index, day_index, i;  KHE_TIME_GROUP tg;
  KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT autc;

  /* make and add a monitor object corresponding to autm (even if empty) */
  mfm = KheMfMonitorMake((KHE_MONITOR) autm, m_index, mf);
  HaArrayAddLast(mfr->monitors, mfm);

  /* get the range of days and add it to mfm */
  autc = KheAvoidUnavailableTimesMonitorConstraint(autm);
  tg = KheAvoidUnavailableTimesConstraintUnavailableTimes(autc);
  if( KheTimeGroupTimeCount(tg) > 0 )
  {
    /* non-empty set of days */
    first_time = KheTimeGroupTime(tg, 0);
    last_time = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
    first_day_index = KheFrameTimeIndex(mf->frame, first_time);
    last_day_index = KheFrameTimeIndex(mf->frame, last_time);
  }
  else
  {
    /* empty set of days */
    first_day_index =  0;
    last_day_index = -1;
  }
  KheMfMonitorSetDayIndexRange(mfm, first_day_index, last_day_index, mf);

  /* add the items and the days they are affected (NB just item 0 here) */
  for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
  {
    time = KheTimeGroupTime(tg, i);
    day_index = KheFrameTimeIndex(mf->frame, time);
    KheMfMonitorSetItemAffectedByDay(mfm, 0, day_index);
  }

  /* add the monitor as an item to the resource's days affected */
  KheMfMonitorSetAffectedDays(mfm, mfr, mf);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceAddClusterBusyTimesMonitor(KHE_MF_RESOURCE mfr,        */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int m_index,KHE_MONITOR_FRAMER mf)*/
/*                                                                           */
/*  Update mfr's data structures to take account of cluster busy times       */
/*  monitor cbtm, whose index is m_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceAddClusterBusyTimesMonitor(KHE_MF_RESOURCE mfr,
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int m_index, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR mfm;  KHE_TIME first_time, last_time, time;  KHE_POLARITY po;
  int first_day_index, last_day_index, day_index, i, j;  KHE_TIME_GROUP tg;

  /* make and add a monitor object corresponding to cbtm (even if empty) */
  mfm = KheMfMonitorMake((KHE_MONITOR) cbtm, m_index, mf);
  HaArrayAddLast(mfr->monitors, mfm);

  /* get the range of days and add it to mfm */
  KheClusterBusyTimesMonitorRange(cbtm, &first_time, &last_time);
  if( first_time != NULL )
  {
    /* non-empty set of days */
    first_day_index = KheFrameTimeIndex(mf->frame, first_time);
    last_day_index = KheFrameTimeIndex(mf->frame, last_time);
  }
  else
  {
    /* empty set of days */
    first_day_index =  0;
    last_day_index = -1;
  }
  KheMfMonitorSetDayIndexRange(mfm, first_day_index, last_day_index, mf);

  /* add the items and the days they are affected */
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      time = KheTimeGroupTime(tg, j);
      day_index = KheFrameTimeIndex(mf->frame, time);
      KheMfMonitorSetItemAffectedByDay(mfm, i, day_index);
    }
  }

  /* add the monitor as an item to the resource's days affected */
  KheMfMonitorSetAffectedDays(mfm, mfr, mf);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceAddLimitBusyTimesMonitor(KHE_MF_RESOURCE mfr,          */
/*    KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, int m_index, KHE_MONITOR_FRAMER mf) */
/*                                                                           */
/*  Update mfr's data structures to take account of limit busy times         */
/*  monitor lbtm, whose index is m_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceAddLimitBusyTimesMonitor(KHE_MF_RESOURCE mfr,
  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, int m_index, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR mfm;  KHE_TIME first_time, last_time, time;  KHE_TIME_GROUP tg;
  int first_day_index, last_day_index, day_index, i, j, busy_count;

  /* make and add a monitor object corresponding to lbtm (even if empty) */
  mfm = KheMfMonitorMake((KHE_MONITOR) lbtm, m_index, mf);
  HaArrayAddLast(mfr->monitors, mfm);

  /* get the range of days and add it to mfm */
  KheLimitBusyTimesMonitorRange(lbtm, &first_time, &last_time);
  if( first_time != NULL )
  {
    /* non-empty set of days */
    first_day_index = KheFrameTimeIndex(mf->frame, first_time);
    last_day_index = KheFrameTimeIndex(mf->frame, last_time);
  }
  else
  {
    /* empty set of days */
    first_day_index =  0;
    last_day_index = -1;
  }
  KheMfMonitorSetDayIndexRange(mfm, first_day_index, last_day_index, mf);

  /* add the items and the days they are affected */
  for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
  {
    tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &busy_count);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      time = KheTimeGroupTime(tg, j);
      day_index = KheFrameTimeIndex(mf->frame, time);
      KheMfMonitorSetItemAffectedByDay(mfm, i, day_index);
    }
  }

  /* add the monitor as an item to the resource's days affected */
  KheMfMonitorSetAffectedDays(mfm, mfr, mf);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceAddLimitWorkloadMonitor(KHE_MF_RESOURCE mfr,           */
/*    KHE_LIMIT_WORKLOAD_MONITOR lwm, int m_index, KHE_MONITOR_FRAMER mf)    */
/*                                                                           */
/*  Update mfr's data structures to take account of limit workload           */
/*  monitor lwm, whose index is m_index.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceAddLimitWorkloadMonitor(KHE_MF_RESOURCE mfr,
  KHE_LIMIT_WORKLOAD_MONITOR lwm, int m_index, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR mfm;  KHE_TIME first_time, last_time, time;  float workload;
  KHE_TIME_GROUP tg;  int first_day_index, last_day_index, day_index, i, j;

  /* make and add a monitor object corresponding to lwm (even if empty) */
  mfm = KheMfMonitorMake((KHE_MONITOR) lwm, m_index, mf);
  HaArrayAddLast(mfr->monitors, mfm);

  /* get the range of days and add it to mfm */
  KheLimitWorkloadMonitorRange(lwm, &first_time, &last_time);
  if( first_time != NULL )
  {
    /* non-empty set of days */
    first_day_index = KheFrameTimeIndex(mf->frame, first_time);
    last_day_index = KheFrameTimeIndex(mf->frame, last_time);
  }
  else
  {
    /* empty set of days */
    first_day_index =  0;
    last_day_index = -1;
  }
  KheMfMonitorSetDayIndexRange(mfm, first_day_index, last_day_index, mf);

  /* add the items and the days they are affected */
  for( i = 0;  i < KheLimitWorkloadMonitorTimeGroupCount(lwm);  i++ )
  {
    tg = KheLimitWorkloadMonitorTimeGroup(lwm, i, &workload);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      time = KheTimeGroupTime(tg, j);
      day_index = KheFrameTimeIndex(mf->frame, time);
      KheMfMonitorSetItemAffectedByDay(mfm, i, day_index);
    }
  }

  /* add the monitor as an item to the resource's days affected */
  KheMfMonitorSetAffectedDays(mfm, mfr, mf);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceAddLimitActiveIntervalsMonitor(KHE_MF_RESOURCE mfr,    */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int m_index,                  */
/*    KHE_MONITOR_FRAMER mf)                                                 */
/*                                                                           */
/*  Update mfr's data structures to take account of limit active intervals   */
/*  monitor laim, whose index is m_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceAddLimitActiveIntervalsMonitor(KHE_MF_RESOURCE mfr,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int m_index, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_MONITOR mfm;  KHE_TIME first_time, last_time, time;  KHE_POLARITY po;
  int first_day_index, last_day_index, day_index, i, j;  KHE_TIME_GROUP tg;

  /* make and add a monitor object corresponding to laim (even if empty) */
  mfm = KheMfMonitorMake((KHE_MONITOR) laim, m_index, mf);
  HaArrayAddLast(mfr->monitors, mfm);

  /* get the range of days and add it to mfm */
  KheLimitActiveIntervalsMonitorRange(laim, &first_time, &last_time);
  if( first_time != NULL )
  {
    /* non-empty set of days */
    first_day_index = KheFrameTimeIndex(mf->frame, first_time);
    last_day_index = KheFrameTimeIndex(mf->frame, last_time);
  }
  else
  {
    /* empty set of days */
    first_day_index =  0;
    last_day_index = -1;
  }
  KheMfMonitorSetDayIndexRange(mfm, first_day_index, last_day_index, mf);

  /* add the items and the days they are affected */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      time = KheTimeGroupTime(tg, j);
      day_index = KheFrameTimeIndex(mf->frame, time);
      KheMfMonitorSetItemAffectedByDay(mfm, i, day_index);
    }
  }

  /* add the monitor as an item to the resource's days affected */
  KheMfMonitorSetAffectedDays(mfm, mfr, mf);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceAddMonitor(KHE_MF_RESOURCE mfr, KHE_MONITOR m,         */
/*    int m_index, KHE_MONITOR_FRAMER mf)                                    */
/*                                                                           */
/*  Add monitor m (with index m_index) to mf.                                */
/*                                                                           */
/*  Some monitors are skipped because we don't handle them.  However, it     */
/*  is important for each monitor m to appear in mfr's monitors array at     */
/*  its index position m_index, so we add NULL entries to monitors there.    */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceAddMonitor(KHE_MF_RESOURCE mfr, KHE_MONITOR m,
  int m_index, KHE_MONITOR_FRAMER mf)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_AVOID_CLASHES_MONITOR_TAG:

      /* omitting these ones */
      HaArrayAddLast(mfr->monitors, NULL);
      break;

    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:

      KheMfResourceAddAvoidUnavailableTimesMonitor(mfr,
	(KHE_AVOID_UNAVAILABLE_TIMES_MONITOR) m, m_index, mf);
      break;

    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:

      /* omitting these ones */
      HaArrayAddLast(mfr->monitors, NULL);
      break;

    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      KheMfResourceAddClusterBusyTimesMonitor(mfr,
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m, m_index, mf);
      break;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      KheMfResourceAddLimitBusyTimesMonitor(mfr,
	(KHE_LIMIT_BUSY_TIMES_MONITOR) m, m_index, mf);
      break;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      KheMfResourceAddLimitWorkloadMonitor(mfr,
	(KHE_LIMIT_WORKLOAD_MONITOR) m, m_index, mf);
      break;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      KheMfResourceAddLimitActiveIntervalsMonitor(mfr,
	(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m, m_index, mf);
      break;

    default:

      HnAbort("KheMfResourceAddMonitor internal error (tag %d)",
	KheMonitorTag(m));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_RESOURCE KheMfResourceMake(KHE_RESOURCE r, KHE_MONITOR_FRAMER mf) */
/*                                                                           */
/*  Make a new resource object for r.  This includes building a set of       */
/*  monitor indexes for r on each day of the cycle.                          */
/*                                                                           */
/*****************************************************************************/

static KHE_MF_RESOURCE KheMfResourceMake(KHE_RESOURCE r, KHE_MONITOR_FRAMER mf)
{
  KHE_MF_RESOURCE res;  int i;  KHE_MONITOR m;

  /* make the basic object */
  HaMake(res, mf->arena);
  res->resource = r;
  HaArrayInit(res->monitors, mf->arena);
  HaArrayInit(res->days, mf->arena);
  KheSetInit(res->monitor_indexes, mf->arena);
  res->seq_num = -1;  /* signifies monitor_indexes not up to date */

  /* add one resource on day object for each day of the cycle */
  for( i = 0;  i < KheFrameTimeGroupCount(mf->frame);  i++ )
    HaArrayAddLast(res->days, KheResourceOnDayMake(mf));

  /* add r's monitors to res */
  for( i = 0;  i < KheSolnResourceMonitorCount(mf->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(mf->soln, r, i);
    KheMfResourceAddMonitor(res, m, i, mf);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceBringMonitorIndexesUpToDate(KHE_MF_RESOURCE mfr,       */
/*    KHE_MONITOR_FRAMER mf)                                                 */
/*                                                                           */
/*  Bring mfr->monitor_indexes up to date.  It needs to contain the union,   */
/*  over the days of interest d, of mfr's monitor indexes on day d.          */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceBringMonitorIndexesUpToDate(KHE_MF_RESOURCE mfr,
  KHE_MONITOR_FRAMER mf)
{
  int day_index, i;  KHE_MF_RESOURCE_ON_DAY mfrd;
  if( mfr->seq_num < mf->seq_num )
  {
    KheSetClear(mfr->monitor_indexes);
    KheSetForEach(mf->day_indexes, day_index, i)
    {
      mfrd = HaArray(mfr->days, day_index);
      KheSetUnion(mfr->monitor_indexes, mfrd->monitor_indexes);
    }
    mfr->seq_num = mf->seq_num;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMfResourceDebug(KHE_MF_RESOURCE mfr, KHE_MONITOR_FRAMER mf,      */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of mfr onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheMfResourceDebug(KHE_MF_RESOURCE mfr, KHE_MONITOR_FRAMER mf,
  int verbosity, int indent, FILE *fp)
{
  KHE_MF_MONITOR mfm;  int i;  KHE_MF_RESOURCE_ON_DAY mfrd;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ %s\n", indent, "", KheResourceId(mfr->resource));
    HaArrayForEach(mfr->monitors, mfm, i)
      KheMfMonitorDebug(mfm, mf, verbosity, indent + 2, fp);
    HaArrayForEach(mfr->days, mfrd, i)
      KheMfResourceOnDayDebug(mfrd, i, mfr, mf, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "%s", KheResourceId(mfr->resource));
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_MONITOR_FRAMER"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_MF_RESOURCE KheMfGetResource(KHE_MONITOR_FRAMER mf, KHE_RESOURCE r)  */
/*                                                                           */
/*  Get the mf resource object corresponding to mf; possibly make it first.  */
/*                                                                           */
/*****************************************************************************/

static KHE_MF_RESOURCE KheMfGetResource(KHE_MONITOR_FRAMER mf, KHE_RESOURCE r)
{
  int index;  KHE_MF_RESOURCE res;
  index = KheResourceInstanceIndex(r);
  HaArrayFill(mf->resources, index + 1, NULL);
  res = HaArray(mf->resources, index);
  if( res == NULL )
  {
    res = KheMfResourceMake(r, mf);
    HaArrayPut(mf->resources, index, res);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public functions"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_MONITOR_FRAMER KheMonitorFramerMake(KHE_SOLN soln, KHE_FRAME frame,  */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Create a new monitor framer object with these attributes.                */
/*                                                                           */
/*****************************************************************************/

KHE_MONITOR_FRAMER KheMonitorFramerMake(KHE_SOLN soln, KHE_FRAME frame,
  HA_ARENA a)
{
  KHE_MONITOR_FRAMER res;
  HaMake(res, a);
  res->arena = a;
  res->soln = soln;
  res->frame = frame;
  KheSetInit(res->day_indexes, a);
  res->state = KHE_MONITOR_FRAMER_INIT;
  res->seq_num = 0;
  HaArrayInit(res->resources, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFramerDaysBegin(KHE_MONITOR_FRAMER mf)                    */
/*                                                                           */
/*  Begin adding some selected days to mf.                                   */
/*                                                                           */
/*****************************************************************************/

void KheMonitorFramerDaysBegin(KHE_MONITOR_FRAMER mf)
{
  HnAssert(mf->state != KHE_MONITOR_FRAMER_LOADING,
    "KheMonitorFramerDaysBegin called out of order");
  KheSetClear(mf->day_indexes);
  mf->state = KHE_MONITOR_FRAMER_LOADING;
  mf->seq_num++;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFramerDaysAdd(KHE_MONITOR_FRAMER mf,                      */
/*    int first_day_index, int last_day_index)                               */
/*                                                                           */
/*  Add selected days first_day_index .. last_day_index to mf.               */
/*                                                                           */
/*****************************************************************************/

void KheMonitorFramerDaysAdd(KHE_MONITOR_FRAMER mf,
  int first_day_index, int last_day_index)
{
  int i;
  HnAssert(mf->state == KHE_MONITOR_FRAMER_LOADING,
    "KheMonitorFramerDaysAdd called out of order");
  for( i = first_day_index;  i <= last_day_index;  i++ )
    KheSetInsert(mf->day_indexes, i);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFramerDaysEnd(KHE_MONITOR_FRAMER mf)                      */
/*                                                                           */
/*  End adding some selected days to mf.                                     */
/*                                                                           */
/*****************************************************************************/

void KheMonitorFramerDaysEnd(KHE_MONITOR_FRAMER mf)
{
  HnAssert(mf->state == KHE_MONITOR_FRAMER_LOADING,
    "KheMonitorFramerDaysEnd called out of order");
  mf->state = KHE_MONITOR_FRAMER_COMPLETE;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheMonitorFramerTimeSelected(KHE_MONITOR_FRAMER mf, KHE_TIME time)  */
/*                                                                           */
/*  Return true if time is on one of the selected days of mf.                */
/*                                                                           */
/*****************************************************************************/

bool KheMonitorFramerTimeSelected(KHE_MONITOR_FRAMER mf, KHE_TIME time)
{
  int day_index;
  HnAssert(mf->state == KHE_MONITOR_FRAMER_COMPLETE,
    "KheMonitorFramerTimeSelected called out of order");
  day_index = KheFrameTimeIndex(mf->frame, time);
  return KheSetContains(mf->day_indexes, day_index);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorFramerResourceMonitorCount(KHE_MONITOR_FRAMER mf,          */
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Return the number of r's monitors affected by the selected days.         */
/*                                                                           */
/*****************************************************************************/

int KheMonitorFramerResourceMonitorCount(KHE_MONITOR_FRAMER mf, KHE_RESOURCE r)
{
  KHE_MF_RESOURCE mfr;
  HnAssert(mf->state == KHE_MONITOR_FRAMER_COMPLETE,
    "KheMonitorFramerResourceMonitorCount called out of order");
  mfr = KheMfGetResource(mf, r);
  KheMfResourceBringMonitorIndexesUpToDate(mfr, mf);
  return KheSetCount(mfr->monitor_indexes);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorFramerResourceMonitor(KHE_MONITOR_FRAMER mf,               */
/*     KHE_RESOURCE r, int i)                                                */
/*                                                                           */
/*  Return the index of r's i'th affected monitor.                           */
/*                                                                           */
/*****************************************************************************/

int KheMonitorFramerResourceMonitor(KHE_MONITOR_FRAMER mf,
  KHE_RESOURCE r, int i)
{
  KHE_MF_RESOURCE mfr;
  HnAssert(mf->state == KHE_MONITOR_FRAMER_COMPLETE,
    "KheMonitorFramerResourceMonitor called out of order");
  mfr = KheMfGetResource(mf, r);
  KheMfResourceBringMonitorIndexesUpToDate(mfr, mf);
  return KheSetGet(mfr->monitor_indexes, i);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorFramerAffectedItemCount(KHE_MONITOR_FRAMER mf,             */
/*    KHE_RESOURCE r, int i)                                                 */
/*                                                                           */
/*  Return the number of affected items of the i'th monitor of r.            */
/*                                                                           */
/*****************************************************************************/

int KheMonitorFramerAffectedItemCount(KHE_MONITOR_FRAMER mf,
  KHE_RESOURCE r, int i)
{
  KHE_MF_RESOURCE mfr;  KHE_MF_MONITOR mfm;
  HnAssert(mf->state == KHE_MONITOR_FRAMER_COMPLETE,
    "KheMonitorFramerAffectedItemCount called out of order");
  mfr = KheMfGetResource(mf, r);
  KheMfResourceBringMonitorIndexesUpToDate(mfr, mf);
  mfm = HaArray(mfr->monitors, i);
  KheMfMonitorBringItemIndexesUpToDate(mfm, mf);
  return KheSetCount(mfm->item_indexes);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorFramerAffectedItem(KHE_MONITOR_FRAMER mf,                  */
/*    KHE_RESOURCE r, int i, int j)                                          */
/*                                                                           */
/*  Return the j'th affected item of the i'th monitor of r.                  */
/*                                                                           */
/*****************************************************************************/

int KheMonitorFramerAffectedItem(KHE_MONITOR_FRAMER mf, KHE_RESOURCE r,
  int i, int j)
{
  KHE_MF_RESOURCE mfr;  KHE_MF_MONITOR mfm;
  HnAssert(mf->state == KHE_MONITOR_FRAMER_COMPLETE,
    "KheMonitorFramerAffectedItem called out of order");
  mfr = KheMfGetResource(mf, r);
  KheMfResourceBringMonitorIndexesUpToDate(mfr, mf);
  mfm = HaArray(mfr->monitors, i);
  KheMfMonitorBringItemIndexesUpToDate(mfm, mf);
  return KheSetGet(mfm->item_indexes, j);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFramerDebug(KHE_MONITOR_FRAMER mf, int verbosity,         */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of mf with the given verbosity and indent.                   */
/*                                                                           */
/*****************************************************************************/

void KheMonitorFramerDebug(KHE_MONITOR_FRAMER mf, int verbosity,
  int indent, FILE *fp)
{
  KHE_MF_RESOURCE mr;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ MonitorFramer(%s, %s, seq %d)\n", indent, "",
      KheInstanceId(KheSolnInstance(mf->soln)), KheMfStateShow(mf->state),
      mf->seq_num);
    if( mf->state == KHE_MONITOR_FRAMER_COMPLETE )
      fprintf(fp, "%*s  selected days: %s\n", indent, "",
	KheSetShow(mf->day_indexes));
    HaArrayForEach(mf->resources, mr, i)
      KheMfResourceDebug(mr, mf, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "[ Monitor Framer ]");
}
