
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_dynamic_resource.c                                  */
/*  DESCRIPTION:  Dynamic resource solving                                   */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_p.h"
#include <limits.h>
#include <float.h>
#include <math.h>

#define TESTING 1
#define COL_WIDTH  12		/* when debugging solutions */
#define COL_COUNT   7		/* when debugging solutions */

#define bool_show(x) ((x) ? "true" : "false")

#define UNDEF_INT -1
#define UNDEF_FLOAT -1.0

#define DEBUG_MONITOR_ID "Constraint:2/TR_28/2Fri3"
#define DEBUG_MONITOR(m) (true &&					\
  strcmp(KheMonitorId((KHE_MONITOR) m), DEBUG_MONITOR_ID) == 0)

#define	RERUN  1		/* rerun                        */
#define RERUN_MONITOR_ID ""

#define DEBUG1 1		/* main functions		*/
#define DEBUG2 0		/* final solutions		*/
#define DEBUG3 0		/* task classes                 */
#define DEBUG4 0		/* KheDynamicResourceSolverMake */
#define DEBUG5 0		/* hash table efficiency      	*/
#define DEBUG6 0		/* trie dominance tests 	*/
#define DEBUG7 0		/* ResourceMake                 */
#define DEBUG8 0		/* expression construction      */
#define DEBUG9 0		/* KheDrsSolveOpen (costs)      */
#define DEBUG10 0		/* KheDrsExprOpen               */
#define DEBUG11 0		/* KheDrsResourceOpen           */
#define DEBUG12 0		/* KheDrsMakeEvaluateAndAddSoln */
#define DEBUG14 0		/* int seq cost child has opened*/
#define DEBUG15 0		/* KheDrsTaskAssign             */
#define DEBUG16 0		/* KheTaskUnAssignResource      */
#define DEBUG17 1		/* Search                       */
#define DEBUG18 0		/* KheDrsDayOpen                */
#define DEBUG19 0		/* KheDrsTaskClassOpen          */
#define DEBUG20 0		/* KheDrsAUIntervalFindRight    */
#define DEBUG21 0		/* Search statistics            */
#define DEBUG22 0		/* SolnExtend and SolnDoExtend  */
#define DEBUG23 0		/* set signatures               */
#define DEBUG24 0		/* rerun costs                  */
#define DEBUG25 0		/* dominating and dominated     */
#define DEBUG26 0		/* trie stats                   */
#define DEBUG27 0		/* trie stats                   */
#define DEBUG28 0		/* ancestor frequencies         */
#define DEBUG29 0		/* soln set on each day         */
#define DEBUG30 1		/* dominance debug              */

#define DEBUG_LAST_EDGE (false && (drs)->solve_soln_limit == 0)


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations"                                            */
/*                                                                           */
/*  This submodule contains forward type declarations, when needed.          */
/*                                                                           */
/*  There are many types and no obvious best ordering, so here and           */
/*  thoughout this file we've followed the traditional ordering, except      */
/*  that we handle resources before days, because the implementation         */
/*  opens resources before days.  Here is the order:                         */
/*                                                                           */
/*    KHE entities        Corresponding DRS entities                         */
/*    -----------------------------------------------------------------      */
/*    Resources           DRS resources                                      */
/*    Times               DRS days                                           */
/*    Events              DRS tasks                                          */
/*    Constraints         DRS dominance tests and expressions                */
/*    Solutions           DRS solutions and the solver                       */
/*    -----------------------------------------------------------------      */
/*                                                                           */
/*****************************************************************************/

/* type declarations (resources) */
typedef struct khe_drs_resource_on_day_rec *KHE_DRS_RESOURCE_ON_DAY;
typedef HA_ARRAY(KHE_DRS_RESOURCE_ON_DAY) ARRAY_KHE_DRS_RESOURCE_ON_DAY;

typedef struct khe_drs_resource_rec *KHE_DRS_RESOURCE;
typedef HA_ARRAY(KHE_DRS_RESOURCE) ARRAY_KHE_DRS_RESOURCE;

/* type declarations (times) */
typedef struct khe_drs_day_rec *KHE_DRS_DAY;
typedef HA_ARRAY(KHE_DRS_DAY) ARRAY_KHE_DRS_DAY;

/* type declarations (events) */
typedef struct khe_drs_task_on_day_rec *KHE_DRS_TASK_ON_DAY;
typedef HA_ARRAY(KHE_DRS_TASK_ON_DAY) ARRAY_KHE_DRS_TASK_ON_DAY;

typedef struct khe_drs_task_rec *KHE_DRS_TASK;
typedef HA_ARRAY(KHE_DRS_TASK) ARRAY_KHE_DRS_TASK;

typedef struct khe_drs_task_class_rec *KHE_DRS_TASK_CLASS;
typedef HA_ARRAY(KHE_DRS_TASK_CLASS) ARRAY_KHE_DRS_TASK_CLASS;

/* type declarations (constraints) */
typedef struct khe_drs_dom_test_rec *KHE_DRS_DOM_TEST;
typedef HA_ARRAY(KHE_DRS_DOM_TEST) ARRAY_KHE_DRS_DOM_TEST;

typedef struct khe_drs_expr_rec *KHE_DRS_EXPR;
typedef HA_ARRAY(KHE_DRS_EXPR) ARRAY_KHE_DRS_EXPR;

/* type declarations (signatures and solutions) */
typedef struct khe_drs_signature_rec *KHE_DRS_SIGNATURE;
typedef HA_ARRAY(KHE_DRS_SIGNATURE) ARRAY_KHE_DRS_SIGNATURE;

typedef struct khe_drs_soln_rec *KHE_DRS_SOLN;
typedef HA_ARRAY(KHE_DRS_SOLN) ARRAY_KHE_DRS_SOLN;
typedef HP_TABLE(KHE_DRS_SOLN) TABLE_KHE_DRS_SOLN;

typedef struct khe_drs_soln_set_rec *KHE_DRS_SOLN_SET;
typedef HA_ARRAY(KHE_DRS_SOLN_SET) ARRAY_KHE_DRS_SOLN_SET;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - resources"                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_RESOURCE_ON_DAY - what one resource is doing on one day     */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_resource_on_day_rec {
  KHE_DRS_RESOURCE		encl_dr;		/* the resource      */
  KHE_DRS_DAY			day;			/* the day           */
  KHE_DRS_TASK_ON_DAY		closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_EXPR		external_today;		/* external today    */
  ARRAY_KHE_DRS_EXPR		internal_today;		/* internal today    */
  /* ARRAY_KHE_DRS_DOM_TEST	dom_tests; */		/* for each sig pos  */
  /* HA_ARRAY_INT	    eq_dom_test_indexes; */	/* where = tests are */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_RESOURCE - one resource                                     */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_resource_rec {
  KHE_RESOURCE			resource;		/* the resource      */
  int				open_resource_index;	/* index when open   */
  ARRAY_KHE_DRS_RESOURCE_ON_DAY	days;			/* doing each day    */
  ARRAY_KHE_DRS_SIGNATURE	extend_signatures;	/* for current day   */
  KHE_DRS_TASK_ON_DAY		extend_fixed_dtd;	/* if fixed today    */
  KHE_DRS_SIGNATURE		extend_fixed_signature;	/* if fixed today    */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - times"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DAY_RANGE - a range of days                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_day_range_rec {
  int			first;
  int			last;
} KHE_DRS_DAY_RANGE;

typedef HA_ARRAY(KHE_DRS_DAY_RANGE) ARRAY_KHE_DRS_DAY_RANGE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DAY - one day                                               */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_day_rec {
  int				frame_index;		/* index in frame    */
  int				open_day_index;		/* index when open   */
  KHE_TIME_GROUP		time_group;		/* day's time group  */
  ARRAY_KHE_DRS_TASK_CLASS	task_classes;		/* today's tasks     */
  /* ARRAY_KHE_DRS_EXPR		internal_today; */	/* open today        */
  ARRAY_KHE_DRS_EXPR		nr_internal_today;	/* open today        */
  ARRAY_KHE_DRS_DOM_TEST	dom_tests;		/* for each sig pos  */
  HA_ARRAY_INT			eq_dom_test_indexes;	/* where = tests are */
  KHE_DRS_SOLN_SET		soln_set;		/* today's solns     */
  /* KHE_COST		single_day_curr_cost; */	/* today so far      */
  /* KHE_COST		single_day_min_cost;  */	/* min curr_cost     */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - events"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK_ON_DAY - what one proper root task is doing on one day */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_on_day_rec {
  KHE_DRS_TASK			encl_dt;		/* enclosing task    */
  KHE_DRS_DAY			day;			/* the day           */
  KHE_TASK			task;			/* atomic task today */
  KHE_TIME			time;			/* time on day       */
  KHE_DRS_RESOURCE_ON_DAY	closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_EXPR		external_today;		/* leaf today        */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK - one proper root task                                 */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_rec {
  KHE_DRS_TASK_CLASS		encl_dtc;
  int				index_in_encl_dtc;
  bool				extend_must_assign;	/* used by SolnExtend*/
  bool				open;			/* true when open    */
  KHE_TASK			task;			/* the proper root   */
  KHE_COST			asst_cost;		/* assignment cost   */
  KHE_COST			non_asst_cost;		/* non-asst cost     */
  KHE_DRS_RESOURCE		closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_TASK_ON_DAY	days;
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK_CLASS - a set of tasks, taken from one KHE_TASK_CLASS  */
/*                                                                           */
/*  For efficiency, unassigned_tasks is managed lazily.  Every unassigned    */
/*  task is in it, but not in order, and there may be assigned tasks in it.  */
/*  These defects are corrected when the task class is opened for solving.   */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_class_rec {
  KHE_TASK_CLASS		orig_task_class;	/* origin of this    */
  KHE_DRS_DAY_RANGE		day_range;		/* days running      */
  ARRAY_KHE_DRS_TASK		all_tasks;		/* all tasks in class*/
  ARRAY_KHE_DRS_TASK		unassigned_tasks;	/* unassigned tasks  */
  int				extend_used;		/* used by SolnExtend*/
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - constraints"                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_KIND - the kind of dominance testing to use             */
/*                                                                           */
/*****************************************************************************/

/* *** defined in khe_solvers.h
typedef enum {
  KHE_DRS_DOM_NONE,
  KHE_DRS_DOM_WEAK,
  KHE_DRS_DOM_MEDIUM,
  KHE_DRS_DOM_STRONG,
  KHE_DRS_DOM_TRIE,
  KHE_DRS_DOM_STRONG_WITH_TRADEOFF,
  KHE_DRS_DOM_TRIE_WITH_TRADEOFF,
  KHE_DRS_DOM_INDEXED_WITH_TRADEOFF
} KHE_DRS_DOM_KIND;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_TYPE - the type of a dominance test                     */
/*                                                                           */
/*****************************************************************************/

/* *** no longer used
typedef enum {
  KHE_DRS_DOM_UNUSED,		** "-"      should never be used      **
  KHE_DRS_DOM_GE,		** ">="     min limit only            **
  KHE_DRS_DOM_LE,		** "<="     max limit only            **
  KHE_DRS_DOM_EQ,		** "==",    for uncertain cases       **
  KHE_DRS_DOM_GE_LOOSE,		** ">=*"    min limit with loosening  **
  KHE_DRS_DOM_EQ_LOOSE		** "==*"    max + min with loosening  **
} KHE_DRS_DOM_TYPE;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_TEST - a dominance test                                 */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_dom_test_rec {
  bool			allow_zero;		/* allow zero ("Z" in doc)   */
  int			min_limit;              /* min limit (0 if none)     */
  int			max_limit;              /* max limit (INT_MAX if ne) */
  bool			tradeoff_allowed;	/* true if tradeoff allowed  */
  int			a;			/* "a" in doc                */
  int			b;			/* "b" in doc                */
  KHE_COST		tradeoff_cost;		/* tradeoff cost, if tradeoff*/
  KHE_MONITOR		monitor;		/* monitor, if any           */
};

/* *** obsolete version
struct khe_drs_dom_test_rec {
  KHE_DRS_DOM_TYPE	type;			** type of test              **
  bool			tradeoff;		** true if accepts tradeoff  **
  int			min_limit;		** GE_LOOSE and EQ_LOOSE     **
  KHE_COST		tradeoff_cost;		** tradeoff cost, if tradeoff**
};
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_TAG                                                    */
/*                                                                           */
/*  Tags for expression types.                                               */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_EXPR_ASSIGNED_TASK_TAG,
  KHE_DRS_EXPR_BUSY_TIME_TAG,
  KHE_DRS_EXPR_FREE_TIME_TAG,
  KHE_DRS_EXPR_WORK_TIME_TAG,
  KHE_DRS_EXPR_BUSY_DAY_TAG,
  KHE_DRS_EXPR_FREE_DAY_TAG,
  KHE_DRS_EXPR_WORK_DAY_TAG,
  KHE_DRS_EXPR_OR_TAG,
  KHE_DRS_EXPR_AND_TAG,	
  KHE_DRS_EXPR_INT_SUM_TAG,
  KHE_DRS_EXPR_FLOAT_SUM_TAG,
  KHE_DRS_EXPR_INT_DEV_TAG,
  KHE_DRS_EXPR_FLOAT_DEV_TAG,
  KHE_DRS_EXPR_INT_SUM_COST_TAG,
  KHE_DRS_EXPR_INT_SEQ_COST_TAG
} KHE_DRS_EXPR_TAG;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_ADJUST_TYPE - types of signature value adjustments          */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_ADJUST_ORDINARY,
  KHE_DRS_ADJUST_NO_MAX,
  KHE_DRS_ADJUST_LINEAR,
  KHE_DRS_ADJUST_STEP
} KHE_DRS_ADJUST_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_VALUE - the value of an expression; could be int or float   */
/*                                                                           */
/*****************************************************************************/

typedef union {
  int			int_val;		/* value, if int   */
  float			float_val;		/* value, if float */
} KHE_DRS_VALUE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_PARENT - a expression's parent                              */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_parent_rec {
  KHE_DRS_EXPR		expr;			/* the parent expression */
  int			index;			/* index in parent       */
} KHE_DRS_PARENT;

typedef HA_ARRAY(KHE_DRS_PARENT) ARRAY_KHE_DRS_PARENT;


/*****************************************************************************/
/*                                                                           */
/*  INHERIT_KHE_DRS_EXPR - fields inherited by all expression types          */
/*                                                                           */
/*  For each open day di of e, except its very last open day,                */
/*                                                                           */
/*    e->sig_index[di - e->open_day_range.first]                             */
/*                                                                           */
/*  is the position of e's value in the signatures of solutions ending on    */
/*  day di.  For each open day di of e,                                      */
/*                                                                           */
/*    e->open_child_indexes[di - e->open_day_range.first]                    */
/*                                                                           */
/*  is the index of the first child whose last open day index is di or more. */
/*  For more on this, see the header of submodule "open child indexes".      */
/*                                                                           */
/*****************************************************************************/

#define INHERIT_KHE_DRS_EXPR						\
  KHE_DRS_EXPR_TAG	tag;			/* tag field      */	\
  bool			gathered;		/* for opening    */	\
  /* bool		single_day; */		/* single day     */	\
  /* KHE_DRS_DOM_TEST	dom_test; */		/* dom test (obs) */	\
  int			postorder_index;	/* expr index     */	\
  KHE_DRS_RESOURCE	resource;		/* if resource    */	\
  KHE_DRS_VALUE		u;			/* value          */	\
  ARRAY_KHE_DRS_PARENT	parents;		/* parent exprs   */	\
  ARRAY_KHE_DRS_EXPR	children;		/* child exprs    */	\
  KHE_DRS_DAY_RANGE	open_day_range;		/* open day range */	\
  ARRAY_KHE_DRS_EXPR	open_children;		/* open children  */	\
  HA_ARRAY_INT		open_child_indexes;	/* for each day   */	\
  HA_ARRAY_INT		sig_indexes;		/* for each day   */	\
  /* HA_ARRAY_INT	signature_indexes; */	/* for each day   */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR - abstract supertype of expressions                    */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_expr_rec {
  INHERIT_KHE_DRS_EXPR
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_ASSIGNED_TASK                                          */
/*                                                                           */
/*  An expression derived from a task on day, whose value is 1 if the task   */
/*  is assigned a resource from the given resource group, and 0 otherwise.   */
/*                                                                           */
/*  We include the task on day because we merge equal expressions of this    */
/*  kind, and that is only possible when the task is in the expression.      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_assigned_task_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_TASK_ON_DAY		task_on_day;
  KHE_RESOURCE_GROUP		resource_group;
} *KHE_DRS_EXPR_ASSIGNED_TASK;

typedef HA_ARRAY(KHE_DRS_EXPR_ASSIGNED_TASK) ARRAY_KHE_DRS_EXPR_ASSIGNED_TASK;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_BUSY_TIME                                              */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy at a given      */
/*  time, and 0 when the resource is free then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time we're busy at    */
} *KHE_DRS_EXPR_BUSY_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_TIME) ARRAY_KHE_DRS_EXPR_BUSY_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FREE_TIME                                              */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free at a given      */
/*  time, and 0 when the resource is busy then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time we're free at    */
} *KHE_DRS_EXPR_FREE_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_FREE_TIME) ARRAY_KHE_DRS_EXPR_FREE_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_WORK_TIME                                              */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource at    */
/*  a given time.  This will be 0.0 if the resource is free then.  The       */
/*  expression stores the time, but only for debugging.                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time of this workload */
} *KHE_DRS_EXPR_WORK_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_WORK_TIME) ARRAY_KHE_DRS_EXPR_WORK_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_BUSY_DAY                                               */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy on a given      */
/*  day, and 0 when the resource is free then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're busy at     */
} *KHE_DRS_EXPR_BUSY_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_DAY) ARRAY_KHE_DRS_EXPR_BUSY_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FREE_DAY                                               */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free on a given      */
/*  day, and 0 when the resource is busy then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_FREE_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_FREE_DAY) ARRAY_KHE_DRS_EXPR_FREE_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_WORK_DAY                                               */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource on    */
/*  a given day.  This will be 0.0 if the resource is free then.  The        */
/*  expression stores the day (as a time group), but only for debugging.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_WORK_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_WORK_DAY) ARRAY_KHE_DRS_EXPR_WORK_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_OR                                                     */
/*                                                                           */
/*  An expression whose value is the logical `or' of its children's values.  */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*  The closed state here is the number of closed children with value 1.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_or_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_OR;

typedef HA_ARRAY(KHE_DRS_EXPR_OR) ARRAY_KHE_DRS_EXPR_OR;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_AND                                                    */
/*                                                                           */
/*  An expression whose value is the logical `and' of its children's values. */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*  The closed state here is the number of closed children with value 0.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_and_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_AND;

typedef HA_ARRAY(KHE_DRS_EXPR_AND) ARRAY_KHE_DRS_EXPR_AND;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SUM                                                */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.           */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_INT_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM) ARRAY_KHE_DRS_EXPR_INT_SUM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT_SUM                                              */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.  The      */
/*  expression's value, and its children's values, have type float.          */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_sum_rec {
  INHERIT_KHE_DRS_EXPR
  float				closed_state;
} *KHE_DRS_EXPR_FLOAT_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_SUM) ARRAY_KHE_DRS_EXPR_FLOAT_SUM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_DEV                                                */
/*                                                                           */
/*  An expression whose value is the deviation of its integer-valued child.  */
/*                                                                           */
/*  The closed state here is the value of the closed child.  It is not       */
/*  stored here explicitly; it is retrieved from the child when required.    */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_dev_rec {
  INHERIT_KHE_DRS_EXPR
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
} *KHE_DRS_EXPR_INT_DEV;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_DEV) ARRAY_KHE_DRS_EXPR_INT_DEV;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT_DEV                                              */
/*                                                                           */
/*  An expression whose value is the integer-valued deviation of its         */
/*  float-valued child.                                                      */
/*                                                                           */
/*  The closed state here is the value of the closed child.  It is not       */
/*  stored here explicitly; it is retrieved from the child when required.    */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_dev_rec {
  INHERIT_KHE_DRS_EXPR
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
} *KHE_DRS_EXPR_FLOAT_DEV;

typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_DEV) ARRAY_KHE_DRS_EXPR_FLOAT_DEV;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_COST                                                   */
/*                                                                           */
/*  This type is the abstract supertype of KHE_DRS_EXPR_INT_SUM_COST and     */
/*  KHE_DRS_EXPR_INT_SEQ_COST.  It is used, but not much.                    */
/*                                                                           */
/*****************************************************************************/

#define INHERIT_KHE_DRS_EXPR_COST					   \
  INHERIT_KHE_DRS_EXPR							   \
  KHE_MONITOR		monitor;		/* monitor (debug only) */ \
  KHE_COST_FUNCTION	cost_fn;		/* cost function        */ \
  KHE_COST		combined_weight;	/* combined weight      */ \
  int			min_limit;		/* minimum limit        */ \
  int			max_limit;		/* maximum limit        */ \
  bool			allow_zero;		/* allow zero flag      */ \
  KHE_DRS_ADJUST_TYPE	adjust_type;		/* type of adjustment   */ \
  int			history_before;		/* history value before */ \
  int			history_after;		/* history value after  */ \
  int			history;		/* history value        */ \
  KHE_COST		rerun_open_and_search_cost;  /* reruns only     */ \
  KHE_COST		rerun_open_and_close_cost;   /* reruns only     */


typedef struct khe_drs_expr_cost_rec {
  INHERIT_KHE_DRS_EXPR_COST
  int				closed_state;
} *KHE_DRS_EXPR_COST;

typedef HA_ARRAY(KHE_DRS_EXPR_COST) ARRAY_KHE_DRS_EXPR_COST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SUM_COST                                           */
/*                                                                           */
/*  Combines INT_SUM, INT_DEV, and COST into a single expression.  This      */
/*  allows important optimizations in some cases, depending on adjust_type.  */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_cost_rec {
  INHERIT_KHE_DRS_EXPR_COST
  int				closed_state;
} *KHE_DRS_EXPR_INT_SUM_COST;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM_COST) ARRAY_KHE_DRS_EXPR_INT_SUM_COST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_CLOSED_SEQ                                                  */
/*                                                                           */
/*  A sequence of consecutive closed children, summarized.  The fields are   */
/*                                                                           */
/*    start_index                                                            */
/*      The index in the sequence of all children where the sequence starts. */
/*                                                                           */
/*    stop_index                                                             */
/*      The index in the sequence of all children where the sequence         */
/*      stops; that is, the first child index not included in it.  So        */
/*      start_index <= stop_index, and when they are equal, the sequence     */
/*      is empty.                                                            */
/*                                                                           */
/*    active_at_left                                                         */
/*      The length of the a-interval at the extreme left, or 0 if none.      */
/*                                                                           */
/*    active_at_right                                                        */
/*      The length of the a-interval at the extreme right, or 0 if none.     */
/*                                                                           */
/*  If there is a single a-interval running the full length of the           */
/*  closed sequence (if all its children are active), then                   */
/*                                                                           */
/*    active_at_left == active_at_right == stop_index - start_index          */
/*                                                                           */
/*****************************************************************************/

typedef struct {
  int		start_index;
  int		stop_index;
  int		active_at_left;
  int		active_at_right;
} *KHE_DRS_CLOSED_SEQ;

typedef HA_ARRAY(KHE_DRS_CLOSED_SEQ) ARRAY_KHE_DRS_CLOSED_SEQ;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_A_INTERVAL - one a-interval                                 */
/*                                                                           */
/*  This may include indexes from history.                                   */
/*                                                                           */
/*  Note.  If the length of ai is 0, it does not matter whether an           */
/*  unassigned child precedes it or not, because its cost will be 0          */
/*  anyway.  However we must keep the value correct because there may        */
/*  be a merge later that makes it relevant.                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct {
  int		start_index;		/* index of first child              */
  int		stop_index;		/* index after last child            */
  bool		unassigned_precedes;	/* true if unassigned child precedes */
} KHE_DRS_A_INTERVAL;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_AU_INTERVAL - one au-interval                               */
/*                                                                           */
/*  This may include indexes from history and history_after.                 */
/*                                                                           */
/*****************************************************************************/

typedef struct {
  int		start_index;		/* index of first child              */
  int		stop_index;		/* index after last child            */
  bool		has_active_child;	/* true if contains an active child  */
} KHE_DRS_AU_INTERVAL;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SEQ_COST                                           */
/*                                                                           */
/*  Combines INT_SEQ, INT_DEV (without allow_zero), and COST into a single   */
/*  expression, for the same optimizations as KHE_DRS_EXPR_INT_SUM_COST.     */
/*                                                                           */
/*  The first_open_child_index contains the index in open_children of the    */
/*  first open child.  This will usually be 0, but during closing the        */
/*  children are closed in increasing index order and this field increases.  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_seq_cost_rec {
  INHERIT_KHE_DRS_EXPR_COST
  int				first_open_child_index;
  ARRAY_KHE_DRS_CLOSED_SEQ	closed_seqs;	/* closed sequences          */
} *KHE_DRS_EXPR_INT_SEQ_COST;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SEQ_COST) ARRAY_KHE_DRS_EXPR_INT_SEQ_COST;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - solutions and solver"                      */
/*                                                                           */
/*****************************************************************************/

#define INHERIT_KHE_DRS_SIGNATURE					\
  KHE_COST		cost;			/* cost          */	\
  HA_ARRAY_INT		sig;			/* signature     */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SIGNATURE - a signature                                          */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_signature_rec {
  INHERIT_KHE_DRS_SIGNATURE
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN - a partial solution, part of the search tree          */
/*                                                                           */
/*  All valid values are non-NULL.  In the root solution the prev_soln and   */
/*  day fields are NULL, and prev_tasks is empty.                            */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_soln_rec {
  INHERIT_KHE_DRS_SIGNATURE
  KHE_DRS_SOLN		prev_soln;		/* edge from prev_soln       */
  KHE_DRS_DAY		day;			/* day containing this soln  */
  ARRAY_KHE_DRS_TASK_ON_DAY prev_tasks;		/* indexed by open resources */
#if TESTING
  int			sorted_rank;		/* rank when solns sorted    */
#endif
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_LIST - a list of solns; has several uses               */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_list_rec {
  ARRAY_KHE_DRS_SOLN	solns;
} *KHE_DRS_SOLN_LIST;

typedef HA_ARRAY(KHE_DRS_SOLN_LIST) ARRAY_KHE_DRS_SOLN_LIST;
typedef HP_TABLE(KHE_DRS_SOLN_LIST) TABLE_KHE_DRS_SOLN_LIST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_TRIE - a trie of solns                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_trie_rec *KHE_DRS_SOLN_TRIE;
typedef HA_ARRAY(KHE_DRS_SOLN_TRIE) ARRAY_KHE_DRS_SOLN_TRIE;

struct khe_drs_soln_trie_rec {
  KHE_DRS_SOLN			soln;
  ARRAY_KHE_DRS_SOLN_TRIE	children;
  int				base;		/* first index of children */
};


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_INDEXED_SOLN_SET - an array of soln lists, indexed by cost       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_indexed_soln_set_rec {
  KHE_COST			base;
  KHE_COST			increment;
  ARRAY_KHE_DRS_SOLN_LIST	soln_lists;
} *KHE_DRS_INDEXED_SOLN_SET;

typedef HA_ARRAY(KHE_DRS_INDEXED_SOLN_SET) ARRAY_KHE_DRS_INDEXED_SOLN_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_SET_PART - one part of a solution set                       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_set_part_rec {
  int				count;			/* solns now in part */
  KHE_DRS_DOM_KIND		dom_kind;		/* kind of dominance */
  union {
    KHE_DRS_SOLN_LIST		none_dom_set;		/* none dominance    */
    TABLE_KHE_DRS_SOLN		weak_dom_set;		/* weak dominance    */
    TABLE_KHE_DRS_SOLN_LIST	medium_dom_set;		/* medium dominance  */
    KHE_DRS_SOLN_LIST		strong_dom_set;		/* strong dominance  */
    KHE_DRS_SOLN_TRIE		trie_dom_set;		/* trie dominance    */
    KHE_DRS_INDEXED_SOLN_SET	indexed_dom_set;	/* indexed           */
  } u;
} *KHE_DRS_SOLN_SET_PART;

typedef HA_ARRAY(KHE_DRS_SOLN_SET_PART) ARRAY_KHE_DRS_SOLN_SET_PART;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_SET - a set of solns, with a choice of implementations */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_soln_set_rec {
  int				made_count;		/* solns made        */
  KHE_DRS_SOLN_SET_PART		cache;			/* cache (optional)  */
  KHE_DRS_SOLN_SET_PART		main;			/* main part         */
#if TESTING
  ARRAY_KHE_DRS_SOLN		sorted_solns;
#endif
};


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN_DAY - one day of a packed solution                   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_packed_soln_day_rec {
  KHE_DRS_DAY			day;			/* day of this obj   */
  ARRAY_KHE_DRS_TASK_ON_DAY	prev_tasks;		/* tasks             */
} *KHE_DRS_PACKED_SOLN_DAY;

typedef HA_ARRAY(KHE_DRS_PACKED_SOLN_DAY) ARRAY_KHE_DRS_PACKED_SOLN_DAY;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN - a packed solution object                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_packed_soln_rec {
  KHE_COST			cost;			/* soln cost         */
  ARRAY_KHE_DRS_PACKED_SOLN_DAY	days;			/* days of rerun     */
} *KHE_DRS_PACKED_SOLN;

typedef HA_ARRAY(KHE_DRS_PACKED_SOLN) ARRAY_KHE_DRS_PACKED_SOLN;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DYNAMIC_RESOURCE_SOLVER - a dynamic resource solver             */
/*                                                                           */
/*****************************************************************************/

struct khe_dynamic_resource_solver_rec {

  /* fields which are constant throughout the lifetime of the solver */
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_RESOURCE_TYPE		resource_type;
  KHE_OPTIONS			options;
  KHE_FRAME			days_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_TASK_CLASS_SOLVER		task_class_solver;
  ARRAY_KHE_DRS_RESOURCE	all_resources;
  ARRAY_KHE_DRS_DAY		all_days;
  ARRAY_KHE_DRS_TASK		all_root_tasks;
  ARRAY_KHE_DRS_EXPR_COST	all_cost_exprs;
  int				postorder_count;
  KHE_COST			increment;		/* gcd of weights */

  /* free list fields */
  ARRAY_KHE_DRS_CLOSED_SEQ	free_closed_seqs;
  ARRAY_KHE_DRS_DOM_TEST	free_dom_tests;
  ARRAY_KHE_DRS_SIGNATURE	free_signatures;
  ARRAY_KHE_DRS_SOLN		free_solns;
  ARRAY_KHE_DRS_SOLN_LIST	free_soln_lists;
  ARRAY_KHE_DRS_SOLN_TRIE	free_soln_tries;
  ARRAY_KHE_DRS_SOLN_SET_PART	free_soln_set_parts;
  ARRAY_KHE_DRS_INDEXED_SOLN_SET free_indexed_soln_sets;
  ARRAY_KHE_DRS_SOLN_SET	free_soln_sets;
  ARRAY_KHE_DRS_PACKED_SOLN_DAY	free_packed_soln_days;
  ARRAY_KHE_DRS_PACKED_SOLN	free_packed_solns;

  /* fields that vary with the solve */
  KHE_RESOURCE_SET		selected_resource_set; /* selected for solve */
  ARRAY_KHE_DRS_DAY_RANGE	selected_day_ranges;   /* selected for solve */
  int				solve_soln_limit;
  KHE_COST			solve_init_cost;	/* before solving    */
  KHE_COST			solve_start_cost;	/* before first day  */
  ARRAY_KHE_DRS_RESOURCE	open_resources;
  ARRAY_KHE_DRS_DAY		open_days;
  ARRAY_KHE_DRS_EXPR		open_exprs;
  KHE_DRS_PACKED_SOLN		rerun;			/* non-NULL if rerun */

  /* fields used by KheDrsSolnDoExtend */
  ARRAY_KHE_DRS_TASK_ON_DAY	extend_prev_tasks;
  ARRAY_KHE_DRS_SIGNATURE	extend_prev_signatures;
  KHE_COST			extend_cost;
  int				extend_must_assign_count;

  /* fields for recording running time */
#if TESTING
  KHE_TIMER			timer;
  HA_ARRAY_INT			table_size_per_day;
  HA_ARRAY_FLOAT		running_time_per_day;
  HA_ARRAY_INT			ancestor_freq;
#endif
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule " miscellaneous functions"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int min(int a, int b)                                                    */
/*                                                                           */
/*  Return the smaller of a and b, or either if they are equal.              */
/*                                                                           */
/*****************************************************************************/

static int min(int a, int b)
{
  return a < b ? a : b;
}


/*****************************************************************************/
/*                                                                           */
/*  int max(int a, int b)                                                    */
/*                                                                           */
/*  Return the larger of a and b, or either if they are equal.               */
/*                                                                           */
/*****************************************************************************/

static int max(int a, int b)
{
  return a > b ? a : b;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SIGNATURE"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SIGNATURE KheDrsSignatureMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)   */
/*                                                                           */
/*  Make a new, empty signature object.                                      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SIGNATURE KheDrsSignatureMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SIGNATURE res;
  if( HaArrayCount(drs->free_signatures) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_signatures);
    HaArrayClear(res->sig);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->sig, drs->arena);
  }
  res->cost = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SIGNATURE KheDrsSignatureBuild(KHE_DRS_RESOURCE_ON_DAY drd,      */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_SOLN prev_soln,                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Build a sig seg holding what happens when drd is assigned to dtd.        */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsExprLeafClear(KHE_DRS_EXPR e,KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsExprEvalSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,
  int next_di, KHE_DRS_SIGNATURE ss, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static KHE_DRS_SIGNATURE KheDrsSignatureBuild(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_SOLN prev_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SIGNATURE res;  KHE_DRS_EXPR e;  int i, next_di;

  /* make a new signature object */
  res = KheDrsSignatureMake(drs);

  /* relevant part of what KheDrsPushAsst does (or used to do) */
  if( dtd != NULL )
  {
    /* update leaf expressions of the resource on day */
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd, drs);
  }

  /* relevant part of what KheDrsMakeEvaluateAndAddSoln does */
  next_di = drd->day->open_day_index;
  HaArrayForEach(drd->internal_today, e, i)
    KheDrsExprEvalSignature(e, prev_soln, next_di, res, drs);

  /* relevant part of what KheDrsPopAsst does (or used to do) */
  if( dtd != NULL )
  {
    /* update leaf expressions of the resource */
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafClear(e, drs);
  }

  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSignatureDebug(KHE_DRS_SIGNATURE sig, int verbosity,          */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of sig with the given verbosity and indent.                  */
/*  This function will also print a NULL sig, as "-".                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSignatureDebug(KHE_DRS_SIGNATURE sig, int verbosity,
  int indent, FILE *fp)
{
  int val, i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( sig != NULL )
  {
    fprintf(fp, "<%.5f", KheCostShow(sig->cost));
    HaArrayForEach(sig->sig, val, i)
      fprintf(fp, " %d", val);
    fprintf(fp, ">");
  }
  else
    fprintf(fp, "-");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE_ON_DAY"                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(KHE_DRS_RESOURCE dr,     */
/*    KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  Make a new resource on day object with these attributes.                 */
/*  This function assumes that dr is not assigned any task on day.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY res;
  HaMake(res, drs->arena);
  res->encl_dr = dr;
  res->day = day;
  res->closed_asst = NULL;
  HaArrayInit(res->external_today, drs->arena);
  HaArrayInit(res->internal_today, drs->arena);
  /* HaArrayInit(res->dom_tests, drs->arena); */
  /* HaArrayInit(res->eq_dom_test_indexes, drs->arena); */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,         */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If drd has a BUSY_TIME, FREE_TIME, or WORK_TIME expression with the      */
/*  given tag, dom_test, and time, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/
/* ***
static bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2);
*** */

static bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, /* KHE_DRS_DOM_TEST dom_test, */ KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_EXPR e;  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_EXPR_FREE_TIME eft;  KHE_DRS_EXPR_WORK_TIME ewt;
  HaArrayForEach(drd->external_today, e, i)
    if( e->tag == tag /* && KheDrsDomTestEqual(e->dom_test, dom_test) */ )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_TIME_TAG:

	  ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
	  if( ebt->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_TIME_TAG:

	  eft = (KHE_DRS_EXPR_FREE_TIME) e;
	  if( eft->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_TIME_TAG:

	  ewt = (KHE_DRS_EXPR_WORK_TIME) e;
	  if( ewt->time == time )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsResourceOnDayHasTimeExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,          */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If drd has a BUSY_DAY, FREE_DAY, or WORK_DAY expression with the given   */
/*  tag, dom_test, and time group, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, /* KHE_DRS_DOM_TEST dom_test, */ KHE_TIME_GROUP tg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_EXPR e;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;  KHE_DRS_EXPR_WORK_DAY ewd;
  HaArrayForEach(drd->external_today, e, i)
    if( e->tag == tag /* && KheDrsDomTestEqual(e->dom_test, dom_test) */ )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_DAY_TAG:

	  ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
	  if( KheTimeGroupEqual(ebd->time_group, tg) )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_DAY_TAG:

	  efd = (KHE_DRS_EXPR_FREE_DAY) e;
	  if( KheTimeGroupEqual(efd->time_group, tg) )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_DAY_TAG:

	  ewd = (KHE_DRS_EXPR_WORK_DAY) e;
	  if( KheTimeGroupEqual(ewd->time_group, tg) )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsResourceHasDayExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,             */
/*    KHE_DRS_EXPR e)                                                        */
/*                                                                           */
/*  Add e, known to be new, to drd.                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR e)
{
  HaArrayAddLast(drd->external_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayAddOpenExpr(KHE_DRS_RESOURCE_ON_DAY drd,         */
/*    KHE_DRS_EXPR e)                                                        */
/*                                                                           */
/*  Add internal expression e to drd.                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceOnDayAddOpenExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR e)
{
  HaArrayAddLast(drd->internal_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,           */
/*    KHE_DRS_DOM_TEST dom_test)                                             */
/*                                                                           */
/*  Add dom_test to drd.                                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_DOM_TEST dom_test)
{
  int res;

  ** add dom_test to drd **
  res = HaArrayCount(drd->dom_tests);
  HnAssert(dom_test.type != KHE_DRS_DOM_UNUSED,
    "KheDrsDayAddDomTest internal error");
  HaArrayAddLast(drd->dom_tests, dom_test);

  ** if dom_test is equality, record it in eq_dom_test_indexes **
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(drd->eq_dom_test_indexes, res);

  ** return dom_test's index **
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayIsFixed(KHE_DRS_RESOURCE_ON_DAY drd,             */
/*    KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY *dtd)                           */
/*                                                                           */
/*  If the assignment of drd is fixed, return true and set *dtd to the task  */
/*  on day that it is fixed to.                                              */
/*                                                                           */
/*  There are two reasons why drd may be fixed:                              */
/*                                                                           */
/*    * drd may have a closed assignment;                                    */
/*                                                                           */
/*    * drd's resource may have been assigned to some task on the previous   */
/*      day, and that task may still be running                              */
/*                                                                           */
/*  We can find out whether drd's resource has been assigned to some task    */
/*  on the previous day by consulting the edge leading into soln.            */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,
  KHE_DRS_RESOURCE dr, KHE_DRS_TASK_ON_DAY *dtd);
static bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,
  KHE_DRS_TASK_ON_DAY *dtd);

static bool KheDrsResourceOnDayIsFixed(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY *dtd)
{
  KHE_DRS_TASK_ON_DAY dtd1, dtd2;

  /* if drd has a closed assignment, it's fixed to that */
  if( drd->closed_asst != NULL )
    return *dtd = drd->closed_asst, true;

  /* if drd's resource is assigned to a task in soln which is still running, */
  /* then drd is fixed to that */
  if( KheDrsSolnResourceIsAssigned(soln, drd->encl_dr, &dtd1) &&
	KheDrsTaskBusyOnDay(dtd1->encl_dt, drd->day, &dtd2) )
    return *dtd = dtd2, true;

  /* otherwise drd has no fixed assignment */
  return *dtd = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,               */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of drd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,
  int verbosity, int indent, FILE *fp);
static char *KheDrsDayId(KHE_DRS_DAY day);

static void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "%s: ", KheDrsDayId(drd->day));
  if( drd->closed_asst != NULL )
    KheDrsTaskOnDayDebug(drd->closed_asst, 1, -1, fp);
  else
    fprintf(fp, "-");
  fprintf(fp, " (external_today %d)", HaArrayCount(drd->external_today));
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOpen(KHE_DRS_RESOURCE dr, int open_resource_index,    */
/*    KHE_DRS_PACKED_SOLN init_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)        */
/*                                                                           */
/*  Open dr.  This involves unassigning all the tasks assigned dr which      */
/*  lie in the selected day ranges of drs, and gathering for opening the     */
/*  expressions that depend on what dr is doing on those days, including     */
/*  on days where dr remains assigned.                                       */
/*                                                                           */
/*  Also, if init_soln != NULL, included the unassignments as assignments    */
/*  in init_soln, so that they can be undone later if required.              */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task);
static bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE si1, KHE_DRS_DAY_RANGE si2);
static KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last);
static void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,
  int indent, FILE *fp);
static void KheDrsPackedSolnSetTaskOnDay(KHE_DRS_PACKED_SOLN packed_soln,
  int open_day_index, int open_resource_index, KHE_DRS_TASK_ON_DAY dtd);

static void KheDrsResourceOpen(KHE_DRS_RESOURCE dr, int open_resource_index,
  KHE_DRS_PACKED_SOLN init_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY_RANGE ddr, open_day_range;  int i, j, k, open_day_index;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_TASK dt;  KHE_DRS_EXPR e;
  KHE_DRS_TASK_ON_DAY dtd;
  if( DEBUG11 )
  {
    fprintf(stderr, "[ KheDrsResourceOpen(%s, %d, drs)\n",
      KheResourceId(dr->resource), open_resource_index);
    KheDrsResourceDebug(dr, 2, 2, stderr);
  }
  dr->open_resource_index = open_resource_index;
  open_day_index = 0;
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      /* unassign any task assigned on drd, if it lies entirely in ddr */
      drd = HaArray(dr->days, j);
      dtd = drd->closed_asst;
      if( dtd != NULL )
      {
	dt = dtd->encl_dt;
	if( KheDrsDayRangeSubset(dt->encl_dtc->day_range, ddr) )
	{
	  KheDrsTaskUnAssign(dt, true);
	  if( init_soln != NULL )
            KheDrsPackedSolnSetTaskOnDay(init_soln, open_day_index,
	      open_resource_index, dtd);
	}
      }

      /* clear the internal_today array */
      HaArrayClear(drd->internal_today);

      /* gather for opening expressions depending on drd (unassigned or not) */
      open_day_range = KheDrsDayRangeMake(open_day_index, open_day_index);
      HaArrayForEach(drd->external_today, e, k)
      {
	e->open_day_range = open_day_range;
	KheDrsExprGatherForOpening(e, drs);
      }

      /* increase open_day_index */
      open_day_index++;
    }
  if( DEBUG11 )
  {
    fprintf(stderr, "  at end of KheDrsResourceOpen(%s, %d, drs):\n",
      KheResourceId(dr->resource), open_resource_index);
    KheDrsResourceDebug(dr, 2, 2, stderr);
    fprintf(stderr, "] KheDrsResourceOpen\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceClose(KHE_DRS_RESOURCE dr)                            */
/*                                                                           */
/*  Close dr.  There is nothing to do except clear dr->open_resource_index.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceClose(KHE_DRS_RESOURCE dr)
{
  dr->open_resource_index = -1;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLimit(KHE_MONITOR m)                                       */
/*                                                                           */
/*  Return the maximum limit of m if there is one, or else the minimum       */
/*  limit if there is one, or else 0.                                        */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLimit(KHE_MONITOR m)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;  int val;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      cbtc = KheClusterBusyTimesMonitorConstraint(
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m);
      val = KheClusterBusyTimesConstraintMaximum(cbtc);
      if( val < KheClusterBusyTimesConstraintTimeGroupCount(cbtc) )
	return val;
      val = KheClusterBusyTimesConstraintMinimum(cbtc);
      return val;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      lbtc = KheLimitBusyTimesMonitorConstraint(
	(KHE_LIMIT_BUSY_TIMES_MONITOR) m);
      val = KheLimitBusyTimesConstraintMaximum(lbtc);
      if( val < INT_MAX )
	return val;
      val = KheLimitBusyTimesConstraintMinimum(lbtc);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      /* could do better here */
      return 0;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      val = KheLimitActiveIntervalsMonitorMaximum(laim);
      if( val < KheLimitActiveIntervalsMonitorTimeGroupCount(laim) )
	return val;
      val = KheLimitActiveIntervalsMonitorMinimum(laim);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_AVOID_CLASHES_MONITOR_TAG:
    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:
    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:
    default:

      return 0;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)       */
/*                                                                           */
/*  Comparison function for sorting monitors so that those with the          */
/*  largest limits come first.                                               */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) p1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) p2;
  return KheMonitorLimit(m2) - KheMonitorLimit(m1);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new resource object with these attributes.                        */
/*  Return NULL if unsuccessful.                                             */
/*                                                                           */
/*****************************************************************************/
static void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m, KHE_DRS_RESOURCE dr);

static KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE res;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_DAY day;  /* KHE_MONITOR m; */

  /* initialize res, including one resource on day object for each day */
  if( DEBUG7 )
    fprintf(stderr, "[ KheDrsResourceMake(%s, drs)\n", KheResourceId(r));
  HaMake(res, drs->arena);
  res->resource = r;
  res->open_resource_index = -1;
  HaArrayInit(res->days, drs->arena);
  HaArrayForEach(drs->all_days, day, i)
  {
    drd = KheDrsResourceOnDayMake(res, day, drs);
    HaArrayAddLast(res->days, drd);
  }
  HaArrayInit(res->extend_signatures, drs->arena);

  /* add expressions based on the resource's monitors */
  /* *** moved to KheDrsResourceAddMonitors to fix sequencing problem
  KheSolnResourceMonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    if( DEBUG7 )
      fprintf(stderr, "  %s monitor %s\n", KheResourceId(r), KheMonitorId(m));
    KheDrsAddResourceMonitor(drs, m, res);
  }
  *** */
  if( DEBUG7 )
    fprintf(stderr, "] KheDrsResourceMake returning\n");
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddMonitors(KHE_DRS_RESOURCE dr,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add monitors that monitor resource dr.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddMonitors(KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_RESOURCE r;  int i;  KHE_MONITOR m;
  r = dr->resource;
  KheSolnResourceMonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    if( DEBUG7 )
      fprintf(stderr, "  %s monitor %s\n", KheResourceId(r), KheMonitorId(m));
    KheDrsAddResourceMonitor(drs, m, dr);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDay(KHE_DRS_RESOURCE dr,         */
/*    KHE_DRS_DAY day)                                                       */
/*                                                                           */
/*  Return the resource on day object representing dr on day.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDay(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY day)
{
  return HaArray(dr->days, day->frame_index);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayAtTime(KHE_DRS_RESOURCE dr,   */
/*    KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Return the resource on day object whose resource is dr and whose day     */
/*  is time's day.                                                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayAtTime(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int day_index;
  day_index = KheFrameTimeIndex(drs->days_frame, time);
  /* ***
  if( DEBUG4 )
    fprintf(stderr, "  in KheDrsResourceOnDayAtTime(dr %p %s, %s, drs)"
      " day_index = %d, dr->days = %d\n",
      (void *) dr, KheResourceId(dr->resource),
      KheTimeId(time), day_index, HaArrayCount(dr->days));
  *** */
  return HaArray(dr->days, day_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceSetSignatures(KHE_DRS_RESOURCE dr,                    */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Set the sig segs for dr on day when extending prev_soln.                 */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,
  KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt);
static void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc);
static KHE_DRS_TASK_ON_DAY KheDrsPackedSolnTaskOnDay(
  KHE_DRS_PACKED_SOLN packed_soln, KHE_DRS_DAY open_day,
  KHE_DRS_RESOURCE open_resource);
static bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si);
static void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp);

static void KheDrsResourceSetSignatures(KHE_DRS_RESOURCE dr,
  KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_SIGNATURE sig;  KHE_DRS_TASK dt;  KHE_DRS_TASK_ON_DAY dtd;
  KHE_DRS_TASK_CLASS dtc;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_COST moved_cost;

  /* delete and free any old signatures */
  HaArrayForEach(dr->extend_signatures, sig, i)
    if( sig != NULL )
      HaArrayAddLast(drs->free_signatures, sig);
  HaArrayClear(dr->extend_signatures);
  if( dr->extend_fixed_signature != NULL )
    HaArrayAddLast(drs->free_signatures, dr->extend_fixed_signature);

  drd = KheDrsResourceOnDay(dr, next_day);
  if( KheDrsResourceOnDayIsFixed(drd, prev_soln, &dtd) )
  {
    /* fixed assignment; set extend_fixed_dtd and extend_fixed_signature */
    dr->extend_fixed_dtd = dtd;
    dr->extend_fixed_signature = KheDrsSignatureBuild(drd, dtd, prev_soln, drs);

    /* move any unavoidable cost into drs->extend_cost */
    moved_cost = dr->extend_fixed_signature->cost;
    drs->extend_cost += moved_cost;
    dr->extend_fixed_signature->cost = 0;
  }
  else if( RERUN && drs->rerun != NULL )
  {
    /* this is a rerun, take sole dtd (possibly NULL) from drs->rerun */
    dr->extend_fixed_dtd = dtd =
      KheDrsPackedSolnTaskOnDay(drs->rerun, next_day, dr);
    dr->extend_fixed_signature = KheDrsSignatureBuild(drd, dtd, prev_soln, drs);

    /* move any unavoidable cost into drs->extend_cost */
    moved_cost = dr->extend_fixed_signature->cost;
    drs->extend_cost += moved_cost;
    dr->extend_fixed_signature->cost = 0;
  }
  else
  {
    /* ordinary assignment; build one signature for each task class */
    dr->extend_fixed_dtd = NULL;
    dr->extend_fixed_signature = NULL;
    HaArrayForEach(next_day->task_classes, dtc, i)
    {
      if( KheDrsTaskClassAcceptResourceBegin(dtc, dr, &dt) )
      {
	if( !KheDrsTaskBusyOnDay(dt, next_day, &dtd) )
	  HnAbort("KheDrsResourceSetSignatures internal error");
	sig = KheDrsSignatureBuild(drd, dtd, prev_soln, drs);
	KheDrsTaskClassAcceptResourceEnd(dtc);
      }
      else
	sig = NULL;
      HaArrayAddLast(dr->extend_signatures, sig);
    }

    /* plus one signature for non-assignment */
    sig = KheDrsSignatureBuild(drd, NULL, prev_soln, drs);
    HaArrayAddLast(dr->extend_signatures, sig);

    /* move any unavoidable cost into drs->extend_cost */
    moved_cost = KheCost(INT_MAX, INT_MAX);
    HaArrayForEach(dr->extend_signatures, sig, i)
      if( sig != NULL && sig->cost < moved_cost )
	moved_cost = sig->cost;
    if( moved_cost < KheCost(INT_MAX, INT_MAX) )
    {
      drs->extend_cost += moved_cost;
      HaArrayForEach(dr->extend_signatures, sig, i)
	if( sig != NULL )
	  sig->cost -= moved_cost;
    }
  }
  if( DEBUG23 )
  {
    KHE_DRS_EXPR e;  int si;
    fprintf(stderr, "  [ KheDrsResourceSetSignatures(%s, -, %s, drs), "
      "moved_cost %.5f\n", KheResourceId(dr->resource),
      KheDrsDayId(next_day), KheCostShow(moved_cost));
    HaArrayForEach(drd->internal_today, e, i)
      if( KheDrsExprInSignature(e, next_day, &si) )
      {
	fprintf(stderr, "    %d: ", si);
	KheDrsExprDoDebug(e, NULL, stderr);
	fprintf(stderr, "\n");
      }
    if( dr->extend_fixed_signature != NULL )
    {
      fprintf(stderr, "    extend_fixed_signature: ");
      KheDrsSignatureDebug(dr->extend_fixed_signature, 1, 0, stderr);
    }
    HaArrayForEach(dr->extend_signatures, sig, i)
      KheDrsSignatureDebug(sig, 1, 4, stderr);
    fprintf(stderr, "  ] KheDrsResourceSetSignatures\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,             */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dr onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,
  int indent, FILE *fp)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  int i;
  if( verbosity > 1 && indent >= 0 )
  {
    /* show timetable in full */
    fprintf(fp, "%*s[ %s", indent, "", KheResourceId(dr->resource));
    if( dr->open_resource_index >= 0 )
      fprintf(fp, " (open %d)", dr->open_resource_index);
    fprintf(fp, "\n");
    HaArrayForEach(dr->days, drd, i)
      KheDrsResourceOnDayDebug(drd, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "%s", KheResourceId(dr->resource));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY_RANGE"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last)                */
/*                                                                           */
/*  Return a new interval with these endpoints.                              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last)
{
  KHE_DRS_DAY_RANGE res;
  res.first = first;
  res.last = last;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeAdd(KHE_DRS_DAY_RANGE *ddr, int index)                */
/*                                                                           */
/*  Update *ddr to include index.                                            */
/*                                                                           */
/*  Implementation note.  Both cases are needed here, as proved by a trace   */
/*  of Add([1, 0], 4).  The second case wrongly produces [1, 4].             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeAdd(KHE_DRS_DAY_RANGE *ddr, int index)
{
  if( ddr->first > ddr->last )
  {
    /* ddr is empty, so make it cover just index */
    ddr->first = ddr->last = index;
  }
  else
  {
    /* ddr is non-empty, so expand it to include index */
    if( index < ddr->first )
      ddr->first = index;
    if( index > ddr->last )
      ddr->last = index;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,            */
/*    KHE_DRS_DAY_RANGE ddr)                                                 */
/*                                                                           */
/*  Merge ddr into *day_ranges, a sorted array of day ranges.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  KHE_DRS_DAY_RANGE ddr)
{
  KHE_DRS_DAY_RANGE ddr2;  int i;

  /* if ddr is empty, it changes nothing (this can't happen, actually) */
  if( ddr.first > ddr.last )
    return;

  HaArrayForEach(*day_ranges, ddr2, i)
  {
    /***********************************************************************/
    /*                                                                     */
    /*  Invariant: day_ranges and ddr may have changed, but it remains     */
    /*  true that merging ddr with day_ranges gives the desired result.    */
    /*  Also, day_ranges[0..i-1] precede and do not interact with ddr.     */
    /*                                                                     */
    /***********************************************************************/

    if( ddr2.last + 1 < ddr.first )
    {
      /* ddr2 precedes ddr and does not interact with it; continue */
      continue;
    }

    if( ddr.last + 1 < ddr2.first )
    {
      /* ddr precedes ddr2 and does not interact with it, or with day_ranges */
      HaArrayAdd(*day_ranges, i, ddr);
      return;
    }

    /* ddr2 interacts with ddr, so expand ddr to include ddr2, then */
    /* delete ddr2 and continue */
    if( ddr2.first < ddr.first )
      ddr.first = ddr2.first;
    if( ddr2.last > ddr.last )
      ddr.last = ddr2.last;
    HaArrayDeleteAndShift(*day_ranges, i);
    i--;
  }

  /* ddr does not interact with any day_ranges, and it follows them all */
  HaArrayAddLast(*day_ranges, ddr);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeContains(KHE_DRS_DAY_RANGE ddr, int i)                */
/*                                                                           */
/*  Return true if ddr contains i.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsDayRangeContains(KHE_DRS_DAY_RANGE ddr, int i)
{
  return ddr.first <= i && i <= ddr.last;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE ddr1, KHE_DRS_DAY_RANGE ddr2)*/
/*                                                                           */
/*  Return true if ddr1 is a subset of ddr2.                                 */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE ddr1, KHE_DRS_DAY_RANGE ddr2)
{
  return ddr1.first >= ddr2.first && ddr1.last <= ddr2.last;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeDebug(KHE_DRS_DAY_RANGE ddr,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of ddr onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeDebug(KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_TIME_GROUP tg1, tg2;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  tg1 = KheFrameTimeGroup(drs->days_frame, ddr.first);
  tg2 = KheFrameTimeGroup(drs->days_frame, ddr.last);
  if( tg1 == tg2 )
    fprintf(fp, "%s", KheTimeGroupId(tg1));
  else
    fprintf(fp, "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,           */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of *day_ranges onto fp with the given verbosity and indent.  */
/*                                                                           */
/*****************************************************************************/

/* *** good but currently unused
static void KheDrsDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY_RANGE ddr;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[");
  HaArrayForEach(*day_ranges, ddr, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    KheDrsDayRangeDebug(ddr, drs, 1, -1, fp);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY KheDrsDayMake(int frame_index,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new day object with the given attributes.  Its task classes,      */
/*  expressions, and dominance tests will be added later.                    */
/*                                                                           */
/*  Here frame_index is the index of the new day in the common frame.        */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DAY KheDrsDayMake(int frame_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY res;
  HaMake(res, drs->arena);
  res->frame_index = frame_index;
  res->open_day_index = -1;  /* signals not currently open for solving */
  res->time_group = KheFrameTimeGroup(drs->days_frame, frame_index);
  HaArrayInit(res->task_classes, drs->arena);
  /* HaArrayInit(res->internal_today, drs->arena); */
  HaArrayInit(res->nr_internal_today, drs->arena);
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  res->soln_set = NULL;  /* initialized at the start of each solve */
  /* res->single_day_curr_cost = res->single_day_min_cost = 0; */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsDayId(KHE_DRS_DAY day)                                       */
/*                                                                           */
/*  Return the Id of day.  This is just the Id of its time group.            */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsDayId(KHE_DRS_DAY day)
{
  return KheTimeGroupId(day->time_group);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayOpen(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,               */
/*    int open_day_index, KHE_DRS_DOM_KIND dom_kind,                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Open day for solving.  It lies within day range ddr, its index in the    */
/*  the sequence of all open days is open_day_index, and dom_kind is the     */
/*  kind of dominance testing to use on this day.                            */
/*                                                                           */
/*  NB We open the day's task classes separately, because they may contain   */
/*  tasks for later days and opening assumes that those days are open.       */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
static KHE_DRS_SOLN_SET KheDrsSolnSetMake(KHE_DRS_DOM_KIND main_dom_kind,
  bool use_cache, KHE_DRS_DOM_KIND cache_dom_kind,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsDayOpen(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,
  int open_day_index, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,
  KHE_DRS_DOM_KIND cache_dom_kind, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* KHE_DRS_TASK_CLASS dtc;  int i; */
  if( DEBUG18 )
    fprintf(stderr, "[ KheDrsDayOpen(%s, %d-%d, %d)\n", KheDrsDayId(day),
      ddr.first, ddr.last, open_day_index);
  day->open_day_index = open_day_index;
  /* *** doing this separately now
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassOpen(dtc, ddr, drs);
  *** */
  /* HaArrayClear(day->internal_today); */
  HaArrayClear(day->nr_internal_today);
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  day->soln_set = KheDrsSolnSetMake(main_dom_kind, use_cache,
    cache_dom_kind, drs);
  if( DEBUG18 )
    fprintf(stderr, "] KheDrsDayOpen\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayOpenTaskClasses(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Open the task classes of day.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayOpenTaskClasses(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  if( DEBUG18 )
    fprintf(stderr, "[ KheDrsDayOpenTaskClasses(%s, %d-%d)\n",
      KheDrsDayId(day), ddr.first, ddr.last);
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassOpen(dtc, ddr, drs);
  if( DEBUG18 )
    fprintf(stderr, "] KheDrsDayOpenTaskClasses\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e)               */
/*                                                                           */
/*  Add open expression e to day, and also to a resource on day if needed.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e)
{
  /* HaArrayAddLast(day->internal_today, e); */
  HnAssert(e->resource == NULL, "KheDrsDayAddOpenExpr internal error");
  HaArrayAddLast(day->nr_internal_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)      */
/*                                                                           */
/*  Add dom_test to day and return its index in day->dom_tests.              */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsDomTestIsStrictEquality(KHE_DRS_DOM_TEST dom_test);

static int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)
{
  int res;

  /* add dom_test to day */
  res = HaArrayCount(day->dom_tests);
  /* ***
  HnAssert(dom_test.type != KHE_DRS_DOM_UNUSED,
    "KheDrsDayAddDomTest internal error");
  *** */
  HaArrayAddLast(day->dom_tests, dom_test);

  /* if dom_test is equality, record it in eq_dom_test_indexes */
  /* ***
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
  *** */
  if( KheDrsDomTestIsStrictEquality(dom_test) )
    HaArrayAddLast(day->eq_dom_test_indexes, res);

  /* return dom_test's index */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClose(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Close day.                                                               */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc);
static void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsDayClose(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  day->open_day_index = -1;  /* signals no longer open */
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassClose(dtc);
  /* HaArrayClear(day->internal_today); */
  HaArrayClear(day->nr_internal_today);
  HaArrayAppend(drs->free_dom_tests, day->dom_tests, i);
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  KheDrsSolnSetFree(day->soln_set, drs);
  day->soln_set = NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDaySingleCostBegin(KHE_DRS_DAY day)                           */
/*                                                                           */
/*  Begin keeping track of the single cost on day.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDaySingleCostBegin(KHE_DRS_DAY day)
{
  if( DEBUG22 )
    day->single_day_min_cost = KheCost(INT_MAX, INT_MAX);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDaySingleCostEnd(KHE_DRS_DAY day)                             */
/*                                                                           */
/*  End keeping track of the single cost on day.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDaySingleCostEnd(KHE_DRS_DAY day)
{
  if( DEBUG22 )
  {
    if( day->single_day_min_cost < KheCost(INT_MAX, INT_MAX) )
      fprintf(stderr, "  at end of day %s, single_day_min_cost = %.5f\n",
	KheTimeGroupId(day->time_group), KheCostShow(day->single_day_min_cost));
    else
      fprintf(stderr, "  at end of day %s, single_day_min_cost = inf\n",
	KheTimeGroupId(day->time_group));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDaySingleCostSolnBegin(KHE_DRS_DAY day)                       */
/*                                                                           */
/*  Begin recording the single cost of one solution on day.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDaySingleCostSolnBegin(KHE_DRS_DAY day)
{
  if( DEBUG22 )
    day->single_day_curr_cost = 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDaySingleCostSolnEnd(KHE_DRS_DAY day)                         */
/*                                                                           */
/*  End recording the single cost of one solution on day.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDaySingleCostSolnEnd(KHE_DRS_DAY day)
{
  if( DEBUG22 )
  {
    if( day->single_day_curr_cost < day->single_day_min_cost )
      day->single_day_min_cost = day->single_day_curr_cost;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDaySingleCostAddCost(KHE_DRS_DAY day, KHE_COST cost)          */
/*                                                                           */
/*  Add cost to the current single cost on day.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDaySingleCostAddCost(KHE_DRS_DAY day, KHE_COST cost)
{
  if( DEBUG22 )
    day->single_day_curr_cost += cost;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,              */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of the dom tests of day.                                     */
/*                                                                           */
/*****************************************************************************/
static char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100]);
/* ***
static KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type,
  bool tradeoff, KHE_COST tradeoff_cost);
*** */

/* *** can't conveniently do this any more
static void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,
  int indent, FILE *fp)
{
  int dom_test_freq[KHE_DRS_DOM_EQ_LOOSE + 1];  int i;
  KHE_DRS_DOM_TEST dom_test;  char buff[100];

  ** initialize dom_test_freq **
  for( i = KHE_DRS_DOM_UNUSED;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
    dom_test_freq[i] = 0;

  ** accumulate dom_test_freq **
  HaArrayForEach(day->dom_tests, dom_test, i)
    dom_test_freq[dom_test.type]++;

  ** display frequencies **
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ Day %s dom tests: ", KheDrsDayId(day));
  for( i = 0;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
  {
    if( i > 0 )
      fprintf(fp, ", ");
    fprintf(fp, "%d %s", dom_test_freq[i],
      KheDrsDomTestShow(KheDrsDomTestMake(i, false, 0), buff));
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of day onto fp.                                              */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);
static void KheDrsSolnSetDebug(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);

static void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Day %s (frame %d, ", indent, "",
      KheDrsDayId(day), day->frame_index);
    if( day->open_day_index >= 0 )
      fprintf(fp, "open %d)\n", day->open_day_index);
    else
      fprintf(fp, "closed)\n");
    if( verbosity >= 2 )
    {
      HaArrayForEach(day->task_classes, dtc, i)
	KheDrsTaskClassDebug(dtc, drs, verbosity, indent + 2, fp);
      /* ***
      if( day->open_day_index >= 0 )
	KheDrsDayDomTestsDebug(day, verbosity, indent + 2, fp);
      *** */
    }
    if( verbosity >= 3 && day->soln_set != NULL )
      KheDrsSolnSetDebug(day->soln_set, drs, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "%s", KheDrsDayId(day));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDebugSignatureKey(KHE_DRS_DAY day,                         */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of a key to the signature on open day day.                   */
/*                                                                           */
/*****************************************************************************/
/* ***
static bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si);
*** */
static void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp);

/* *** perfectly good, but currently unused
static void KheDrsDayDebugSignatureKey(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_EXPR e;  int i, si;  bool print;
  KHE_DRS_EXPR_INT_SUM_COST e1;
  KHE_DRS_EXPR_INT_SEQ_COST e2;
  HnAssert(day->open_day_index >= 0,
    "KheDrsDayDebugSignatureKey internal error");
  if( verbosity == 1 )
  {
    ** print only the signature of the expression we are tracking **
    HaArrayForEach(day->internal_today, e, i)
      if( KheDrsExprInSignature(e, day, &si) )
      {
	switch( e->tag )
	{
	  case KHE_DRS_EXPR_INT_SUM_COST_TAG:

	    e1 = (KHE_DRS_EXPR_INT_SUM_COST) e;
	    print = (strcmp(KheMonitorId(e1->monitor), RERUN_MONITOR_ID)==0);
	    break;

	  case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

	    e2 = (KHE_DRS_EXPR_INT_SEQ_COST) e;
	    print = (strcmp(KheMonitorId(e2->monitor), RERUN_MONITOR_ID)==0);
	    break;

	  default:

	    print = false;
	    break;
	}

	** e has an entry in the signature for this day **
	if( print )
	{
	  fprintf(fp, "%*s  %2d: ", indent, "", si);
	  KheDrsExprDebug(e, drs, 1, -1, fp);
	  fprintf(fp, "\n");
	}
      }
  }
  else
  {
    ** print the whole thing **
    fprintf(fp, "%*s[ Signature of %s:\n", indent, "", KheDrsDayId(day));
    HaArrayForEach(day->internal_today, e, i)
      if( KheDrsExprInSignature(e, day, &si) )
      {
	** e has an entry in the signature for this day **
	fprintf(fp, "%*s  %2d: ", indent, "", si);
	KheDrsExprDebug(e, drs, 1, -1, fp);
	fprintf(fp, "\n");
      }
    fprintf(fp, "%*s]\n", indent, "");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_ON_DAY"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_ON_DAY KheDrsTaskOnDayMake(KHE_DRS_TASK encl_dt,            */
/*    KHE_DRS_DAY day, KHE_TASK task, KHE_TIME time,                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new task on day object with these attributes.                     */
/*  This function assumes that the task is not assigned anything.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK_ON_DAY KheDrsTaskOnDayMake(KHE_DRS_TASK encl_dt,
  KHE_DRS_DAY day, KHE_TASK task, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY res;
  HaMake(res, drs->arena);
  res->encl_dt = encl_dt;
  res->day = day;
  res->task = task;
  res->time = time;
  res->closed_asst = NULL;
  HaArrayInit(res->external_today, drs->arena);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskOnDayCmp(const void *t1, const void *t2)                   */
/*                                                                           */
/*  Comparison function for sorting an array of task on day objects by       */
/*  increasing day index.                                                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskOnDayCmp(const void *t1, const void *t2)
{
  KHE_DRS_TASK_ON_DAY dtd1, dtd2;
  dtd1 = * (KHE_DRS_TASK_ON_DAY *) t1;
  dtd2 = * (KHE_DRS_TASK_ON_DAY *) t2;
  return dtd1->day->frame_index - dtd2->day->frame_index;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsOptionalResourceGroupEqual(KHE_RESOURCE_GROUP rg1,            */
/*    KHE_RESOURCE_GROUP rg2)                                                */
/*                                                                           */
/*  Return true if rg1 and rg2 are both NULL, or both non-NULL and equal.    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsOptionalResourceGroupEqual(KHE_RESOURCE_GROUP rg1,
  KHE_RESOURCE_GROUP rg2)
{
  if( rg1 == NULL )
    return rg2 == NULL;
  else
    return rg2 != NULL && KheResourceGroupEqual(rg1, rg2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,         */
/*    KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res)                */
/*                                                                           */
/*  If dtd has an assigned task expression with resource group attribute     */
/*  rg, return true with *res set to that expression; else return false.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res)
{
  KHE_DRS_EXPR e;  int i;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  HaArrayForEach(dtd->external_today, e, i)
  {
    eat = (KHE_DRS_EXPR_ASSIGNED_TASK) e;
    if( KheDrsOptionalResourceGroupEqual(rg, eat->resource_group) )
      return *res = eat, true;
  }
  return *res = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,         */
/*    KHE_DRS_EXPR_ASSIGNED_TASK eat)                                        */
/*                                                                           */
/*  Add eat (assumed new) to dtd.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_EXPR_ASSIGNED_TASK eat)
{
  HaArrayAddLast(dtd->external_today, (KHE_DRS_EXPR) eat);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskOnDayIsLast(KHE_DRS_TASK_ON_DAY dtd)                      */
/*                                                                           */
/*  Return true if dtd is the last task on day of its task.                  */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskOnDayIsLast(KHE_DRS_TASK_ON_DAY dtd)
{
  return dtd == HaArrayLast(dtd->encl_dt->days);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,                       */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of dtd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( dtd->task != NULL )
  {
    KheTaskDebug(dtd->task, 1, -1, fp);
    fprintf(fp, "(%s:%s:%s)", KheTimeId(dtd->time), KheDrsDayId(dtd->day),
      dtd->closed_asst == NULL ? "-" :
      KheResourceId(dtd->closed_asst->encl_dr->resource));
  }
  else
  {
    fprintf(fp, "-(%s)", KheDrsDayId(dtd->day));
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskAssign(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr, bool task)   */
/*                                                                           */
/*  Assign dr (which must be non-NULL) to dt, by setting the close_asst      */
/*  fields in dt, dt->days, and dr.  If task is true, also assign dt's KHE   */
/*  task to dr's KHE resource, aborting if that does not succeed.            */
/*                                                                           */
/*  The enclosing task class has a list of unassigned tasks; but do not      */
/*  remove dt from it, because we are being lazy:  the unassigned tasks      */
/*  will be sorted out later, by KheDrsTaskClassOrganizeUnassignedTasks.     */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp);

static void KheDrsTaskAssign(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr, bool task)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  if( DEBUG15 )
    fprintf(stderr, "[ KheDrsTaskAssign(%s (%d days), %s, %s)\n",
      KheTaskId(dt->task), HaArrayCount(dt->days),
      KheResourceId(dr->resource), bool_show(task));
  HnAssert(dt->closed_asst == NULL, "KheDrsTaskAssign internal error 1");
  HnAssert(dr != NULL, "KheDrsTaskAssign internal error 2");
  if( task && !KheTaskAssignResource(dt->task, dr->resource) )
    HnAbort("KheDrsTaskAssign internal error 3 (cannot assign %s to %s)",
      KheResourceId(dr->resource), KheTaskId(dt->task));
  dt->closed_asst = dr;
  HaArrayForEach(dt->days, dtd, i)
  {
    drd = KheDrsResourceOnDay(dr, dtd->day);
    HnAssert(dtd->closed_asst == NULL, "KheDrsTaskAssign internal error 4");
    HnAssert(drd->closed_asst == NULL, "KheDrsTaskAssign internal error 5");
    dtd->closed_asst = drd;
    drd->closed_asst = dtd;
  }
  if( DEBUG15 )
    fprintf(stderr, "] KheDrsTaskAssign returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task)                      */
/*                                                                           */
/*  Unassign dt (which must be assigned some drs resource dr), by clearing   */
/*  the closed_asst fields in dt, dt->days, and dr.  If task is true, also   */
/*  unassign dt's KHE task, or do nothing if the task will not unassign.     */
/*                                                                           */
/*  The enclosing task class has a list of unassigned tasks; add dt to it.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;
  dr = dt->closed_asst;
  HnAssert(dr != NULL, "KheDrsTaskUnAssign internal error 1");
  if( DEBUG15 )
    fprintf(stderr, "[ KheDrsTaskUnAssign(%s, %s, %s)\n", KheTaskId(dt->task),
      KheResourceId(dr->resource), bool_show(task));
  if( task )
  {
    HnAssert(KheTaskAsstResource(dt->task) == dr->resource,
      "KheDrsTaskUnAssign internal error 2");
    if( !KheTaskUnAssignResource(dt->task) )
    {
      if( DEBUG16 )
	fprintf(stderr, "  failed to unassign %s\n", KheTaskId(dt->task));
      return;
    }
    if( DEBUG9 )
      fprintf(stderr, "  after unassigning %s, cost = %.5f\n",
	KheTaskId(dt->task), KheCostShow(KheSolnCost(KheTaskSoln(dt->task))));
  }
  HaArrayForEach(dt->days, dtd, i)
  {
    drd = KheDrsResourceOnDay(dr, dtd->day);
    HnAssert(dtd->closed_asst == drd, "KheDrsTaskUnAssign internal error 3");
    HnAssert(drd->closed_asst == dtd, "KheDrsTaskUnAssign internal error 4");
    dtd->closed_asst = NULL;
    drd->closed_asst = NULL;
  }
  dt->closed_asst = NULL;
  HaArrayAddLast(dt->encl_dtc->unassigned_tasks, dt);
  if( DEBUG15 )
    fprintf(stderr, "] KheDrsTaskUnAssign returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskAddDays(KHE_TASK task, KHE_DRS_TASK dt,                      */
/*    int first_solve_index, int last_solve_index,                           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add the days covered by task and its descendants to dt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheTaskAddDays(KHE_TASK task, KHE_DRS_TASK dt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_TASK child_task;  int i, durn, day_index;  KHE_DRS_DAY day;
  KHE_MEET meet;  KHE_TIME start_time, time;  KHE_DRS_TASK_ON_DAY dtd;

  /* do it for task */
  meet = KheTaskMeet(task);
  HnAssert(meet != NULL, "KheTaskAddDays internal error (no meet)");
  start_time = KheMeetAsstTime(meet);
  HnAssert(start_time != NULL,  "KheTaskAddDays internal error (no time)");
  durn = KheMeetDuration(meet);
  for( i = 0;  i < durn;  i++ )
  {
    time = KheTimeNeighbour(start_time, i);
    day_index = KheFrameTimeIndex(drs->days_frame, time);
    day = HaArray(drs->all_days, day_index);
    dtd = KheDrsTaskOnDayMake(dt, day, task, time, drs);
    HaArrayAddLast(dt->days, dtd);
  }

  /* do it for task's descendants */
  for( i = 0;  i < KheTaskAssignedToCount(task);  i++ )
  {
    child_task = KheTaskAssignedTo(task, i);
    KheTaskAddDays(child_task, dt, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOrganizeDays(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Assuming that dt->days are sorted by increasing day index, organize      */
/*  them so that there is one per day, by inserting NULL entries where       */
/*  there are gaps.  There can be no cases of two on one day, thanks to an   */
/*  earlier call to KheTaskClassNoOverlap.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOrganizeDays(KHE_DRS_TASK dt)
{
  int init_len, reqd_len, i, final_pos;
  KHE_DRS_TASK_ON_DAY first_dtd, last_dtd, dtd;

  /* quit early if the number of days is fine as is */
  first_dtd = HaArrayFirst(dt->days);
  last_dtd = HaArrayLast(dt->days);
  reqd_len = last_dtd->day->frame_index - first_dtd->day->frame_index + 1;
  init_len = HaArrayCount(dt->days);
  if( reqd_len == init_len )
    return;

  /* add the required number of NULL entries, initially at the array's end */
  HaArrayFill(dt->days, reqd_len, NULL);

  /* move the entries to their final positions, working right to left */
  for( i = init_len - 1;  i >= 0;  i-- )
  {
    dtd = HaArray(dt->days, i);
    final_pos = dtd->day->frame_index - first_dtd->day->frame_index;
    if( final_pos != i )
    {
      HaArrayPut(dt->days, final_pos, dtd);
      HaArrayPut(dt->days, i, NULL);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assuming that r has the correct resource type, return its corresponding  */
/*  drs resource.                                                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(KheResourceResourceTypeIndex(r) < HaArrayCount(drs->all_resources),
    "KheDrsResource internal error");
  return HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskMake(KHE_DRS_TASK_CLASS encl_dtc,                 */
/*    int index_in_encl_dtc, KHE_TASK task, KHE_COST asst_cost,              */
/*    KHE_COST non_asst_cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Make a new drs task with these attributes, add it to drs->all_root_tasks */
/*  at the same index as it has in the solution, and return it.              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheDrsTaskMake(KHE_DRS_TASK_CLASS encl_dtc,
  int index_in_encl_dtc, KHE_TASK task, KHE_COST asst_cost,
  KHE_COST non_asst_cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;  int task_index;  KHE_RESOURCE r;

  /* get the basic object and initialize its fields */
  HnAssert(KheTaskIsProperRoot(task), "KheDrsTaskMake internal error");
  HaMake(res, drs->arena);
  res->encl_dtc = encl_dtc;
  res->index_in_encl_dtc = index_in_encl_dtc;
  res->extend_must_assign = false;
  res->open = false;
  res->task = task;
  res->asst_cost = asst_cost;
  res->non_asst_cost = non_asst_cost;
  res->closed_asst = NULL;
  HaArrayInit(res->days, drs->arena);

  /* add, sort, and organize the task on day objects */
  KheTaskAddDays(task, res, drs);
  HaArraySort(res->days, &KheDrsTaskOnDayCmp);
  KheDrsTaskOrganizeDays(res);

  /* add res to drs->all_root_tasks */
  task_index = KheTaskSolnIndex(task);
  HaArrayFill(drs->all_root_tasks, task_index + 1, NULL);
  HnAssert(HaArray(drs->all_root_tasks, task_index) == NULL,
    "KheDrsTaskMake internal error");
  HaArrayPut(drs->all_root_tasks, task_index, res);

  /* if task is assigned, make that assignment in res */
  r = KheTaskAsstResource(task);
  if( r != NULL )
    KheDrsTaskAssign(res, KheDrsResource(r, drs), false);

  /* and also return res */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskIndexInEnclCmp(const void *t1, const void *t2)             */
/*                                                                           */
/*  Comparison function for sorting an array of drs tasks into increasing    */
/*  order of their index in their enclosing task class.                      */
/*                                                                           */
/*  In addition, it moves any assigned tasks to the end.                     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskIndexInEnclCmp(const void *t1, const void *t2)
{
  KHE_DRS_TASK dt1 = * (KHE_DRS_TASK *) t1;
  KHE_DRS_TASK dt2 = * (KHE_DRS_TASK *) t2;
  if( dt1->closed_asst != NULL )
  {
    if( dt2->closed_asst != NULL )
    {
      /* dt1 is assigned, dt2 is assigned */
      return 0;
    }
    else
    {
      /* dt1 is assigned, dt2 is not assigned */
      return 1;
    }
  }
  else
  {
    if( dt2->closed_asst != NULL )
    {
      /* dt1 is not assigned, dt2 is assigned */
      return -1;
    }
    else
    {
      /* dt1 is not assigned, dt2 is not assigned */
      return dt1->index_in_encl_dtc - dt2->index_in_encl_dtc;
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOpen(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Open task for solving.  We'll have checked previously that its days      */
/*  all lie in the current open interval, and it must currently be closed    */
/*  but unassigned; so opening dt basically just gathers for opening all     */
/*  expressions that depend on dt's assignment.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOpen(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR e;  int i, j, di;
  KHE_DRS_DAY_RANGE open_day_range;

  /* open dt */
  if( DEBUG19 )
    fprintf(stderr, "[ KheDrsTaskOpen(%s, drs)\n", KheTaskId(dt->task));
  HnAssert(!dt->open, "KheDrsTaskOpen internal error 1");
  HnAssert(dt->closed_asst == NULL, "KheDrsTaskOpen internal error 2");
  dt->open = true;

  /* gather external expressions for opening */
  HaArrayForEach(dt->days, dtd, i)
  {
    di = dtd->day->open_day_index;
    HnAssert(di >= 0, "KheDrsTaskOpen internal error 3");
    open_day_range = KheDrsDayRangeMake(di, di);
    HaArrayForEach(dtd->external_today, e, j)
    {
      e->open_day_range = open_day_range;
      KheDrsExprGatherForOpening(e, drs);
    }
  }
  if( DEBUG19 )
    fprintf(stderr, "] KheDrsTaskOpen\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)               */
/*                                                                           */
/*  If dt is open, close it, with a call to KheDrsTaskAssign if dr != NULL.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)
{
  if( dt->open )
  {
    if( DEBUG19 )
      fprintf(stderr, "[ KheDrsTaskClose(%s, drs)\n", KheTaskId(dt->task));
    dt->open = false;
    if( dr != NULL )
      KheDrsTaskAssign(dt, dr, true);
    if( DEBUG19 )
      fprintf(stderr, "] KheDrsTaskClose\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,               */
/*    KHE_DRS_TASK_ON_DAY *dtd)                                              */
/*                                                                           */
/*  If dt is running on day, return true with *dtd set to its task on day    */
/*  object for that day.  Otherwise return false with *dtd set to NULL.      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,
  KHE_DRS_TASK_ON_DAY *dtd)
{
  KHE_DRS_TASK_ON_DAY dtd2;  int i;
  HaArrayForEach(dt->days, dtd2, i)
    if( dtd2->day == day )
      return *dtd = dtd2, true;
  return *dtd = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,                     */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dt onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  KheTaskDebug(dt->task, 1, -1, fp);
  fprintf(fp, " asst_cost %.5f, non_asst_cost %.5f ",
    KheCostShow(dt->asst_cost), KheCostShow(dt->non_asst_cost));
  fprintf(fp, "(%d, %s, %s; ", dt->index_in_encl_dtc,
    dt->open ? "open" : "closed",
    dt->closed_asst == NULL ? "-" : KheResourceId(dt->closed_asst->resource));
  HaArrayForEach(dt->days, dtd, i)
    KheDrsTaskOnDayDebug(dtd, verbosity, -1, fp);
  fprintf(fp, ")");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_CLASS"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassMake(KHE_TASK_CLASS tc,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task class based on tc and add it to its first day.       */
/*  Also make drs tasks corresponding to tc's tasks and add them to the      */
/*  drs class task and to drs->all_root_tasks.                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassMake(KHE_TASK_CLASS tc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS res;  int i;  KHE_TASK task;  KHE_DRS_TASK dt;
  KHE_DRS_DAY day;  KHE_COST asst_cost, non_asst_cost;

  /* get or make a drs task class object */
  HaMake(res, drs->arena);
  res->orig_task_class = tc;
  res->day_range = KheDrsDayRangeMake(KheTaskClassFirstDayIndex(tc),
    KheTaskClassLastDayIndex(tc));
  HaArrayInit(res->all_tasks, drs->arena);
  HaArrayInit(res->unassigned_tasks, drs->arena);
  res->extend_used = 0;

  /* make drs tasks for its khe tasks and add them */
  for( i = 0;  i < KheTaskClassTaskCount(tc);  i++ )
  {
    task = KheTaskClassTask(tc, i, &asst_cost, &non_asst_cost);
    dt = KheDrsTaskMake(res, i, task, asst_cost, non_asst_cost, drs);
    HaArrayAddLast(res->all_tasks, dt);
    if( dt->closed_asst == NULL )
      HaArrayAddLast(res->unassigned_tasks, dt);
  }

  /* make res one of the task classes of its first day */
  day = HaArray(drs->all_days, res->day_range.first);
  HaArrayAddLast(day->task_classes, res);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassOrganizeUnassignedTasks(KHE_DRS_TASK_CLASS dtc)      */
/*                                                                           */
/*  Organize the unassigned tasks of dtc for solving.                        */
/*                                                                           */
/*  When this function is called, unassigned_tasks may be out of order,      */
/*  and, because unassigned_tasks is handled lazily, some of them may in     */
/*  fact be assigned.  This function fixes both of those problems.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassOrganizeUnassignedTasks(KHE_DRS_TASK_CLASS dtc)
{
  KHE_DRS_TASK dt;

  /* sort unassigned_tasks into order with assigned tasks at the end */
  HaArraySortUnique(dtc->unassigned_tasks, &KheDrsTaskIndexInEnclCmp);

  /* remove assigned tasks from the end */
  while( HaArrayCount(dtc->unassigned_tasks) > 0 )
  {
    dt = HaArrayLast(dtc->unassigned_tasks);
    if( dt->closed_asst == NULL )
      break;
    HaArrayDeleteLast(dtc->unassigned_tasks);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  If dtc is suitable (if its day range lies within ddr, and its resource   */
/*  domain includes at least one open resource), then open it.  Opening dtc  */
/*  basically means organizing its unassigned tasks into decreasing          */
/*  desirability order and opening them.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  int i;
  if( DEBUG19 )
    fprintf(stderr, "[ KheDrsTaskClassOpen(dtc, %d-%d)\n", ddr.first, ddr.last);
  if( KheDrsDayRangeSubset(dtc->day_range, ddr) &&
      !KheResourceSetDisjointGroup(drs->selected_resource_set,
	KheTaskClassDomain(dtc->orig_task_class)) )
  {
    /* dtc can open; organize and open unassigned_tasks; say none are used */
    KheDrsTaskClassOrganizeUnassignedTasks(dtc);
    HaArrayForEach(dtc->unassigned_tasks, dt, i)
      KheDrsTaskOpen(dt, drs);
    dtc->extend_used = 0;
  }
  else
  {
    /* dtc can't open; set dtc->extend_used to make that clear */
    dtc->extend_used = INT_MAX;
  }
  if( DEBUG19 )
    fprintf(stderr, "] KheDrsTaskClassOpen\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Close dtc.  Some of its tasks may be already closed; that's OK.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc)
{
  KHE_DRS_TASK dt;  int i;
  HaArrayForEach(dtc->unassigned_tasks, dt, i)
    KheDrsTaskClose(dt, NULL);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,          */
/*    KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt)                                 */
/*                                                                           */
/*  If dtc accepts dr (because dr lies in its tasks' domains, and not all    */
/*  the tasks are assigned), then return true and set *dt to one of dtc's    */
/*  tasks that dr can be assigned to.  Otherwise return false.               */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,
  KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt)
{
  if( dtc->extend_used < HaArrayCount(dtc->unassigned_tasks) &&
      KheResourceGroupContains(KheTaskClassDomain(dtc->orig_task_class),
	dr->resource) )
  {
    *dt = HaArray(dtc->unassigned_tasks, dtc->extend_used);
    dtc->extend_used++;
    return true;
  }
  else
    return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc)            */
/*                                                                           */
/*  Undo the effect of a corresponding KheDrsTaskClassAcceptResourceBegin    */
/*  which returned true.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc)
{
  dtc->extend_used--;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc, int verbosity,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of dtc onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK dt;  int i, pos;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Task class ", indent, "");
    KheDrsDayRangeDebug(dtc->day_range, drs, 1, 0, fp);
    HaArrayForEach(dtc->all_tasks, dt, i)
    {
      fprintf(fp, "%*s  %s ", indent, "",
	HaArrayContains(dtc->unassigned_tasks, dt, &pos) ?
	"unassigned" : "assigned");
      KheDrsTaskDebug(dt, 1, 0, fp);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    fprintf(fp, "TaskClass(");
    KheDrsDayRangeDebug(dtc->day_range, drs, 1, -1, fp);
    fprintf(fp, ", %d tasks)", HaArrayCount(dtc->all_tasks));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DOM_KIND"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  char *KheDomKindShow(KHE_DRS_DOM_KIND dom_kind)                          */
/*                                                                           */
/*  Return a suitable label for dom_kind.                                    */
/*                                                                           */
/*****************************************************************************/

static char *KheDomKindShow(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:			return "None";
    case KHE_DRS_DOM_WEAK:			return "Weak";
    case KHE_DRS_DOM_MEDIUM:			return "Medium";
    case KHE_DRS_DOM_STRONG:			return "Strong";
    case KHE_DRS_DOM_TRIE:			return "Trie";
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:	return "StrongWithTradeoff";
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:	return "TrieWithTradeoff";
    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:	return "IndexedWithTradeoff";

    default:

      HnAbort("KheDomKindShow: unknown dom_kind (%d)", dom_kind);
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DOM_TEST"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMake(bool allow_zero, int min_limit,       */
/*    int max_limit, int a, int b, bool tradeoff_allowed,                    */
/*    KHE_COST tradeoff_cost, KHE_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Make and return a new dominance test object.                             */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMake(bool allow_zero, int min_limit,
  int max_limit, int a, int b, bool tradeoff_allowed,
  KHE_COST tradeoff_cost, KHE_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DOM_TEST res;
  if( HaArrayCount(drs->free_dom_tests) > 0 )
    res = HaArrayLastAndDelete(drs->free_dom_tests);
  else
    HaMake(res, drs->arena);
  res->allow_zero = allow_zero;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->tradeoff_allowed = tradeoff_allowed;
  res->a = a;
  res->b = b;
  res->tradeoff_cost = tradeoff_cost;
  res->monitor = m;
  return res;
}


/* *** old version
static KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type,
  bool tradeoff, KHE_COST tradeoff_cost)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.tradeoff = tradeoff;
  res.min_limit = -1;
  res->tradeoff_cost = tradeoff ? tradeoff_cost : 0;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMakeUnused(void)                           */
/*                                                                           */
/*  Make an unused dom test.                                                 */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused
static KHE_DRS_DOM_TEST KheDrsDomTestMakeUnused(void)
{
  return KheDrsDomTestMake(false, 0, INT_MAX, false, 0);
}
*** */

/* *** old version
static KHE_DRS_DOM_TEST KheDrsDomTestMakeUnused(void)
{
  KHE_DRS_DOM_TEST res;
  res.type = KHE_DRS_DOM_UNUSED;
  res.tradeoff = false;
  res.min_limit = -1;
  res->tradeoff_cost = 0;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val,  */
/*    bool tradeoff, KHE_COST tradeoff_cost)                                 */
/*                                                                           */
/*  Make and return a new dominance test object with a min limit attribute.  */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val,
  bool tradeoff, KHE_COST tradeoff_cost)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.tradeoff = tradeoff;
  res.min_limit = val;
  res->tradeoff_cost = tradeoff ? tradeoff_cost : 0;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)      */
/*                                                                           */
/*  Return true if dt1 and dt2 are equal.  This function assumes, as it      */
/*  may, that when the min_limit field is unused its value is always -1,     */
/*  and that when the tradeoff field is false the tradeoff_cost field is 0.  */
/*                                                                           */
/*****************************************************************************/

/* *** old version, and now no longer used
static bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)
{
  return dt1.type == dt2.type && dt1.min_limit == dt2.min_limit &&
    dt1.tradeoff == dt2.tradeoff && dt1->tradeoff_cost == dt2->tradeoff_cost;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)     */
/*                                                                           */
/*  Return true if val1 dominates val2, according to dt.                     */
/*                                                                           */
/*  Implementation note.  This formula is derived in the Appendix to the     */
/*  User's Guide, where it is called dom(val1, val2).                        */
/*                                                                           */
/*****************************************************************************/

static bool tilde(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  return (dt->allow_zero ? (val1 == 0) == (val2 == 0) : true);
}

static bool KheDrsDomMax(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  return val1 <= dt->a || val1 <= val2;
}

static bool KheDrsDomMin(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  return val1 >= dt->b || (tilde(dt, val1, val2) && val1 >= val2);
}

static bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  return KheDrsDomMax(dt, val1, val2) && KheDrsDomMin(dt, val1, val2);
}

/* *** old version
static bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      ** should never be called **
      HnAbort("KheDrsDomTestDominates internal error (KHE_DRS_DOM_UNUSED)");
      return false;  ** keep compiler happy **

    case KHE_DRS_DOM_GE:

      ** min limit **
      return val1 >= val2;

    case KHE_DRS_DOM_LE:

      ** max limit **
      return val1 <= val2;

    case KHE_DRS_DOM_EQ:

      ** values must be equal **
      return val1 == val2;

    case KHE_DRS_DOM_GE_LOOSE:

      ** loose min limit **
      return val1 >= val2 || val1 >= dt.min_limit;

    case KHE_DRS_DOM_EQ_LOOSE:

      ** max and loose min; both conditions must hold (basically equality) **
      return (val1 <= val2) && (val1 >= val2 || val1 >= dt.min_limit);

    default:

      HnAbort("KheDrsDomTestDominates internal error (%d)", dt.type);
      return false;  ** keep compiler happy **
  }
}
*** */

/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTryTradeoff(KHE_DRS_DOM_TEST dt, int delta_val,               */
/*    KHE_COST *cost1, KHE_COST cost2)                                       */
/*                                                                           */
/*  If dt->tradeoff_allowed, carry out tradeoff dominance with value change  */
/*  delta_val.  Otherwise return false.                                      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTryTradeoff(KHE_DRS_DOM_TEST dt, int delta_val,
  KHE_COST *cost1, KHE_COST cost2)
{
  KHE_COST cost;
  cost = *cost1 + dt->tradeoff_cost * delta_val;
  return cost <= cost2 ? (*cost1 = cost, true) : false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestDominatesWithTradeoff(KHE_DRS_DOM_TEST dt,             */
/*    int val1, KHE_COST *cost1, int val2, KHE_COST cost2)                   */
/*                                                                           */
/*  Return true if val1 dominates val2, according to dt, allowing for a      */
/*  tradeoff.                                                                */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestDominatesWithTradeoff(KHE_DRS_DOM_TEST dt,
  int val1, KHE_COST *cost1, int val2, KHE_COST cost2)
{
  if( dt->tradeoff_allowed )
  {
    /* try it with tradeoff */
    if( KheDrsDomMax(dt, val1, val2) )
    {
      if( KheDrsDomMin(dt, val1, val2) )
	return true;
      else if( dt->allow_zero )
	return false;
      else
	return KheDrsTryTradeoff(dt, val2 - val1, cost1, cost2);
    }
    else
    {
      if( KheDrsDomMin(dt, val1, val2) )
	return KheDrsTryTradeoff(dt, val1 - val2, cost1, cost2);
      else
	return false;
    }
  }
  else
  {
    /* try it without tradeoff */
    return KheDrsDomTestDominates(dt, val1, val2);
  }
}

/* *** sti ll to do
static bool KheDrsDomTestDominatesWithTra deoff(KHE_DRS_DOM_TEST dt,
  int val1, KHE_COST *cost1, int val2, KHE_COST cost2)
{
  KHE_COST cost;
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      ** should never be called **
      HnAbort("KheDrsDomTestDominates internal error (KHE_DRS_DOM_UNUSED)");
      return false;  ** keep compiler happy **

    case KHE_DRS_DOM_GE:

      ** min limit **
      if( val1 >= val2 )
	return true;
      else if( !dt.tradeoff )
	return false;
      else
      {
	cost = *cost1 + dt->tradeoff_cost * (val2 - val1);
	return cost <= cost2 ? (*cost1 = cost, true) : false;
      }

    case KHE_DRS_DOM_LE:

      ** max limit **
      if( val1 <= val2 )
	return true;
      else if( !dt.tradeoff )
	return false;
      else
      {
	cost = *cost1 + dt->tradeoff_cost * (val1 - val2);
	return cost <= cost2 ? (*cost1 = cost, true) : false;
      }

    case KHE_DRS_DOM_EQ:

      ** values must be equal **
      if( val1 == val2 )
	return true;
      else if( !dt.tradeoff )
	return false;
      else
      {
	if( val1 < val2 )
	  cost = *cost1 + dt->tradeoff_cost * (val2 - val1);
	else
	  cost = *cost1 + dt->tradeoff_cost * (val1 - val2);
	return cost <= cost2 ? (*cost1 = cost, true) : false;
      }

    case KHE_DRS_DOM_GE_LOOSE:

      ** loose min limit **
      if( val1 >= val2 || val1 >= dt.min_limit )
	return true;
      else if( !dt.tradeoff )
	return false;
      else
      {
	cost = *cost1 + dt->tradeoff_cost * (val2 - val1);
	return cost <= cost2 ? (*cost1 = cost, true) : false;
      }

    case KHE_DRS_DOM_EQ_LOOSE:

      ** max and loose min; both conditions must hold (basically equality) **
      if( val1 <= val2 && (val1 >= val2 || val1 >= dt.min_limit) )
	return true;
      else if( !dt.tradeoff )
	return false;
      else
      {
	if( val1 < val2 )
	  cost = *cost1 + dt->tradeoff_cost * (val2 - val1);
	else
	  cost = *cost1 + dt->tradeoff_cost * (val1 - val2);
	return cost <= cost2 ? (*cost1 = cost, true) : false;
      }


    default:

      HnAbort("KheDrsDomTestDominates internal error (%d)", dt.type);
      return false;  ** keep compiler happy **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestIsStrictEquality(KHE_DRS_DOM_TEST dom_test)            */
/*                                                                           */
/*  Return true if dom_test amounts to a strict equality test.  As shown     */
/*  in the documentation, the allow_zero flag does not affect the result.    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestIsStrictEquality(KHE_DRS_DOM_TEST dom_test)
{
  return dom_test->a < 0 && dom_test->b >= INT_MAX;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatingSet(KHE_DRS_DOM_TEST dt, int x,              */
/*    int *from_val, int *to_val, bool *also_zero)                           */
/*                                                                           */
/*  Set [*from_val, *to_val] and *also_zero to the set of non-negative       */
/*  values that dominate x under dt.  That is, set it to                     */
/*                                                                           */
/*    { v | v >= 0 && KheDrsDomTestDominates(dt, v, x) }                     */
/*                                                                           */
/*  The formula for this is derived in the User's Guide.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatingSet(KHE_DRS_DOM_TEST dt, int x,
  int *from_val, int *to_val, bool *also_zero)
{
  if( dt->allow_zero && x == 0 )
  {
    *from_val = max(0, dt->b);
    *to_val = max(dt->a, 0);
    *also_zero = (*from_val > 0);
  }
  else
  {
    *from_val = min(dt->b, x);
    if( *from_val < 0 )  *from_val = 0;
    *to_val = max(dt->a, x);
    *also_zero = false;
  }
}


/* *** old version
static void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      ** should never be called **
      HnAbort("KheDrsDomTestDominatingRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  ** keep compiler happy **
      break;

    case KHE_DRS_DOM_GE:

      ** largest range of values x such that x >= val **
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_LE:

      ** largest range of values x such that x <= val **
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ:

      ** largest range of values x such that x == val **
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      ** largest range of values x such that x >= val || x >= dt.min_limit **
      ** i.e. such that x >= min(val, dt.min_limit) **
      *from_val = min(val, dt.min_limit), *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      ** largest range of values x st x <= val && x >= min(val, dt.min_limit) **
      *from_val = min(val, dt.min_limit), *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatingRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  ** keep compiler happy **
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatingRangeWithTrad eoff(KHE_DRS_DOM_TEST dt,      */
/*    int val, int *from_val, int *to_val, int *extra_val)                   */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that dominate val        */
/*  under dt.  That is, set it to the largest range of values x such that    */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, x, val)                                     */
/*                                                                           */
/*  Also set *extra_val to one extra val, just before *from_val or just      */
/*  after *to_val, that dominates at the cost of dt->tradeoff_cost.  Or      */
/*  if that can't be done, set *extra_val to -1.                             */
/*                                                                           */
/*****************************************************************************/

/* *** sti ll to do
static void KheDrsDomTestDominatingRangeWithTr adeoff(KHE_DRS_DOM_TEST dt,
  int val, int *from_val, int *to_val, int *extra_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      ** should never be called **
      HnAbort("KheDrsDomTestDominatingRange internal error (DOM_UNUSED)");
      *from_val = *to_val = *extra_val = 0;  ** keep compiler happy **
      break;

    case KHE_DRS_DOM_GE:

      ** largest range of values x such that x >= val **
      *from_val = val, *to_val = INT_MAX;
      *extra_val = (*from_val > 0 && dt.tradeoff ? *from_val - 1 : -1);
      break;

    case KHE_DRS_DOM_LE:

      ** largest range of values x such that x <= val **
      *from_val = 0, *to_val = val;
      *extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      break;

    case KHE_DRS_DOM_EQ:

      ** largest range of values x such that x == val **
      *from_val = val, *to_val = val;
      *extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      ** largest range of values x such that x >= val || x >= dt.min_limit **
      ** i.e. such that x >= min(val, dt.min_limit) **
      *from_val = min(val, dt.min_limit), *to_val = INT_MAX;
      *extra_val = (*from_val > 0 && dt.tradeoff ? *from_val - 1 : -1);
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      ** largest range of values x st x <= val && x >= min(val, dt.min_limit) **
      *from_val = min(val, dt.min_limit), *to_val = val;
      *extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      break;

    default:

      HnAbort("KheDrsDomTestDominatingRange internal error (%d)", dt.type);
      *from_val = *to_val = *extra_val = 0;  ** keep compiler happy **
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatedSet(KHE_DRS_DOM_TEST dt, int x,               */
/*    int *from_val, int *to_val, bool *also_zero)                           */
/*                                                                           */
/*  Set [*from_val, *to_val] and *also_zero to the set of non-negative       */
/*  values that x dominates under dt.  That is, set it to                    */
/*                                                                           */
/*    { v | v >= 0 && KheDrsDomTestDominates(dt, x, val) }                   */
/*                                                                           */
/*  This formula is derived in the Appendix to the User's Guide.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatedSet(KHE_DRS_DOM_TEST dt, int x,
  int *from_val, int *to_val, bool *also_zero)
{
  if( dt->allow_zero && x >= 1 )
  {
    *from_val = (x <= dt->a ? 1 : max(1, x));
    *to_val = (x >= dt->b ? INT_MAX : x);
    if( x <= max(dt->a, 0) && x >= dt->b )
    {
      /* make sure 0 is in the set */
      if( *from_val == 1 )
	*from_val = 0;
      *also_zero = (*from_val > 0);
    }
    else
      *also_zero = false;
  }
  else
  {
    *from_val = (x <= dt->a ? 0 : x);
    *to_val = (x >= dt->b ? INT_MAX : x);
    *also_zero = false;
  }
}


/* *** old version
static void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      ** should never be called **
      HnAbort("KheDrsDomTestDominatedRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  ** keep compiler happy **
      break;

    case KHE_DRS_DOM_GE:

      ** largest range of values x such that val >= x **
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_LE:

      ** largest range of values x such that val <= x **
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ:

      ** largest range of values x such that val == x **
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      ** largest range of values x such that val >= x || val >= dt.min_limit **
      if( val >= dt.min_limit )
        *from_val = 0, *to_val = INT_MAX;
      else
        *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      ** largest range of x st val <= x && (val >= x || val >= dt.min_limit) **
      if( val >= dt.min_limit )
	*from_val = val, *to_val = INT_MAX;
      else
        *from_val = val, *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatedRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  ** keep compiler happy **
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatedRangeWithTrade off(KHE_DRS_DOM_TEST dt,       */
/*    int val, int *from_val, int *to_val, int *extra_val)                   */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that are dominated by    */
/*  val under dt.  That is, set it to the largest range of values x s.t.     */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, val, x)                                     */
/*                                                                           */
/*  Also set *extra_val to one extra val, just before *from_val or just      */
/*  after *to_val, that is dominated at the cost of dt->tradeoff_cost.  Or   */
/*  if that can't be done, set *extra_val to -1.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDomTestDominatedRangeWithTra deoff(KHE_DRS_DOM_TEST dt,
  int val, int *from_val, int *to_val, int *extra_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      ** should never be called **
      HnAbort("KheDrsDomTestDominatedRange internal error (DOM_UNUSED)");
      *from_val = *to_val = *extra_val = 0;  ** keep compiler happy **
      break;

    case KHE_DRS_DOM_GE:

      ** largest range of values x such that val >= x **
      *from_val = 0, *to_val = val;
      *extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      break;

    case KHE_DRS_DOM_LE:

      ** largest range of values x such that val <= x **
      *from_val = val, *to_val = INT_MAX;
      *extra_val = (*from_val > 0 && dt.tradeoff ? *from_val - 1 : -1);
      break;

    case KHE_DRS_DOM_EQ:

      ** largest range of values x such that val == x **
      *from_val = val, *to_val = val;
      *extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      ** largest range of values x such that val >= x || val >= dt.min_limit **
      if( val >= dt.min_limit )
      {
        *from_val = 0, *to_val = INT_MAX,
	*extra_val = -1;
      }
      else
      {
        *from_val = 0, *to_val = val;
	*extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      }
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      ** largest range of x st val <= x && (val >= x || val >= dt.min_limit) **
      if( val >= dt.min_limit )
      {
	*from_val = val, *to_val = INT_MAX;
	*extra_val = (*from_val > 0 && dt.tradeoff ? *from_val - 1 : -1);
      }
      else
      {
        *from_val = val, *to_val = val;
	*extra_val = (dt.tradeoff ? *to_val + 1 : -1);
      }
      break;

    default:

      HnAbort("KheDrsDomTestDominatedRange internal error (%d)", dt.type);
      *from_val = *to_val = *extra_val = 0;  ** keep compiler happy **
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])             */
/*                                                                           */
/*  Show dt, using buff for scratch storage.                                 */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])
{
  snprintf(buff, 100, "[Z %s, a %d, b %d]", bool_show(dt->allow_zero),
    dt->a, dt->b);
  /* ***
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      sprintf(buff, "-");
      break;

    case KHE_DRS_DOM_GE:

      sprintf(buff, ">=");
      break;

    case KHE_DRS_DOM_LE:

      sprintf(buff, "<=");
      break;

    case KHE_DRS_DOM_EQ:

      sprintf(buff, "==");
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      sprintf(buff, ">=(%d)", dt.min_limit);
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      sprintf(buff, "=(%d)", dt.min_limit);
      break;

    default:

      sprintf(buff, "?");
      break;
  }
  *** */
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSet(char *label, int x, int from_val, int to_val,             */
/*    bool also_zero, int indent, FILE *fp)                                  */
/*                                                                           */
/*  Print label(x) = {set} onto fp with the given indent.                    */
/*                                                                           */
/*****************************************************************************/

static void KhePrintInt(int val, FILE *fp)
{
  if( val == INT_MAX )
    fprintf(fp, "inf");
  else
    fprintf(fp, "%d", val);
}

static void KheDrsSet(char *label, int x, int from_val, int to_val,
  bool also_zero, int indent, FILE *fp)
{
  fprintf(fp, "%*s%s(%d) = {%s", indent, "", label, x, also_zero ? "0, " : "");
  KhePrintInt(from_val, fp);
  fprintf(fp, " .. ");
  KhePrintInt(to_val, fp);
  fprintf(fp, "}\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDebug(KHE_DRS_DOM_TEST dom_test, int verbosity,        */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dom_test, including dominating and dominated.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDebug(KHE_DRS_DOM_TEST dom_test, int verbosity,
  int indent, FILE *fp)
{
  char buff[100];  int x, from_val, to_val;  bool also_zero;
  if( indent < 0 )
    fprintf(fp, "%s", KheDrsDomTestShow(dom_test, buff));
  else
  {
    fprintf(fp, "%*s[ %s\n", indent, "", KheDrsDomTestShow(dom_test, buff));
    if( verbosity >= 3 )
    {
      for( x = 0;  x < 10;  x++ )
      {
	KheDrsDomTestDominatingSet(dom_test, x, &from_val, &to_val, &also_zero);
        KheDrsSet("dominating", x, from_val, to_val, also_zero, indent + 2, fp);
      }
      for( x = 0;  x < 10;  x++ )
      {
	KheDrsDomTestDominatedSet(dom_test, x, &from_val, &to_val, &also_zero);
        KheDrsSet("dominated", x, from_val, to_val, also_zero, indent + 2, fp);
      }
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnMake(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,     */
/*    KHE_COST cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Make a new soln with these attributes.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnMake(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,
  KHE_COST cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;

  /* get the basic object */
  if( HaArrayCount(drs->free_solns) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_solns);
    HaArrayClear(res->prev_tasks);
    HaArrayClear(res->sig);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->prev_tasks, drs->arena);
    HaArrayInit(res->sig, drs->arena);
  }
  
  /* initialize the fields */
  res->cost = cost;
  res->prev_soln = prev_soln;
  res->day = day;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnFree(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Free soln.                                                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnFree(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_solns, soln);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignatureFullHash(KHE_DRS_SOLN soln)                       */
/*                                                                           */
/*  Hash function for hashing the full signature of soln.                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignatureFullHash(KHE_DRS_SOLN soln)
{
  int res, val, i;
  res = 0;
  HaArrayForEach(soln->sig, val, i)
    res = (res + val) * 3;
  return (res < 0 ? - res : res);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignatureFullHashUntyped(void *p)                          */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignatureHash.                              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignatureFullHashUntyped(void *p)
{
  return KheDrsSolnSignatureFullHash((KHE_DRS_SOLN) p);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignaturePartialHash(KHE_DRS_SOLN soln)                    */
/*                                                                           */
/*  Hash function for hashing the partial signature of soln, that is, only   */
/*  the parts of it whose dominance tests are equality (or loose equality).  */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignaturePartialHash(KHE_DRS_SOLN soln)
{
  int res, index, val, i;  KHE_DRS_DAY day;
  res = 0;
  day = soln->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
  {
    val = HaArray(soln->sig, index);
    res = (res + val) * 3;
  }
  if( res < 0 )
    res = - res;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignatureFullEqual(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)*/
/*                                                                           */
/*  Return true if the signatures of soln1 and soln2 are equal.  The lengths */
/*  of their signatures must be equal.                                       */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignatureFullEqual(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)
{
  int i;
  HnAssert(HaArrayCount(soln1->sig) == HaArrayCount(soln2->sig),
    "KheDrsSolnSignaturesEqual: signatures have different lengths");
  for( i = 0;  i < HaArrayCount(soln1->sig);  i++ )
    if( HaArray(soln1->sig, i) != HaArray(soln2->sig, i) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignatureFullEqualUntyped(void *p1, void *p2)             */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignaturesEqual.                            */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignatureFullEqualUntyped(void *p1, void *p2)
{
  return KheDrsSolnSignatureFullEqual((KHE_DRS_SOLN) p1,
    (KHE_DRS_SOLN) p2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignaturePartialEqual(KHE_DRS_SOLN soln1,                 */
/*    KHE_DRS_SOLN soln2)                                                    */
/*                                                                           */
/*  Return true if the signatures of soln1 and soln2 are partially equal,    */
/*  that is, the entries whose dominance tests are equality (or loose        */
/*  equality) are equal.                                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignaturePartialEqual(KHE_DRS_SOLN soln1,
  KHE_DRS_SOLN soln2)
{
  int i, index;  KHE_DRS_DAY day;
  HnAssert(HaArrayCount(soln1->sig) == HaArrayCount(soln2->sig),
   "KheDrsSolnSignaturePartialEqual: signatures have different lengths");
  day = soln1->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
    if( HaArray(soln1->sig, index) != HaArray(soln2->sig, index) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSignatureDebug(KHE_DRS_SOLN soln, FILE *fp)               */
/*                                                                           */
/*  Debug print of the signature of soln onto fp (no verbosity or indent).   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSignatureDebug(KHE_DRS_SOLN soln, FILE *fp)
{
  int i, val;
  fprintf(fp, "[");
  HaArrayForEach(soln->sig, val, i)
  {
    if( i > 0 )
      fprintf(fp, " ");
    fprintf(fp, "%d", val);
  }
  fprintf(fp, "]");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSignatureDebugUntyped(void *p, FILE *fp)                  */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignatureDebug.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSignatureDebugUntyped(void *p, FILE *fp)
{
  KheDrsSolnSignatureDebug((KHE_DRS_SOLN) p, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2,         */
/*    int start_depth)                                                       */
/*                                                                           */
/*  Return true if soln1 dominates soln2.  They come from the same day.      */
/*  Parameter start_depth says how far along the signature to start.         */
/*                                                                           */
/*****************************************************************************/

/* *** old non-tradeoff version, no longer used
static bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2,
  int start_depth)
{
  KHE_DRS_DOM_TEST dt;  int i, val1, val2, sig_len;  KHE_DRS_DAY day;
  HnAssert(soln1->day == soln2->day,
    "KheDrsSolnDominates internal error: soln1->day = %s, soln2->day = %s",
    soln1->day != NULL ? KheDrsDayId(soln1->day) : "-",
    soln2->day != NULL ? KheDrsDayId(soln2->day) : "-");
  if( soln2->cost < soln1->cost )
    return false;
  day = soln1->day;
  sig_len = HaArrayCount(soln1->sig);
  for( i = start_depth;  i < sig_len;  i++ )
  {
    dt = HaArray(day->dom_tests, i);
    val1 = HaArray(soln1->sig, i);
    val2 = HaArray(soln2->sig, i);
    if( !KheDrsDomTestDominates(dt, val1, val2) )
      return false;
  }
  return true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsDistanceToAncestor(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)     */
/*                                                                           */
/*  Return the distance to the first common ancestor.  If soln1 == soln2     */
/*  this will be 0; otherwise, if they have the same parent solution, it     */
/*  will be 1; and so on.  The distance must be the same from each soln.     */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static int KheDrsDistanceToAncestor(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)
{
  int res;
  res = 0;
  while( soln1 != soln2 )
    soln1 = soln1->prev_soln, soln2 = soln2->prev_soln, res++;
  return res;
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, bool with_tradeoff,         */
/*    KHE_COST extra_cost, KHE_DRS_SOLN soln2, int start_depth,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return true if soln1 dominates soln2, using strong dominance if          */
/*  with_tradeoff is false, and tradeoff dominance if with_tradeoff is       */
/*  true.  If with_tradeoff is true, we have already traded off extra_cost   */
/*  for soln1.  Parameter start_depth says how far along to start.           */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, bool with_tradeoff,
  KHE_COST extra_cost, KHE_DRS_SOLN soln2, int start_depth,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DOM_TEST dt;  int i, v1, v2, sig_len;  KHE_DRS_DAY day;
  KHE_COST cost1;
  HnAssert(soln1->day == soln2->day, "KheDrsSolnDominates "
    "internal error: soln1->day = %s, soln2->day = %s",
    soln1->day != NULL ? KheDrsDayId(soln1->day) : "-",
    soln2->day != NULL ? KheDrsDayId(soln2->day) : "-");
  if( with_tradeoff )
  {
    cost1 = soln1->cost + extra_cost;
    if( cost1 > soln2->cost )
      return false;
    day = soln1->day;
    sig_len = HaArrayCount(soln1->sig);
    for( i = start_depth;  i < sig_len;  i++ )
    {
      dt = HaArray(day->dom_tests, i);
      v1 = HaArray(soln1->sig, i);
      v2 = HaArray(soln2->sig, i);
      if( !KheDrsDomTestDominatesWithTradeoff(dt, v1, &cost1, v2, soln2->cost) )
	return false;
    }
  }
  else
  {
    if( soln2->cost < soln1->cost )
      return false;
    day = soln1->day;
    sig_len = HaArrayCount(soln1->sig);
    for( i = start_depth;  i < sig_len;  i++ )
    {
      dt = HaArray(day->dom_tests, i);
      v1 = HaArray(soln1->sig, i);
      v2 = HaArray(soln2->sig, i);
      if( !KheDrsDomTestDominates(dt, v1, v2) )
	return false;
    }
  }
#if TESTING
  {
    int dist = KheDrsDistanceToAncestor(soln1, soln2);
    HaArrayFill(drs->ancestor_freq, dist + 1, 0);
    HaArray(drs->ancestor_freq, dist)++;
  }
#endif
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDominatesDebug(KHE_DRS_SOLN soln1, bool with_tradeoff,    */
/*    KHE_COST extra_cost, KHE_DRS_SOLN soln2, int start_depth,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of what KheDrsSolnDominates is doing with these arguments.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDominatesDebug(KHE_DRS_SOLN soln1, bool with_tradeoff,
  KHE_COST extra_cost, KHE_DRS_SOLN soln2, int start_depth,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY day;  KHE_DRS_DOM_TEST dom_test;  int i, val1, val2;
  bool still_going;  KHE_COST cost1;
  day = soln1->day;
  cost1 = soln1->cost + extra_cost;
  fprintf(fp, "%*s[ KheDrsSolnDominatesDebug(soln1 %.5f, soln2 %.5f)\n",
    indent, "", KheCostShow(soln1->cost), KheCostShow(soln2->cost));
  fprintf(fp, "%*s  %5s %5s %8s | %5s %5s | %8s | %s\n", indent, "",
    "Min", "Max", "Weight", "Soln1", "Soln2", "Dominate", "Monitor");
  fprintf(fp, "%*s  %5s-%5s-%8s-+-%5s-%5s-+-%8s-+-%s\n", indent, "",
    "-----", "-----", "--------", "-----", "-----", "--------",
    "----------------------------------");
  still_going = true;
  HaArrayForEach(day->dom_tests, dom_test, i)
  {
    val1 = HaArray(soln1->sig, i);
    val2 = HaArray(soln2->sig, i);
    fprintf(fp, "%*s  ", indent, "");

    /* Min */
    if( dom_test->min_limit == 0 )
      fprintf(fp, "%5s", " ");
    else if( dom_test->allow_zero )
      fprintf(fp, "%4dz", dom_test->min_limit);
    else
      fprintf(fp, "%5d", dom_test->min_limit);
    fprintf(fp, " ");

    /* Max */
    if( dom_test->max_limit == INT_MAX )
      fprintf(fp, "%5s", " ");
    else
      fprintf(fp, "%5d", dom_test->max_limit);
    fprintf(fp, " ");

    /* Weight */
    if( dom_test->tradeoff_allowed )
      fprintf(fp, "%8.5f", KheCostShow(dom_test->tradeoff_cost));
    else
      fprintf(fp, "%8s", "");
    fprintf(fp, " | ");

    /* soln1 and soln2 */
    fprintf(fp, "%5d %5d", val1, val2);
    fprintf(fp, " | ");

    /* dominates */
    if( !still_going )
      fprintf(fp, "%8s", "XXXXXXXX");
    else
    {
      still_going = KheDrsDomTestDominatesWithTradeoff(dom_test, val1, &cost1,
	val2, soln2->cost);
      if( still_going )
	fprintf(fp, "%8.5f", KheCostShow(cost1));
      else
	fprintf(fp, "%8s", "XXXXXXXX");
    }
    fprintf(fp, " | ");

    /* monitor */
    if( dom_test->monitor != NULL )
      fprintf(fp, "%s", KheMonitorId(dom_test->monitor));
    fprintf(fp, "\n");
  }
  fprintf(fp, "%*s  %5s-%5s-%8s-+-%5s-%5s-+-%8s-+-%s\n", indent, "",
    "-----", "-----", "--------", "-----", "-----", "--------",
    "----------------------------------");
  fprintf(fp, "%*s] %s\n", indent, "", bool_show(still_going));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnOverWrite(KHE_DRS_SOLN dest_soln, KHE_DRS_SOLN src_soln)  */
/*                                                                           */
/*  Overwrite src_soln onto dest_soln.  Their days and signatures are known  */
/*  to be equal, so those don't need to be touched.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnOverWrite(KHE_DRS_SOLN dest_soln, KHE_DRS_SOLN src_soln)
{
  int i;
  dest_soln->prev_soln = src_soln->prev_soln;
  HaArrayClear(dest_soln->prev_tasks);
  HaArrayAppend(dest_soln->prev_tasks, src_soln->prev_tasks, i);
  dest_soln->cost = src_soln->cost;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,                     */
/*    KHE_DRS_RESOURCE dr, KHE_DRS_TASK_ON_DAY *dtd)                         */
/*                                                                           */
/*  If dr is assigned in soln->prev_tasks, return true and set *dtd to what  */
/*  it is assigned to; otherwise return false and set *dtd to NULL.          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,
  KHE_DRS_RESOURCE dr, KHE_DRS_TASK_ON_DAY *dtd)
{
  if( soln->prev_soln == NULL )
  {
    /* this is the root solution, so there can be no assignment */
    return *dtd = NULL, false;
  }
  else
  {
    /* non-root solution, get assignment from soln->prev_tasks */
    *dtd = HaArray(soln->prev_tasks, dr->open_resource_index);
    return *dtd != NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPushAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,                     */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Push the assignment of drd to dtd onto drs, and update leaf exprs.       */
/*  Here drd must be non-NULL, but dtd could be NULL, in which case no leaf  */
/*  expr updates are required, because the default state is correct then.    */
/*                                                                           */
/*****************************************************************************/

/* *** old version
static void KheDrsPushAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  ** add the task to the edge **
  HaArrayAddLast(drs->extend_prev_tasks, dtd);

  if( dtd != NULL )
  {
    ** update leaf expressions of the resource on day **
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd, drs);

    ** update leaf expressions of the task on day **
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPopAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,                      */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Undo the effect of the corresponding KheDrsPushAsst.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsPopAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  ** remove the task from the edge **
  HaArrayDeleteLast(drs->extend_prev_tasks);

  if( dtd != NULL )
  {
    ** update leaf expressions of the resource **
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafClear(e, drs);

    ** update leaf expressions of the task **
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafClear(e, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPushAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,                     */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,                  */
/*    KHE_DRS_SIGNATURE sig)                                                 */
/*                                                                           */
/*  Push the assignment of drd to dtd onto drs, and update leaf exprs.       */
/*  Here drd must be non-NULL, but dtd could be NULL, in which case no leaf  */
/*  expr updates are required, because the default state is correct then.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPushAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_SIGNATURE sig)
{
  KHE_DRS_EXPR e;  int i;

  /* add the task to the edge */
  HaArrayAddLast(drs->extend_prev_tasks, dtd);
  HaArrayAddLast(drs->extend_prev_signatures, sig);

  if( dtd != NULL )
  {
    /* update leaf expressions of the task on day */
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPopAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,                      */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,                  */
/*    KHE_DRS_SIGNATURE sig)                                                 */
/*                                                                           */
/*  Undo the effect of the corresponding KheDrsPushAsst.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPopAsst(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_SIGNATURE sig)
{
  KHE_DRS_EXPR e;  int i;

  /* remove the task from the edge */
  HaArrayDeleteLast(drs->extend_prev_tasks);
  HaArrayDeleteLast(drs->extend_prev_signatures);

  if( dtd != NULL )
  {
    /* update leaf expressions of the task */
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafClear(e, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTryAsst(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,     */
/*    KHE_COST cost, KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assign dtd to drd, recurse, and unassign.                                */
/*                                                                           */
/*  Implementation note.  After the call to KheDrsPushAsst, there will be    */
/*  avail_resources resources available for further assignments, and there   */
/*  will be drs->extend_must_assign_count must-assign tasks remaining to     */
/*  assign.  Before the call to KheDrsPushAsst, therefore, the condition     */
/*  avail_resources >= drs->extend_must_assign_count must hold.              */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnDoExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

/* ***
static void KheDrsSolnTryAsst(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int save_extend_must_assign_count, avail_resources;
  KHE_COST save_extend_cost;  KHE_DRS_TASK dt;

  ** save drs->extend_cost and drs->extend_must_assign_count **
  save_extend_cost = drs->extend_cost;
  save_extend_must_assign_count = drs->extend_must_assign_count;

  ** update drs->extend_cost and drs->extend_must_assign_count **
  if( dtd != NULL )
  {
    dt = dtd->encl_dt;
    if( dtd == HaArrayFirst(dt->days) )
      drs->extend_cost += dt->asst_cost;
    if( dt->extend_must_assign )
      drs->extend_must_assign_count--;
  }

  ** push dtd assignment, recurse, and pop dtd assignment **
  avail_resources = HaArrayCount(drs->open_resources) -
    (HaArrayCount(drs->extend_prev_tasks) + 1);
  if( drs->extend_cost < drs->solve_init_cost &&
      avail_resources >= drs->extend_must_assign_count )
  {
    KheDrsPushAsst(drs, dtd, drd);
    KheDrsSolnDoExtend(prev_soln, next_day, drs);
    KheDrsPopAsst(drs, dtd, drd);
  }

  ** restore drs->extend_cost and drs->extend_must_assign_count **
  drs->extend_must_assign_count = save_extend_must_assign_count;
  drs->extend_cost = save_extend_cost;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTryAsst(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,     */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,                  */
/*    KHE_DRS_SIGNATURE sig, KHE_DYNAMIC_RESOURCE_SOLVER drs)                */
/*                                                                           */
/*  Assign dtd to drd, recurse, and unassign.                                */
/*                                                                           */
/*  Implementation note.  After the call to KheDrsPushAsst, there will be    */
/*  avail_resources resources available for further assignments, and there   */
/*  will be drs->extend_must_assign_count must-assign tasks remaining to     */
/*  assign.  Before the call to KheDrsPushAsst, therefore, the condition     */
/*  avail_resources >= drs->extend_must_assign_count must hold.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTryAsst(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_SIGNATURE sig, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int save_extend_must_assign_count, avail_resources;
  KHE_COST save_extend_cost;  KHE_DRS_TASK dt;

  /* save drs->extend_cost and drs->extend_must_assign_count */
  save_extend_cost = drs->extend_cost;
  save_extend_must_assign_count = drs->extend_must_assign_count;

  /* update drs->extend_cost and drs->extend_must_assign_count */
  if( dtd != NULL )
  {
    dt = dtd->encl_dt;
    if( dtd == HaArrayFirst(dt->days) )
      drs->extend_cost += dt->asst_cost;
    if( dt->extend_must_assign )
      drs->extend_must_assign_count--;
  }
  drs->extend_cost += sig->cost;

  /* push dtd assignment, recurse, and pop dtd assignment */
  avail_resources = HaArrayCount(drs->open_resources) -
    (HaArrayCount(drs->extend_prev_tasks) + 1);
  if( drs->extend_cost < drs->solve_init_cost &&
      avail_resources >= drs->extend_must_assign_count )
  {
    KheDrsPushAsst(drs, dtd, drd, sig);
    KheDrsSolnDoExtend(prev_soln, next_day, drs);
    KheDrsPopAsst(drs, dtd, drd, sig);
  }

  /* restore drs->extend_cost and drs->extend_must_assign_count */
  drs->extend_must_assign_count = save_extend_must_assign_count;
  drs->extend_cost = save_extend_cost;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMakeEvaluateAndAddSoln(KHE_DRS_SOLN prev_soln,                */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Make, evaluate, and possibly add to next_day a solution for next_day     */
/*  made from prev_soln and drs->extend_prev_tasks.                          */
/*                                                                           */
/*****************************************************************************/
/* ***
static void KheDrsExprEvalDay(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,
  KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs);
*** */
static void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET dns, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp);

/* *** old version
static void KheDrsMakeEvaluateAndAddSoln(KHE_DRS_SOLN prev_soln,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN next_soln;  KHE_DRS_EXPR e;  int i;

  ** quit early if over solve_soln_limit **
  if( --drs->solve_soln_limit < 0 )
    return;

  ** make soln for next_day from prev_soln and drs->extend_prev_tasks **
  HnAssert(prev_soln != NULL, "KheDrsMakeEvaluateAndAddSoln internal error");
  next_soln = KheDrsSolnMake(prev_soln, next_day, prev_soln->cost, drs);
  HaArrayAppend(next_soln->prev_tasks, drs->extend_prev_tasks, i);

  ** detailed debug of the last edge **
  if( DEBUG_LAST_EDGE )
  {
    fprintf(stderr, "[ KheDrsMakeEvaluateAndAddSoln last edge:\n");
    fprintf(stderr, "\n  last edge, prev_soln:\n");
    KheDrsSolnDebug(prev_soln, drs, 2, 2, stderr);
    fprintf(stderr, "\n  last edge, next_soln:\n");
    KheDrsSolnDebug(next_soln, drs, 2, 2, stderr);
  }

  ** set signature and update cost **
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, "\n  last edge, cost and signature updating:\n");
  ** KheDrsDaySingleCostSolnBegin(next_day); **
  HaArrayForEach(next_day->internal_today, e, i)
    KheDrsExprEvalDay(e, prev_soln, next_soln, drs);
  ** KheDrsDaySingleCostSolnEnd(next_day); **
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, "]\n");

  ** depending on cost, either add to next_day or free soln **
  if( DEBUG12 )
  {
    fprintf(stderr, "  KheDrsMakeEvaluateAndAddSoln adding soln:\n");
    KheDrsSolnDebug(next_soln, drs, 2, 2, stderr);
  }
  if( next_soln->cost < drs->solve_init_cost )
    KheDrsSolnSetAddSoln(next_day->soln_set, next_soln, drs);
  else
    KheDrsSolnFree(next_soln, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMakeEvaluateAndAddSoln(KHE_DRS_SOLN prev_soln,                */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Make, evaluate, and possibly add to next_day a solution for next_day     */
/*  made from prev_soln and drs->extend_prev_tasks.                          */
/*                                                                           */
/*****************************************************************************/
static int KheDrsSolnOpenDayIndex(KHE_DRS_SOLN soln);

static void KheDrsMakeEvaluateAndAddSoln(KHE_DRS_SOLN prev_soln,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN next_soln;  KHE_DRS_EXPR e;  int i, j, di;
  KHE_DRS_SIGNATURE sig;

  /* quit early if over solve_soln_limit */
  if( --drs->solve_soln_limit < 0 )
    return;

  /* make soln for next_day from prev_soln and drs->extend_prev_tasks */
  HnAssert(prev_soln != NULL, "KheDrsMakeEvaluateAndAddSoln internal error");
  next_soln = KheDrsSolnMake(prev_soln, next_day, drs->extend_cost, drs);
  HaArrayAppend(next_soln->prev_tasks, drs->extend_prev_tasks, i);

  /* detailed debug of the last edge */
  if( DEBUG_LAST_EDGE )
  {
    fprintf(stderr, "[ KheDrsMakeEvaluateAndAddSoln last edge:\n");
    fprintf(stderr, "\n  last edge, prev_soln:\n");
    KheDrsSolnDebug(prev_soln, drs, 2, 2, stderr);
    fprintf(stderr, "\n  last edge, next_soln:\n");
    KheDrsSolnDebug(next_soln, drs, 2, 2, stderr);
  }

  /* set signature and update cost */
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, "\n  last edge, cost and signature updating:\n");
  HnAssert(HaArrayCount(drs->extend_prev_signatures) ==
    HaArrayCount(drs->open_resources),
    "KheDrsMakeEvaluateAndAddSoln internal error");
  HaArrayForEach(drs->extend_prev_signatures, sig, i)
    HaArrayAppend(next_soln->sig, sig->sig, j);
  /* KheDrsDaySingleCostSolnBegin(next_day); */
  di = KheDrsSolnOpenDayIndex(next_soln);
  HaArrayForEach(next_day->nr_internal_today, e, i)
    KheDrsExprEvalSignature(e, prev_soln, di, (KHE_DRS_SIGNATURE)next_soln,drs);
    /* KheDrsExprEvalDay(e, prev_soln, next_soln, drs); */
  /* KheDrsDaySingleCostSolnEnd(next_day); */
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, "]\n");

  /* depending on cost, either add to next_day or free soln */
  if( DEBUG12 )
  {
    fprintf(stderr, "  KheDrsMakeEvaluateAndAddSoln adding soln:\n");
    KheDrsSolnDebug(next_soln, drs, 2, 2, stderr);
  }
  if( next_soln->cost < drs->solve_init_cost )
    KheDrsSolnSetAddSoln(next_day->soln_set, next_soln, drs);
  else
    KheDrsSolnFree(next_soln, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDoExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Carry out the main part of KheDrsSolnExtend.                             */
/*                                                                           */
/*****************************************************************************/

/* *** old version
static void KheDrsSolnDoExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE dr;  int i;  KHE_DRS_TASK_CLASS dtc;
  KHE_DRS_TASK dt;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_TASK_ON_DAY dtd;

  if( HaArrayCount(drs->extend_prev_tasks) >= HaArrayCount(drs->open_resources))
  {
    ** drs->extend_prev_tasks is ready to make into a solution and evaluate **
    KheDrsMakeEvaluateAndAddSoln(prev_soln, next_day, drs);
  }
  else
  {
    i = HaArrayCount(drs->extend_prev_tasks);
    dr = HaArray(drs->open_resources, i);
    drd = KheDrsResourceOnDay(dr, next_day);
    if( RERUN && drs->rerun != NULL )
    {
      ** this is a rerun, take sole dtd (possibly NULL) from drs->rerun **
      dtd = KheDrsPackedSolnTaskOnDay(drs->rerun, next_day->open_day_index, i);
      KheDrsSolnTryAsst(prev_soln, next_day, dtd, drd, drs);
    }
    else if( KheDrsResourceOnDayIsFixed(drd, prev_soln, &dtd) )
    {
      ** try dtd only **
      KheDrsSolnTryAsst(prev_soln, next_day, dtd, drd, drs);
    }
    else
    {
      ** try a task from any class that will accept dr **
      HaArrayForEach(next_day->task_classes, dtc, i)
	if( KheDrsTaskClassAcceptResourceBegin(dtc, dr, &dt) )
	{
	  if( !KheDrsTaskBusyOnDay(dt, next_day, &dtd) )
	    HnAbort("KheDrsSolnDoExtend internal error");
	  KheDrsSolnTryAsst(prev_soln, next_day, dtd, drd, drs);
	  KheDrsTaskClassAcceptResourceEnd(dtc);
	}

      ** try a free day **
      KheDrsSolnTryAsst(prev_soln, next_day, NULL, drd, drs);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskClassContainsTask(KHE_DRS_TASK_CLASS dtc, KHE_DRS_TASK dt)*/
/*                                                                           */
/*  Return true if dtc contain unassigned task dt.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTaskClassContainsTask(KHE_DRS_TASK_CLASS dtc, KHE_DRS_TASK dt)
{
  int pos;
  return HaArrayContains(dtc->unassigned_tasks, dt, &pos);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SIGNATURE KheDrsSolnGetTaskSignature(KHE_DRS_TASK_ON_DAY dtd,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Get the sig seg corresponding to dtd.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_SIGNATURE KheDrsSolnGetTaskSignature(KHE_DRS_RESOURCE dr,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_DAY next_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{

  int i;  KHE_DRS_TASK_CLASS dtc;
  if( dtd == NULL )
    return HaArrayLast(dr->extend_signatures);
  else
  {
    HaArrayForEach(next_day->task_classes, dtc, i)
      if( KheDrsTaskClassContainsTask(dtc, dtd->encl_dt) )
	return HaArray(dr->extend_signatures, i);
    HnAbort("KheDrsSolnGetTaskSignature internal error");
    return NULL;  ** keep compiler happy **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDoExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Carry out the main part of KheDrsSolnExtend.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDoExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE dr;  int i, count;  KHE_DRS_TASK_CLASS dtc;
  KHE_DRS_SIGNATURE sig; KHE_DRS_TASK dt;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_TASK_ON_DAY dtd;

  count = HaArrayCount(drs->extend_prev_tasks);
  if( DEBUG22 )
    fprintf(stderr, "%*s[ KheDrsSolnDoExtend(-, %s, drs) extend_cost %.5f\n",
      2 * count, "", KheDrsDayId(next_day), KheCostShow(drs->extend_cost));
  if( HaArrayCount(drs->extend_prev_tasks) >= HaArrayCount(drs->open_resources))
  {
    /* drs->extend_prev_tasks is ready to make into a solution and evaluate */
    KheDrsMakeEvaluateAndAddSoln(prev_soln, next_day, drs);
  }
  else
  {
    i = HaArrayCount(drs->extend_prev_tasks);
    dr = HaArray(drs->open_resources, i);
    drd = KheDrsResourceOnDay(dr, next_day);
    if( KheDrsResourceOnDayIsFixed(drd, prev_soln, &dtd) )
    {
      /* try dtd only */
      HnAssert(dtd == dr->extend_fixed_dtd,
	"KheDrsSolnDoExtend internal error 1");
      KheDrsSolnTryAsst(prev_soln, next_day, dtd, drd,
	dr->extend_fixed_signature, drs);
    }
    else if( RERUN && drs->rerun != NULL )
    {
      /* this is a rerun, take sole dtd (possibly NULL) from drs->rerun */
      dtd = KheDrsPackedSolnTaskOnDay(drs->rerun, next_day, dr);
      HnAssert(dtd == dr->extend_fixed_dtd,
	"KheDrsSolnDoExtend internal error 2");
      KheDrsSolnTryAsst(prev_soln, next_day, dtd, drd,
	dr->extend_fixed_signature, drs);
      /* sig = KheDrsSolnGetTaskSignature(dr, dtd, next_day, drs); */
    }
    else
    {
      /* try a task from any class that will accept dr */
      HaArrayForEach(next_day->task_classes, dtc, i)
	if( KheDrsTaskClassAcceptResourceBegin(dtc, dr, &dt) )
	{
	  if( !KheDrsTaskBusyOnDay(dt, next_day, &dtd) )
	    HnAbort("KheDrsSolnDoExtend internal error 3");
	  sig = HaArray(dr->extend_signatures, i);
	  KheDrsSolnTryAsst(prev_soln, next_day, dtd, drd, sig, drs);
	  KheDrsTaskClassAcceptResourceEnd(dtc);
	}

      /* try a free day */
      sig = HaArray(dr->extend_signatures, i);
      KheDrsSolnTryAsst(prev_soln, next_day, NULL, drd, sig, drs);
    }
  }
  if( DEBUG22 )
    fprintf(stderr, "%*s] KheDrsSolnDoExtend returning\n", 2 * count, "");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Build drs->extend_prev_tasks up to full length in all possible ways,     */
/*  then for each way, use it to extend prev_soln to a solution for new_day. */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnSetBeginCacheSegment(KHE_DRS_SOLN_SET soln_set,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsSolnSetEndCacheSegment(KHE_DRS_SOLN_SET soln_set,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY next_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  KHE_DRS_TASK dt;  int i, j;  KHE_DRS_RESOURCE dr;

  /* begin caching extensions of prev_soln */
  if( DEBUG22 )
  {
    fprintf(stderr, "[ KheDrsSolnExtend(-, %s, drs)\n", KheDrsDayId(next_day));
    fprintf(stderr, "  solve_init_cost %.5f, solve_start_cost %.5f\n",
      KheCostShow(drs->solve_init_cost), KheCostShow(drs->solve_start_cost));
    fprintf(stderr, "  prev_soln->cost %.5f, solve_soln_limit %d\n",
      KheCostShow(prev_soln->cost), drs->solve_soln_limit);
  }
  KheDrsSolnSetBeginCacheSegment(next_day->soln_set, drs);

  /* initialize drs->extend fields other than signatures */
  HaArrayClear(drs->extend_prev_tasks);  /* not strictly necessary */
  drs->extend_cost = prev_soln->cost;
  drs->extend_must_assign_count = 0;
  HaArrayForEach(next_day->task_classes, dtc, i)
    HaArrayForEach(dtc->unassigned_tasks, dt, j)
      if( prev_soln->cost + dt->non_asst_cost >= drs->solve_init_cost )
      {
	/* dt must be assigned, otherwise cost will be too high */
	dt->extend_must_assign = true;
	drs->extend_must_assign_count++;
      }
      else
	dt->extend_must_assign = false;

  /* initialize signatures */
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceSetSignatures(dr, prev_soln, next_day, drs);

  /* carry out the main part of the extend */
  /* KheDrsSolnDoExtend(prev_soln, next_day, drs); */
  KheDrsSolnDoExtend(prev_soln, next_day, drs);

  /* end caching extensions of prev_soln */
  KheDrsSolnSetEndCacheSegment(next_day->soln_set, drs);
  if( DEBUG22 )
    fprintf(stderr, "] KheDrsSolnExtend returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KhePrintCol(char *str, bool rule_at_right, FILE *fp)                */
/*                                                                           */
/*  Print str (truncated if required) onto fp in a column.  If rule_at_right */
/*  is true, include a vertical rule at the right.                           */
/*                                                                           */
/*****************************************************************************/

static void KhePrintCol(char *str, bool rule_at_right, FILE *fp)
{
  int i;  char *p;

  /* print initial margin */
  fprintf(fp, " ");

  /* print exactly COL_WIDTH - 3 characters, starting with str */
  for( i = 0, p = str;  i < COL_WIDTH - 3;  i++ )
  {
    if( *p != '\0' )
      fprintf(fp, "%c", *p), p++;
    else
      fprintf(fp, " ");
  }

  /* print final margin, optionally including a rule */
  fprintf(fp, rule_at_right ? " |" : "  ");
}


/*****************************************************************************/
/*                                                                           */
/*  void KhePrintRule(int col_count, int indent, FILE *fp)                   */
/*                                                                           */
/*  Print a rule, wide enough for col_count columns.                         */
/*                                                                           */
/*****************************************************************************/

static void KhePrintRule(int col_count, int indent, FILE *fp)
{
  int i, j;

  /* indent */
  fprintf(fp, "%*s", indent, "");

  /* header column */
  fprintf(fp, "%*s+", COL_WIDTH - 1, "");

  /* ordinary columns */
  for( i = 0;  i < col_count;  i++ )
  {
    for( j = 1;  j < COL_WIDTH;  j++ )
      fprintf(fp, "-");
    fprintf(fp, "+");
  }

  /* final newline */
  fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  char *ShortName(char *name, char *prefix)                                */
/*                                                                           */
/*  Return name with prefix removed from the front, if present.              */
/*                                                                           */
/*****************************************************************************/

static char *ShortName(char *name, char *prefix)
{
  if( strstr(name, prefix) != name )
    return name;
  else
  {
    name = &name[strlen(prefix)];
    if( name[0] == ':' )
      return &name[1];
    else
      return name;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebugResourceTimetable(KHE_DRS_RESOURCE dr,               */
/*    KHE_DRS_SOLN soln, int first_day_index, int last_day_index,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of the full timetable of dr from first_day_index to          */
/*  last_day_index inclusive.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebugResourceTimetable(KHE_DRS_RESOURCE dr,
  KHE_DRS_SOLN soln, int first_day_index, int last_day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN soln2;  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE_ON_DAY drd;
  int i;  KHE_DRS_DAY day;

  /* indent */
  fprintf(fp, "%*s", indent, "");

  /* header column */
  KhePrintCol(KheResourceId(dr->resource), true, fp);

  /* ordinary columns */
  for( i = first_day_index;  i <= last_day_index;  i++ )
  {
    day = HaArray(drs->all_days, i);
    if( day->open_day_index == -1 )
    {
      /* a closed day, get timetable from dr */
      drd = HaArray(dr->days, i);
      dtd = drd->closed_asst;
      if( dtd == NULL || dtd->task == NULL )
        KhePrintCol("", true, fp);
      else
	KhePrintCol(ShortName(KheTaskId(dtd->task), KheDrsDayId(day)),
	  KheDrsTaskOnDayIsLast(dtd), fp);
    }
    else if( soln->day == NULL ||
	     day->open_day_index > soln->day->open_day_index )
    {
      /* an open day whose assignment is not yet known */
      KhePrintCol("?", true, fp);
    }
    else
    {
      /* an open day whose assignment is known */
      for( soln2 = soln;  soln2 != NULL;  soln2 = soln2->prev_soln )
	if( soln2->day == day )
	  break;
      HnAssert(soln2 != NULL,
	"KheDrsSolnDebugResourceTimetable internal error 1");
      HnAssert(0 <= dr->open_resource_index &&
	dr->open_resource_index < HaArrayCount(soln2->prev_tasks),
	"KheDrsSolnDebugResourceTimetable internal error 2");
      dtd = HaArray(soln2->prev_tasks, dr->open_resource_index);
      if( dtd == NULL || dtd->task == NULL )
        KhePrintCol("", true, fp);
      else
	KhePrintCol(ShortName(KheTaskId(dtd->task), KheDrsDayId(day)),
	  KheDrsTaskOnDayIsLast(dtd), fp);
    }
  }

  /* final newline */
  fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebugTimetable(KHE_DRS_SOLN soln, int first_day_index,    */
/*    int last_day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs,                   */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the timetables of the open resources from first_day_index */
/*  to last_day_index inclusive.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebugTimetable(KHE_DRS_SOLN soln, int first_day_index,
  int last_day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int i;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE dr;  char buff[200];

  /* header row and first rule */
  fprintf(fp, "%*s", indent + COL_WIDTH, "");
  for( i = first_day_index;  i <= last_day_index;  i++ )
  {
    day = HaArray(drs->all_days, i);
    if( day->open_day_index != -1 )
    {
      snprintf(buff, 200, "*%s*", KheDrsDayId(day));
      KhePrintCol(buff, false, fp);
    }
    else
      KhePrintCol(KheDrsDayId(day), false, fp);
  }
  fprintf(fp, "\n");
  KhePrintRule(last_day_index - first_day_index + 1, indent, fp);

  /* resource rows and other rules */
  HaArrayForEach(drs->open_resources, dr, i)
  {
    KheDrsSolnDebugResourceTimetable(dr, soln, first_day_index, last_day_index,
      drs, verbosity, indent, fp);
    KhePrintRule(last_day_index - first_day_index + 1, indent, fp);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of soln onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int i, j;
  for( i = 0;  i < HaArrayCount(drs->all_days) - 1;  i += COL_COUNT )
  {
    fprintf(fp, "\n");  /* even the first block has a preceding blank line */
    j = min(i + COL_COUNT - 1, HaArrayCount(drs->all_days) - 1);
    KheDrsSolnDebugTimetable(soln, i, j, drs, verbosity, indent, fp);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_LIST"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_LIST KheDrsSolnListMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Return a new, empty KHE_DRS_SOLN_LIST object.                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_LIST KheDrsSolnListMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST res;
  if( HaArrayCount(drs->free_soln_lists) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_lists);
    HaArrayClear(res->solns);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->solns, drs->arena);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListFree(KHE_DRS_SOLN_LIST soln_list,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ns and the solns in it.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListFree(KHE_DRS_SOLN_LIST soln_list,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(soln_list->solns, soln, i)
    KheDrsSolnFree(soln, drs);
  HaArrayAddLast(drs->free_soln_lists, soln_list);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnListSignaturePartialHashUntyped(void *p)                   */
/*                                                                           */
/*  Hash function for soln lists, being the partial hash of the list's       */
/*  first element.                                                           */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnListSignaturePartialHashUntyped(void *p)
{
  KHE_DRS_SOLN_LIST soln_list;  KHE_DRS_SOLN soln;
  soln_list = (KHE_DRS_SOLN_LIST) p;
  soln = HaArrayFirst(soln_list->solns);
  return KheDrsSolnSignaturePartialHash(soln);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnListSignaturePartialEqualUntyped(void *p1, void *p2)      */
/*                                                                           */
/*  Equality function for soln lists, being the partial equality of the      */
/*  lists' first elements.                                                   */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnListSignaturePartialEqualUntyped(void *p1, void *p2)
{
  KHE_DRS_SOLN_LIST soln_list1, soln_list2;  KHE_DRS_SOLN soln1, soln2;
  soln_list1 = (KHE_DRS_SOLN_LIST) p1;
  soln_list2 = (KHE_DRS_SOLN_LIST) p2;
  soln1 = HaArrayFirst(soln_list1->solns);
  soln2 = HaArrayFirst(soln_list2->solns);
  return KheDrsSolnSignaturePartialEqual(soln1, soln2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListAddSoln(KHE_DRS_SOLN_LIST soln_list,                  */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *count)        */
/*                                                                           */
/*  Carry out a strong dominance insert of soln into soln_list.              */
/*                                                                           */
/*****************************************************************************/

/* *** old non-tradeoff version, no longer used
static void KheDrsSolnListAddSoln(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *count)
{
  KHE_DRS_SOLN other_soln;  int i;

  ** if soln is dominated by anything else, delete soln and return **
  HaArrayForEachReverse(soln_list->solns, other_soln, i)
    if( KheDrsSolnDominates(other_soln, soln, 0) )
    {
      KheDrsSolnFree(soln, drs);
      return;
    }

  ** remove other solns that soln dominates **
  HaArrayForEach(soln_list->solns, other_soln, i)
    if( KheDrsSolnDominates(soln, other_soln, 0) )
    {
      KheDrsSolnFree(other_soln, drs);
      HaArrayDeleteAndPlug(soln_list->solns, i);
      (*count)--;
      i--;
    }

  ** add soln to soln_list **
  HaArrayAddLast(soln_list->solns, soln);
  (*count)++;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnListDominates(KHE_DRS_SOLN_LIST soln_list,                */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Return true if soln_list contains a solution that dominates soln.        */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnListDominates(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN other_soln;  int i;
  HaArrayForEach(soln_list->solns, other_soln, i)
    if( KheDrsSolnDominates(other_soln, with_tradeoff, 0, soln, 0, drs) )
      return true;
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListRemoveDominated(KHE_DRS_SOLN_LIST soln_list,          */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int *count)                                                            */
/*                                                                           */
/*  Remove those elements of soln_list that are dominated by soln.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListRemoveDominated(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *count)
{
  KHE_DRS_SOLN other_soln;  int i;
  HaArrayForEach(soln_list->solns, other_soln, i)
    if( KheDrsSolnDominates(soln, with_tradeoff, 0, other_soln, 0, drs) )
    {
      KheDrsSolnFree(other_soln, drs);
      HaArrayDeleteAndPlug(soln_list->solns, i);
      (*count)--;
      i--;
    }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListAddSoln(KHE_DRS_SOLN_LIST soln_list,                  */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int *count)                                                            */
/*                                                                           */
/*  Carry out a strong dominance insert of soln into soln_list, with or      */
/*  without tradeoff depending on with_tradeoff.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListAddSoln(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *count)
{
  if( KheDrsSolnListDominates(soln_list, soln, with_tradeoff, drs) )
  {
    /* soln is dominated by some other soln, so delete soln */
    KheDrsSolnFree(soln, drs);
  }
  else
  {
    /* remove other solns that soln dominates, and add soln to soln_list */
    KheDrsSolnListRemoveDominated(soln_list, soln, with_tradeoff, drs, count);
    HaArrayAddLast(soln_list->solns, soln);
    (*count)++;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListExtend(KHE_DRS_SOLN_LIST soln_list,                   */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Extend soln_list.                                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListExtend(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN prev_soln;  int i;
  HaArrayForEach(soln_list->solns, prev_soln, i)
    KheDrsSolnExtend(prev_soln, next_day, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListAddAll(KHE_DRS_SOLN_LIST soln_list,                   */
/*    KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Add the elements of soln_list to to_ssp.                                 */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnSetPartAddSoln(KHE_DRS_SOLN_SET_PART ssp,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnListAddAll(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(soln_list->solns, soln, i)
    KheDrsSolnSetPartAddSoln(to_ssp, soln, drs);
  HaArrayClear(soln_list->solns);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListGather(KHE_DRS_SOLN_LIST soln_list,                   */
/*    KHE_DRS_SOLN_SET soln_set)                                             */
/*                                                                           */
/*  Gather the elements of soln_list into soln_set->sorted_solns.            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListGather(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(soln_list->solns, soln, i)
    HaArrayAddLast(soln_set->sorted_solns, soln);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnListFirstSoln(KHE_DRS_SOLN_LIST soln_list)        */
/*                                                                           */
/*  Return the first soln of soln_list, or NULL if none.                     */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnListFirstSoln(KHE_DRS_SOLN_LIST soln_list)
{
  return HaArrayCount(soln_list->solns) == 0 ? NULL :
    HaArrayFirst(soln_list->solns);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnListLastSoln(KHE_DRS_SOLN_LIST soln_list)         */
/*                                                                           */
/*  Return the last soln of soln_list, or NULL if none.                      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnListLastSoln(KHE_DRS_SOLN_LIST soln_list)
{
  return HaArrayCount(soln_list->solns) == 0 ? NULL :
    HaArrayLast(soln_list->solns);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListExtractFinalSoln(KHE_DRS_SOLN_LIST soln_list,         */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Extract the final solution from soln_list, or do nothing if empty.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListExtractFinalSoln(KHE_DRS_SOLN_LIST soln_list,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(soln_list->solns, soln, i)
  {
    HnAssert(*final_soln == NULL,
      "KheDrsSolnListExtractFinalSoln internal error");
    *final_soln = soln;
  }
}

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListSignatureDebugUntyped(void *p, FILE *fp)              */
/*                                                                           */
/*  Debug print of the signature of soln set p.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListSignatureDebugUntyped(void *p, FILE *fp)
{
  KHE_DRS_SOLN_LIST soln_list;  KHE_DRS_SOLN soln;
  soln_list = (KHE_DRS_SOLN_LIST) p;
  soln = HaArrayFirst(soln_list->solns);
  KheDrsSolnSignatureDebug(soln, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListDebug(KHE_DRS_SOLN_LIST soln_list,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of soln_list onto fp with the given verbosity and indent.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListDebug(KHE_DRS_SOLN_LIST soln_list,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN soln;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ SolnList(%d solns)\n", indent, "",
      HaArrayCount(soln_list->solns));
    HaArrayForEach(soln_list->solns, soln, i)
    {
      KheDrsSolnDebug(soln, drs, verbosity, indent + 2, fp);
      if( i > 50 )
      {
	fprintf(fp, "%*s... omitting the rest\n", indent + 2, "");
	break;
      }
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "SolnList(%d solns)", HaArrayCount(soln_list->solns));
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_TRIE"                                            */
/*                                                                           */
/*  Abstraction function                                                     */
/*  --------------------                                                     */
/*                                                                           */
/*  A soln_trie represents a set of solns.  If soln_trie == NULL, the set    */
/*  is empty.  Otherwise the set consists of soln_trie->soln (if non-NULL)   */
/*  plus the sets stored in the children.                                    */
/*                                                                           */
/*  Invariant                                                                */
/*  ---------                                                                */
/*                                                                           */
/*  Each soln appears in the trie in a place determined by the sequence      */
/*  of values in its signature, in the usual trie manner.  Exactly one       */
/*  of these three cases applies:                                            */
/*                                                                           */
/*    (1)  soln_trie == NULL                                                 */
/*                                                                           */
/*    (2)  soln_trie != NULL && soln_trie->soln != NULL &&                   */
/*           HaArrayCount(soln_trie->children) == 0                          */
/*                                                                           */
/*    (3)  soln_trie != NULL && soln_trie->soln == NULL &&                   */
/*           HaArrayCount(soln_trie->children) > 0                           */
/*                                                                           */
/*  In case (3), the first child of soln_trie->children is non-NULL, and     */
/*  the last child of soln_trie->children is non-NULL.  These may be the     */
/*  same child.  The concrete index of the first child is 0 as usual,        */
/*  but it represents a child with abstract index soln_trie->base, with      */
/*  subsequent children also having abstract indexes increased by this.      */
/*  This allows the trie to handle indexes which are all far above zero      */
/*  without becoming burdened by many initial NULL children.                 */
/*                                                                           */
/*  There is no requirement that two or more solns be present in case (3).   */
/*  If there are no solns, case (1) always applies; if there are two or      */
/*  more solns, case (3) always applies; but if there is exactly one soln,   */
/*  then either of cases (2) or (3) applies.  This is because it would be    */
/*  too slow and tedious to detect situations where a case (3) soln could    */
/*  be simplified to a case (2) soln.                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_TRIE KheDrsSolnTrieMake(KHE_DRS_SOLN soln,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new trie object containing soln.  It satisfies case (2) above.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_TRIE KheDrsSolnTrieMake(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE res;
  HnAssert(soln != NULL, "KheDrsSolnTrieMake internal error (soln == NULL)");
  if( HaArrayCount(drs->free_soln_tries) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_tries);
    HaArrayClear(res->children);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->children, drs->arena);
  }
  res->soln = soln;
  res->base = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieFree(KHE_DRS_SOLN_TRIE *soln_trie,                    */
/*    bool free_solns, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  Free *soln_trie and all its descendants, setting *soln_trie to NULL.     */
/*  If free_solns, free the solutions in the trie as well as the trie.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieFree(KHE_DRS_SOLN_TRIE *soln_trie,
  bool free_solns, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;
  if( *soln_trie == NULL )
  {
    /* case (1), do nothing */
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2), delete *soln_trie and optionally (*soln_trie)->soln */
    if( free_solns )
      KheDrsSolnFree((*soln_trie)->soln, drs);
    HaArrayAddLast(drs->free_soln_tries, *soln_trie);
    *soln_trie = NULL;
  }
  else
  {
    /* case (3), delete children and *soln_trie */
    for( i = 0;  i < HaArrayCount((*soln_trie)->children);  i++ )
      KheDrsSolnTrieFree(&HaArray((*soln_trie)->children, i), free_solns, drs);
    HaArrayAddLast(drs->free_soln_tries, *soln_trie);
    *soln_trie = NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieAddChild(KHE_DRS_SOLN_TRIE soln_trie,                 */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert soln in its appropriate place among the children of soln_trie.    */
/*                                                                           */
/*  Implementation note.  This is a helper function; it neither assumes      */
/*  that the invariant is true before nor guarantees that it is true after.  */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnTrieAdd(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnTrieAddChild(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int index;
  index = HaArray(soln->sig, depth);
  if( HaArrayCount(soln_trie->children) == 0 )
  {
    /* array is empty, set base to index and insert at 0 */
    HaArrayAddLast(soln_trie->children, NULL);
    soln_trie->base = index;
    KheDrsSolnTrieAdd(&HaArray(soln_trie->children, 0), soln, depth + 1, drs);
  }
  else if( index < soln_trie->base )
  {
    /* shift the children right, decrease the base to index, and insert at 0 */
    HaArrayShiftRight(soln_trie->children, soln_trie->base - index, NULL);
    soln_trie->base = index;
    KheDrsSolnTrieAdd(&HaArray(soln_trie->children, 0), soln, depth + 1, drs);
  }
  else
  {
    /* make sure there is a position for soln, and insert it there */
    index -= soln_trie->base;
    HaArrayFill(soln_trie->children, index + 1, NULL);
    KheDrsSolnTrieAdd(&HaArray(soln_trie->children, index), soln, depth+1, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieAdd(KHE_DRS_SOLN_TRIE *soln_trie, KHE_DRS_SOLN soln,  */
/*    int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)                            */
/*                                                                           */
/*  Insert soln into *soln_trie, possibly changing *soln_trie as a result.   */
/*  The depth parameter indicates how far along the signature we are.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieAdd(KHE_DRS_SOLN_TRIE *soln_trie, KHE_DRS_SOLN soln,
  int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( *soln_trie == NULL )
  {
    /* case (1), make a new soln containing just soln, -> case (2) */
    *soln_trie = KheDrsSolnTrieMake(soln, drs);
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2), add soln and (*soln_trie)->soln to children, -> case (3) */
    KheDrsSolnTrieAddChild(*soln_trie, soln, depth, drs);
    soln = (*soln_trie)->soln;
    (*soln_trie)->soln = NULL;
    KheDrsSolnTrieAddChild(*soln_trie, soln, depth, drs);
  }
  else
  {
    /* case (3), add soln to children, -> case (3) */
    KheDrsSolnTrieAddChild(*soln_trie, soln, depth, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,                */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Return true if there is a soln in soln_trie that dominates soln.         */
/*                                                                           */
/*****************************************************************************/

/* *** non-tradeoff version, no longer used
static bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int val, from_val, to_val, i;  bool also_zero;  KHE_DRS_DOM_TEST dom_test;
  if( soln_trie == NULL )
  {
    ** case (1); no solns in soln_trie, so none dominate **
    return false;
  }
  else if( soln_trie->soln != NULL )
  {
    ** case (2); the only possible dominator is soln_trie->soln **
    return KheDrsSolnDominates(soln_trie->soln, soln, depth);
  }
  else
  {
    ** case (3); find the dominating set for the current signature value **
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->sig, depth);
    KheDrsDomTestDominatingSet(dom_test, val, &from_val, &to_val, &also_zero);

    ** convert dominating set to dominating indexes **
    from_val -= soln_trie->base;
    to_val -= soln_trie->base;
    if( from_val < 0 )
      from_val = 0;
    if( to_val >= HaArrayCount(soln_trie->children) )
      to_val = HaArrayCount(soln_trie->children) - 1;
    also_zero = also_zero && soln_trie->base == 0;

    ** search for dominating solutions at the dominating indexes **
    for( i = from_val;  i <= to_val;  i++ )
      if( KheDrsSolnTrieDominates(HaArray(soln_trie->children, i), soln,
	  depth + 1, drs) )
	return true;
    if( also_zero )
      if( KheDrsSolnTrieDominates(HaArray(soln_trie->children, 0), soln,
	  depth + 1, drs) )
	return true;
    return false;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheSetHasExtraValueAtLeft(int val, int from_val, int to_val,        */
/*    bool also_zero, int *extra_val)                                        */
/*                                                                           */
/*  If the interval [from_val .. to_val] has val at the left end, return     */
/*  true and set *extra_val to the adjacent value outside the interval.      */
/*                                                                           */
/*****************************************************************************/

static bool KheSetHasExtraValueAtLeft(int val, int from_val, int to_val,
  bool also_zero, int *extra_val)
{
  if( from_val <= to_val && from_val == val && from_val >= (also_zero ? 2 : 1) )
    return *extra_val = from_val - 1, true;
  else
    return *extra_val = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheSetHasExtraValueAtRight(int val, int from_val, int to_val,       */
/*    bool also_zero, int *extra_val)                                        */
/*                                                                           */
/*  If the interval [from_val .. to_val] has val at the right end, return    */
/*  true and set *extra_val to the adjacent value outside the interval.      */
/*                                                                           */
/*****************************************************************************/

static bool KheSetHasExtraValueAtRight(int val, int from_val, int to_val,
  bool also_zero, int *extra_val)
{
  if( from_val <= to_val && to_val == val )
    return *extra_val = to_val + 1, true;
  else
    return *extra_val = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,                */
/*    bool with_tradeoff, KHE_COST extra_cost, KHE_DRS_SOLN soln, int depth, */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return true if there is a soln in soln_trie that dominates soln, given   */
/*  (if with_tradeoff is true) that we have already paid extra_cost for the  */
/*  solutions in soln_trie.                                                  */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,
  bool with_tradeoff, KHE_COST extra_cost, KHE_DRS_SOLN soln, int depth,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;  bool also_zero;
  if( soln_trie == NULL )
  {
    /* case (1); no solns in soln_trie, so none dominate */
    return false;
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); the only possible dominator is soln_trie->soln */
    return KheDrsSolnDominates(soln_trie->soln, with_tradeoff,
      extra_cost, soln, depth, drs);
  }
  else
  {
    /* case (3); find the dominating set for the current signature value */
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->sig, depth);
    KheDrsDomTestDominatingSet(dom_test, val, &from_val, &to_val, &also_zero);

    /* convert dominating set to dominating indexes actually present */
    from_val -= soln_trie->base;
    to_val -= soln_trie->base;
    if( from_val < 0 )
      from_val = 0;
    if( to_val >= HaArrayCount(soln_trie->children) )
      to_val = HaArrayCount(soln_trie->children) - 1;
    also_zero = also_zero && soln_trie->base == 0;

    /* search for dominating solutions at the dominating indexes */
    for( i = from_val;  i <= to_val;  i++ )
      if( KheDrsSolnTrieDominates(HaArray(soln_trie->children, i),
	    with_tradeoff, extra_cost, soln, depth + 1, drs) )
	return true;
    if( also_zero && KheDrsSolnTrieDominates(HaArray(soln_trie->children, 0),
	    with_tradeoff, extra_cost, soln, depth + 1, drs) )
	return true;

    if( with_tradeoff && dom_test->tradeoff_allowed )
    {
      /* if there is an extra value at right, convert that to a dominating */
      /* index and search for dominating solutions at that dominating index */
      if( KheSetHasExtraValueAtRight(val, from_val, to_val, also_zero, &i))
      {
	i -= soln_trie->base;
	if( i >= 0 && i < HaArrayCount(soln_trie->children) &&
	    KheDrsSolnTrieDominates(HaArray(soln_trie->children, i),
	      with_tradeoff, extra_cost + dom_test->tradeoff_cost,
	      soln, depth + 1, drs) )
	  return true;
      }

      /* if there is an extra value at left, convert that to a dominating */
      /* index and search for dominating solutions at that dominating index */
      if( KheSetHasExtraValueAtLeft(val, from_val, to_val, also_zero, &i))
      {
	/* extra value at left */
	i -= soln_trie->base;
	if( i >= 0 && i < HaArrayCount(soln_trie->children) &&
	    KheDrsSolnTrieDominates(HaArray(soln_trie->children, i),
	      with_tradeoff, extra_cost + dom_test->tradeoff_cost,
	      soln, depth + 1, drs) )
	  return true;
      }
    }

    /* all done, no luck */
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,         */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,         */
/*    int *count)                                                            */
/*                                                                           */
/*  Remove those elements of *soln_trie that are dominated by soln.  This    */
/*  can change *soln_trie.                                                   */
/*                                                                           */
/*****************************************************************************/

/* *** non-tradeoff version, no longer used
static void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *count)
{
  int val, from_val, to_val, i;  bool also_zero;  KHE_DRS_DOM_TEST dom_test;
  KHE_DRS_SOLN_TRIE child_trie;
  if( *soln_trie == NULL )
  {
    ** case (1); there is nothing to remove **
  }
  else if( (*soln_trie)->soln != NULL )
  {
    ** case (2); if (*soln_trie)->soln is dominated, remove whole trie **
    if( KheDrsSolnDominates(soln, (*soln_trie)->soln, depth) )
    {
      KheDrsSolnTrieFree(soln_trie, true, drs);
      (*count)--;
    }
  }
  else
  {
    ** case (3); find the dominated set for the current signature value **
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->sig, depth);
    KheDrsDomTestDominatedSet(dom_test, val, &from_val, &to_val, &also_zero);

    ** convert dominated set to dominated indexes **
    from_val -= (*soln_trie)->base;
    to_val -= (*soln_trie)->base;
    if( from_val < 0 )
      from_val = 0;
    if( to_val >= HaArrayCount((*soln_trie)->children) )
      to_val = HaArrayCount((*soln_trie)->children) - 1;
    also_zero = also_zero && (*soln_trie)->base == 0;

    ** remove dominated solutions at the dominated indexes **
    for( i = from_val;  i <= to_val;  i++ )
      KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, i), soln,
	depth + 1, drs, count);
    if( also_zero )
      KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, 0), soln,
	depth + 1, drs, count);

    ** remove trailing NULL children **
    while( HaArrayCount((*soln_trie)->children) > 0 &&
	HaArrayLast((*soln_trie)->children) == NULL )
      HaArrayDeleteLast((*soln_trie)->children);

    if( HaArrayCount((*soln_trie)->children) == 0 )
    {
      ** if children array is empty then the trie is empty, so delete it **
      KheDrsSolnTrieFree(soln_trie, true, drs);
    }
    else
    {
      ** if children array is non-empty, remove initial NULL children **
      HaArrayForEach((*soln_trie)->children, child_trie, i)
	if( child_trie != NULL )
	  break;
      if( i > 0 )
      {
	HaArrayShiftLeft((*soln_trie)->children, i);
        (*soln_trie)->base += i;
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,         */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_COST extra_cost, int depth, */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int *count)                           */
/*                                                                           */
/*  Remove those elements of *soln_trie that are dominated by soln.  If      */
/*  with_tradeoff is true, we have already paid extra_cost for soln.  This   */
/*  can change *soln_trie.                                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_COST extra_cost, int depth,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int *count)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  bool also_zero;  KHE_DRS_SOLN_TRIE child_trie;
  if( *soln_trie == NULL )
  {
    /* case (1); there is nothing to remove */
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2); if (*soln_trie)->soln is dominated, remove whole trie */
    if( KheDrsSolnDominates(soln, with_tradeoff, extra_cost,
	  (*soln_trie)->soln, depth, drs) )
    {
      KheDrsSolnTrieFree(soln_trie, true, drs);
      (*count)--;
    }
  }
  else
  {
    /* case (3); remove dominated solns among the children */
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->sig, depth);
    KheDrsDomTestDominatedSet(dom_test, val, &from_val, &to_val, &also_zero);

    /* convert dominated set to dominated indexes */
    from_val -= (*soln_trie)->base;
    to_val -= (*soln_trie)->base;
    if( from_val < 0 )
      from_val = 0;
    if( to_val >= HaArrayCount((*soln_trie)->children) )
      to_val = HaArrayCount((*soln_trie)->children) - 1;
    also_zero = also_zero && (*soln_trie)->base == 0;

    /* remove dominated solutions at the dominated indexes */
    for( i = from_val;  i <= to_val;  i++ )
      KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, i),
	soln, with_tradeoff, extra_cost, depth + 1, drs, count);
    if( also_zero )
      KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, 0),
	soln, with_tradeoff, extra_cost, depth + 1, drs, count);

    if( with_tradeoff && dom_test->tradeoff_allowed )
    {
      /* if there is an extra value at right, convert that to a dominated */
      /* index and search for dominated solutions at that dominated index */
      if( KheSetHasExtraValueAtRight(val, from_val, to_val, also_zero, &i) )
      {
	i -= (*soln_trie)->base;
	if( i >= 0 && i < HaArrayCount((*soln_trie)->children) )
	  KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, i),
	    soln, with_tradeoff, extra_cost + dom_test->tradeoff_cost,
	    depth + 1, drs, count);
      }

      /* if there is an extra value at left, convert that to a dominated */
      /* index and search for dominated solutions at that dominated index */
      if( KheSetHasExtraValueAtLeft(val, from_val, to_val, also_zero, &i) )
      {
	i -= (*soln_trie)->base;
	if( i >= 0 && i < HaArrayCount((*soln_trie)->children) )
	  KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, i),
	    soln, with_tradeoff, extra_cost + dom_test->tradeoff_cost,
	    depth + 1, drs, count);
      }
    }

    /* remove trailing NULL children */
    while( HaArrayCount((*soln_trie)->children) > 0 &&
	HaArrayLast((*soln_trie)->children) == NULL )
      HaArrayDeleteLast((*soln_trie)->children);

    if( HaArrayCount((*soln_trie)->children) == 0 )
    {
      /* the children array is empty so the trie is empty; delete it */
      KheDrsSolnTrieFree(soln_trie, true, drs);
    }
    else
    {
      /* the children array is non-empty, so remove initial NULL children */
      HaArrayForEach((*soln_trie)->children, child_trie, i)
	if( child_trie != NULL )
	  break;
      if( i > 0 )
      {
	HaArrayShiftLeft((*soln_trie)->children, i);
        (*soln_trie)->base += i;
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieAddSoln(KHE_DRS_SOLN_TRIE *soln_trie,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *count)        */
/*                                                                           */
/*  Carry out a strong dominance insert of soln into *soln_trie.             */
/*                                                                           */
/*****************************************************************************/

/* *** non-tradeoff version, no longer used
static void KheDrsSolnTrieAddSoln(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *count)
{
  if( DEBUG6 )
    fprintf(stderr, "[ KheDrsSolnTrieAddSoln(soln for %s)\n",
      KheDrsDayId(soln->day));
  if( KheDrsSolnTrieDominates(*soln_trie, soln, 0, drs) )
  {
    ** soln is dominated by some other soln, so delete soln **
    if( DEBUG6 )
      fprintf(stderr, "  trie deleting soln for %s\n",
	KheDrsDayId(soln->day));
    KheDrsSolnFree(soln, drs);
  }
  else
  {
    ** remove solns that soln dominates, and insert soln **
    if( DEBUG6 )
      fprintf(stderr, "  trie inserting soln for %s\n",
	KheDrsDayId(soln->day));
    KheDrsSolnTrieRemoveDominated(soln_trie, soln, 0, drs, count);
    KheDrsSolnTrieAdd(soln_trie, soln, 0, drs);
    (*count)++;
  }
  if( DEBUG6 )
    fprintf(stderr, "] KheDrsSolnTrieAddSoln returning\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieAddSoln(KHE_DRS_SOLN_TRIE *soln_trie,                 */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int *count)                                                            */
/*                                                                           */
/*  Carry out a strong dominance insert of soln into *soln_trie.  Use        */
/*  tradeoff dominance if with_tradeoff is true.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieAddSoln(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *count)
{
  if( DEBUG6 )
    fprintf(stderr, "[ KheDrsSolnTrieAddSoln(soln for %s)\n",
      KheDrsDayId(soln->day));
  if( KheDrsSolnTrieDominates(*soln_trie, with_tradeoff, 0, soln, 0, drs) )
  {
    /* soln is dominated by some other soln, so delete soln */
    if( DEBUG6 )
      fprintf(stderr, "  trie deleting soln for %s\n",
	KheDrsDayId(soln->day));
    KheDrsSolnFree(soln, drs);
  }
  else
  {
    /* remove solns that soln dominates, and insert soln */
    if( DEBUG6 )
      fprintf(stderr, "  trie inserting soln for %s\n",
	KheDrsDayId(soln->day));
    KheDrsSolnTrieRemoveDominated(soln_trie, soln, with_tradeoff, 0, 0, drs,
      count);
    KheDrsSolnTrieAdd(soln_trie, soln, 0, drs);
    (*count)++;
  }
  if( DEBUG6 )
    fprintf(stderr, "] KheDrsSolnTrieAddSoln returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieExtend(KHE_DRS_SOLN_TRIE prev_soln_trie,              */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Extend the solutions in prev_soln_trie to next_day.                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieExtend(KHE_DRS_SOLN_TRIE prev_soln_trie,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( prev_soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( prev_soln_trie->soln != NULL )
  {
    /* case (2); extend soln_trie->soln */
    KheDrsSolnExtend(prev_soln_trie->soln, next_day, drs);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(prev_soln_trie->children, child_trie, i)
      KheDrsSolnTrieExtend(child_trie, next_day, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieAddAll(KHE_DRS_SOLN_TRIE soln_trie,                   */
/*    KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Add each solution from soln_trie to to_ssp, without copying it or        */
/*  deleting it from soln_trie.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieAddAll(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); add soln_trie->soln to to_ssp */
    KheDrsSolnSetPartAddSoln(to_ssp, soln_trie->soln, drs);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieAddAll(child_trie, to_ssp, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieGather(KHE_DRS_SOLN_TRIE soln_trie,                   */
/*    KHE_DRS_SOLN_SET soln_set)                                             */
/*                                                                           */
/*  Gather the solutions of soln_trie into soln_set->sorted_solns, without   */
/*  disturbing soln_trie.                                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieGather(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); add soln_trie->soln to soln_set->sorted_solns */
    HaArrayAddLast(soln_set->sorted_solns, soln_trie->soln);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieGather(child_trie, soln_set);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieExtractFinalSoln(KHE_DRS_SOLN_TRIE soln_trie,         */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Extract the unique solution (if any) from soln_trie.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieExtractFinalSoln(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); extract soln_trie->soln */
    HnAssert(*final_soln == NULL,
      "KheDrsSolnTrieExtractFinalSoln internal error 1");
    *final_soln = soln_trie->soln;
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieExtractFinalSoln(child_trie, final_soln);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieInitNullStats(KHE_DRS_SOLN_TRIE soln_trie,            */
/*    int *node_count, int *init_null_count)                                 */
/*                                                                           */
/*  Find the number of initial NULL children.                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieInitNullStats(KHE_DRS_SOLN_TRIE soln_trie,
  int *node_count, int *init_null_count)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); just a node, no initial NULL children */
    *node_count += 1;
  }
  else
  {
    /* case (3); count initial NULL children here in soln_trie */
    *node_count += 1;
    HaArrayForEach(soln_trie->children, child_trie, i)
    {
      if( child_trie != NULL )
	break;
      *init_null_count += 1;
    }

    /* do the job in the children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieInitNullStats(child_trie, node_count, init_null_count);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDebugNullInitialChildren(KHE_DRS_SOLN_TRIE soln_trie) */
/*                                                                           */
/*  Debug print of the number of NULL initial children in soln_trie.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDebugNullInitialChildren(KHE_DRS_SOLN_TRIE soln_trie)
{
  int node_count, init_null_count;
  node_count = init_null_count = 0;
  KheDrsSolnTrieInitNullStats(soln_trie, &node_count, &init_null_count);
  fprintf(stderr, "  initial NULL children: %d nodes, %d count, average %.1f\n",
    node_count, init_null_count, node_count == 0 ? 0.0 :
    (float) init_null_count / (float) node_count);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDoDebug(KHE_DRS_SOLN_TRIE soln_trie, int index,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Carry out the recursive part of KheDrsSolnTrieDebug.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDoDebug(KHE_DRS_SOLN_TRIE soln_trie, int index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); soln_trie->soln */
    fprintf(fp, "%*s%d: ", indent, "", index);
    KheDrsSolnDebug(soln_trie->soln, drs, verbosity, indent, fp);
  }
  else
  {
    /* case (3); handle children */
    fprintf(fp, "%*s[ %d (base %d):\n", indent, "", index, soln_trie->base);
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieDoDebug(child_trie, i, drs, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDebug(KHE_DRS_SOLN_TRIE soln_trie,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of soln_trie onto fp with the given verbosity and indent.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDebug(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ SolnTrie\n", indent, "");
    if( soln_trie == NULL )
    {
      /* case (1); do nothing */
    }
    else if( soln_trie->soln != NULL )
    {
      /* case (2); soln_trie->soln */
      KheDrsSolnDebug(soln_trie->soln, drs, verbosity, indent + 2, fp);
    }
    else
    {
      /* case (3); handle children */
      fprintf(fp, "%*s  (base %d):\n", indent, "", soln_trie->base);
      HaArrayForEach(soln_trie->children, child_trie, i)
	KheDrsSolnTrieDoDebug(child_trie, i, drs, verbosity, indent + 2, fp);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_INDEXED_SOLN_SET"                                     */
/*                                                                           */
/*  Invariant - base is undefined when soln_lists is empty; increment is     */
/*  always well-defined (and constant).                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_INDEXED_SOLN_SET KheDrsIndexedSolnSetMake(                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new, empty indexed soln set.                                      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_INDEXED_SOLN_SET KheDrsIndexedSolnSetMake(
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_INDEXED_SOLN_SET res;

  /* make the object */
  if( HaArrayCount(drs->free_indexed_soln_sets) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_indexed_soln_sets);
    HaArrayClear(res->soln_lists);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->soln_lists, drs->arena);
  }

  /* initialize the base and increment */
  res->base = 0;
  HnAssert(drs->increment > 0, "KheDrsIndexedSolnSetMake internal error");
  res->increment = drs->increment;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedDrsSolnSetFree(KHE_DRS_INDEXED_SOLN_SET iss,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free iss.                                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetFree(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;
  HaArrayForEach(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
      KheDrsSolnListFree(soln_list, drs);
  HaArrayClear(iss->soln_lists);
  HaArrayAddLast(drs->free_indexed_soln_sets, iss);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsIndexedSolnSetDominates(KHE_DRS_INDEXED_SOLN_SET iss,         */
/*    bool with_tradeoff, KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Return true if iss contains a solution that dominates soln.              */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsIndexedSolnSetDominates(KHE_DRS_INDEXED_SOLN_SET iss,
  bool with_tradeoff, KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, pos;  KHE_DRS_SOLN_LIST soln_list;
  if( HaArrayCount(iss->soln_lists) == 0 )
    return false;
  else
  {
    pos = (soln->cost - iss->base) / iss->increment;
    if( pos >= HaArrayCount(iss->soln_lists) )
      pos = HaArrayCount(iss->soln_lists) - 1;
    for( i = 0;  i <= pos;  i++ )
    {
      soln_list = HaArray(iss->soln_lists, i);
      if( soln_list != NULL &&
	  KheDrsSolnListDominates(soln_list, soln, with_tradeoff, drs) )
	return true;
    }
  }
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetRemoveDominated(KHE_DRS_INDEXED_SOLN_SET iss,   */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int *count)                                                            */
/*                                                                           */
/*  Remove from iss all solutions that are dominated by soln.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetRemoveDominated(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *count)
{
  int i, pos;  KHE_DRS_SOLN_LIST soln_list;
  if( HaArrayCount(iss->soln_lists) > 0 )
  {
    pos = (soln->cost - iss->base) / iss->increment;
    if( pos < 0 )
      pos = 0;
    for( i = pos;  i < HaArrayCount(iss->soln_lists);  i++ )
    {
      soln_list = HaArray(iss->soln_lists, i);
      if( soln_list != NULL )
	KheDrsSolnListRemoveDominated(soln_list, soln, with_tradeoff, drs,
          count);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetAdd(KHE_DRS_INDEXED_SOLN_SET iss,               */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)                    */
/*                                                                           */
/*  Add soln to iss, assuming all the dominance testing is done.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetAdd(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST soln_list;  int shift, pos;

  /* make sure the array contains soln_list, ready to receive soln */
  if( HaArrayCount(iss->soln_lists) == 0 )
  {
    /* array is empty, so set the base and add a soln list for soln */
    iss->base = soln->cost;
    soln_list = KheDrsSolnListMake(drs);
    HaArrayAddLast(iss->soln_lists, soln_list);
  }
  else if( iss->base > soln->cost )
  {
    /* shift the array right, reset the base, and add a soln list for soln */
    shift = (iss->base - soln->cost) / iss->increment;
    HnAssert(shift > 0, "KheDrsIndexedSolnSetAddSoln internal error");
    HaArrayShiftRight(iss->soln_lists, shift, NULL);
    iss->base = soln->cost;
    soln_list = KheDrsSolnListMake(drs);
    HaArrayPut(iss->soln_lists, 0, soln_list);
  }
  else
  {
    /* no shift or base reset, but may need to extend the array to the right */
    pos = (soln->cost - iss->base) / iss->increment;
    HaArrayFill(iss->soln_lists, pos + 1, NULL);
    soln_list = HaArray(iss->soln_lists, pos);
    if( soln_list == NULL )
    {
      soln_list = KheDrsSolnListMake(drs);
      HaArrayPut(iss->soln_lists, pos, soln_list);
    }
  }

  /* add soln to soln_list */
  HaArrayAddLast(soln_list->solns, soln);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetAddSoln(KHE_DRS_INDEXED_SOLN_SET iss,           */
/*    KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int *count)                                                            */
/*                                                                           */
/*  Add soln to iss, including tradeoff dominance testing.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetAddSoln(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_SOLN soln, bool with_tradeoff, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *count)
{
  if( KheDrsIndexedSolnSetDominates(iss, with_tradeoff, soln, drs) )
  {
    /* soln is dominated by some other soln, so delete soln */
    KheDrsSolnFree(soln, drs);
  }
  else
  {
    /* remove solns that soln dominates, and insert soln */
    KheDrsIndexedSolnSetRemoveDominated(iss, soln, with_tradeoff, drs, count);
    KheDrsIndexedSolnSetAdd(iss, soln, drs);
    (*count)++;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetExtend(KHE_DRS_INDEXED_SOLN_SET iss,            */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Extend iss.                                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetExtend(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;
  HaArrayForEach(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
      KheDrsSolnListExtend(soln_list, next_day, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetAddAll(KHE_DRS_INDEXED_SOLN_SET iss,            */
/*    KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Add all the elements of iss to to_ssp.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetAddAll(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;
  HaArrayForEach(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
      KheDrsSolnListAddAll(soln_list, to_ssp, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetGather(KHE_DRS_INDEXED_SOLN_SET iss,            */
/*    KHE_DRS_SOLN_SET soln_set)                                             */
/*                                                                           */
/*  Gather the solutions of iss into soln_set->sorted_solns.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetGather(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;
  HaArrayForEach(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
      KheDrsSolnListGather(soln_list, soln_set);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsIndexedSolnSetFirstSoln(KHE_DRS_INDEXED_SOLN_SET iss) */
/*                                                                           */
/*  Return the first soln of iss, or NULL if none.                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsIndexedSolnSetFirstSoln(KHE_DRS_INDEXED_SOLN_SET iss)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;  KHE_DRS_SOLN res;
  HaArrayForEach(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
    {
      res = KheDrsSolnListFirstSoln(soln_list);
      if( res != NULL )
	return res;
    }
  return NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsIndexedSolnSetLastSoln(KHE_DRS_INDEXED_SOLN_SET iss)  */
/*                                                                           */
/*  Return the last soln of iss, or NULL if none.                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsIndexedSolnSetLastSoln(KHE_DRS_INDEXED_SOLN_SET iss)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;  KHE_DRS_SOLN res;
  HaArrayForEachReverse(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
    {
      res = KheDrsSolnListFirstSoln(soln_list);  /* not an error */
      if( res != NULL )
	return res;
    }
  return NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetExtractFinalSoln(KHE_DRS_INDEXED_SOLN_SET iss,  */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Set *final_soln to the unique final solution stored in iss.  It will     */
/*  have already been initialized to NULL, and it should stay that way if    */
/*  iss is empty.                                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetExtractFinalSoln(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;
  HaArrayForEach(iss->soln_lists, soln_list, i)
    if( soln_list != NULL )
      KheDrsSolnListExtractFinalSoln(soln_list, final_soln);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsIndexedSolnSetDebug(KHE_DRS_INDEXED_SOLN_SET iss,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of iss onto fp with the given verbosity and index.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsIndexedSolnSetDebug(KHE_DRS_INDEXED_SOLN_SET iss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN_LIST soln_list;  int i;  KHE_DRS_SOLN soln;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ IndexSolnSet(base %.5f, increment %.5f)\n",
      indent, "", KheCostShow(iss->base), KheCostShow(iss->increment));
    HaArrayForEach(iss->soln_lists, soln_list, i)
    {
      fprintf(fp, "%*s  [%.5f .. %.5f] ", indent, "",
	KheCostShow(iss->base + i * iss->increment),
	KheCostShow(iss->base + (i + 1) * iss->increment));
      if( soln_list == NULL )
	fprintf(fp, "%3s", "-");
      else if( HaArrayCount(soln_list->solns) == 0 )
	fprintf(fp, "%3d", 0);
      else
      {
	fprintf(fp, "%3d: ", HaArrayCount(soln_list->solns));
	soln = HaArrayFirst(soln_list->solns);
	KheDrsSolnSignatureDebug(soln, fp);
      }
      fprintf(fp, "\n");
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_SET_PART"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_SET_PART KheDrsSolnSetPartMake(KHE_DRS_DOM_KIND dom_kind,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a new KHE_DRS_SOLN_SET_PART object.                               */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_SET_PART KheDrsSolnSetPartMake(KHE_DRS_DOM_KIND dom_kind,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_SET_PART res;
  if( HaArrayCount(drs->free_soln_set_parts) > 0 )
    res = HaArrayLastAndDelete(drs->free_soln_set_parts);
  else
    HaMake(res, drs->arena);
  res->count = 0;
  res->dom_kind = dom_kind;
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      res->u.none_dom_set = KheDrsSolnListMake(drs);
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableInit(res->u.weak_dom_set,
	&KheDrsSolnSignatureFullHashUntyped,
	&KheDrsSolnSignatureFullEqualUntyped,
	&KheDrsSolnSignatureDebugUntyped, drs->arena);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableInit(res->u.medium_dom_set,
	&KheDrsSolnListSignaturePartialHashUntyped,
	&KheDrsSolnListSignaturePartialEqualUntyped,
	&KheDrsSolnListSignatureDebugUntyped, drs->arena);
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      res->u.strong_dom_set = KheDrsSolnListMake(drs);
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:

      res->u.trie_dom_set = NULL;  /* represents the empty set */
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:

      res->u.indexed_dom_set = KheDrsIndexedSolnSetMake(drs);
      break;

    default:

      HnAbort("KheDrsSolnSetPartMake internal error: dom_kind %d\n", dom_kind);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartFree(KHE_DRS_SOLN_SET_PART ssp,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ssp and everything in it.                                           */
/*                                                                           */
/*  Most of the objects go onto free lists in drs for re-use later.  The     */
/*  exceptions are HpTable objects; they return to the arena they came from. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartFree(KHE_DRS_SOLN_SET_PART ssp,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int pos;  KHE_DRS_SOLN soln;  KHE_DRS_SOLN_LIST ns;
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListFree(ssp->u.none_dom_set, drs);
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(ssp->u.weak_dom_set, soln, pos)
	KheDrsSolnFree(soln, drs);
      HpTableFree(ssp->u.weak_dom_set);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(ssp->u.medium_dom_set, ns, pos)
	KheDrsSolnListFree(ns, drs);
      HpTableFree(ssp->u.medium_dom_set);
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      KheDrsSolnListFree(ssp->u.strong_dom_set, drs);
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:

      if( DEBUG26 )
	KheDrsSolnTrieDebugNullInitialChildren(ssp->u.trie_dom_set);
      KheDrsSolnTrieFree(&ssp->u.trie_dom_set, true, drs);
      HnAssert(ssp->u.trie_dom_set == NULL,
	"KheDrsSolnSetPartFree internal error");
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:

      KheDrsIndexedSolnSetFree(ssp->u.indexed_dom_set, drs);
      break;

    default:

      HnAbort("KheDrsSolnSetPartFree internal error: unknown dom_kind (%d)\n",
	ssp->dom_kind);
  }
  HaArrayAddLast(drs->free_soln_set_parts, ssp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartAddSoln(KHE_DRS_SOLN_SET_PART ssp,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)                    */
/*                                                                           */
/*  Add soln to ssp, using whatever kind of dominance ssp is set up for.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartAddSoln(KHE_DRS_SOLN_SET_PART ssp,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN other_soln;  KHE_DRS_SOLN_LIST soln_set, other_soln_set;
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      /* no dominance testing, just add soln */
      HaArrayAddLast(ssp->u.none_dom_set->solns, soln);
      ssp->count++;
      break;

    case KHE_DRS_DOM_WEAK:

      /* do equivalence testing */
      if( HpTableAddUnique(ssp->u.weak_dom_set, (void *) soln, soln,other_soln))
	ssp->count++;
      else
      {
	/* ensure other_soln is the better of other_soln and soln */
	if( soln->cost < other_soln->cost )
	  KheDrsSolnOverWrite(other_soln, soln);

	/* free soln */
	KheDrsSolnFree(soln, drs);
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      /* make soln_set holding just soln and try adding it to medium_dom_set */
      soln_set = KheDrsSolnListMake(drs);
      HaArrayAddLast(soln_set->solns, soln);
      if( HpTableAddUnique(ssp->u.medium_dom_set, (void *) soln_set, soln_set,
	    other_soln_set) )
      {
	/* soln_set added, so that's all that's needed */
	ssp->count++;
      }
      else
      {
	/* remove soln_set and do strong dom. of soln with other_soln_set */
	HaArrayClear(soln_set->solns);
	KheDrsSolnListFree(soln_set, drs);
        KheDrsSolnListAddSoln(other_soln_set, soln, false, drs, &ssp->count);
      }
      break;

    case KHE_DRS_DOM_STRONG:
    
      /* carry out strong dominance test between soln_set and soln */
      KheDrsSolnListAddSoln(ssp->u.strong_dom_set, soln, false,drs,&ssp->count);
      /* ***
      KheDrsSolnListAddSoln(ssp->u.strong_dom_set, soln, drs,&ssp->count);
      *** */
      break;

    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:
    
      /* strong dominance test between soln_set and soln, with tradeoff  */
      KheDrsSolnListAddSoln(ssp->u.strong_dom_set, soln, true, drs,&ssp->count);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieAddSoln(&ssp->u.trie_dom_set, soln, false, drs,&ssp->count);
      /* ***
      KheDrsSolnTrieAddSoln(&ssp->u.trie_dom_set, soln, drs, &ssp->count);
      *** */
      break;
    
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:

      /* trie dominance test between soln_set and soln, with tradeoff  */
      KheDrsSolnTrieAddSoln(&ssp->u.trie_dom_set, soln, true, drs, &ssp->count);
      break;
    
    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:

      KheDrsIndexedSolnSetAddSoln(ssp->u.indexed_dom_set, soln, true, drs,
	&ssp->count);
      break;

    default:

      HnAbort("KheDrsSolnSetPartAddSoln internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartExtend(KHE_DRS_SOLN_SET_PART prev_ssp,             */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Extend prev_ssp by adding tasks from next_day.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartExtend(KHE_DRS_SOLN_SET_PART prev_ssp,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN prev_soln;  int i;  KHE_DRS_SOLN_LIST prev_soln_list;
  switch( prev_ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListExtend(prev_ssp->u.none_dom_set, next_day, drs);
      /* ***
      HaArrayForEach(prev_ssp->u.none_dom_set->solns, prev_soln, j)
	KheDrsSolnExtend(prev_soln, next_day, drs);
      *** */
      break;

    case KHE_DRS_DOM_WEAK:

      if( DEBUG5 )
      {
	float probes;  int table_size;
	probes = HpTableProbeLength(prev_ssp->u.weak_dom_set);
	table_size = HpTableSize(prev_ssp->u.weak_dom_set);
	fprintf(stderr, "  hash table: count %d, size %d, probes %.2f\n",
	  prev_ssp->count, table_size, probes);
	if( probes > 14.0 )
	  HpTableDebug(prev_ssp->u.weak_dom_set, 2, stderr);
      }
      HpTableForEachValue(prev_ssp->u.weak_dom_set, prev_soln, i)
	KheDrsSolnExtend(prev_soln, next_day, drs);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(prev_ssp->u.medium_dom_set, prev_soln_list, i)
	KheDrsSolnListExtend(prev_soln_list, next_day, drs);
	/* ***
	HaArrayForEach(prev_soln_list->solns, prev_soln, j)
	  KheDrsSolnExtend(prev_soln, next_day, drs);
	*** */
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      KheDrsSolnListExtend(prev_ssp->u.strong_dom_set, next_day, drs);
      /* ***
      HaArrayForEach(prev_ssp->u.strong_dom_set->solns, prev_soln, j)
	KheDrsSolnExtend(prev_soln, next_day, drs);
      *** */
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      KheDrsSolnTrieExtend(prev_ssp->u.trie_dom_set, next_day, drs);
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      KheDrsIndexedSolnSetExtend(prev_ssp->u.indexed_dom_set, next_day, drs);
      break;
    
    default:

      HnAbort("KheDrsSolnSetPartExtend internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartMoveAll(KHE_DRS_SOLN_SET_PART from_ssp,            */
/*    KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Move every solution out of from_ssp and into to_ssp, leaving             */
/*  from_ssp well-defined but empty.                                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartMoveAll(KHE_DRS_SOLN_SET_PART from_ssp,
  KHE_DRS_SOLN_SET_PART to_ssp, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_list;
  switch( from_ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListAddAll(from_ssp->u.none_dom_set, to_ssp, drs);
      /* ***
      HaArrayForEach(from_ssp->u.none_dom_set->solns, soln, j)
	KheDrsSolnSetPartAddSoln(to_ssp, soln, drs);
      HaArrayClear(from_ssp->u.none_dom_set->solns);
      *** */
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(from_ssp->u.weak_dom_set, soln, i)
        KheDrsSolnSetPartAddSoln(to_ssp, soln, drs);
      HpTableClear(from_ssp->u.weak_dom_set);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(from_ssp->u.medium_dom_set, soln_list, i)
      {
	HaArrayForEach(soln_list->solns, soln, j)
	  KheDrsSolnSetPartAddSoln(to_ssp, soln, drs);
	HaArrayAddLast(drs->free_soln_lists, soln_list);
      }
      HpTableClear(from_ssp->u.medium_dom_set);
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      KheDrsSolnListAddAll(from_ssp->u.strong_dom_set, to_ssp, drs);
      /* ***
      HaArrayForEach(from_ssp->u.strong_dom_set->solns, soln, j)
	KheDrsSolnSetPartAddSoln(to_ssp, soln, drs);
      HaArrayClear(from_ssp->u.strong_dom_set->solns);
      *** */
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      /* add each from_ssp solution to to_ssp, then free (== clear) from_ssp */
      KheDrsSolnTrieAddAll(from_ssp->u.trie_dom_set, to_ssp, drs);
      KheDrsSolnTrieFree(&from_ssp->u.trie_dom_set, false, drs);
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      KheDrsIndexedSolnSetAddAll(from_ssp->u.indexed_dom_set, to_ssp, drs);
      KheDrsIndexedSolnSetFree(from_ssp->u.indexed_dom_set, drs);
      from_ssp->u.indexed_dom_set = NULL;
      break;
    
    default:

      HnAbort("KheDrsSolnSetPartMoveAll internal error");
      break;
  }
  from_ssp->count = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartGather(KHE_DRS_SOLN_SET_PART ssp,                  */
/*    KHE_DRS_SOLN_SET soln_set)                                             */
/*                                                                           */
/*  Gather the solutions of ssp into soln_set->sorted_solns, without         */
/*  disturbing ssp.                                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartGather(KHE_DRS_SOLN_SET_PART ssp,
  KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_list;
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListGather(ssp->u.none_dom_set, soln_set);
      /* ***
      HaArrayForEach(ssp->u.none_dom_set->solns, soln, j)
	HaArrayAddLast(soln_set->sorted_solns, soln);
      *** */
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(ssp->u.weak_dom_set, soln, i)
	HaArrayAddLast(soln_set->sorted_solns, soln);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(ssp->u.medium_dom_set, soln_list, i)
      {
	HaArrayForEach(soln_list->solns, soln, j)
	  HaArrayAddLast(soln_set->sorted_solns, soln);
      }
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      KheDrsSolnListGather(ssp->u.strong_dom_set, soln_set);
      /* ***
      HaArrayForEach(ssp->u.strong_dom_set->solns, soln, j)
	HaArrayAddLast(soln_set->sorted_solns, soln);
      *** */
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      KheDrsSolnTrieGather(ssp->u.trie_dom_set, soln_set);
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      KheDrsIndexedSolnSetGather(ssp->u.indexed_dom_set, soln_set);
      break;
    
    default:

      HnAbort("KheDrsSolnSetPartGather internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnSetPartFirstSoln(KHE_DRS_SOLN_SET_PART ssp)       */
/*                                                                           */
/*  Return the first solution of ssp, or NULL if none.                       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnSetPartFirstSoln(KHE_DRS_SOLN_SET_PART ssp)
{
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      return KheDrsSolnListFirstSoln(ssp->u.none_dom_set);

    case KHE_DRS_DOM_WEAK:

      /* stil l to do (possibly never) */
      return NULL;

    case KHE_DRS_DOM_MEDIUM:

      /* stil l to do (possibly never) */
      return NULL;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      return KheDrsSolnListFirstSoln(ssp->u.strong_dom_set);

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      /* stil l to do (possibly never) */
      return NULL;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      return KheDrsIndexedSolnSetFirstSoln(ssp->u.indexed_dom_set);
    
    default:

      HnAbort("KheDrsSolnSetPartFirstSoln internal error");
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnSetPartLastSoln(KHE_DRS_SOLN_SET_PART ssp)       */
/*                                                                           */
/*  Return the first solution of ssp, or NULL if none.                       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnSetPartLastSoln(KHE_DRS_SOLN_SET_PART ssp)
{
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      return KheDrsSolnListLastSoln(ssp->u.none_dom_set);

    case KHE_DRS_DOM_WEAK:

      /* stil l to do (possibly never) */
      return NULL;

    case KHE_DRS_DOM_MEDIUM:

      /* stil l to do (possibly never) */
      return NULL;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      return KheDrsSolnListLastSoln(ssp->u.strong_dom_set);

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      /* stil l to do (possibly never) */
      return NULL;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      return KheDrsIndexedSolnSetLastSoln(ssp->u.indexed_dom_set);
    
    default:

      HnAbort("KheDrsSolnSetPartLastSoln internal error");
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartExtractSingleSoln(KHE_DRS_SOLN_SET_PART ssp,       */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Assuming that ssp contains at most one solution, set *final_soln to      */
/*  that solution, or to NULL if there is no solution.                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartExtractSingleSoln(KHE_DRS_SOLN_SET_PART ssp,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_set;
  *final_soln = NULL;
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListExtractFinalSoln(ssp->u.none_dom_set, final_soln);
      /* ***
      HaArrayForEach(ssp->u.none_dom_set->solns, soln, j)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetPartExtractSingleSoln internal error 3");
	*final_soln = soln;
      }
      *** */
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(ssp->u.weak_dom_set, soln, i)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetPartExtractSingleSoln internal error 1");
	*final_soln = soln;
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(ssp->u.medium_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	{
	  HnAssert(*final_soln == NULL,
	    "KheDrsSolnSetPartExtractSingleSoln internal error 2");
	  *final_soln = soln;
	}
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      KheDrsSolnListExtractFinalSoln(ssp->u.strong_dom_set, final_soln);
      /* ***
      HaArrayForEach(ssp->u.strong_dom_set->solns, soln, j)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetPartExtractSingleSoln internal error 3");
	*final_soln = soln;
      }
      *** */
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      KheDrsSolnTrieExtractFinalSoln(ssp->u.trie_dom_set, final_soln);
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      KheDrsIndexedSolnSetExtractFinalSoln(ssp->u.indexed_dom_set, final_soln);
      break;
    
    default:

      HnAbort("KheDrsSolnSetPartExtractSingleSoln internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetPartDebug(KHE_DRS_SOLN_SET_PART ssp,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of ssp onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetPartDebug(KHE_DRS_SOLN_SET_PART ssp,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN soln;  int i;  KHE_DRS_SOLN_LIST soln_list;
  switch( ssp->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListDebug(ssp->u.none_dom_set, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_DOM_WEAK:

      fprintf(fp, "%*s[ WeakTable\n", indent, "");
      HpTableForEachValue(ssp->u.weak_dom_set, soln, i)
	KheDrsSolnDebug(soln, drs, verbosity, indent + 2, fp);
      fprintf(fp, "%*s]\n", indent, "");
      break;

    case KHE_DRS_DOM_MEDIUM:

      fprintf(fp, "%*s[ MediumTable\n", indent, "");
      HpTableForEachValue(ssp->u.medium_dom_set, soln_list, i)
	KheDrsSolnListDebug(soln_list, drs, verbosity, indent, fp);
      fprintf(fp, "%*s]\n", indent, "");
      break;

    case KHE_DRS_DOM_STRONG:
    case KHE_DRS_DOM_STRONG_WITH_TRADEOFF:

      KheDrsSolnListDebug(ssp->u.strong_dom_set, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_DOM_TRIE:
    case KHE_DRS_DOM_TRIE_WITH_TRADEOFF:
    
      KheDrsSolnTrieDebug(ssp->u.trie_dom_set, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_DOM_INDEXED_WITH_TRADEOFF:
    
      KheDrsIndexedSolnSetDebug(ssp->u.indexed_dom_set, drs, verbosity,
	indent, fp);
      break;
    
    default:

      HnAbort("KheDrsSolnSetPartDebug internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_SET"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetBeginCacheSegment(KHE_DRS_SOLN_SET soln_set,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Begin a cache segment, that is, if we are caching at all.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetBeginCacheSegment(KHE_DRS_SOLN_SET soln_set,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* actually there is nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetEndCacheSegment(KHE_DRS_SOLN_SET soln_set)             */
/*                                                                           */
/*  End a cache segment, that is, if we are caching at all.  This is done    */
/*  by deleting every solution from the cache and inserting it into the      */
/*  main table.                                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetEndCacheSegment(KHE_DRS_SOLN_SET soln_set,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( soln_set->cache != NULL )
    KheDrsSolnSetPartMoveAll(soln_set->cache, soln_set->main, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_SET KheDrsSolnSetMake(KHE_TIME_GROUP tg,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a new KHE_DRS_SOLN_SET object.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_SET KheDrsSolnSetMake(KHE_DRS_DOM_KIND main_dom_kind,
  bool use_cache, KHE_DRS_DOM_KIND cache_dom_kind,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_SET res;
  if( HaArrayCount(drs->free_soln_sets) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_sets);
#if TESTING
    HaArrayClear(res->sorted_solns);
#endif
  }
  else
  {
    HaMake(res, drs->arena);
#if TESTING
    HaArrayInit(res->sorted_solns, drs->arena);
#endif
  }
  res->made_count = 0;
  res->cache = (use_cache ? KheDrsSolnSetPartMake(cache_dom_kind, drs) : NULL);
  res->main = KheDrsSolnSetPartMake(main_dom_kind, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free dns and everything in it.                                           */
/*                                                                           */
/*  Most of the objects go onto free lists in drs for re-use later.  The     */
/*  exceptions are HpTable objects; they return to the arena they came from. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetFree(KHE_DRS_SOLN_SET soln_set,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( soln_set->cache != NULL )
    KheDrsSolnSetPartFree(soln_set->cache, drs);
  KheDrsSolnSetPartFree(soln_set->main, drs);
  HaArrayAddLast(drs->free_soln_sets, soln_set);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET soln_set, KHE_DRS_SOLN soln,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add soln to soln_set, into the cache if present, otherwise the main.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET soln_set, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  soln_set->made_count++;
  if( soln_set->cache != NULL )
    KheDrsSolnSetPartAddSoln(soln_set->cache, soln, drs);
  else
    KheDrsSolnSetPartAddSoln(soln_set->main, soln, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetGatherAndSort(KHE_DRS_SOLN_SET soln_set)               */
/*                                                                           */
/*  Gather and sort the solutions of soln_set.                               */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static int KheDrsSolnCmp(const void *t1, const void *t2)
{
  KHE_DRS_SOLN soln1 = * (KHE_DRS_SOLN *) t1;
  KHE_DRS_SOLN soln2 = * (KHE_DRS_SOLN *) t2;
  return KheCostCmp(soln1->cost, soln2->cost);
}

static void KheDrsSolnSetGatherAndSort(KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN soln;  int i;

  /* gather the solutions of soln_set into soln_set->sorted_solns */
  HaArrayClear(soln_set->sorted_solns);
  if( soln_set->cache != NULL )
    KheDrsSolnSetPartGather(soln_set->cache, soln_set);
  KheDrsSolnSetPartGather(soln_set->main, soln_set);

  /* sort the solutions by increasing cost */
  HaArraySort(soln_set->sorted_solns, &KheDrsSolnCmp);

  /* assign a rank to each solution */
  HaArrayForEach(soln_set->sorted_solns, soln, i)
    soln->sorted_rank = i + 1;
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetExtend(KHE_DRS_SOLN_SET prev_soln_set,                 */
/*    KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Extend prev_soln_set by adding tasks from next_day.                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetExtend(KHE_DRS_SOLN_SET prev_soln_set,
  KHE_DRS_DAY next_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( prev_soln_set->cache != NULL )
    KheDrsSolnSetPartExtend(prev_soln_set->cache, next_day, drs);
  KheDrsSolnSetPartExtend(prev_soln_set->main, next_day, drs);
#if TESTING
  KheDrsSolnSetGatherAndSort(next_day->soln_set);
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetExtractSingleSoln(KHE_DRS_SOLN_SET dns,                */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Assuming that dns contains at most one solution, set *final_soln to      */
/*  that solution, or to NULL if there is no solution.                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetExtractSingleSoln(KHE_DRS_SOLN_SET soln_set,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN soln1, soln2;
  if( soln_set->cache != NULL )
    KheDrsSolnSetPartExtractSingleSoln(soln_set->cache, &soln1);
  else
    soln1 = NULL;
  KheDrsSolnSetPartExtractSingleSoln(soln_set->main, &soln2);
  if( soln1 == NULL )
    *final_soln = soln2;
  else if( soln2 == NULL )
    *final_soln = soln1;
  else
  {
    HnAbort("KheDrsSolnSetExtractSingleSoln internal error");
    *final_soln = NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnSetFirstSoln(KHE_DRS_SOLN_SET soln_set)           */
/*                                                                           */
/*  Return the first solution of soln_set, whatever that is.                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnSetFirstSoln(KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN res;
  if( soln_set->cache != NULL )
  {
    res = KheDrsSolnSetPartFirstSoln(soln_set->cache);
    if( res != NULL )
      return res;
  }
  return KheDrsSolnSetPartFirstSoln(soln_set->main);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnSetLastSoln(KHE_DRS_SOLN_SET soln_set)           */
/*                                                                           */
/*  Return the first solution of soln_set, whatever that is.                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnSetLastSoln(KHE_DRS_SOLN_SET soln_set)
{
  KHE_DRS_SOLN res;
  res = KheDrsSolnSetPartLastSoln(soln_set->main);
  if( res != NULL )
    return res;
  else if( soln_set->cache != NULL )
    return KheDrsSolnSetPartLastSoln(soln_set->cache);
  else
    return NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSetKeptCount(KHE_DRS_SOLN_SET soln_set)                    */
/*                                                                           */
/*  Return the number of solutions now kept in soln_set.                     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSetKeptCount(KHE_DRS_SOLN_SET soln_set)
{
  int res;
  res = soln_set->main->count;
  if( soln_set->cache != NULL )
    res += soln_set->cache->count;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetDebug(KHE_DRS_SOLN_SET dns,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of dns onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetDebug(KHE_DRS_SOLN_SET soln_set,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  fprintf(fp, "%*s[ SolnSet(made %d)\n", indent, "", soln_set->made_count);
  if( soln_set->cache != NULL )
  {
    fprintf(fp, "%*sCache:\n", indent + 2, "");
    KheDrsSolnSetPartDebug(soln_set->cache, drs, verbosity, indent + 2, fp);
  }
  fprintf(fp, "%*sMain:\n", indent + 2, "");
  KheDrsSolnSetPartDebug(soln_set->main, drs, verbosity, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_PACKED_SOLN" (solutions in an easier format)          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN_DAY KhePackedSolnDayMakeEmpty(KHE_DRS_DAY day,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an empty packed solution day object for day.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_PACKED_SOLN_DAY KhePackedSolnDayMakeEmpty(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PACKED_SOLN_DAY res;
  if( HaArrayCount(drs->free_packed_soln_days) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_packed_soln_days);
    HaArrayClear(res->prev_tasks);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->prev_tasks, drs->arena);
  }
  res->day = day;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN_DAY KheDrsPackedSolnDayMakeFromSoln(                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)                    */
/*                                                                           */
/*  Make a packed solution day object from soln.                             */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_PACKED_SOLN_DAY KheDrsPackedSolnDayMakeFromSoln(
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_PACKED_SOLN_DAY res;
  res = KhePackedSolnDayMakeEmpty(soln->day, drs);
  HaArrayAppend(res->prev_tasks, soln->prev_tasks, i);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN KheDrsPackedSolnMake(KHE_COST cost,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new, empty packed solution object with this cost.                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_PACKED_SOLN KheDrsPackedSolnMake(KHE_COST cost,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PACKED_SOLN res;
  if( HaArrayCount(drs->free_packed_solns) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_packed_solns);
    HaArrayClear(res->days);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->days, drs->arena);
  }
  res->cost = cost;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPackedSolnDelete(KHE_DRS_PACKED_SOLN packed_soln,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete packed_soln.                                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPackedSolnDelete(KHE_DRS_PACKED_SOLN packed_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;
  HaArrayAppend(drs->free_packed_soln_days, packed_soln->days, i);
  HaArrayAddLast(drs->free_packed_solns, packed_soln);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN KheDrsPackedSolnBuildEmpty(                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Build an empty packed soln object for the open days and resources of     */
/*  drs, and cost KheSolnCost(drs->soln).                                    */
/*                                                                           */
/*  Implementation note.  This function is called early during opening,      */
/*  when only selected_resource_set and selected_day_ranges are defined.     */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_PACKED_SOLN KheDrsPackedSolnBuildEmpty(
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PACKED_SOLN res;  KHE_DRS_PACKED_SOLN_DAY rd;
  KHE_DRS_DAY_RANGE ddr;  KHE_DRS_DAY day;  int i, j, count;

  res = KheDrsPackedSolnMake(KheSolnCost(drs->soln), drs);
  count = KheResourceSetResourceCount(drs->selected_resource_set);
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      /* make a day for j */
      day = HaArray(drs->all_days, j);
      rd = KhePackedSolnDayMakeEmpty(day, drs);
      HaArrayFill(rd->prev_tasks, count, NULL);
      HaArrayAddLast(res->days, rd);
    }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_PACKED_SOLN KheDrsPackedSolnBuildFromSoln(KHE_DRS_SOLN soln,     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Build a packed soln object from soln.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_PACKED_SOLN KheDrsPackedSolnBuildFromSoln(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PACKED_SOLN res;  KHE_DRS_SOLN soln2;  KHE_DRS_PACKED_SOLN_DAY rd;
  int i, j;

  /* build the packed soln but with the days in reverse order */
  res = KheDrsPackedSolnMake(soln->cost, drs);
  for( soln2 = soln;  soln2->prev_soln != NULL;  soln2 = soln2->prev_soln )
  {
    rd = KheDrsPackedSolnDayMakeFromSoln(soln2, drs);
    HaArrayAddLast(res->days, rd);
  }

  /* reverse the order of the days and return */
  for( i = 0, j = HaArrayCount(res->days) - 1;  i < j;  i++, j-- )
    HaArraySwap(res->days, i, j, rd);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_ON_DAY KheDrsPackedSolnTaskOnDay(                           */
/*    KHE_DRS_PACKED_SOLN packed_soln, int open_day_index,                   */
/*    int open_resource_index)                                               */
/*                                                                           */
/*  Return the task on day object (possibly NULL) held in packed_soln for    */
/*  the given open day index and open resource index.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK_ON_DAY KheDrsPackedSolnTaskOnDay(
  KHE_DRS_PACKED_SOLN packed_soln, KHE_DRS_DAY open_day,
  KHE_DRS_RESOURCE open_resource)
{
  KHE_DRS_PACKED_SOLN_DAY rd;  int open_day_index, open_resource_index;
  open_day_index = open_day->open_day_index;
  HnAssert(0 <= open_day_index &&
    open_day_index < HaArrayCount(packed_soln->days),
    "KheDrsPackedSolnTaskOnDay internal error (day %d out of range 0 .. %d)",
    open_day_index, HaArrayCount(packed_soln->days) - 1);
  rd = HaArray(packed_soln->days, open_day_index);
  open_resource_index = open_resource->open_resource_index;
  HnAssert(0 <= open_resource_index &&
      open_resource_index < HaArrayCount(rd->prev_tasks),
   "KheDrsPackedSolnTaskOnDay internal error (resource %d out of range 0 .. %d",
    open_resource_index, HaArrayCount(rd->prev_tasks) - 1);
  return HaArray(rd->prev_tasks, open_resource_index);
}

/* *** old version
static KHE_DRS_TASK_ON_DAY KheDrsPackedSolnTaskOnDay(
  KHE_DRS_PACKED_SOLN packed_soln, int open_day_index, int open_resource_index)
{
  KHE_DRS_PACKED_SOLN_DAY rd;
  HnAssert(0 <= open_day_index &&
    open_day_index < HaArrayCount(packed_soln->days),
    "KheDrsPackedSolnTaskOnDay internal error (day %d out of range 0 .. %d)",
    open_day_index, HaArrayCount(packed_soln->days) - 1);
  rd = HaArray(packed_soln->days, open_day_index);
  HnAssert(0 <= open_resource_index &&
      open_resource_index < HaArrayCount(rd->prev_tasks),
   "KheDrsPackedSolnTaskOnDay internal error (resource %d out of range 0 .. %d",
    open_resource_index, HaArrayCount(rd->prev_tasks) - 1);
  return HaArray(rd->prev_tasks, open_resource_index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPackedSolnSetTaskOnDay(KHE_DRS_PACKED_SOLN packed_soln,       */
/*    int open_day_index, int open_resource_index, KHE_DRS_TASK_ON_DAY dtd)  */
/*                                                                           */
/*  Set the task on day object (possibly NULL) held in packed_soln for       */
/*  the given open day index and open resource index.                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPackedSolnSetTaskOnDay(KHE_DRS_PACKED_SOLN packed_soln,
  int open_day_index, int open_resource_index, KHE_DRS_TASK_ON_DAY dtd)
{
  KHE_DRS_PACKED_SOLN_DAY rd;
  HnAssert(0 <= open_day_index &&
    open_day_index < HaArrayCount(packed_soln->days),
    "KheDrsPackedSolnSetTaskOnDay internal error (day %d out of range 0 .. %d)",
    open_day_index, HaArrayCount(packed_soln->days) - 1);
  rd = HaArray(packed_soln->days, open_day_index);
  HnAssert(0 <= open_resource_index && open_resource_index <
    HaArrayCount(rd->prev_tasks), "KheDrsPackedSolnSetTaskOnDay "
    "internal error (resource %d out of range 0 .. %d)",
    open_resource_index, HaArrayCount(rd->prev_tasks) - 1);
  HaArrayPut(rd->prev_tasks, open_resource_index, dtd);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPackedSolnDebugResourceTimetable(KHE_DRS_RESOURCE dr,         */
/*    KHE_DRS_PACKED_SOLN packed_soln, int first_day_index, last_day_index,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of the full timetable of dr from first_day_index to          */
/*  last_day_index inclusive.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPackedSolnDebugResourceTimetable(KHE_DRS_RESOURCE dr,
  KHE_DRS_PACKED_SOLN packed_soln, int first_day_index, int last_day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE_ON_DAY drd;
  int i;  KHE_DRS_DAY day;  KHE_DRS_PACKED_SOLN_DAY rd;

  /* indent */
  fprintf(fp, "%*s", indent, "");

  /* header column */
  KhePrintCol(KheResourceId(dr->resource), true, fp);

  /* ordinary columns */
  for( i = first_day_index;  i <= last_day_index;  i++ )
  {
    day = HaArray(drs->all_days, i);
    if( day->open_day_index == -1 )
    {
      /* a closed day, get timetable from dr */
      drd = HaArray(dr->days, i);
      dtd = drd->closed_asst;
      if( dtd == NULL || dtd->task == NULL )
        KhePrintCol("", true, fp);
      else
	KhePrintCol(ShortName(KheTaskId(dtd->task), KheDrsDayId(day)),
	  KheDrsTaskOnDayIsLast(dtd), fp);
    }
    else
    {
      /* an open day whose assignment is known */
      rd = HaArray(packed_soln->days, day->open_day_index);
      HnAssert(0 <= dr->open_resource_index &&
	dr->open_resource_index < HaArrayCount(rd->prev_tasks),
	"KheDrsPackedSolnDebugResourceTimetable internal error 2");
      dtd = HaArray(rd->prev_tasks, dr->open_resource_index);
      if( dtd == NULL || dtd->task == NULL )
        KhePrintCol("", true, fp);
      else
	KhePrintCol(ShortName(KheTaskId(dtd->task), KheDrsDayId(day)),
	  KheDrsTaskOnDayIsLast(dtd), fp);
    }
  }

  /* final newline */
  fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPackedSolnDebugTimetable(KHE_DRS_PACKED_SOLN packed_soln,     */
/*    int first_day_index, int last_day_index,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of the timetables of the open resources from first_day_index */
/*  to last_day_index inclusive.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPackedSolnDebugTimetable(KHE_DRS_PACKED_SOLN packed_soln,
  int first_day_index, int last_day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  int i;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE dr;  char buff[200];

  /* header row and first rule */
  fprintf(fp, "%*s", indent, "");
  if( first_day_index == 0 )
    fprintf(fp, "%*.5f  ", COL_WIDTH - 2, KheCostShow(packed_soln->cost));
  else
    fprintf(fp, "%*s", COL_WIDTH, "");
  for( i = first_day_index;  i <= last_day_index;  i++ )
  {
    day = HaArray(drs->all_days, i);
    if( day->open_day_index != -1 )
    {
      snprintf(buff, 200, "*%s*", KheDrsDayId(day));
      KhePrintCol(buff, false, fp);
    }
    else
      KhePrintCol(KheDrsDayId(day), false, fp);
  }
  fprintf(fp, "\n");
  KhePrintRule(last_day_index - first_day_index + 1, indent, fp);

  /* resource rows and other rules */
  HaArrayForEach(drs->open_resources, dr, i)
  {
    KheDrsPackedSolnDebugResourceTimetable(dr, packed_soln,
      first_day_index, last_day_index, drs, verbosity, indent, fp);
    KhePrintRule(last_day_index - first_day_index + 1, indent, fp);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPackedSolnDebug(KHE_DRS_PACKED_SOLN packed_soln, char *str,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of packed_soln in timetable format.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsPackedSolnDebug(KHE_DRS_PACKED_SOLN packed_soln, char *str,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  int i, j;
  if( str != NULL && str[0] != '\0' )
    fprintf(fp, "%*s%s:\n", indent, "", str);
  for( i = 0;  i < HaArrayCount(drs->all_days) - 1;  i += COL_COUNT )
  {
    fprintf(fp, "\n");  /* even the first block has a preceding blank line */
    j = min(i + COL_COUNT - 1, HaArrayCount(drs->all_days) - 1);
    KheDrsPackedSolnDebugTimetable(packed_soln, i, j, drs, verbosity,
      indent, fp);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "Solving"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs, bool testing,      */
/*    int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,        */
/*    KHE_DRS_DOM_KIND cache_dom_kind, KHE_DRS_PACKED_SOLN *init_soln)       */
/*                                                                           */
/*  Open drs for solving, including setting *init_soln to the initial soln.  */
/*                                                                           */
/*  If testing is true, clear statistics variables for a new test.           */
/*                                                                           */
/*****************************************************************************/
static int KheDrsExprPostorderCmp(const void *t1, const void *t2);
static void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolveOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs, bool testing,
  int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,
  KHE_DRS_DOM_KIND cache_dom_kind, KHE_DRS_PACKED_SOLN *init_soln)
{
  KHE_DRS_DAY_RANGE ddr;  KHE_DRS_DAY day;  int i, j;  KHE_DRS_EXPR e;
  KHE_DRS_RESOURCE dr;  KHE_RESOURCE r;

  /* debug print */
  if( DEBUG9 )
  {
    fprintf(stderr, "[ KheDrsSolveOpen(drs, soln_limit ");
    if( soln_limit == 0 )
      fprintf(stderr, "(none)");
    else
      fprintf(stderr, "%d", soln_limit);
    fprintf(stderr, ", %s, %s, %s), cost %.5f\n",
      KheDomKindShow(main_dom_kind), bool_show(use_cache),
      KheDomKindShow(cache_dom_kind), KheCostShow(KheSolnCost(drs->soln)));
  }

  /* initialize testing fields */
#if TESTING
  if( testing )
  {
    HaArrayClear(drs->table_size_per_day);
    HaArrayClear(drs->running_time_per_day);
    KheTimerResetStartTime(drs->timer);
    HaArrayClear(drs->ancestor_freq);
  }
#endif

  /* initialize fields that vary with the solve */
  drs->solve_soln_limit = (soln_limit <= 0 ? INT_MAX : soln_limit);
  drs->solve_init_cost = drs->solve_start_cost = KheSolnCost(drs->soln);
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);

  /* open selected resources (grows drs->open_resources and drs->open_exprs) */
  *init_soln = KheDrsPackedSolnBuildEmpty(drs);
  for( i = 0; i < KheResourceSetResourceCount(drs->selected_resource_set); i++ )
  {
    r = KheResourceSetResource(drs->selected_resource_set, i);
    dr = HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
    KheDrsResourceOpen(dr, HaArrayCount(drs->open_resources), *init_soln, drs);
    HaArrayAddLast(drs->open_resources, dr);
  }

  /* open selected days (grows drs->open_days and drs->open_exprs) */
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      day = HaArray(drs->all_days, j);
      KheDrsDayOpen(day, ddr, HaArrayCount(drs->open_days), main_dom_kind,
	use_cache, cache_dom_kind, drs);
      HaArrayAddLast(drs->open_days, day);
    }

  /* open the task classes on the selected days (days must be open first) */
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      day = HaArray(drs->all_days, j);
      KheDrsDayOpenTaskClasses(day, ddr, drs);
    }

  /* sort drs->open_exprs by increasing postorder index, then open them */
  HaArraySort(drs->open_exprs, &KheDrsExprPostorderCmp);
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprOpen(e, drs);
  if( DEBUG9 )
    fprintf(stderr, "] KheDrsSolveOpen soln cost %.5f, solve_start_cost %.5f\n",
      KheCostShow(KheSolnCost(drs->soln)), KheCostShow(drs->solve_start_cost));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostCheckRerunCost(KHE_DYNAMIC_RESOURCE_SOLVER drs,       */
/*    int first, int last)                                                   */
/*                                                                           */
/*  Check the rerun costs of drs->all_cost_exprs[first .. last].  These      */
/*  are all derived from the same monitor.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostCheckRerunCost(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int first, int last)
{
  KHE_COST rerun_open_and_search_cost, rerun_open_and_close_cost;  int i;
  KHE_DRS_EXPR_COST ec;

  /* sum the costs of drs->all_cost_exprs[first .. last] */
  rerun_open_and_search_cost = rerun_open_and_close_cost = 0;
  for( i = first;  i <= last;  i++ )
  {
    ec = HaArray(drs->all_cost_exprs, i);
    rerun_open_and_search_cost += ec->rerun_open_and_search_cost;
    rerun_open_and_close_cost += ec->rerun_open_and_close_cost;
  }

  /* check the costs against the monitor costs */
  ec = HaArray(drs->all_cost_exprs, first);
  if( rerun_open_and_search_cost != KheMonitorCost(ec->monitor) )
    fprintf(stderr, "  open_and_search_cost %.5f, monitor cost %.5f (%s)\n",
      KheCostShow(rerun_open_and_search_cost),
      KheCostShow(KheMonitorCost(ec->monitor)), KheMonitorId(ec->monitor));
  if( rerun_open_and_close_cost != KheMonitorCost(ec->monitor) )
    fprintf(stderr, "  open_and_close_cost  %.5f, monitor cost %.5f (%s)\n",
      KheCostShow(rerun_open_and_close_cost),
      KheCostShow(KheMonitorCost(ec->monitor)), KheMonitorId(ec->monitor));

  /* ***
  switch( e->tag )
  {
    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostCheckRerunCost((KHE_DRS_EXPR_INT_SUM_COST) e);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostCheckRerunCost((KHE_DRS_EXPR_INT_SEQ_COST) e);
      break;

    default:

      ** do nothing **
      break;
  }
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs,                   */
/*    KHE_DRS_PACKED_SOLN soln, bool check_rerun_costs)                      */
/*                                                                           */
/*  Close drs for solving, using soln to assign tasks.                       */
/*                                                                           */
/*  If check_rerun_costs is true, this is part of a rerun and we want        */
/*  to check monitor costs at the end of the close.                          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_PACKED_SOLN soln, bool check_rerun_costs)
{
  KHE_DRS_DAY day;  int i, j;  KHE_DRS_EXPR e;  KHE_DRS_EXPR_COST e1, e2;
  KHE_DRS_RESOURCE dr;  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_PACKED_SOLN_DAY rd;
  KHE_MONITOR m;

  /* traverse soln, closing assigned tasks */
  HaArrayForEachReverse(soln->days, rd, i)
    HaArrayForEach(rd->prev_tasks, dtd, j)
      if( dtd != NULL )
      {
	dr = HaArray(drs->open_resources, j);
	KheDrsTaskClose(dtd->encl_dt, dr);
      }

  /* close the open days, including closing unassigned tasks */
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayClose(day, drs);

  /* close the open expressions */
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprClose(e, drs);

  /* close the open resources */
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceClose(dr);

  /* close drs */
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);

  /* optionally check rerun costs */
  if( check_rerun_costs )
  {
    fprintf(stderr, "[ KheDrsRerun checking costs:\n");
    for( i = 0;  i < HaArrayCount(drs->all_cost_exprs);  i = j )
    {
      e1 = HaArray(drs->all_cost_exprs, i);
      for( j = i + 1;  j < HaArrayCount(drs->all_cost_exprs);  j++ )
      {
	e2 = HaArray(drs->all_cost_exprs, j);
	if( e1->monitor != e2->monitor )
	  break;
      }
      KheDrsExprCostCheckRerunCost(drs, i, j - 1);
    }
    if( DEBUG24 )
    {
      for( i = 0;  i < KheSolnMonitorCount(drs->soln);  i++ )
      {
	m = KheSolnMonitor(drs->soln, i);
	if( KheMonitorTag(m) != KHE_GROUP_MONITOR_TAG &&
	    KheMonitorCost(m) >= KheCost(1, 0) )
	  fprintf(stderr, "  large monitor cost %.5f (%s)\n",
	    KheCostShow(KheMonitorCost(m)), KheMonitorId(m));
      }
    }
    fprintf(stderr, "]\n");
  }

  HnAssert(KheSolnCost(drs->soln) == soln->cost,
    "KheDrsSolveClose internal error: soln %.5f != packed %.5f",
    KheCostShow(KheSolnCost(drs->soln)), KheCostShow(soln->cost));
  /* *** this seems to be rubbish here
  HnAssert(KheSolnCost(drs->soln) <= drs->solve_start_cost,
    "KheDrsSolveClose internal error: soln %.5f > init %.5f",
    KheCostShow(KheSolnCost(drs->soln)), KheCostShow(drs->solve_init_cost));
  *** */
}


/* ***
static void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_MARK mark,
  bool testing_only)
{
  KHE_DRS_DAY day;  int i;  KHE_DRS_EXPR e;  KHE_DRS_SOLN soln, final_soln;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE dr;
  ** KHE_DRS_RESOURCE_ON_DAY drd; **

  ** extract final_soln **
  day = HaArrayLast(drs->open_days);
  final_soln = NULL;
  KheDrsSolnSetExtractSingleSoln(day->soln_set, &final_soln);
  if( DEBUG2 )
  {
    if( final_soln != NULL )
    {
      fprintf(stderr, "  final solution:\n");
      KheDrsSolnDebug(final_soln, drs, 1, 2, stderr);
    }
    else
      fprintf(stderr, "  no final solution\n");
  }

  if( testing_only || final_soln == NULL )
  {
    ** testing, or no final soln (no improvement); return to initial soln **
    KheMarkEnd(mark, true);
    HnAssert(KheSolnCost(drs->soln) == drs->solve_init_cost,
      "KheDynamicResourceSolveClose internal error 1");
  }
  else
  {
    ** improvement; traverse final solution, closing assigned tasks **
    for( soln = final_soln;  soln->prev_soln != NULL;  soln = soln->prev_soln )
      HaArrayForEach(soln->prev_tasks, dtd, i)
	if( dtd != NULL )
	{
	  dr = HaArray(drs->open_resources, i);
	  KheDrsTaskClose(dtd->encl_dt, dr);
	  ** ***
	  drd = KheDrsResourceOnDay(dr, soln->day);
	  KheDrsTaskClose(dtd->encl_dt, drd->encl_dr);
	  *** **
	}
    KheMarkEnd(mark, false);

#if RERUN
    ** make final checks, but do not abort if they fail **
    if( KheSolnCost(drs->soln) != drs->solve_start_cost )
      fprintf(stderr, "  KheDynamicResourceSolveClose soln %.5f, start %.5f\n",
	KheCostShow(KheSolnCost(drs->soln)),KheCostShow(drs->solve_start_cost));
    if( KheSolnCost(drs->soln) != final_soln->cost )
      fprintf(stderr, "  KheDynamicResourceSolveClose soln %.5f, final %.5f\n",
	KheCostShow(KheSolnCost(drs->soln)), KheCostShow(final_soln->cost));
    if( KheSolnCost(drs->soln) >= drs->solve_init_cost )
      fprintf(stderr, "  KheDynamicResourceSolveClose soln %.5f, init %.5f\n",
	KheCostShow(KheSolnCost(drs->soln)), KheCostShow(drs->solve_init_cost));

    ** if this is not a rerun, the next run will be, so make drs->rerun **
    if( drs->rerun == NULL )
      drs->rerun = KheDrsPackedSolnBuildFromSoln(final_soln, drs);
#else
    ** make final checks, aborting if they fail **
    HnAssert(KheSolnCost(drs->soln) == drs->solve_start_cost,
      "KheDynamicResourceSolveClose internal error 2 (soln %.5f, start %.5f)",
      KheCostShow(KheSolnCost(drs->soln)), KheCostShow(drs->solve_start_cost));
    HnAssert(KheSolnCost(drs->soln) == final_soln->cost,
      "KheDynamicResourceSolveClose internal error 3 (soln %.5f, final %.5f)",
      KheCostShow(KheSolnCost(drs->soln)), KheCostShow(final_soln->cost));
    HnAssert(KheSolnCost(drs->soln) < drs->solve_init_cost,
      "KheDynamicResourceSolveClose internal error 4 (soln %.5f, init %.5f)",
      KheCostShow(KheSolnCost(drs->soln)), KheCostShow(drs->solve_init_cost));
#endif
  }

  ** close the open days, including closing unassigned tasks **
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayClose(day, drs);

  ** close the open expressions **
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprClose(e, drs);

  ** close the open resources **
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceClose(dr);

  ** close drs **
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,                        */
/*    int soln_limit, KHE_DRS_DOM_KIND dom_kind, bool testing_only)          */
/*                                                                           */
/*  Carry out a solve, including opening, solving proper, and closing.       */
/*  If testing_only is true, do not change the solution.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND dom_kind, bool testing_only)
{
  KHE_DRS_PACKED_SOLN init_soln, final_soln;
  init_soln = KheDrsPackedSolnBuildEmpty(drs);
  KheDrsSolveOpen(drs, soln_limit, dom_kind, init_soln);
  KheDrsSolveSearch(drs, &final_soln);
  KheDrsSolveClose(drs, init_soln);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsRerun(KHE_DYNAMIC_RESOURCE_SOLVER drs,                        */
/*    int soln_limit, KHE_DRS_DOM_KIND dom_kind)                             */
/*                                                                           */
/*  Rerun drs.                                                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsRerun(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND dom_kind)
{
  KHE_DRS_EXPR e;  int i;

  ** initialize the rerun cost fields in all cost expressions **
  HaArrayForEach(drs->all_cost_exprs, e, i)
    KheDrsExprCostInitRerunCost(e);

  ** do the solve **
  KheDrsSolve(drs, soln_limit, dom_kind, false);

  ** check the rerun cost fields in all cost expressions **
  fprintf(stderr, "[ KheDrsPackedSoln checking cost expressions:\n");
  HaArrayForEach(drs->all_cost_exprs, e, i)
    KheDrsExprCostCheckRerunCost(e);
  fprintf(stderr, "] KheDrsPackedSoln\n");

  ** clean up and exit **
  KheDrsPackedSolnDelete(drs->rerun, drs);
  drs->rerun = NULL;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsPrintSearchStatistics(KHE_DRS_SOLN soln,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Print statistics about the search that has just ended, leading to        */
/*  new solution soln.                                                       */
/*                                                                           */
/*****************************************************************************/

static float percent(int num, int denom)
{
  return denom == 0 ? -1.0 : (float) (100 * num) / (float) denom;
}

static int KheDrsViableSolns(KHE_DRS_SOLN_SET soln_set,
  KHE_COST max_viable_cost)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(soln_set->sorted_solns, soln, i)
    if( soln->cost > max_viable_cost )
      break;
  return i;
}

static void KheDrsPrintSearchStatistics(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
#if TESTING
  KHE_DRS_SOLN soln2, solni, solnj;  KHE_DRS_SOLN_SET soln_set;
  int i, j, count, rank, viable, total_count, total_rank, total_viable;
  if( DEBUG21 )
  {
    /* rank and viable solutions */
    fprintf(stderr, "  %4s  %10s %10s %6s    %10s %10s %6s\n",
      "Day", "Rank", "Total", "%", "Viable", "Total", "%");
    fprintf(stderr, "%s", "  -------------------------------------------"
      "-----------------------\n");
    total_count = total_viable = total_rank = 0;
    for( soln2 = soln;  soln2->prev_soln != NULL;  soln2 = soln2->prev_soln )
    {
      soln_set = soln2->day->soln_set;
      count = HaArrayCount(soln_set->sorted_solns);
      rank = soln2->sorted_rank;
      viable = KheDrsViableSolns(soln_set, soln->cost);
      total_count += count;
      total_rank += rank;
      total_viable += viable;
      fprintf(stderr, "  %4s  %10d %10d %6.1f    %10d %10d %6.1f\n",
	KheDrsDayId(soln2->day), rank, count, percent(rank, count),
	viable, count, percent(viable, count));
    }
    fprintf(stderr, "%s", "  -------------------------------------------"
      "-----------------------\n");
    fprintf(stderr, "  %4s  %10d %10d %6.1f    %10d %10d %6.1f\n",
      "Tot.", total_rank, total_count, percent(total_rank, total_count),
      total_viable, total_count, percent(total_viable, total_count));
  }

  if( DEBUG27 )
  {
    /* cost distributions */
    for( soln2 = soln;  soln2->prev_soln != NULL;  soln2 = soln2->prev_soln )
    {
      fprintf(stderr, "  %s cost distributions:\n", KheDrsDayId(soln2->day));
      soln_set = soln2->day->soln_set;
      for( i = 0;  i < HaArrayCount(soln_set->sorted_solns);  i = j )
      {
	solni = HaArray(soln_set->sorted_solns, i);
	for( j = i + 1;  j < HaArrayCount(soln_set->sorted_solns);  j++ )
	{
	  solnj = HaArray(soln_set->sorted_solns, j);
	  if( solni->cost != solnj->cost )
	    break;
	}
	fprintf(stderr, "    %.5f: %d\n", KheCostShow(solni->cost), j - i);
      }
    }
  }
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolveSearch(KHE_DYNAMIC_RESOURCE_SOLVER drs, bool testing,    */
/*    KHE_DRS_PACKED_SOLN *final_soln)                                       */
/*                                                                           */
/*  Carry out the actual solve.  If it succeeded in finding a new best       */
/*  solution, set *final_soln to that solution and return true, otherwise    */
/*  set *final_soln to NULL and return false.                                */
/*                                                                           */
/*  If testing is true, update the testing statistics as the run proceeds.   */
/*                                                                           */
/*****************************************************************************/
/* ***
static void KheDrsSolnDebugBrief(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);
*** */

static bool KheDrsSolveSearch(KHE_DYNAMIC_RESOURCE_SOLVER drs, bool testing,
  KHE_DRS_PACKED_SOLN *final_soln)
{
  KHE_DRS_DAY next_day, day;  KHE_DRS_SOLN_SET prev_soln_set;
  int i;  KHE_DRS_SOLN root_soln, soln /* , soln2 */;

  /* do the search */
  if( DEBUG17 )
    fprintf(stderr, "[ KheDrsSolveSearch(%d resources, %d days)\n",
      HaArrayCount(drs->open_resources), HaArrayCount(drs->open_days));
  prev_soln_set = KheDrsSolnSetMake(KHE_DRS_DOM_NONE, false,
    KHE_DRS_DOM_NONE, drs);
  root_soln = KheDrsSolnMake(NULL, NULL, drs->solve_start_cost, drs);
  KheDrsSolnSetAddSoln(prev_soln_set, root_soln, drs);
#if TESTING
  if( testing )
  {
    HaArrayAddLast(drs->table_size_per_day,
      KheDrsSolnSetKeptCount(prev_soln_set));
    HaArrayAddLast(drs->running_time_per_day, KheTimerElapsedTime(drs->timer));
  }
#endif
  HaArrayForEach(drs->open_days, next_day, i)
  {
    /* KheDrsDaySingleCostBegin(next_day); */
    KheDrsSolnSetExtend(prev_soln_set, next_day, drs);
    /* KheDrsDaySingleCostEnd(next_day); */
    prev_soln_set = next_day->soln_set;
#if TESTING
    if( testing )
    {
      HaArrayAddLast(drs->table_size_per_day,
	KheDrsSolnSetKeptCount(prev_soln_set));
      HaArrayAddLast(drs->running_time_per_day,KheTimerElapsedTime(drs->timer));
    }
#endif
    if( DEBUG17 )
    {
      fprintf(stderr, "  KheDrsSolveSearch ending day %s (%d solns made, "
	"%d kept)\n", KheDrsDayId(next_day), prev_soln_set->made_count,
	KheDrsSolnSetKeptCount(prev_soln_set));
      if( DEBUG29 )
	KheDrsSolnSetDebug(prev_soln_set, drs, 2, 2, stderr);
      /* ***
      if( prev_soln_set->kept_count == 1 )
      {
	KheDrsSolnSetExtractSingleSoln(prev_soln_set, &soln2);
        KheDrsSolnDebugBrief(soln2, drs, 1, 2, stderr);
      }
      *** */
      if( DEBUG30 )
      {
	KHE_DRS_SOLN soln1, soln2;
	soln1 = KheDrsSolnSetFirstSoln(prev_soln_set);
	soln2 = KheDrsSolnSetLastSoln(prev_soln_set);
	if( soln1 != NULL && soln2 != NULL )
	  KheDrsSolnDominatesDebug(soln1, true, 0, soln2, 0, drs, 2, 2, stderr);
      }
    }
    if( DEBUG28 )
    {
      int freq, j;
      fprintf(stderr, "  ancestor frequencies:");
      HaArrayForEach(drs->ancestor_freq, freq, j)
	fprintf(stderr, " %d", freq);
      fprintf(stderr, "\n");
    }
    HaArrayClear(drs->ancestor_freq);
  }

  /* set soln to final solution, or to NULL if none */
  if( HaArrayCount(drs->open_days) > 0 )
  {
    day = HaArrayLast(drs->open_days);
    KheDrsSolnSetExtractSingleSoln(day->soln_set, &soln);
  }
  else
    soln = NULL;

  if( soln != NULL )
  {
    /* have solution, convert to packed form and return true */
    KheDrsPrintSearchStatistics(soln, drs);
    if( DEBUG17 )
    {
      /* ***
      for( soln2 = soln;  soln2->prev_soln != NULL;  soln2 = soln2->prev_soln )
        KheDrsSolnDebugBrief(soln2, drs, 1, 2, stderr);
      *** */
      fprintf(stderr, "] KheDrsSolveSearch returning true\n");
    }
    *final_soln = KheDrsPackedSolnBuildFromSoln(soln, drs);
    return true;
  }
  else
  {
    /* no solution, return false */
    if( DEBUG17 )
      fprintf(stderr, "] KheDrsSolveSearch returning false\n");
    *final_soln = NULL;
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - construction"                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprInitBegin(KHE_DRS_EXPR e, KHE_DRS_EXPR_TAG tag,           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DRS_RESOURCE dr,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Shared generic code for creating an expression object and                */
/*  initializing those fields common to all expressions.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprInitBegin(KHE_DRS_EXPR e, KHE_DRS_EXPR_TAG tag,
  /* KHE_DRS_DOM_TEST dom_test, */ KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* HaMake(e, drs->arena); */
  e->tag = tag;
  e->gathered = false;
  /* e->dom_test = dom_test; no longer used */
  e->postorder_index = -1;
  e->resource = dr;
  HaArrayInit(e->parents, drs->arena);
  HaArrayInit(e->children, drs->arena);
  e->open_day_range = KheDrsDayRangeMake(1, 0);
  HaArrayInit(e->open_children, drs->arena);
  HaArrayInit(e->open_child_indexes, drs->arena);
  HaArrayInit(e->sig_indexes, drs->arena);
  /* HaArrayInit(e->signature_indexes, drs->arena); */
}

/* ***
#define KheDrsExprInitBegin(res, tg, dt, drs)				\
  HaMake(res, drs->arena);						\
  res->tag = tg;							\
  res->gathered = false;						\
  res->dom_test = dt;							\
  res->postorder_index = -1;						\
  HaArrayInit(res->parents, drs->arena);				\
  HaArrayInit(res->children, drs->arena);				\
  res->open_day_range = KheDrsDayRangeMake(1, 0);			\
  HaArrayInit(res->open_children, drs->arena);				\
  HaArrayInit(res->open_child_indexes, drs->arena);			\
  HaArrayInit(res->sig_indexes, drs->arena);
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprIsSingleDay(KHE_DRS_EXPR e, KHE_DRS_DAY *day)             */
/*                                                                           */
/*  Return true if e is derived from and event resource constraint on a      */
/*  single day.  Initially *day is NULL, but if an ASSIGNED_TASK is found,   */
/*  *day is set to its day.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsExprIsSingleDay(KHE_DRS_EXPR e, KHE_DRS_DAY *day)
{
  KHE_DRS_EXPR child_e;  int i;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_BUSY_TIME ebt;  KHE_DRS_EXPR_FREE_TIME eft;
  KHE_DRS_EXPR_WORK_TIME ewt;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;  KHE_DRS_EXPR_WORK_DAY ewd;
  KHE_DRS_DAY e_day;

  ** check the children **
  HaArrayForEach(e->children, child_e, i)
    if( !KheDrsExprIsSingleDay(child_e, day) )
      return false;

  ** check e itself **
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      eat = (KHE_DRS_EXPR_ASSIGNED_TASK) e;
      e_day = eat->task_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
      e_day = ebt->resource_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      eft = (KHE_DRS_EXPR_FREE_TIME) e;
      e_day = eft->resource_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      ewt = (KHE_DRS_EXPR_WORK_TIME) e;
      e_day = ewt->resource_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
      e_day = ebd->resource_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      efd = (KHE_DRS_EXPR_FREE_DAY) e;
      e_day = efd->resource_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      ewd = (KHE_DRS_EXPR_WORK_DAY) e;
      e_day = ewd->resource_on_day->day;
      return (*day == NULL ? *day = e_day, true : *day == e_day);

    case KHE_DRS_EXPR_OR_TAG:
    case KHE_DRS_EXPR_AND_TAG:	
    case KHE_DRS_EXPR_INT_SUM_TAG:
    case KHE_DRS_EXPR_FLOAT_SUM_TAG:
    case KHE_DRS_EXPR_INT_DEV_TAG:
    case KHE_DRS_EXPR_FLOAT_DEV_TAG:
    case KHE_DRS_EXPR_INT_SUM_COST_TAG:
    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      ** these don't prevent anything **
      return true;

    default:

      HnAbort("KheDrsExprIsSingleDay: internal error");
      return false;  ** keep compiler happy **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KheDrsExprInitEnd(res, drs)                                              */
/*                                                                           */
/*  Shared generic code for ending the intialization of res by setting       */
/*  its postorder index and closed value.                                    */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsExprInitEnd(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* ***
  KHE_DRS_DAY day;
  day = NULL;
  e->single_day = KheDrsExprIsSingleDay(e, &day);
  *** */
  e->postorder_index = drs->postorder_count++;
  KheDrsExprSetClosedValue(e, drs);
}

/* ***
#define KheDrsExprInitEnd(res, drs)					\
  (res)->postorder_index = (drs)->postorder_count++,			\
  KheDrsExprSetClosedValue((KHE_DRS_EXPR) (res), (drs))
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make child a child of parent.                                            */
/*                                                                           */
/*  This function includes a switch on the expression type and then calls a  */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - opening and closing"                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Gather e into drs->open_exprs, the list of expressions that are to be    */
/*  opened.  This does not actually open e; that comes later.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  if( !e->gathered )
  {
    e->gathered = true;
    HaArrayAddLast(drs->open_exprs, e);
    HaArrayForEach(e->parents, prnt, i)
      KheDrsExprGatherForOpening(prnt.expr, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprPostorderCmp(const void *t1, const void *t2)               */
/*                                                                           */
/*  Comparison function for sorting an array of expressions by increasing    */
/*  postorder index.                                                         */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprPostorderCmp(const void *t1, const void *t2)
{
  KHE_DRS_EXPR e1 = * (KHE_DRS_EXPR *) t1;
  KHE_DRS_EXPR e2 = * (KHE_DRS_EXPR *) t2;
  return e1->postorder_index - e2->postorder_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,      */
/*    int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  As part of opening e for solving, inform e that its child_index'th       */
/*  child, child_e, has opened.                                              */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,
  int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprChildDomTest(KHE_DRS_EXPR e,                  */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to a child of e on its open_child_index'th      */
/*  open day.                                                                */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprChildDomTest(KHE_DRS_EXPR e,
  int open_child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprDomTest(KHE_DRS_EXPR e, int open_child_index) */
/*                                                                           */
/*  Return a dom test suited to e on its open_child_index'th open day.       */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprDomTest(KHE_DRS_EXPR e,
  int open_child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprInitOpenChildIndexes(KHE_DRS_EXPR e)                      */
/*                                                                           */
/*  Initialize the open child index array of e.                              */
/*                                                                           */
/*****************************************************************************/

/* *** inlined into KheDrsExprOpen now
static void KheDrsExprInitOpenChildIndexes(KHE_DRS_EXPR e)
{
  int i, di;  KHE_DRS_EXPR child_e;
  HaArrayClear(e->open_child_indexes);
  di = e->open_day_range.first - 1;
  HaArrayForEach(e->open_children, child_e, i)
  {
    while( child_e->open_day_range.last > di )
    {
      di++;
      HaArrayAddLast(e->open_child_indexes, i);

      *******************************************************************
      **                                                               **
      **  Here we have just added i to index di of open_child_indexes, **
      **  and the child at index i is the first to have a larger last  **
      **  open day index than di.                                      **
      **                                                               **
      *******************************************************************
    }
  }
  HaArrayAddLast(e->open_child_indexes, i);  ** number of open children **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Open e for solving.                                                      */
/*                                                                           */
/*  By the time this is called, all of e's open children will have           */
/*  called KheDrsExprChildHasOpened.                                         */
/*                                                                           */
/*  Implementation note - how e->open_day_range is set                       */
/*  --------------------------------------------------                       */
/*  In external expressions, e->open_day_range receives its final value      */
/*  when KheDrsExprGatherForOpening is called.  In internal expressions,     */
/*  its value is initially empty and is updated by each call to              */
/*  KheDrsExprChildHasOpened until it reaches its final value.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  KHE_DRS_DAY day;  int i, di;  KHE_DRS_EXPR child_e;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DOM_TEST dom_test;

  /* inform e's parents that e is open */
  if( DEBUG10 )
  {
    fprintf(stderr, "[ KheDrsExprOpen\n");
    KheDrsExprDebug(e, drs, 2, 2, stderr);
  }
  e->gathered = false;
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasOpened(prnt.expr, e, prnt.index, drs);

  /* what to do now depends on whether e is external or internal */
  if( e->tag <= KHE_DRS_EXPR_WORK_DAY_TAG )
  {
    /* external expression; clear its value */
    KheDrsExprLeafClear(e, drs);
  }
  else
  {
    /* internal expression; build open_child_indexes into e's open children */
    /* KheDrsExprInitOpenChildIndexes(e); */
    HaArrayClear(e->open_child_indexes);
    di = e->open_day_range.first - 1;
    HaArrayForEach(e->open_children, child_e, i)
    {
      while( child_e->open_day_range.last > di )
      {
	di++;
	HaArrayAddLast(e->open_child_indexes, i);

	/*****************************************************************/
	/*                                                               */
	/*  Here we have just added i to index di of open_child_indexes, */
	/*  and the child at index i is the first to have a larger last  */
	/*  open day index than di.                                      */
	/*                                                               */
	/*****************************************************************/
      }
    }
    HaArrayAddLast(e->open_child_indexes, i);  /* number of open children */

    /* inform e's days that e is open */
    HaArrayClear(e->sig_indexes);
    for( i = e->open_day_range.first;  i <= e->open_day_range.last;  i++ )
    {
      /* add e to either its day or its resource on day */
      HnAssert( 0 <= i && i < HaArrayCount(drs->open_days),
	"KheDrsExprOpen: open day index (%d) out of range (0 .. %d)",
	i, HaArrayCount(drs->open_days) - 1);
      day = HaArray(drs->open_days, i);
      if( e->resource == NULL )
	KheDrsDayAddOpenExpr(day, e);
      else
      {
	drd = KheDrsResourceOnDay(e->resource, day);
	KheDrsResourceOnDayAddOpenExpr(drd, e);
      }

      /* signatures and dom tests (needed on each day except the last) */
      if( i < e->open_day_range.last )
      {
	dom_test = KheDrsExprDomTest(e, i, drs);
	if( DEBUG25 && e->tag == KHE_DRS_EXPR_INT_SUM_COST_TAG )
	{
	  KHE_DRS_EXPR_INT_SUM_COST eisc = (KHE_DRS_EXPR_INT_SUM_COST) e;
	  fprintf(stderr, "[ dom_test for %s on day %s:\n",
	    KheMonitorId(eisc->monitor), KheDrsDayId(day));
	  KheDrsDomTestDebug(dom_test, 3, 2, stderr);
	  fprintf(stderr, "]\n");
	}
	HaArrayAddLast(e->sig_indexes, KheDrsDayAddDomTest(day, dom_test));
      }
    }

    /* ***
    ** if e is a resource expression, inform its drd's that it is open **
    if( e->resource != NULL )
    {
      ** HaArrayClear(e->signature_indexes); **
      for( i = e->open_day_range.first;  i <= e->open_day_range.last;  i++ )
      {
	** add e to drd's list of open internal expressions **
	HnAssert( 0 <= i && i < HaArrayCount(drs->open_days),
	  "KheDrsExprOpen: open day index (%d) out of range (0 .. %d)",
	  i, HaArrayCount(drs->open_days) - 1);
	day = HaArray(drs->open_days, i);
	drd = KheDrsResourceOnDay(e->resource, day);
	KheDrsResourceOnDayAddOpenExpr(drd, e);

	** signatures and dom tests (needed on each day except the last) **
	** sorting this out sti ll to do; we don't need a dom test, but  **
	** we do need to add to e->signature_indexes, surely?            **
	** ***
	if( i < e->open_day_range.last )
	  HaArrayAddLast(e->signature_indexes,
	    KheDrsResourceOnDayAddDomTest(drd, e->dom_test));
	*** **
      }
    }
    *** */
  }
  if( DEBUG10 )
  {
    KheDrsExprDebug(e, drs, 2, 2, stderr);
    fprintf(stderr, "] KheDrsExprOpen\n");
  }
}



/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,                            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Inform internal expression e that its child_index'th child, child_e,     */
/*  has closed.  The child will have its correct closed_state and value      */
/*  fields when this is called.                                              */
/*                                                                           */
/*  Although KheDrsExprChildHasOpened adds child_e to e->open_children,      */
/*  KheDrsExprChildHasClosed does not remove child_e from e->open_children.  */
/*  This is because children are closed in the same order they were opened,  */
/*  but there is no efficient way to remove them in this order, short of     */
/*  something too complicated such as a circular list.                       */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Close e, which may be internal or external.  Whatever it needs to work   */
/*  out its closed value (an assignment, closed children, or closed state)   */
/*  is up to date.  At the end, inform e's parents that e has closed.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;

  /* set e's closed value */
  KheDrsExprSetClosedValue(e, drs);

  /* clear fields that are meaningful only when e is open */
  e->open_day_range = KheDrsDayRangeMake(1, 0);
  HaArrayClear(e->open_children);
  HaArrayClear(e->open_child_indexes);
  HaArrayClear(e->sig_indexes);

  /* inform e's parents that e has closed */
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasClosed(prnt.expr, e, prnt.index, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing e, set its value suitably for its closed state.       */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - open day indexes"                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprOpenDayIsFirst(KHE_DRS_EXPR e, int open_day_index)        */
/*                                                                           */
/*  Return true if open_day_index is the index of the first open day of e.   */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsExprOpenDayIsFirst(KHE_DRS_EXPR e, int open_day_index)
{
  return open_day_index == e->open_day_range.first;
}

/* ***
#define KheDrsExprOpenDayIsFirst(e, open_day_index)			\
  ((open_day_index) == (e)->open_day_range.first)
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprOpenDayIsLast(KHE_DRS_EXPR e, int open_day_index)         */
/*                                                                           */
/*  Return true if open_day_index is the index of the last open day of e.    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsExprOpenDayIsLast(KHE_DRS_EXPR e, int open_day_index)
{
  return open_day_index == e->open_day_range.last;
}

/* ***
#define KheDrsExprOpenDayIsLast(e, open_day_index)			\
  ((open_day_index) == (e)->open_day_range.last)
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprOpenChildrenBefore(KHE_DRS_EXPR e, int di)                 */
/*                                                                           */
/*  Return the number of open children of e whose last open day precedes     */
/*  the day with open day index di.  Or di could be one plus the index of    */
/*  the last open day, in which case return the number of open children.     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprOpenChildrenBefore(KHE_DRS_EXPR e, int di)
{
  return HaArray(e->open_child_indexes, di - e->open_day_range.first);
}

/* ***
#define KheDrsExprOpenChildrenBefore(e, di)			\
  HaArray((e)->open_child_indexes, (di) - (e)->open_day_range.first)
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprOpenChildrenAtOrAfter(KHE_DRS_EXPR e, int di)              */
/*                                                                           */
/*  Return the number of open children of e whose last open day is at or     */
/*  after the day with open day index di.  Or di could be one plus the       */
/*  index of the last open day, in which case return 0.                      */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprOpenChildrenAtOrAfter(KHE_DRS_EXPR e, int di)
{
  return HaArrayCount(e->open_children) - KheDrsExprOpenChildrenBefore(e, di);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprForEachOpenDayChild(KHE_DRS_EXPR e, int di,               */
/*    KHE_DRS_EXPR x, int i)                                                 */
/*                                                                           */
/*  Iterator which visits every open child of e whose last open day is di,   */
/*  setting x to each child in turn and i to its index in e->open_children.  */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprForEachOpenDayChild(e, di, x, i)			  \
  i1 = KheDrsExprOpenChildrenBefore((KHE_DRS_EXPR) (e), (di));		  \
  i2 = KheDrsExprOpenChildrenBefore((KHE_DRS_EXPR) (e), (di) + 1);	  \
  for( (i) = i1;						          \
       (i) < i2 ? ((x) = HaArray((e)->open_children, (i)), true) : false; \
       (i)++ )


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - accessing signatures"                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprSigVal(KHE_DRS_EXPR e, int open_day_index,                 */
/*    KHE_DRS_SOLN soln)                                                     */
/*                                                                           */
/*  Retrieve from soln the signature of e on open day open_day_index.        */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprSigVal(KHE_DRS_EXPR e, int open_day_index,
  KHE_DRS_SOLN soln)
{
  int pos;
  pos = HaArray(e->sig_indexes, open_day_index - e->open_day_range.first);
  return HaArray(soln->sig, pos);
}

/* ***
#define KheDrsExprSigVal(e, open_day_index, soln)			\
  HaArray((soln)->sig,						\
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first))
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprPutSigVal(KHE_DRS_EXPR e, int open_day_index,             */
/*    KHE_DRS_SOLN soln, int sig_val)                                        */
/*                                                                           */
/*  Add sig_val, the signature of e on open_day_index, to soln.              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprPutSigVal(KHE_DRS_EXPR e, int open_day_index,
  KHE_DRS_SOLN soln, int sig_val)
{
  int pos;
  pos = HaArray(e->sig_indexes, open_day_index - e->open_day_range.first);
  HnAssert(HaArrayCount(soln->sig) == pos,
    "KheDrsExprPutSigVal internal error"),
  HaArrayAddLast(soln->sig, sig_val);
}
*** */

/* ***
#define KheDrsExprPutSigVal(e, open_day_index, soln, sig_val)		     \
  HnAssert(HaArrayCount((soln)->sig) ==				     \
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first), \
    "KheDrsExprPutSigVal internal error"),				     \
  HaArrayAddLast((soln)->sig, (sig_val));
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprPutSigVal(KHE_DRS_EXPR e, int open_day_index,             */
/*    KHE_DRS_SIGNATURE sig, int sig_val)                                    */
/*                                                                           */
/*  Add sig_val, the signature of e on open_day_index, to sig.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprPutSigVal(KHE_DRS_EXPR e, int open_day_index,
  KHE_DRS_SIGNATURE sig, int sig_val)
{
  /* ***
  int pos;
  pos = HaArray(e->signature_indexes, open_day_index - e->open_day_range.first);
  HnAssert(HaArrayCount(sig->sig) == pos,
    "KheDrsExprPutSigVal internal error"),
  *** */
  HaArrayAddLast(sig->sig, sig_val);
}

/* ***
#define KheDrsExprPutSigVal(e, open_day_index, soln, sig_val)		     \
  HnAssert(HaArrayCount((soln)->sig) ==				     \
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first), \
    "KheDrsExprPutSigVal internal error"),				     \
  HaArrayAddLast((soln)->sig, (sig_val));
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si)     */
/*                                                                           */
/*  If e is in the signature on day, return true and set *si to its          */
/*  position in the signature.  Otherwise return false.                      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si)
{
  int i = day->open_day_index;
  if( e->open_day_range.first <= i && i < e->open_day_range.last )
    return *si = HaArray(e->sig_indexes, i - e->open_day_range.first), true;
  else
    return *si = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - searching"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,          */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Inform leaf expression e that dtd has been assigned drd.                 */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafClear(KHE_DRS_EXPR e)                                 */
/*                                                                           */
/*  Inform leaf expression e that the assignment of drd to dtd carried out   */
/*  by KheDrsExprLeafSet above has been removed.                             */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafClear(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEvalDay(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,           */
/*    KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Evaluate expression e in next_soln, whose day is one of e's open days.   */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprEvalDay(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,
  KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs);
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - other"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp)      */
/*                                                                           */
/*  Unindented debug print of that part of e which is specific to it.        */
/*  If soln != NULL, also print e's signature (if any) in soln.              */
/*                                                                           */
/*  This function switches on the expression type and then calls a           */
/*  version specific to that type.  Like all version-specific expression     */
/*  functions, its implementation is deferred to after the subtypes.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp);


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebug(KHE_DRS_EXPR e,                                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebug(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  /* char buff2[100]; */  KHE_DRS_EXPR child_e;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprDoDebug(e, NULL, fp);
    fprintf(fp, " pi %d oc %d", e->postorder_index,
      HaArrayCount(e->open_children));
    /* , KheDrsDomTestShow(e->dom_test, buff2)); */
    /* ***
    fprintf(fp, " pi %d sd %s oc %d dt %s", e->postorder_index,
      bool_show(e->single_day), HaArrayCount(e->open_children),
      KheDrsDomTestShow(e->dom_test, buff2));
    *** */
    if( HaArrayCount(e->children) > 0 )
    {
      fprintf(fp, "\n");
      HaArrayForEach(e->children, child_e, i)
	KheDrsExprDebug(child_e, drs, verbosity, indent + 2, fp);
      fprintf(fp, "%*s]\n", indent, "");
    }
    else
      fprintf(fp, " ]\n");
  }
  else
    KheDrsExprDoDebug(e, NULL, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugSoln(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.  Here      */
/*  e is known to be active in soln.                                         */
/*                                                                           */
/*****************************************************************************/

/* *** not currently used
static void KheDrsExprDebugSoln(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  int verbosity, int indent, FILE *fp)
{
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  KheDrsExprDoDebug(e, soln, fp);
  if( indent > 0 )
    fprintf(fp, "\n");
}
*** */

/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, */
/*    int dev)                                                               */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int dev)
{
  switch( cost_fn )
  {
    case KHE_STEP_COST_FUNCTION:

        return dev > 0 ? combined_weight : 0;

    case KHE_LINEAR_COST_FUNCTION:

        return dev * combined_weight;

    case KHE_QUADRATIC_COST_FUNCTION:

	return dev * dev * combined_weight;

    default:

	HnAbort("KheDrsCost internal error");
	return KheCost(0, 0);  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,         */
/*    bool is_float, FILE *fp)                                               */
/*                                                                           */
/*  If soln != NULL and e has a signature entry in soln, debug that entry.   */
/*                                                                           */
/*  NB soln == NULL does not mean that there is a valid solution whose       */
/*  value is NULL, it just means that we are doing this debug without the    */
/*  benefit of a current solution, and so this function has nothing to do.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebugSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  bool is_float, FILE *fp)
{
  int si;
  if( soln != NULL && soln->day != NULL &&
      KheDrsExprInSignature(e, soln->day, &si) )
  {
    if( is_float )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(soln->sig, si) / 100.0);
    else
      fprintf(fp, " $%d:%d$", si, HaArray(soln->sig, si));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "open child indexes"                                           */
/*                                                                           */
/*  For each open day di of e,                                               */
/*                                                                           */
/*    e->open_child_indexes[di - e->open_day_range.first]                    */
/*                                                                           */
/*  is the index of the first open child whose last open day index is di or  */
/*  more.  Equivalently, it is the number of open children whose last open   */
/*  day index is strictly less than di.                                      */
/*                                                                           */
/*  There is also a sentinel value, so for each open day di, the iterator    */
/*                                                                           */
/*    start_child_index = child_index[di - e->open_day_range.first];         */
/*    stop_child_index  = child_index[di + 1 - e->open_day_range.first];     */
/*    for( i = start_child_index;  i < stop_child_index;  i++ )              */
/*      child_e = HaArray(e->children, i);                                   */
/*                                                                           */
/*  visits the children of e whose last_day_index is di.                     */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_ASSIGNED_TASK"                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(                   */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_RESOURCE_GROUP rg,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an assigned task expression object with these attributes.           */
/*  Or if there is already such an expression in dtd's list, use that.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(
  KHE_DRS_TASK_ON_DAY dtd, KHE_RESOURCE_GROUP rg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_ASSIGNED_TASK res;  /* KHE_DRS_DOM_TEST dom_test; */
  /* dom_test = KheDrsDomTestMakeUnused(); */
  HnAssert(dtd->task != NULL, "KheDrsExprAssignedTaskMake internal error");
  if( KheDrsTaskOnDayHasAssignedTaskExpr(dtd, rg, &res) )
    return res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_ASSIGNED_TASK_TAG,
    /* dom_test, */ NULL, drs);
  res->task_on_day = dtd;
  res->resource_group = rg;
  KheDrsTaskOnDayAddAssignedTaskExpr(dtd, res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);  /* will set postorder_index and u.int_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprAssignedTaskValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,          */
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Return the value of eat when its task is assigned r (possibly NULL).     */
/*  This will be 1 when r is non-NULL and lies in eat's resource group.      */
/*  For efficienty, eat's resource group may be NULL, meaning all resources. */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprAssignedTaskValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_RESOURCE r)
{
  KHE_RESOURCE_GROUP rg;
  if( r == NULL )
    return 0;
  rg = eat->resource_group;
  return (rg == NULL || KheResourceGroupContains(rg, r)) ? 1 : 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskLeafSet(KHE_DRS_EXPR_ASSIGNED_TASK eat,       */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Update the value of eat to reflect the assignment of drd to dtd.  Here   */
/*  dtd is known to be the relevant task (which is why it is unused), and    */
/*  drd is known to represent a non-NULL resource, not a non-assignment.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskLeafSet(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  eat->u.int_val = KheDrsExprAssignedTaskValue(eat, drd->encl_dr->resource);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskLeafClear(KHE_DRS_EXPR_ASSIGNED_TASK eat)     */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprAssignedTaskLeafSet above has been removed.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskLeafClear(KHE_DRS_EXPR_ASSIGNED_TASK eat)
{
  eat->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskSetClosedValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eat, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskSetClosedValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_RESOURCE r;
  drd = eat->task_on_day->closed_asst;
  r = (drd == NULL ? NULL : drd->encl_dr->resource);
  eat->u.int_val = KheDrsExprAssignedTaskValue(eat, r);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDoDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,       */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eat which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskDoDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "ASSIGNED_TASK(");
  HnAssert(eat->task_on_day->task != NULL,
    "KheDrsExprAssignedTaskDoDebug internal error");
  KheTaskDebug(eat->task_on_day->task, 1, -1, fp);
  fprintf(fp, ", %s, %d)", eat->resource_group == NULL ? "-" :
    KheResourceGroupId(eat->resource_group), eat->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eat, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(                           */
/*    KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,                            */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  /* KHE_DRS_DOM_TEST dom_test, */ KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_BUSY_TIME_TAG,
	/* dom_test, */ time, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_TIME) e;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_BUSY_TIME_TAG,
    /* dom_test, */ drd->encl_dr, drs);
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);  /* will set u.int_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeLeafSet(KHE_DRS_EXPR_BUSY_TIME ebt,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ebt, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ebt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeLeafSet(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ebt->u.int_val = (dtd->time == ebt->time ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeLeafClear(KHE_DRS_EXPR_BUSY_TIME ebt)             */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprBusyTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeLeafClear(KHE_DRS_EXPR_BUSY_TIME ebt)
{
  ebt->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeSetClosedValue(KHE_DRS_EXPR_BUSY_TIME ebt,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ebt, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeSetClosedValue(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ebt->resource_on_day->closed_asst;
  ebt->u.int_val = (dtd != NULL && dtd->time == ebt->time ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDoDebug(KHE_DRS_EXPR_BUSY_TIME ebt,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ebt which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeDoDebug(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "BUSY_TIME(%s, %s, iv %d)",
    KheResourceId(ebt->resource_on_day->encl_dr->resource),
    KheTimeId(ebt->time), ebt->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ebt, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  /* KHE_DRS_DOM_TEST dom_test, */ KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_FREE_TIME_TAG,
      /* dom_test, */ time, drs, &e) )
    return (KHE_DRS_EXPR_FREE_TIME) e;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_FREE_TIME_TAG,
    /* dom_test, */ drd->encl_dr, drs);
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);  /* will set u.int_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeLeafSet(KHE_DRS_EXPR_FREE_TIME eft,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of eft, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as eft is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeLeafSet(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  eft->u.int_val = (dtd->time == eft->time ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeLeafClear(KHE_DRS_EXPR_FREE_TIME eft)             */
/*                                                                           */
/*  Inform leaf expression eft that the assignment of drd to dtd carried     */
/*  out by KheDrsExprFreeTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeLeafClear(KHE_DRS_EXPR_FREE_TIME eft)
{
  eft->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeSetClosedValue(KHE_DRS_EXPR_FREE_TIME eft,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eft, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeSetClosedValue(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = eft->resource_on_day->closed_asst;
  eft->u.int_val = (dtd != NULL && dtd->time == eft->time ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDoDebug(KHE_DRS_EXPR_FREE_TIME eft,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eft which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeDoDebug(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FREE_TIME(%s, %s, iv %d)",
    KheResourceId(eft->resource_on_day->encl_dr->resource),
    KheTimeId(eft->time), eft->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eft, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  /* KHE_DRS_DOM_TEST dom_test, */ KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_WORK_TIME_TAG,
	/* dom_test, */ time, drs, &e) )
    return (KHE_DRS_EXPR_WORK_TIME) e;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_WORK_TIME_TAG,
    /* dom_test, */ drd->encl_dr, drs);
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);  /* will set u.float_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeLeafSet(KHE_DRS_EXPR_WORK_TIME ewt,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ewt, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ewt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeLeafSet(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ewt->u.float_val =
    (dtd->time == ewt->time ? KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeLeafClear(KHE_DRS_EXPR_WORK_TIME ewt)             */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprWorkTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeLeafClear(KHE_DRS_EXPR_WORK_TIME ewt)
{
  ewt->u.float_val = 0.0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeSetClosedValue(KHE_DRS_EXPR_WORK_TIME ewt,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ewt, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeSetClosedValue(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ewt->resource_on_day->closed_asst;
  ewt->u.float_val = (dtd != NULL && dtd->time == ewt->time ?
    KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDoDebug(KHE_DRS_EXPR_WORK_TIME ewt,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ewt which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeDoDebug(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "WORK_TIME(%s, %s)",
    KheResourceId(ewt->resource_on_day->encl_dr->resource),
    KheTimeId(ewt->time));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ewt, soln, true, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

#define BUSY_DAY_DEBUG 0
#define BUSY_DAY_TIME_GROUP_ID "1Sat"
#define BUSY_DAY_RESOURCE "TR_25"


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd)                       */
/*                                                                           */
/*  Return true if we want to debug ebd.                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_RESOURCE r;  KHE_TIME_GROUP tg;
  if( RERUN && drs->rerun != NULL && BUSY_DAY_DEBUG )
  {
    r = ebd->resource_on_day->encl_dr->resource;
    tg = ebd->time_group;
    return strcmp(KheResourceId(r), BUSY_DAY_RESOURCE) == 0 &&
      strcmp(KheTimeGroupId(tg), BUSY_DAY_TIME_GROUP_ID) == 0;
  }
  else
    return false;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME_GROUP tg, /* KHE_DRS_DOM_TEST dom_test, */
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasDayExpr(drd, KHE_DRS_EXPR_BUSY_DAY_TAG,
      /* dom_test, */ tg, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_DAY) e;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_BUSY_DAY_TAG,
    /* dom_test, */ drd->encl_dr, drs);
  res->resource_on_day = drd;
  res->time_group = tg;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);  /* will set u.int_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayLeafSet(KHE_DRS_EXPR_BUSY_DAY ebd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ebd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ebt is.               */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprBusyDayDoDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_SOLN soln, FILE *fp);

static void KheDrsExprBusyDayLeafSet(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ebd->u.int_val = 1;
  if( KheDrsBusyDayDebug(ebd, drs) )
  {
    fprintf(stderr, "  KheDrsExprBusyDayLeafSet(");
    KheDrsExprBusyDayDoDebug(ebd, NULL, stderr);
    fprintf(stderr, ")\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayLeafClear(KHE_DRS_EXPR_BUSY_DAY ebd)               */
/*                                                                           */
/*  Inform leaf soln ebd that the assignment of drd to dtd carried out by    */
/*  KheDrsExprBusyDayLeafSet above has been removed.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayLeafClear(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ebd->u.int_val = 0;
  if( KheDrsBusyDayDebug(ebd, drs) )
  {
    fprintf(stderr, "  KheDrsExprBusyDayLeafClear(");
    KheDrsExprBusyDayDoDebug(ebd, NULL, stderr);
    fprintf(stderr, ")\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDaySetClosedValue(KHE_DRS_EXPR_BUSY_DAY ebd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ebd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDaySetClosedValue(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ebd->resource_on_day->closed_asst;
  ebd->u.int_val = (dtd != NULL ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDoDebug(KHE_DRS_EXPR_BUSY_DAY ebd,                 */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ebd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayDoDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "BUSY_DAY(%s, %s, iv %d)",
    KheResourceId(ebd->resource_on_day->encl_dr->resource),
    KheTimeGroupId(ebd->time_group), ebd->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ebd, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME_GROUP tg, /* KHE_DRS_DOM_TEST dom_test, */
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasDayExpr(drd, KHE_DRS_EXPR_FREE_DAY_TAG,
	/* dom_test, */ tg, drs, &e) )
    return (KHE_DRS_EXPR_FREE_DAY) e;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_FREE_DAY_TAG,
    /* dom_test, */ drd->encl_dr, drs);
  res->resource_on_day = drd;
  res->time_group = tg;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);  /* will set u.int_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayLeafSet(KHE_DRS_EXPR_FREE_DAY efd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of efd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as efd is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayLeafSet(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  efd->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayLeafClear(KHE_DRS_EXPR_FREE_DAY efd)               */
/*                                                                           */
/*  Inform leaf expression eft that the assignment of drd to dtd carried     */
/*  out by KheDrsExprFreeDayLeafSet above has been removed.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayLeafClear(KHE_DRS_EXPR_FREE_DAY efd)
{
  efd->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDaySetClosedValue(KHE_DRS_EXPR_FREE_DAY efd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDaySetClosedValue(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = efd->resource_on_day->closed_asst;
  efd->u.int_val = (dtd != NULL ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDoDebug(KHE_DRS_EXPR_FREE_DAY efd,                 */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of efd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayDoDebug(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FREE_DAY(%s, %s, iv %d)",
    KheResourceId(efd->resource_on_day->encl_dr->resource),
    KheTimeGroupId(efd->time_group), efd->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) efd, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused, although I really should be using it
static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsHasDayExpr(drs, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, tg, &e) )
    return (KHE_DRS_EXPR_WORK_DAY) e;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_WORK_DAY_TAG,
    ** dom_test, ** drd->encl_dr, drs);
  res->u.float_val = 0.0;  ** default value, used sometimes **
  res->time_group = tg;
  KheDrsAddDayExpr(drs, tg, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd((KHE_DRS_EXPR) res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayLeafSet(KHE_DRS_EXPR_WORK_DAY ewd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ewd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ewd is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayLeafSet(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ewd->u.float_val = KheTaskWorkloadPerTime(dtd->task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayLeafClear(KHE_DRS_EXPR_WORK_DAY ewd)               */
/*                                                                           */
/*  Inform leaf expression ewt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprWorkDayLeafSet above has been removed.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayLeafClear(KHE_DRS_EXPR_WORK_DAY ewd)
{
  ewd->u.float_val = 0.0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDaySetClosedValue(KHE_DRS_EXPR_WORK_DAY ewd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ewd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDaySetClosedValue(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ewd->resource_on_day->closed_asst;
  ewd->u.int_val = (dtd != NULL ? KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDoDebug(KHE_DRS_EXPR_WORK_DAY ewd,                 */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ewd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayDoDebug(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "WORK_DAY(%s, %s)",
    KheResourceId(ewd->resource_on_day->encl_dr->resource),
    KheTimeGroupId(ewd->time_group));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ewd, soln, true, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_OR"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,              */
/*    KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)                  */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_OR object with these attributes.                 */
/*  The called must call KheDrsExprInitEnd after adding all the children.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_OR KheDrsExprOrMake(/* KHE_DRS_DOM_TEST dom_test, */
  KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_OR_TAG,
    /* dom_test, */ dr, drs);
  res->closed_state = 0;
  /* later call to KheDrsExprInitEnd will set u.int_val */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrAddChild(KHE_DRS_EXPR_OR eo, KHE_DRS_EXPR child_e,      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  New child child_e has just been added to eo.  Update eo's closed         */
/*  state appropriately.  This happens to be the same as closing a child.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrAddChild(KHE_DRS_EXPR_OR eo, KHE_DRS_EXPR child_e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eo for solving, eo's child_index'th child, child_e,   */
/*  has been opened.  Update eo's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of eo is the number of eo's closed children whose       */
/*  value is 1, so this update subtracts 1 from the closed state if          */
/*  child_e's value (which preserves its value before opening) is 1.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state -= 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eo, eo's child_index'th child, child_e, has been      */
/*  closed.  Update eo's closed state to reflect this.                       */
/*                                                                           */
/*  The closed state of eo is the number of eo's closed children whose       */
/*  value is 1, so this update adds 1 to the closed state if child_e's       */
/*  value (which is up to date because it is closed first) is 1.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrSetClosedValue(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eo, set its value suitably for its closed state.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrSetClosedValue(KHE_DRS_EXPR_OR eo,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eo->u.int_val = (eo->closed_state > 0 ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprOrDomTest(KHE_DRS_EXPR_OR eo,                 */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to eo on its open_child_index'th open day.      */
/*                                                                           */
/*****************************************************************************/

/* *** unused
static KHE_DRS_DOM_TEST KheDrsExprOrDomTest(KHE_DRS_EXPR_OR eo,
  int open_child_index)
{
  ** placeholder; sti ll to do **
  return KheDrsDomTestMake(false, INT_MAX, 0, false, 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnOpenDayIndex(KHE_DRS_SOLN soln)                            */
/*                                                                           */
/*  Return the open day index of soln's day.  This day is assumed to exist,  */
/*  which means that KheDrsSolnOpenDayIndex cannot be called on the root     */
/*  solution since it has no day.                                            */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnOpenDayIndex(KHE_DRS_SOLN soln)
{
  return soln->day->open_day_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrEvalDay(KHE_DRS_EXPR_OR eo, KHE_DRS_SOLN prev_soln,     */
/*    KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Evaluate expression eo within next_soln.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrEvalDay(KHE_DRS_EXPR_OR eo, KHE_DRS_SOLN prev_soln,
  KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val, di;  KHE_DRS_EXPR child_e;

  di = KheDrsSolnOpenDayIndex(next_soln);
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eo, di) )
  {
    ** first day, so we have a 0 (false) value here **
    val = 0;
  }
  else
  {
    ** not first day, so retrieve a previous value **
    val = KheDrsExprSigVal((KHE_DRS_EXPR) eo, di - 1, prev_soln);
  }

  ** accumulate the values of the children of eo that finalized today **
  KheDrsExprForEachOpenDayChild(eo, di, child_e, i)
    if( child_e->u.int_val == 1 )
      val = 1;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eo, di) )
  {
    ** last day; incorporate closed state and set value **
    if( eo->closed_state > 0 )
      val = 1;
    eo->u.int_val = val;
  }
  else
  {
    ** not last day; store val in next_soln **
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eo, di, next_soln, val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrEvalSignature(KHE_DRS_EXPR_OR eo,                       */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE sig,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eo within sig.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrEvalSignature(KHE_DRS_EXPR_OR eo,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE sig,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eo, next_di) )
  {
    /* first day, so we have a 0 (false) value here */
    val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = KheDrsExprSigVal((KHE_DRS_EXPR) eo, next_di - 1, prev_soln);
  }

  /* accumulate the values of the children of eo that finalized today */
  KheDrsExprForEachOpenDayChild(eo, next_di, child_e, i)
    if( child_e->u.int_val == 1 )
      val = 1;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eo, next_di) )
  {
    /* last day; incorporate closed state and set value */
    if( eo->closed_state > 0 )
      val = 1;
    eo->u.int_val = val;
  }
  else
  {
    /* not last day; store val in next_soln */
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eo, next_di, sig, val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDoDebug(KHE_DRS_EXPR_OR eo,                             */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eo which is specific to it.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrDoDebug(KHE_DRS_EXPR_OR eo,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "OR(cs %d, iv %d)", eo->closed_state, eo->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eo, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_AND"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,            */
/*    KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)                  */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_AND object with these attributes.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_AND KheDrsExprAndMake(/* KHE_DRS_DOM_TEST dom_test, */
  KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_AND_TAG, 
    /* dom_test, */ dr, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndAddChild(KHE_DRS_EXPR_AND ea, KHE_DRS_EXPR child_e,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  New child child_e has just been added to ea.  Update ea's closed         */
/*  state appropriately.  This happens to be the same as closing a child.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndAddChild(KHE_DRS_EXPR_AND ea, KHE_DRS_EXPR child_e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening ea for solving, ea's child_index'th child, child_e,   */
/*  has been opened.  Update ea's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ea is the number of ea's closed children whose       */
/*  value is 0, so this update subtracts 1 from the closed state if          */
/*  child_e's value (which preserves its value before opening) is 0.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state -= 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing ea, ea's child_index'th child, child_e, has been      */
/*  closed.  Update ea's closed state to reflect this.                       */
/*                                                                           */
/*  The closed state of ea is the number of ea's closed children whose       */
/*  value is 0, so this update adds 1 to the closed state if child_e's       */
/*  value (which is up to date because it is closed first) is 0.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndSetClosedValue(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ea, set its value suitably for its closed state.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndSetClosedValue(KHE_DRS_EXPR_AND ea,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ea->u.int_val = (ea->closed_state > 0 ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprAndDomTest(KHE_DRS_EXPR_AND ea,               */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to ea on its open_child_index'th open day.      */
/*                                                                           */
/*****************************************************************************/

/* *** unused
static KHE_DRS_DOM_TEST KheDrsExprAndDomTest(KHE_DRS_EXPR_AND ea,
  int open_child_index)
{
  ** placeholder; sti ll to do **
  return KheDrsDomTestMake(false, INT_MAX, 0, false, 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndEvalDay(KHE_DRS_EXPR_AND ea, KHE_DRS_SOLN prev_soln,   */
/*    KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Evaluate expression ea within next_soln.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndEvalDay(KHE_DRS_EXPR_AND ea, KHE_DRS_SOLN prev_soln,
  KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val, di;  KHE_DRS_EXPR child_e;

  di = KheDrsSolnOpenDayIndex(next_soln);
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) ea, di) )
  {
    ** first day, so we have a 1 (true) value here **
    val = 1;
  }
  else
  {
    ** not first day, so retrieve a previous value **
    val = KheDrsExprSigVal((KHE_DRS_EXPR) ea, di - 1, prev_soln);
  }

  ** accumulate the values of the children of ea that finalized today **
  KheDrsExprForEachOpenDayChild(ea, di, child_e, i)
    if( child_e->u.int_val == 0 )
      val = 0;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) ea, di) )
  {
    ** last day; incorporate closed_state and set value **
    if( ea->closed_state > 0 )
      val = 0;
    ea->u.int_val = val;
  }
  else
  {
    ** not last day; store value in next_soln **
    KheDrsExprPutSigVal((KHE_DRS_EXPR) ea, di, next_soln, val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndEvalSignature(KHE_DRS_EXPR_AND ea,                     */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression ea within ss.                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndEvalSignature(KHE_DRS_EXPR_AND ea,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) ea, next_di) )
  {
    /* first day, so we have a 1 (true) value here */
    val = 1;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = KheDrsExprSigVal((KHE_DRS_EXPR) ea, next_di - 1, prev_soln);
  }

  /* accumulate the values of the children of ea that finalized today */
  KheDrsExprForEachOpenDayChild(ea, next_di, child_e, i)
    if( child_e->u.int_val == 0 )
      val = 0;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) ea, next_di) )
  {
    /* last day; incorporate closed_state and set value */
    if( ea->closed_state > 0 )
      val = 0;
    ea->u.int_val = val;
  }
  else
  {
    /* not last day; store value in next_soln */
    KheDrsExprPutSigVal((KHE_DRS_EXPR) ea, next_di, ss, val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDoDebug(KHE_DRS_EXPR_AND ea,                           */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ea which is specific to it.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndDoDebug(KHE_DRS_EXPR_AND ea,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "AND(cs %d, iv %d)", ea->closed_state, ea->u.int_val);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ea, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,     */
/*    KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)                  */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(/* KHE_DRS_DOM_TEST dom_test,*/
  KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_INT_SUM_TAG,
    /* dom_test, */ dr, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumAddChild(KHE_DRS_EXPR_INT_SUM eis,                  */
/*    KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  New child child_e has just been added to eis.  Update eis's closed       */
/*  state appropriately.  This happens to be the same as closing a child.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumAddChild(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->closed_state += child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eis for solving, eis's child_index'th child, child_e, */
/*  has been opened.  Update eis's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eis is the sum of the values of eis's s closed       */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->closed_state -= child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eis, eis's child_index'th child, child_e, has been    */
/*  closed.  Update eis's closed state to reflect this.                      */
/*                                                                           */
/*  The closed state of eis is the sum of the values of eis's s closed       */
/*  children, so this update adds child_e's value (which preserves           */
/*  its value before opening) to the closed state.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->closed_state += child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumSetClosedValue(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eis, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumSetClosedValue(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->u.int_val = eis->closed_state;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSumDomTest(KHE_DRS_EXPR_INT_SUM eis,       */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to eis on its open_child_index'th open day.     */
/*                                                                           */
/*****************************************************************************/

/* *** unused
static KHE_DRS_DOM_TEST KheDrsExprIntSumDomTest(KHE_DRS_EXPR_INT_SUM eis,
  int open_child_index)
{
  ** placeholder; sti ll to do **
  return KheDrsDomTestMake(false, INT_MAX, 0, false, 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumEvalDay(KHE_DRS_EXPR_INT_SUM eis,                   */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eis on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumEvalDay(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val, di;  KHE_DRS_EXPR child_e;

  di = KheDrsSolnOpenDayIndex(next_soln);
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eis, di) )
  {
    ** first day, so (ignoring closed_state) we have a 0 value here **
    val = 0;
  }
  else
  {
    ** not first day, so retrieve a previous value **
    val = KheDrsExprSigVal((KHE_DRS_EXPR) eis, di - 1, prev_soln);
  }

  ** accumulate the open values of the children of eis that finalized today **
  KheDrsExprForEachOpenDayChild(eis, di, child_e, i)
    val += child_e->u.int_val;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eis, di) )
  {
    ** last day; add closed_state to the value **
    eis->u.int_val = val + eis->closed_state;
  }
  else
  {
    ** this not eis's last day; store value in next_soln **
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eis, di, next_soln, val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumEvalSignature(KHE_DRS_EXPR_INT_SUM eis,             */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eis within ss.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumEvalSignature(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eis, next_di) )
  {
    /* first day, so (ignoring closed_state) we have a 0 value here */
    val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = KheDrsExprSigVal((KHE_DRS_EXPR) eis, next_di - 1, prev_soln);
  }

  /* accumulate the open values of the children of eis that finalized today */
  KheDrsExprForEachOpenDayChild(eis, next_di, child_e, i)
    val += child_e->u.int_val;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eis, next_di) )
  {
    /* last day; add closed_state to the value */
    eis->u.int_val = val + eis->closed_state;
  }
  else
  {
    /* this not eis's last day; store value in next_soln */
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eis, next_di, ss, val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDoDebug(KHE_DRS_EXPR_INT_SUM eis,                   */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eis which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumDoDebug(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_SUM(%d)", eis->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eis, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_SUM"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(                           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DRS_RESOURCE dr,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_SUM object with these attributes.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(
  /* KHE_DRS_DOM_TEST dom_test, */ KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_SUM res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_FLOAT_SUM_TAG,
    /* dom_test, */ dr, drs);
  res->u.float_val = UNDEF_FLOAT;
  res->closed_state = 0.0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumAddChild(KHE_DRS_EXPR_FLOAT_SUM efs,              */
/*    KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  New child child_e has just been added to efs.  Update efs's closed       */
/*  state appropriately.  This happens to be the same as closing a child.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumAddChild(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->closed_state += child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening efs for solving, efs's child_index'th child, child_e, */
/*  has been opened.  Update efs's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efs is the sum of the values of efs's closed         */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->closed_state -= child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening efs for solving, efs's child_index'th child, child_e, */
/*  has been closed.  Update efs's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efs is the sum of the values of efs's closed         */
/*  children, so this update adds child_e's value (which preserves           */
/*  its value before opening) to the closed state.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->closed_state += child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumSetClosedValue(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efs, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumSetClosedValue(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->u.float_val = efs->closed_state;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprFloatSumDomTest(KHE_DRS_EXPR_FLOAT_SUM efs,   */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to efs on its open_child_index'th open day.     */
/*                                                                           */
/*****************************************************************************/

/* *** unused
static KHE_DRS_DOM_TEST KheDrsExprFloatSumDomTest(KHE_DRS_EXPR_FLOAT_SUM efs,
  int open_child_index)
{
  ** placeholder; sti ll to do **
  return KheDrsDomTestMake(false, INT_MAX, 0, false, 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumEvalDay(KHE_DRS_EXPR_FLOAT_SUM efs,               */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efs on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumEvalDay(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, di;  float val;  KHE_DRS_EXPR child_e;

  di = KheDrsSolnOpenDayIndex(next_soln);
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) efs, di) )
  {
    ** first day, so we have a 0.0 value here **
    val = 0.0;
  }
  else
  {
    ** not first day, so retrieve a previous value **
    val = (float) KheDrsExprSigVal((KHE_DRS_EXPR) efs, di-1, prev_soln) / 100.0;
  }

  ** accumulate the open values of the children of efs that finalized today **
  KheDrsExprForEachOpenDayChild(efs, di, child_e, i)
    val += child_e->u.float_val;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) efs, di) )
  {
    ** last day; add closed_state to the value **
    efs->u.float_val = val + efs->closed_state;
  }
  else
  {
    ** this not efs's last day; store value in next_soln **
    KheDrsExprPutSigVal((KHE_DRS_EXPR) efs, di, next_soln, (int) (val * 100.0));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumEvalSignature(KHE_DRS_EXPR_FLOAT_SUM efs,         */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efs in ss.                                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumEvalSignature(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  float val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) efs, next_di) )
  {
    /* first day, so we have a 0.0 value here */
    val = 0.0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = (float) KheDrsExprSigVal((KHE_DRS_EXPR) efs, next_di - 1, prev_soln)
      / 100.0;
  }

  /* accumulate the open values of the children of efs that finalized today */
  KheDrsExprForEachOpenDayChild(efs, next_di, child_e, i)
    val += child_e->u.float_val;

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) efs, next_di) )
  {
    /* last day; add closed_state to the value */
    efs->u.float_val = val + efs->closed_state;
  }
  else
  {
    /* this not efs's last day; store value in next_soln */
    KheDrsExprPutSigVal((KHE_DRS_EXPR) efs, next_di, ss, (int)(val * 100.0));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDoDebug(KHE_DRS_EXPR_FLOAT_SUM efs,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of efs which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumDoDebug(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FLOAT_SUM(%.2f)", efs->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) efs, soln, true, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_DEV"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprIntDevDev(KHE_DRS_EXPR_INT_DEV eid, int val)               */
/*                                                                           */
/*  Deviation function to use with these expressions.                        */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprIntDevDev(KHE_DRS_EXPR_INT_DEV eid, int val)
{
  if( val == 0 && eid->allow_zero )
    return 0;
  else if( val > eid->max_limit )
    return val - eid->max_limit;
  else if( val < eid->min_limit )
    return eid->min_limit - val;
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit, int max_limit,  */
/*    bool allow_zero, KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs) */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_DEV object with these attributes.            */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit, int max_limit,
  bool allow_zero, KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_DEV res;  /* KHE_DRS_DOM_TEST dom_test; */
  /* dom_test = KheDrsDomTestMakeUnused(); */
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_INT_DEV_TAG,
    /* dom_test, */ dr, drs);
  res->u.int_val = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevAddChild(KHE_DRS_EXPR_INT_DEV eid,                  */
/*    KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  New child child_e has just been added to eid.  Update eid's closed       */
/*  state appropriately.  This happens to be the same as closing a child.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevAddChild(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eid for solving, eid's child_index'th child, child_e, */
/*  has been opened.  Update eid's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eid is the value of eid's single closed child.       */
/*  This is not stored explicitly, so this update does nothing.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eid for solving, eid's child_index'th child, child_e, */
/*  has been closed.  Update eid's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eid is the value of eid's single closed child.       */
/*  This is not stored explicitly, so this update does nothing.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevSetClosedValue(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eid, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevSetClosedValue(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(eid->children);
  eid->u.int_val = KheDrsExprIntDevDev(eid, child_e->u.int_val);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntDevDomTest(KHE_DRS_EXPR_INT_DEV eid,       */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to eid on its open_child_index'th open day.     */
/*                                                                           */
/*****************************************************************************/

/* *** unused
static KHE_DRS_DOM_TEST KheDrsExprIntDevDomTest(KHE_DRS_EXPR_INT_DEV eid,
  int open_child_index)
{
  ** placeholder; sti ll to do **
  return KheDrsDomTestMake(false, INT_MAX, 0, false, 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevEvalDay(KHE_DRS_EXPR_INT_DEV eid,                   */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eid within next_soln.                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevEvalDay(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val, di;  KHE_DRS_EXPR child_e;

  di = KheDrsSolnOpenDayIndex(next_soln);
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eid, di) )
  {
    ** first day **
    val = 0;
  }
  else
  {
    ** not first day, so retrieve a previous value **
    HnAbort("KheDrsExprIntDevEvalDay internal error (non-first day)");
  }

  ** accumulate the open values of the children of eid that finalized today **
  KheDrsExprForEachOpenDayChild(eid, di, child_e, i)
    val += KheDrsExprIntDevDev(eid, child_e->u.int_val);

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eid, di) )
  {
    ** last day; set value **
    eid->u.int_val = val;
  }
  else
  {
    ** not last day; store val in next_soln **
    HnAbort("KheDrsExprIntDevEvalDay internal error (non-last day)");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevEvalSignature(KHE_DRS_EXPR_INT_DEV eid,             */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eid within ss.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevEvalSignature(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eid, next_di) )
  {
    /* first day */
    val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprIntDevEvalSignature internal error (non-first day)");
  }

  /* accumulate the open values of the children of eid that finalized today */
  KheDrsExprForEachOpenDayChild(eid, next_di, child_e, i)
    val += KheDrsExprIntDevDev(eid, child_e->u.int_val);

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eid, next_di) )
  {
    /* last day; set value */
    eid->u.int_val = val;
  }
  else
  {
    /* not last day; store val in next_soln */
    HnAbort("KheDrsExprIntDevEvalSignature internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDoDebug(KHE_DRS_EXPR_INT_DEV eid,                   */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eid which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevDoDebug(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_DEV(%d, %d, %s)", eid->min_limit, eid->max_limit,
    bool_show(eid->allow_zero));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eid, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_DEV"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,             */
/*    int max_limit, bool allow_zero, KHE_DRS_RESOURCE dr,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_DEV object with these attributes.          */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_DEV res;  /* KHE_DRS_DOM_TEST dom_test; */
  /* dom_test = KheDrsDomTestMakeUnused(); */
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_FLOAT_DEV_TAG,
    /* dom_test, */ dr, drs);
  res->u.int_val = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevAddChild(KHE_DRS_EXPR_FLOAT_DEV efd,              */
/*    KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  New child child_e has just been added to efd.  Update efd's closed       */
/*  state appropriately.  This happens to be the same as closing a child.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevAddChild(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    float workload)                                                        */
/*                                                                           */
/*  Return the deviation of workload wrt efd.                                */
/*                                                                           */
/*****************************************************************************/

static int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,
  float workload)
{
  if( efd->allow_zero && workload < 0.001 )
    return 0;
  else if( workload < efd->min_limit - 0.001 )
    return (int) ceil(efd->min_limit - 0.001 - workload);
  else if( workload > efd->max_limit + 0.001 )
    return (int) ceil(workload - efd->max_limit - 0.001);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening efd for solving, efd's child_index'th child, child_e, */
/*  has been opened.  Update efd's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efd is the value of efd's single closed child,       */
/*  so this update needs to do nothing.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing efd, efd's child_index'th child, child_e,             */
/*  has been closed.  Update efd's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efd is the value of efd's single closed child,       */
/*  so this update needs to do nothing.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevSetClosedValue(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevSetClosedValue(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(efd->children);
  efd->u.int_val = KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprFloatDevDomTest(KHE_DRS_EXPR_FLOAT_DEV efd,   */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to efd on its open_child_index'th open day.     */
/*                                                                           */
/*****************************************************************************/

/* *** unused
static KHE_DRS_DOM_TEST KheDrsExprFloatDevDomTest(KHE_DRS_EXPR_FLOAT_DEV efd,
  int open_child_index)
{
  ** placeholder; sti ll to do **
  return KheDrsDomTestMake(false, INT_MAX, 0, false, 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevEvalDay(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efd on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevEvalDay(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val, di;  KHE_DRS_EXPR child_e;

  di = KheDrsSolnOpenDayIndex(next_soln);
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) efd, di) )
  {
    ** first day **
    val = 0;
  }
  else
  {
    ** not first day, so retrieve a previous value **
    HnAbort("KheDrsExprFloatDevEvalDay internal error (non-first day)");
    val = 0;  ** keep compiler happy **
  }

  ** accumulate the open values of the children of efd that finalized today **
  KheDrsExprForEachOpenDayChild(efd, di, child_e, i)
    val += KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) efd, di) )
  {
    ** last day; set value **
    efd->u.int_val = val;
  }
  else
  {
    ** not last day; store value in next_soln **
    HnAbort("KheDrsExprFloatDevEvalDay internal error (non-last day)");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevEvalSignature(KHE_DRS_EXPR_FLOAT_DEV efd,         */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efd within ss.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevEvalSignature(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE ss,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) efd, next_di) )
  {
    /* first day */
    val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprFloatDevEvalSignature internal error (non-first day)");
    val = 0;  /* keep compiler happy */
  }

  /* accumulate the open values of the children of efd that finalized today */
  KheDrsExprForEachOpenDayChild(efd, next_di, child_e, i)
    val += KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);

  if( KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) efd, next_di) )
  {
    /* last day; set value */
    efd->u.int_val = val;
  }
  else
  {
    /* not last day; store value in next_soln */
    HnAbort("KheDrsExprFloatDevEvalSignature internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDoDebug(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of efd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevDoDebug(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FLOAT_DEV(%d, %d, %s)", efd->min_limit, efd->max_limit,
    bool_show(efd->allow_zero));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) efd, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM_COST"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_ADJUST_TYPE KheDrsAdjustType(KHE_COST_FUNCTION cost_fn,          */
/*    int max_limit)                                                         */
/*                                                                           */
/*  Return the adjust type, determined by the cost function and max limit.   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_ADJUST_TYPE KheDrsAdjustType(KHE_COST_FUNCTION cost_fn,
  int max_limit)
{
  if( max_limit == INT_MAX )
    return KHE_DRS_ADJUST_NO_MAX;
  else if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    return KHE_DRS_ADJUST_LINEAR;
  else if( cost_fn == KHE_STEP_COST_FUNCTION )
    return KHE_DRS_ADJUST_STEP;
  else
    return KHE_DRS_ADJUST_ORDINARY;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsAdjustTypeShow(KHE_DRS_ADJUST_TYPE dct)                      */
/*                                                                           */
/*  Return a string representation of dct.                                   */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsAdjustTypeShow(KHE_DRS_ADJUST_TYPE dct)
{
  switch( dct )
  {
    case KHE_DRS_ADJUST_NO_MAX:	  return "NO_MAX";
    case KHE_DRS_ADJUST_LINEAR:	  return "LINEAR";
    case KHE_DRS_ADJUST_STEP:	  return "STEP";
    case KHE_DRS_ADJUST_ORDINARY: return "ORDINARY";
    default:			  return "??";
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprIntSumCostDev(KHE_DRS_EXPR_INT_SUM_COST eisc,              */
/*    int children_with_value_1, int children_with_unknown_value)            */
/*                                                                           */
/*  Return the deviation of eisc, assuming that eisc has the given number    */
/*  of children with value 1, and the given number of children with unknown  */
/*  value.  This is function g from the documentation.                       */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprIntSumCostDev(KHE_DRS_EXPR_INT_SUM_COST eisc,
  int children_with_value_1, int children_with_unknown_value)
{
  int val;
  if( children_with_value_1 == 0 && eisc->allow_zero )
    return 0;
  if( children_with_value_1 > eisc->max_limit )
    return children_with_value_1 - eisc->max_limit;
  val = children_with_value_1 + children_with_unknown_value;
  return (val < eisc->min_limit ? eisc->min_limit - val : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM_COST KheDrsExprIntSumCostMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,    */
/*    int max_limit, bool allow_zero, int history_before, int history_after, */
/*    int history, KHE_DRS_DOM_TEST dom_test, KHE_DRS_RESOURCE dr,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM_COST object with these attributes.       */
/*                                                                           */
/*****************************************************************************/
static void KheDrsAddWeight(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_COST weight);

static KHE_DRS_EXPR_INT_SUM_COST KheDrsExprIntSumCostMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,
  int max_limit, bool allow_zero, int history_before, int history_after,
  int history, /* KHE_DRS_DOM_TEST dom_test, */ KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM_COST res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_INT_SUM_COST_TAG,
    /* dom_test, */ dr, drs);
  res->u.int_val = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  res->adjust_type = KheDrsAdjustType(cost_fn, max_limit);
  res->history_before = history_before;
  res->history_after = history_after;
  res->history = history;
  res->closed_state = 0;
  res->rerun_open_and_search_cost = res->rerun_open_and_close_cost = 0;
  HaArrayAddLast(drs->all_cost_exprs, (KHE_DRS_EXPR_COST) res);
  KheDrsAddWeight(drs, combined_weight);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM_COST KheDrsExprIntSumCostZeroMaxMake(KHE_MONITOR m, */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)                  */
/*                                                                           */
/*  Make an INT_SUM_COST object for the case where there is a maximum limit  */
/*  of 0 (so the dominance test is <=) and no history.  This is the usual    */
/*  (but not the only) replacement for the now omitted EXPR_COST.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM_COST KheDrsExprIntSumCostZeroMaxMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  KHE_DRS_RESOURCE dr, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* ***
  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE,
    cost_fn != KHE_QUADRATIC_COST_FUNCTION, combined_weight);
  *** */
  return KheDrsExprIntSumCostMake(m, cost_fn, combined_weight, 0, 0, false,
    0, 0, 0, /* dom_test, */ dr, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostAddChild(KHE_DRS_EXPR_INT_SUM_COST eisc,        */
/*    KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  New child child_e has just been added to eisc.  Update eisc's state      */
/*  appropriately.  This is similar to closing a child, but not identical.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCostAddChild(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eisc->closed_state += child_e->u.int_val;
  /* not wanted here! KheDrsExprIntSumCostUpdateStartCost(eisc, drs); */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsUpdateRerunCost(KHE_DYNAMIC_RESOURCE_SOLVER drs,              */
/*    KHE_COST *rerun_cost1, KHE_COST *rerun_cost2, KHE_MONITOR m,           */
/*    char *operation, int child_index, char *fmt, ...)                      */
/*                                                                           */
/*  Update *rerun_cost1, and *rerun_cost2 if rerun_cost2 != NULL.  Also      */
/*  produce a debug print about it, if m is RERUN_MONITOR_ID.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsUpdateRerunCost(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_COST *rerun_cost1, KHE_COST *rerun_cost2, KHE_MONITOR m,
  char *operation, int child_index, char *fmt, ...)
{
  if( RERUN && drs->rerun != NULL )
  {
    va_list ap;  char buff[10];  int count;  KHE_COST cost;  bool print;
    print = (strcmp(KheMonitorId(m), RERUN_MONITOR_ID) == 0);
    if( print )
    {
      if( child_index >= 0 )
	snprintf(buff, 9, "%s %d", operation, child_index);
      else
	snprintf(buff, 9, "%s", operation);
      fprintf(stderr, "  %9s: %.5f", buff, KheCostShow(*rerun_cost1));
    }
    va_start(ap, fmt);
    for( count = 0;  fmt[count] != '\0';  count++ )
    {
      cost = va_arg(ap, KHE_COST);
      if( print )
	fprintf(stderr, " %c %.5f", fmt[count], KheCostShow(cost));
      if( fmt[count] == '+' )
	*rerun_cost1 += cost;
      else
	*rerun_cost1 -= cost;
      if( rerun_cost2 != NULL )
      {
	if( fmt[count] == '+' )
	  *rerun_cost2 += cost;
	else
	  *rerun_cost2 -= cost;
      }
    }
    va_end(ap);
    if( print )
    {
      while( count < 4 )
      {
	fprintf(stderr, "          ");
	count++;
      }
      fprintf(stderr, "  (%s)\n", KheMonitorId(m));
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostChildHasOpened(KHE_DRS_EXPR_INT_SUM_COST eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eisc for solving, eisc's child_index'th child,        */
/*  child_e, has been opened.  Update eisc to reflect this, and also         */
/*  drs->solve_start_cost.                                                   */
/*                                                                           */
/*  This function is called after child_e is added to eisc->open_children.   */
/*  The closed state of eisc is the sum of the values of eisc's closed       */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*  Obsolete:                                                                */
/*  So when KheDrsExprIntSumCostUpdateStartCost retrieves the number of      */
/*  open children, it gets the correct (that is, the changed) value.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCostChildHasOpened(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int old_dev, new_dev, i;  KHE_DRS_EXPR e;  KHE_COST old_cost, new_cost;
  if( false && DEBUG14 )
  {
    fprintf(stderr, "[ KheDrsExprIntSumCostChildHasOpened(%s, child_e, %d)\n",
      KheMonitorId(eisc->monitor), child_index);
    HaArrayForEach(eisc->children, e, i)
      KheDrsExprDebug(e, drs, 2, 4, stderr);
  }
  old_dev = KheDrsExprIntSumCostDev(eisc, eisc->history + eisc->closed_state,
    HaArrayCount(eisc->open_children) - 1  + eisc->history_after);
  eisc->closed_state -= child_e->u.int_val;
  new_dev = KheDrsExprIntSumCostDev(eisc, eisc->history + eisc->closed_state,
    HaArrayCount(eisc->open_children) + eisc->history_after);
  if( old_dev != new_dev )
  {
    old_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, old_dev);
    new_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, new_dev);
    drs->solve_start_cost += new_cost - old_cost;
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost,
      &eisc->rerun_open_and_close_cost, eisc->monitor, "open",
      child_index, "+-", new_cost, old_cost);
  }
  if( false && DEBUG14 )
    fprintf(stderr, "] KheDrsExprIntSumCostChildHasOpened\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostChildHasClosed(KHE_DRS_EXPR_INT_SUM_COST eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eisc, eisc's child_index'th child, child_e, has       */
/*  been closed.  Update eisc's closed state to reflect this, and also       */
/*  drs->solve_start_cost.                                                   */
/*                                                                           */
/*  Although KheDrsExprChildHasOpened adds the newly opened child to         */
/*  e->open_children, KheDrsExprChildHasClosed does not remove it.  This     */
/*  is because the removal cannot be done efficiently while preserving       */
/*  the order of the open children.  However that does not matter here,      */
/*  and we need the correct number of open children, so we remove any one.   */
/*                                                                           */
/*  The closed state of eisc is the sum of the values of eisc's closed       */
/*  children, so this update adds child_e's value to the closed state.       */
/*                                                                           */
/*  Obsolete:                                                                */
/*  This is called after child_e is removed from eisc->open_children.        */
/*  So when KheDrsExprIntSumCostUpdateStartCost retrieves the number of      */
/*  open children, it gets the correct (that is, the changed) value.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCostChildHasClosed(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int old_dev, new_dev;  KHE_COST old_cost, new_cost;
  old_dev = KheDrsExprIntSumCostDev(eisc, eisc->history + eisc->closed_state,
    HaArrayCount(eisc->open_children) + eisc->history_after);
  HaArrayDeleteLast(eisc->open_children);
  eisc->closed_state += child_e->u.int_val;
  new_dev = KheDrsExprIntSumCostDev(eisc, eisc->history + eisc->closed_state,
    HaArrayCount(eisc->open_children) + eisc->history_after);
  if( old_dev != new_dev )
  {
    new_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, new_dev);
    old_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, old_dev);
    drs->solve_start_cost += new_cost - old_cost;
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_close_cost, NULL,
      eisc->monitor, "close", child_index, "+-", new_cost, old_cost);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostSetClosedValue(KHE_DRS_EXPR_INT_SUM_COST eisc,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eisc, set its value suitably for its closed state.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCostSetClosedValue(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here, since eisc has start_cost instead of a value */
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSumCostChildDomTest(                       */
/*    KHE_DRS_EXPR_INT_SUM_COST eisc, int open_child_index,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return the dom test for a child of eisc on its open_child_index'th       */
/*  open day.                                                                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprIntSumCostChildDomTest(
  KHE_DRS_EXPR_INT_SUM_COST eisc, int open_child_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int a, b;

  /* calculate a */
  a = (eisc->max_limit < INT_MAX ? -1 : INT_MAX);

  /* calculate b */
  b = (eisc->min_limit > 0 ? INT_MAX : 0);

  /* make and return the dom test */
  return KheDrsDomTestMake(false, 0, INT_MAX, a, b, false, 0, NULL, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSumCostDomTest(                            */
/*    KHE_DRS_EXPR_INT_SUM_COST eisc, int open_child_index)                  */
/*                                                                           */
/*  Return the dom test for e on its open_child_index'th open day.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprIntSumCostDomTest(
  KHE_DRS_EXPR_INT_SUM_COST eisc, int open_child_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int a, b, q;

  /* calculate a */
  if( eisc->max_limit < INT_MAX )
  {
    q = KheDrsExprOpenChildrenAtOrAfter((KHE_DRS_EXPR) eisc, open_child_index);
    a = eisc->max_limit - q;
  }
  else
    a = INT_MAX;

  /* calculate b */
  if( eisc->min_limit > 0 )
    b = eisc->min_limit;
  else
    b = 0;

  /* make and return the dom test */
  return KheDrsDomTestMake(eisc->allow_zero, eisc->min_limit, eisc->max_limit,
    a, b, eisc->cost_fn != KHE_QUADRATIC_COST_FUNCTION, eisc->combined_weight,
    eisc->monitor, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsAdjustedSigVal(int val, KHE_DRS_ADJUST_TYPE adjust_type,       */
/*    int min_limit, int max_limit, int pc, int qc)                          */
/*                                                                           */
/*  Return the adjusted signature type.                                      */
/*                                                                           */
/*****************************************************************************/

static int KheDrsAdjustedSigVal(int val, KHE_DRS_ADJUST_TYPE adjust_type,
  int min_limit, int max_limit, int qc)
{
  int pbar;
  switch( adjust_type )
  {
    case KHE_DRS_ADJUST_ORDINARY:

      return val;

    case KHE_DRS_ADJUST_NO_MAX:

      pbar = max(0, min_limit - qc);
      /* pbar = max(0, min_limit - pc - qc); */
      return min(pbar, val);

    case KHE_DRS_ADJUST_LINEAR:

      pbar = max_limit;
      /* pbar = max(0, max_limit - pc); */
      return min(pbar, val);

    case KHE_DRS_ADJUST_STEP:

      pbar = max_limit + 1;
      /* pbar = max(0, max_limit + 1 - pc); */
      return min(pbar, val);

    default:

      HnAbort("KheDrsAdjustedSigVal internal error (adjusttype)");
      return val;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostEvalDay(KHE_DRS_EXPR_INT_SUM_COST eisc,         */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc in next_soln.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCostEvalDay(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int di, pc, p1, p2, qc, q1, q2, i, i1, i2, dev1, dev2, adjusted_val;
  KHE_DRS_EXPR child_e;  KHE_COST old_cost, new_cost;

  ** get di, the open day index of next_soln **
  di = KheDrsSolnOpenDayIndex(next_soln);

  ** get pc, p1, q1, and dev1 **
  p1 = pc = eisc->history + eisc->closed_state;
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eisc, di) )
  {
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  _%d", p1);
  }
  else
  {
    p1 += KheDrsExprSigVal((KHE_DRS_EXPR) eisc, di - 1, prev_soln);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  $%d:%d",
	KheDrsExprSigVal((KHE_DRS_EXPR) eisc, di-1, prev_soln), p1);
  }
  qc = eisc->history_after;
  q1 = qc + KheDrsExprOpenChildrenAtOrAfter((KHE_DRS_EXPR) eisc, di);
  dev1 = KheDrsExprIntSumCostDev(eisc, p1, q1);
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, " dev1 %d = g(%d, %d)", dev1, p1, q1);

  ** get qc, p2, q2, and dev2 **
  p2 = p1;
  KheDrsExprForEachOpenDayChild(eisc, di, child_e, i)
    p2 += child_e->u.int_val;
  q2 = qc + KheDrsExprOpenChildrenAtOrAfter((KHE_DRS_EXPR) eisc, di + 1);
  dev2 = KheDrsExprIntSumCostDev(eisc, p2, q2);
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, " dev2 %d = g(%d, %d)", dev2, p2, q2);

  ** report the change in cost, if any **
  if( dev2 != dev1 )
  {
    new_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev2);
    old_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev1);
    next_soln->cost += new_cost - old_cost;
    ** KheDrsDaySingleCostAddCost(next_soln->day, new_cost - old_cost); **
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
      eisc->monitor, "search", -1, "+-", new_cost, old_cost);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, " +%.5f -%.5f", KheCostShow(new_cost),
	KheCostShow(old_cost));
  }

  ** if not eisc's last day, store p2 - pc (adjusted) in next_soln **
  if( !KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eisc, di) )
  {
    adjusted_val = KheDrsAdjustedSigVal(p2 - pc, eisc->adjust_type,
      eisc->min_limit, eisc->max_limit, pc, qc);
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eisc, di, next_soln, adjusted_val);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  %d:%d$", p2 - pc, adjusted_val);
  }
  if( DEBUG_LAST_EDGE )
    fprintf(stderr, " %s\n", KheMonitorId(eisc->monitor));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostEvalSignature(KHE_DRS_EXPR_INT_SUM_COST eisc,   */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE sig,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc in sig.                                         */
/*                                                                           */
/*****************************************************************************/

#define DEBUG_ME (false &&						\
  strcmp(KheMonitorId(eisc->monitor), "Constraint:13/NU_6") == 0)

static void KheDrsExprIntSumCostEvalSignature(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE sig,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int p1, p2, qc, q1, q2, i, i1, i2, dev1, dev2, adjusted_val;
  KHE_DRS_EXPR child_e;  KHE_COST old_cost, new_cost;

  /* get p1, q1, and dev1 */
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eisc, next_di) )
  {
    p1 = eisc->history + eisc->closed_state;
  }
  else
  {
    p1 = KheDrsExprSigVal((KHE_DRS_EXPR) eisc, next_di - 1, prev_soln);
    if( DEBUG_LAST_EDGE || DEBUG_ME )
    {
      fprintf(stderr, "  prev_soln %.5f ", KheCostShow(prev_soln->cost));
      KheDrsSolnSignatureDebug(prev_soln, stderr);
      fprintf(stderr, "\n  index %d, pos %d, ",
        next_di - 1 - eisc->open_day_range.first,
	HaArray(eisc->sig_indexes, next_di - 1 - eisc->open_day_range.first));
    }
  }
  qc = eisc->history_after;
  q1 = qc + KheDrsExprOpenChildrenAtOrAfter((KHE_DRS_EXPR) eisc, next_di);
  dev1 = KheDrsExprIntSumCostDev(eisc, p1, q1);
  if( DEBUG_LAST_EDGE || DEBUG_ME )
    fprintf(stderr, "  hi %d cs %d, p1 %d, qc %d, q1 %d, dev1 %d, ",
      eisc->history, eisc->closed_state, p1, qc, q1, dev1);

  /* get qc, p2, q2, and dev2 */
  p2 = p1;
  KheDrsExprForEachOpenDayChild(eisc, next_di, child_e, i)
    p2 += child_e->u.int_val;
  q2 = qc + KheDrsExprOpenChildrenAtOrAfter((KHE_DRS_EXPR) eisc, next_di + 1);
  dev2 = KheDrsExprIntSumCostDev(eisc, p2, q2);
  if( DEBUG_LAST_EDGE || DEBUG_ME )
    fprintf(stderr, "p2 %d, q2 %d, dev2 %d, ", p2, q2, dev2);

  /* report the change in cost, if any */
  if( dev2 != dev1 )
  {
    new_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev2);
    old_cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev1);
    sig->cost += new_cost - old_cost;
    /* KheDrsDaySingleCostAddCost(next_soln->day, new_cost - old_cost); */
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
      eisc->monitor, "search", -1, "+-", new_cost, old_cost);
    if( DEBUG_LAST_EDGE || DEBUG_ME )
      fprintf(stderr, " +%.5f -%.5f, ", KheCostShow(new_cost),
	KheCostShow(old_cost));
  }

  /* if not eisc's last day, store p2 (adjusted) in next_soln */
  if( !KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eisc, next_di) )
  {
    /* store p2 */
    adjusted_val = KheDrsAdjustedSigVal(p2, eisc->adjust_type,
      eisc->min_limit, eisc->max_limit, qc);
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eisc, next_di, sig, adjusted_val);
    if( DEBUG_LAST_EDGE || DEBUG_ME )
      fprintf(stderr, "  p2 %d, adj %d, ", p2, adjusted_val);
  }
  if( DEBUG_LAST_EDGE || DEBUG_ME )
    fprintf(stderr, " %s\n", KheMonitorId(eisc->monitor));
}


/* *** correct but uses old terminology
static void KheDrsExprIntSumCostEvalDay(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i1, i2, ai, bi, ci, xi, prev_dev, yi, dev, di;  KHE_DRS_EXPR child_e;

  ** get ai, bi, and ci **
  di = KheDrsSolnOpenDayIndex(next_soln);
  ai = KheDrsExprNumberOfOpenChildrenBefore(eisc, di);
  bi = KheDrsExprNumberOfOpenChildrenBefore(eisc, di + 1) - ai;
  ci = HaArrayCount(eisc->open_children) - ai - bi;

  ** get xi and prev_dev **
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eisc, di) )
  {
    xi = 0;
    prev_dev = KheDrsExprIntSumCostStartDev(eisc);
  }
  else
  {
    xi = KheDrsExprSigVal((KHE_DRS_EXPR) eisc, di - 1, prev_soln);
    prev_dev = KheDrsExprIntSumCostDev(eisc, eisc->history +
      eisc->closed_state + xi, bi + ci + eisc->history_after);
  }

  ** get yi and dev **
  yi = 0;
  KheDrsExprForEachOpenDayChild(eisc, di, child_e, i1)
    yi += child_e->u.int_val;
  dev = KheDrsExprIntSumCostDev(eisc,
    eisc->history + eisc->closed_state + xi + yi, ci + eisc->history_after);

  ** if not eisc's last day, store xi + yi (or less) in next_soln **
  if( !KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eisc, di) )
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eisc, di, next_soln,
      KheDrsExprIntSumCostAdjustedSigVal(eisc, xi + yi));

  ** report the change in cost c(Ci) - c(C{i-1}) **
  if( dev != prev_dev )
    next_soln->cost += (KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev)
      - KheDrsCost(eisc->cost_fn, eisc->combined_weight, prev_dev));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostInitRerunCost(KHE_DRS_EXPR_INT_SUM_COST eisc)   */
/*                                                                           */
/*  Initialize the rerun cost fields of eisc.                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCostInitRerunCost(KHE_DRS_EXPR_INT_SUM_COST eisc)
{
  eisc->rerun_open_and_search_cost = eisc->rerun_open_and_close_cost =
    KheMonitorCost(eisc->monitor);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostCheckRerunCost(KHE_DRS_EXPR_INT_SUM_COST eisc)  */
/*                                                                           */
/*  Check the rerun cost fields of eisc.                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCostCheckRerunCost(KHE_DRS_EXPR_INT_SUM_COST eisc)
{
  if( eisc->rerun_open_and_search_cost != KheMonitorCost(eisc->monitor) )
    fprintf(stderr, "  open_and_search_cost %.5f, monitor cost %.5f (%s)\n",
      KheCostShow(eisc->rerun_open_and_search_cost),
      KheCostShow(KheMonitorCost(eisc->monitor)), KheMonitorId(eisc->monitor));
  if( eisc->rerun_open_and_close_cost != KheMonitorCost(eisc->monitor) )
    fprintf(stderr, "  open_and_close_cost  %.5f, monitor cost %.5f (%s)\n",
      KheCostShow(eisc->rerun_open_and_close_cost),
      KheCostShow(KheMonitorCost(eisc->monitor)), KheMonitorId(eisc->monitor));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCostDoDebug(KHE_DRS_EXPR_INT_SUM_COST eisc,         */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eisc which is specific to it.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCostDoDebug(KHE_DRS_EXPR_INT_SUM_COST eisc,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_SUM_COST(%s, %s, %d)", KheMonitorId(eisc->monitor),
    KheDrsAdjustTypeShow(eisc->adjust_type), eisc->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eisc, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_CLOSED_SEQ"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_CLOSED_SEQ KheDrsClosedSeqMake(KHE_DYNAMIC_RESOURCE_SOLVER drs,  */
/*    int start_index, int stop_index)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_CLOSED_SEQ object with these attributes.              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_CLOSED_SEQ KheDrsClosedSeqMake(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int start_index, int stop_index)
{
  KHE_DRS_CLOSED_SEQ res;
  if( HaArrayCount(drs->free_closed_seqs) > 0 )
    res = HaArrayLastAndDelete(drs->free_closed_seqs);
  else
    HaMake(res, drs->arena);
  res->start_index = start_index;
  res->stop_index = stop_index;
  res->active_at_left = 0;		/* placeholder (unless empty) */
  res->active_at_right = 0;		/* placeholder (unless empty) */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_CLOSED_SEQ KheDrsClosedSeqMakeInit(                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an initial empty closed sequence.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_CLOSED_SEQ KheDrsClosedSeqMakeInit(
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  return KheDrsClosedSeqMake(drs, 0, 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsClosedSeqDelete(KHE_DRS_CLOSED_SEQ dcs,                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete dcs.                                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsClosedSeqDelete(KHE_DRS_CLOSED_SEQ dcs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_closed_seqs, dcs);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsClosedSeqLength(KHE_DRS_CLOSED_SEQ dcs)                        */
/*                                                                           */
/*  Return the length of dcs.                                                */
/*                                                                           */
/*****************************************************************************/

static int KheDrsClosedSeqLength(KHE_DRS_CLOSED_SEQ dcs)
{
  return dcs->stop_index - dcs->start_index;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsClosedSeqAllActive(KHE_DRS_CLOSED_SEQ dcs)                    */
/*                                                                           */
/*  Return true if every child of dcs is active.                             */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsClosedSeqAllActive(KHE_DRS_CLOSED_SEQ dcs)
{
  return KheDrsClosedSeqLength(dcs) == dcs->active_at_left;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsClosedSeqExtendToRight(KHE_DRS_CLOSED_SEQ dcs, int val)       */
/*                                                                           */
/*  Extend dcs one place to the right, with a child with value val.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsClosedSeqExtendToRight(KHE_DRS_CLOSED_SEQ dcs, int val)
{
  if( val == 0 )
    dcs->active_at_right = 0;
  else
  {
    if( KheDrsClosedSeqAllActive(dcs) )
      dcs->active_at_left++;
    dcs->active_at_right++;
  }
  dcs->stop_index++;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsClosedSeqSetActiveLengths(KHE_DRS_CLOSED_SEQ dcs,             */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc)                                        */
/*                                                                           */
/*  Assuming dcs->start_index and dcs->stop_index are up to date, bring      */
/*  dcs->active_at_left and dcs-> active_at_right up to date.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsClosedSeqSetActiveLengths(KHE_DRS_CLOSED_SEQ dcs,
  KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  KHE_DRS_EXPR child_e;  int i;

  /* bring dcs->active_at_left up to date */
  dcs->active_at_left = 0;
  for( i = dcs->start_index;  i < dcs->stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    if( child_e->u.int_val != 1 )
      break;
    dcs->active_at_left++;
  }

  /* bring dcs->active_at_right up to date */
  dcs->active_at_right = 0;
  for( i = dcs->stop_index - 1;  i >= dcs->start_index;  i-- )
  {
    child_e = HaArray(eisc->children, i);
    if( child_e->u.int_val != 1 )
      break;
    dcs->active_at_right++;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_CLOSED_SEQ KheDrsClosedSeqSplit(KHE_DRS_CLOSED_SEQ dcs,          */
/*    int child_index)                                                       */
/*                                                                           */
/*  Split dcs at child_index.                                                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_CLOSED_SEQ KheDrsClosedSeqSplit(KHE_DRS_CLOSED_SEQ dcs,
  int child_index, KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_CLOSED_SEQ res;

  /* make the new fragment and reset the stop index of the old one */
  HnAssert(dcs->start_index <= child_index,
    "KheDrsClosedSeqSplit internal error 1");
  HnAssert(child_index < dcs->stop_index,
    "KheDrsClosedSeqSplit internal error 2 (child_index %d, dcs->stop_index %d",
    child_index, dcs->stop_index);
  res = KheDrsClosedSeqMake(drs, child_index + 1, dcs->stop_index);
  dcs->stop_index = child_index;

  /* update the active lengths of the two fragements, and return */
  KheDrsClosedSeqSetActiveLengths(dcs, eisc);
  KheDrsClosedSeqSetActiveLengths(res, eisc);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsClosedSeqMerge(KHE_DRS_CLOSED_SEQ dcs1,                       */
/*    KHE_DRS_CLOSED_SEQ dcs2, KHE_DRS_EXPR_INT_SEQ_COST eisc,               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Merge adjacent closed sequences dcs1 and dcs2.  The formerly open        */
/*  child separating them has closed, and its closed value has been set.     */
/*                                                                           */
/*  Place the merged closed sequence in dcs1, and delete dcs2.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsClosedSeqMerge(KHE_DRS_CLOSED_SEQ dcs1,
  KHE_DRS_CLOSED_SEQ dcs2, KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* update dcs1 to hold the merged sequence */
  HnAssert(dcs1->stop_index + 1 == dcs2->start_index,
    "KheDrsClosedSeqMerge internal error 1");
  dcs1->stop_index = dcs2->stop_index;
  KheDrsClosedSeqSetActiveLengths(dcs1, eisc);

  /* delete dcs2 */
  KheDrsClosedSeqDelete(dcs2, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsClosedSeqDebug(KHE_DRS_CLOSED_SEQ dcs, int verbosity,         */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dcs onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsClosedSeqDebug(KHE_DRS_CLOSED_SEQ dcs, int verbosity,
  int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[CLOSED_SEQ %d-%d, active_at_left %d, active_at_right %d]",
    dcs->start_index, dcs->stop_index, dcs->active_at_left,
    dcs->active_at_right);
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_A_INTERVAL"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_A_INTERVAL KheDrsAIntervalMake(int start_index, int stop_index,  */
/*    bool unassigned_precedes)                                              */
/*                                                                           */
/*  Make and return a new a-interval with these attributes.                  */
/*                                                                           */
/*****************************************************************************/

KHE_DRS_A_INTERVAL KheDrsAIntervalMake(int start_index, int stop_index,
  bool unassigned_precedes)
{
  KHE_DRS_A_INTERVAL res;
  res.start_index = start_index;
  res.stop_index = stop_index;
  res.unassigned_precedes = unassigned_precedes;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_A_INTERVAL KheDrsAIntervalMakeLeft(                              */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index, int active_len)        */
/*                                                                           */
/*  Make an a-interval of length active_len which ends just before           */
/*  the open child with index open_index.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_A_INTERVAL KheDrsAIntervalMakeLeft(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index, int active_len)
{
  KHE_DRS_CLOSED_SEQ dcs;
  dcs = HaArray(eisc->closed_seqs, open_index);
  return KheDrsAIntervalMake(dcs->stop_index - active_len,
    dcs->stop_index, false);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_A_INTERVAL KheDrsAIntervalFindLeft(                              */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index)                        */
/*                                                                           */
/*  Find the (possibly empty) a-interval immediately to the left of the      */
/*  child with the given open_index.  Include history if necessary.          */
/*                                                                           */
/*  Implementation note.  This interval consists entirely of the last        */
/*  active_at_right children of the preceding KHE_DRS_CLOSED_SEQ             */
/*  (possibly plus history), even if their number is zero; because they      */
/*  are preceded by another open child, or an inactive child, or nothing.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_A_INTERVAL KheDrsAIntervalFindLeft(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index)
{
  KHE_DRS_CLOSED_SEQ dcs;
  dcs = HaArray(eisc->closed_seqs, open_index);
  if( !KheDrsClosedSeqAllActive(dcs) )
  {
    /* an inactive child precedes the active_at_right active children */
    return KheDrsAIntervalMake(dcs->stop_index - dcs->active_at_right,
      dcs->stop_index, false);
  }
  else if( open_index > 0 )
  {
    /* an unassigned child precedes the active_at_right active children */
    return KheDrsAIntervalMake(dcs->stop_index - dcs->active_at_right,
      dcs->stop_index, true);
  }
  else
  {
    /* nothing but history precedes the active_at_right active children */
    return KheDrsAIntervalMake(dcs->stop_index - dcs->active_at_right
      - eisc->history, dcs->stop_index, false);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_A_INTERVAL KheDrsAIntervalFindRight(                             */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index, bool child_unassigned) */
/*                                                                           */
/*  Find the (possibly empty) a-interval immediately to the right of the     */
/*  child with the given open_index.  That child is assumed to be unassigned */
/*  when child_unassigned is true.                                           */
/*                                                                           */
/*  Implementation note.  This interval consists entirely of the first       */
/*  active_at_left children of the following KHE_DRS_CLOSED_SEQ,             */
/*  even if their number is zero; because they are followed by another       */
/*  open child, or an inactive child, or history_after, or nothing.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_A_INTERVAL KheDrsAIntervalFindRight(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index, bool child_unassigned)
{
  KHE_DRS_CLOSED_SEQ dcs;
  dcs = HaArray(eisc->closed_seqs, open_index + 1);
  return KheDrsAIntervalMake(dcs->start_index,
    dcs->start_index + dcs->active_at_left, child_unassigned);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_A_INTERVAL KheDrsAIntervalMerge(KHE_DRS_A_INTERVAL ai_left,      */
/*    KHE_DRS_A_INTERVAL ai_right)                                           */
/*                                                                           */
/*  Merge ai_left with ai_right.  These intervals are separated by one       */
/*  child (this is checked).  That child is assumed to be active, but that   */
/*  is not checked, and indeed may not be true - we might be building this   */
/*  interval to reflect the old state of that child.                         */
/*                                                                           */
/*  It is safe to set the unassigned_precedes value of the result interval   */
/*  to ai_left.unassigned_precedes.  Even when ai_left is empty, its place   */
/*  is well defined, and ai_left.unassigned_precedes is correct for there.   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_A_INTERVAL KheDrsAIntervalMerge(KHE_DRS_A_INTERVAL ai_left,
  KHE_DRS_A_INTERVAL ai_right)
{
  HnAssert(ai_left.stop_index + 1 == ai_right.start_index,
    "KheDrsAIntervalMerge internal error");
  return KheDrsAIntervalMake(ai_left.start_index, ai_right.stop_index,
    ai_left.unassigned_precedes);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsAIntervalDev(KHE_DRS_A_INTERVAL ai,                            */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc)                                        */
/*                                                                           */
/*  Return the deviation of a-interval ai.  If ai.unassigned_precedes is     */
/*  true, there is an unassigned child immediately before ai.first_index,    */
/*  so return 0 if the cost function is KHE_STEP_COST_FUNCTION.              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsAIntervalDev(KHE_DRS_A_INTERVAL ai,
  KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  int len;

  /* return 0 if an unassigned child immediately precedes this interval */
  /* and the cost function is step                                      */
  if( ai.unassigned_precedes && eisc->cost_fn == KHE_STEP_COST_FUNCTION )
    return 0;

  /* include history if we are at the start */
  /* *** already included where relevant
  if( ai.start_index == 0 )
    ai.start_index -= eisc->history;
  *** */

  /* calculate the length and deviation */
  len = ai.stop_index - ai.start_index;
  return len > eisc->max_limit ? len - eisc->max_limit : 0;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsAIntervalCost(KHE_DRS_A_INTERVAL ai,                      */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc)                                        */
/*                                                                           */
/*  Return the cost of ai.                                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsAIntervalCost(KHE_DRS_A_INTERVAL ai,
  KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  int dev;
  dev = KheDrsAIntervalDev(ai, eisc);
  return dev == 0 ? 0 : KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_AU_INTERVAL"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_AU_INTERVAL KheDrsAUIntervalMake(int start_index,                */
/*    int stop_index, bool has_active_child)                                 */
/*                                                                           */
/*  Make and return an au-interval with these attributes.  However, the      */
/*  has_active_child attribute is ignored when the interval is empty:  an    */
/*  empty interval cannot contain an active child.                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_AU_INTERVAL KheDrsAUIntervalMake(int start_index,
  int stop_index, bool has_active_child)
{
  KHE_DRS_AU_INTERVAL res;
  res.start_index = start_index;
  res.stop_index = stop_index;
  res.has_active_child = (start_index < stop_index && has_active_child);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsAUIntervalLength(KHE_DRS_AU_INTERVAL aui)                      */
/*                                                                           */
/*  Return the length of aui.                                                */
/*                                                                           */
/*****************************************************************************/

static int KheDrsAUIntervalLength(KHE_DRS_AU_INTERVAL aui)
{
  return aui.stop_index - aui.start_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAUIntervalExtendToLeft(KHE_DRS_AU_INTERVAL *aui,              */
/*    int extra_len, bool has_active_child)                                  */
/*                                                                           */
/*  Extend *aui extra_len places to the left.  If extra_len > 0,             */
/*  has_active_child says whether the extra elements are active or not.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAUIntervalExtendToLeft(KHE_DRS_AU_INTERVAL *aui,
  int extra_len, bool has_active_child)
{
  if( extra_len > 0 )
  {
    aui->start_index -= extra_len;
    if( has_active_child )
      aui->has_active_child = true;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAUIntervalExtendToRight(KHE_DRS_AU_INTERVAL *aui,             */
/*    int extra_len, bool has_active_child)                                  */
/*                                                                           */
/*  Extend *aui extra_len places to the right.  If extra_len > 0,            */
/*  has_active_child says whether the extra elements are active or not.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAUIntervalExtendToRight(KHE_DRS_AU_INTERVAL *aui,
  int extra_len, bool has_active_child)
{
  if( extra_len > 0 )
  {
    aui->stop_index += extra_len;
    if( has_active_child )
      aui->has_active_child = true;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_AU_INTERVAL KheDrsAUIntervalMakeLeft(                            */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index, int active_len)        */
/*                                                                           */
/*  Make an au-interval of length active_len which ends just before          */
/*  the open child with index open_index.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_AU_INTERVAL KheDrsAUIntervalMakeLeft(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index, int active_len)
{
  KHE_DRS_CLOSED_SEQ dcs;
  dcs = HaArray(eisc->closed_seqs, open_index);
  return KheDrsAUIntervalMake(dcs->stop_index - active_len,
    dcs->stop_index, true);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_AU_INTERVAL KheDrsAUIntervalFindLeft(                            */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index)                        */
/*                                                                           */
/*  Find the au-interval ending just before the child whose open index       */
/*  is open_index.                                                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_AU_INTERVAL KheDrsAUIntervalFindLeft(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index)
{
  KHE_DRS_CLOSED_SEQ dcs;  KHE_DRS_AU_INTERVAL res;  int i;

  /* initialize res to the interval of active children at the right of dcs */
  dcs = HaArray(eisc->closed_seqs, open_index);
  res = KheDrsAUIntervalMake(dcs->stop_index - dcs->active_at_right,
    dcs->stop_index, true);
  if( !KheDrsClosedSeqAllActive(dcs) )
    return res;

  /* now keep looking to the left of there */
  for( i = open_index - 1;  i >= 0;  i-- )
  {
    /* return early if eisc->min_limit reached */
    if( KheDrsAUIntervalLength(res) >= eisc->min_limit )
      return res;

    /* res includes the open unassigned child before the previous dcs */
    KheDrsAUIntervalExtendToLeft(&res, 1, false);

    /* res includes the active children at the right end of the next dcs */
    dcs = HaArray(eisc->closed_seqs, i);
    KheDrsAUIntervalExtendToLeft(&res, dcs->active_at_right, true);
    if( !KheDrsClosedSeqAllActive(dcs) )
      return res;
  }

  /* at the start, so res includes history */
  KheDrsAUIntervalExtendToLeft(&res, eisc->history, true);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_AU_INTERVAL KheDrsAUIntervalFindRight(                           */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index)                        */
/*                                                                           */
/*  Find the au-interval beginning just after the child whose open index     */
/*  is open_index.                                                           */
/*                                                                           */
/*****************************************************************************/
static char *DrsAUIntervalShow(KHE_DRS_AU_INTERVAL aui);

static KHE_DRS_AU_INTERVAL KheDrsAUIntervalFindRight(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_CLOSED_SEQ dcs;  KHE_DRS_AU_INTERVAL res;  int i;

  /* initialize res to the interval of active children at the left of dcs */
  if( DEBUG20 && drs->rerun != NULL &&
      strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
  {
    fprintf(stderr, "[ KheDrsAUIntervalFindRight(%s, %d, drs)\n",
      KheMonitorId(eisc->monitor), open_index);
    for( i = open_index + 1;  i < HaArrayCount(eisc->closed_seqs);  i++ )
    {
      dcs = HaArray(eisc->closed_seqs, i);
      fprintf(stderr, "  dcs %d ", i);
      KheDrsClosedSeqDebug(dcs, 2, 0, stderr);
    }
  }
  dcs = HaArray(eisc->closed_seqs, open_index + 1);
  res = KheDrsAUIntervalMake(dcs->start_index,
    dcs->start_index + dcs->active_at_left, true);
  if( !KheDrsClosedSeqAllActive(dcs) )
  {
    if( DEBUG20 && drs->rerun != NULL &&
	strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
      fprintf(stderr, "] KheDrsAUIntervalFindRight(eisc, %d) 1= %s\n",
	open_index, DrsAUIntervalShow(res));
    return res;
  }

  /* now keep looking to the right of there */
  for( i = open_index + 2;  i < HaArrayCount(eisc->closed_seqs);  i++ )
  {
    /* return early if eisc->min_limit reached */
    if( KheDrsAUIntervalLength(res) >= eisc->min_limit )
    {
      if( DEBUG20 && drs->rerun != NULL &&
	  strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
	fprintf(stderr, "] KheDrsAUIntervalFindRight(eisc, %d) 2= %s\n",
	  open_index, DrsAUIntervalShow(res));
      return res;
    }

    /* res includes the open unassigned child after the previous dcs */
    KheDrsAUIntervalExtendToRight(&res, 1, false);

    /* res includes the active children at the left end of the next dcs */
    dcs = HaArray(eisc->closed_seqs, i);
    KheDrsAUIntervalExtendToRight(&res, dcs->active_at_left, true);
    if( !KheDrsClosedSeqAllActive(dcs) )
    {
      if( DEBUG20 && drs->rerun != NULL &&
	  strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
	fprintf(stderr, "] KheDrsAUIntervalFindRight(eisc, %d) 3= %s\n",
	  open_index, DrsAUIntervalShow(res));
      return res;
    }
  }

  /* at the end, so res includes history_after */
  KheDrsAUIntervalExtendToRight(&res, eisc->history_after, false);
  if( DEBUG20 && drs->rerun != NULL &&
      strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
    fprintf(stderr, "] KheDrsAUIntervalFindRight(eisc, %d) 4= %s\n",
      open_index, DrsAUIntervalShow(res));
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_AU_INTERVAL KheDrsAUIntervalMerge(KHE_DRS_AU_INTERVAL aui_left,  */
/*    KHE_DRS_AU_INTERVAL aui_right, bool sep_is_active)                     */
/*                                                                           */
/*  Return the merge of aui_left and aui_right.  They are separated by one   */
/*  child, and that child is active if sep_is_active.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_AU_INTERVAL KheDrsAUIntervalMerge(KHE_DRS_AU_INTERVAL aui_left,
  KHE_DRS_AU_INTERVAL aui_right, bool sep_is_active)
{
  HnAssert(aui_left.stop_index + 1 == aui_right.start_index,
    "KheDrsAUIntervalMerge internal error");
  return KheDrsAUIntervalMake(aui_left.start_index, aui_right.stop_index,
    aui_left.has_active_child || aui_right.has_active_child || sep_is_active);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsAUIntervalDev(KHE_DRS_AU_INTERVAL aui,                         */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc)                                        */
/*                                                                           */
/*  Return the deviation of au-interval aui.  If aui.has_active_child is     */
/*  false, the interval has no active child so return 0.  This should        */
/*  also handle the special case of len == 0, where the result is also 0.    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsAUIntervalDev(KHE_DRS_AU_INTERVAL aui,
  KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  int len;

  /* include history if we are at the start */
  /* *** already included where relevant
  if( aui.start_index == 0 && eisc->history > 0 )
  {
    aui.start_index -= eisc->history;
    aui.has_active_child = true;
  }
  *** */

  /* return 0 if the interval has no active child (including len == 0) */
  if( !aui.has_active_child )
    return 0;

  /* include history_after if we are at the end */
  /* *** already included where relevant
  if( aui.stop_index == HaArrayCount(eisc->children) )
    aui.stop_index += eisc->history_after;
  *** */

  /* calculate the length and deviation */
  len = aui.stop_index - aui.start_index;
  return len < eisc->min_limit ? eisc->min_limit - len : 0;
}


/*****************************************************************************/
/*                                                                           */
/*  char *DrsAUIntervalShow(KHE_DRS_AU_INTERVAL aui)                         */
/*                                                                           */
/*  Return a display of aui in static memory.                                */
/*                                                                           */
/*****************************************************************************/

static char *DrsAUIntervalShow(KHE_DRS_AU_INTERVAL aui)
{
  static char buff[20];
  snprintf(buff, 20, "[%d, %d%s]", aui.start_index, aui.stop_index,
    aui.has_active_child ? "a" : "");
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsAUIntervalCost(KHE_DRS_AU_INTERVAL aui,                   */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc)                                        */
/*                                                                           */
/*  Return the cost of aui.                                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsAUIntervalCost(KHE_DRS_AU_INTERVAL aui,
  KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  int dev;
  dev = KheDrsAUIntervalDev(aui, eisc);
  return dev == 0 ? 0 : KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ_COST"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ_COST KheDrsExprIntSeqCostMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,    */
/*    int max_limit, int history_before, int history_after, int history,     */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DRS_RESOURCE dr,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ_COST object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SEQ_COST KheDrsExprIntSeqCostMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,
  int max_limit, int history_before, int history_after, int history,
  /* KHE_DRS_DOM_TEST dom_test, */ KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ_COST res;
  HaMake(res, drs->arena);
  KheDrsExprInitBegin((KHE_DRS_EXPR) res, KHE_DRS_EXPR_INT_SEQ_COST_TAG,
    /* dom_test, */ dr, drs);
  res->u.int_val = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = false;  /* unused field, included by inheritance */
  res->adjust_type = KheDrsAdjustType(cost_fn, max_limit);
  res->history_before = history_before;
  res->history_after = history_after;
  res->history = history;
  res->first_open_child_index = 0;  /* ready for when we open */
  HaArrayInit(res->closed_seqs, drs->arena);
  HaArrayAddLast(res->closed_seqs, KheDrsClosedSeqMakeInit(drs));
  res->rerun_open_and_search_cost = res->rerun_open_and_close_cost = 0;
  HaArrayAddLast(drs->all_cost_exprs, (KHE_DRS_EXPR_COST) res);
  KheDrsAddWeight(drs, combined_weight);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostAddChild(KHE_DRS_EXPR_INT_SEQ_COST eisc,        */
/*    KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  New child child_e has just been added to eisc.  Update eisc's state      */
/*  appropriately.  That is, extend the sole closed sequence one place.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostAddChild(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DRS_EXPR child_e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_CLOSED_SEQ dcs;
  HnAssert(HaArrayCount(eisc->closed_seqs) == 1,
    "KheDrsExprIntSeqCostAddChild internal error (%d closed seqs)",
    HaArrayCount(eisc->closed_seqs));
  dcs = HaArrayFirst(eisc->closed_seqs);
  KheDrsClosedSeqExtendToRight(dcs, child_e->u.int_val);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostClosedSeqDebug(KHE_DRS_EXPR_INT_SEQ_COST eisc,  */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of eisc's closed sequences.                                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostClosedSeqDebug(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_CLOSED_SEQ dcs;  int i;
  HaArrayForEach(eisc->closed_seqs, dcs, i)
    KheDrsClosedSeqDebug(dcs, verbosity, indent, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COST eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Inform internal expression eisc that its child_index'th child, child_e,  */
/*  has opened.  The child will have its correct closed_state and value      */
/*  fields when this is called, and it will have been added to the end of    */
/*  eisc->open_children.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_CLOSED_SEQ dcs, new_dcs;  KHE_DRS_EXPR e;  int i, open_index;
  KHE_DRS_AU_INTERVAL aui_left, aui_right, aui_merged, aui_before, aui_after;
  KHE_DRS_A_INTERVAL ai_left, ai_right, ai_right_before, ai_merged;

  /* find the index of child_e in eisc->open_children (easy: at the end) */
  open_index = HaArrayCount(eisc->open_children) - 1;
  if( DEBUG14 && strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
    fprintf(stderr, "[ KheDrsExprIntSeqCostChildHasOpened(%s, child_e, %d)\n",
      KheMonitorId(eisc->monitor), child_index);

  /* split the last closed seq at child_index */
  dcs = HaArrayLast(eisc->closed_seqs);
  new_dcs = KheDrsClosedSeqSplit(dcs, child_index, eisc, drs);
  HaArrayAddLast(eisc->closed_seqs, new_dcs);
  if( DEBUG14 && strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
  {
    KheDrsExprIntSeqCostClosedSeqDebug(eisc, 2, 2, stderr);
    HaArrayForEach(eisc->children, e, i)
      KheDrsExprDebug(e, drs, 2, 4, stderr);
  }

  /* update the au-intervals and a-intervals to take account of the change */
  if( child_e->u.int_val == 0 )
  {
    /* the opened child moves from inactive to unassigned */
    /* the au-intervals on each side merge */
    aui_left = KheDrsAUIntervalFindLeft(eisc, open_index);
    aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
    aui_merged = KheDrsAUIntervalMerge(aui_left, aui_right, false);
    drs->solve_start_cost += KheDrsAUIntervalCost(aui_merged, eisc)
      - KheDrsAUIntervalCost(aui_left, eisc)
      - KheDrsAUIntervalCost(aui_right, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost,
      &eisc->rerun_open_and_close_cost, eisc->monitor, "open0a", child_index,
      "+--", KheDrsAUIntervalCost(aui_merged, eisc),
      KheDrsAUIntervalCost(aui_left, eisc),
      KheDrsAUIntervalCost(aui_right, eisc));

    /* the a-interval to the right changes its unassigned_precedes */
    if( eisc->cost_fn == KHE_STEP_COST_FUNCTION )
    {
      ai_right_before = KheDrsAIntervalFindRight(eisc, open_index, false);
      ai_right        = KheDrsAIntervalFindRight(eisc, open_index, true);
      drs->solve_start_cost += KheDrsAIntervalCost(ai_right, eisc)
	- KheDrsAIntervalCost(ai_right_before, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost,
        &eisc->rerun_open_and_close_cost, eisc->monitor, "open0b", child_index,
	"+-", KheDrsAIntervalCost(ai_right, eisc),
        KheDrsAIntervalCost(ai_right_before, eisc));
    }
  }
  else
  {
    /* the opened child moves from active to unassigned */
    /* the enclosing au-interval is unchanged, but its cost may change */
    aui_left = KheDrsAUIntervalFindLeft(eisc, open_index);
    aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
    aui_before = KheDrsAUIntervalMerge(aui_left, aui_right, true);
    aui_after = KheDrsAUIntervalMerge(aui_left, aui_right, false);
    drs->solve_start_cost += KheDrsAUIntervalCost(aui_after, eisc)
      - KheDrsAUIntervalCost(aui_before, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost,
      &eisc->rerun_open_and_close_cost, eisc->monitor, "open1a", child_index,
      "+-", KheDrsAUIntervalCost(aui_after, eisc),
      KheDrsAUIntervalCost(aui_before, eisc));

    /* the enclosing a-interval splits */
    /* NB unassigned_precedes is correct when ai_left or ai_right is empty */
    ai_left = KheDrsAIntervalFindLeft(eisc, open_index);
    ai_right = KheDrsAIntervalFindRight(eisc, open_index, true);
    ai_merged = KheDrsAIntervalMerge(ai_left, ai_right);
    drs->solve_start_cost += KheDrsAIntervalCost(ai_left, eisc)
      + KheDrsAIntervalCost(ai_right, eisc)
      - KheDrsAIntervalCost(ai_merged, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost,
      &eisc->rerun_open_and_close_cost, eisc->monitor, "open1b", child_index,
      "++-", KheDrsAIntervalCost(ai_left, eisc),
      KheDrsAIntervalCost(ai_right, eisc),
      KheDrsAIntervalCost(ai_merged, eisc));
  }
  if( DEBUG14 && strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
    fprintf(stderr, "] KheDrsExprIntSeqCostChildHasOpened\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COST eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Inform internal expression eisc that its child_index'th child, child_e,  */
/*  has closed.  The child will have its correct closed_state and value      */
/*  fields when this is called; but it will not have been removed from       */
/*  e->open_children.                                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int open_index;  KHE_DRS_CLOSED_SEQ dcs1, dcs2;
  KHE_DRS_AU_INTERVAL aui_left, aui_right, aui_merged, aui_before, aui_after;
  KHE_DRS_A_INTERVAL ai_left, ai_right, ai_right_before, ai_merged;

  /* find the index of child_e in eisc->open_children */
  open_index = eisc->first_open_child_index;
  HnAssert(HaArray(eisc->open_children, open_index) == child_e,
    "KheDrsExprIntSeqCostChildHasClosed internal error 1");
  eisc->first_open_child_index++;

  /* update the au-intervals and a-intervals to take account of the change */
  if( child_e->u.int_val == 0 )
  {
    /* the closed child moves from unassigned to inactive */
    /* the enclosing au-interval splits */
    aui_left = KheDrsAUIntervalFindLeft(eisc, 0 /* not open_index */);
    aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
    aui_merged = KheDrsAUIntervalMerge(aui_left, aui_right, false);
    drs->solve_start_cost += KheDrsAUIntervalCost(aui_left, eisc)
      + KheDrsAUIntervalCost(aui_right, eisc)
      - KheDrsAUIntervalCost(aui_merged, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_close_cost, NULL,
      eisc->monitor, "close0a", child_index, "++-",
      KheDrsAUIntervalCost(aui_left, eisc),
      KheDrsAUIntervalCost(aui_right, eisc),
      KheDrsAUIntervalCost(aui_merged, eisc));

    /* the a-interval to the right changes its unassigned_precedes */
    ai_right_before = KheDrsAIntervalFindRight(eisc, open_index, true);
    ai_right        = KheDrsAIntervalFindRight(eisc, open_index, false);
    drs->solve_start_cost += KheDrsAIntervalCost(ai_right, eisc)
      - KheDrsAIntervalCost(ai_right_before, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_close_cost, NULL,
      eisc->monitor, "close0b", child_index, "+-",
      KheDrsAIntervalCost(ai_right, eisc),
      KheDrsAIntervalCost(ai_right_before, eisc));
  }
  else
  {
    /* the closed child moves from unassigned to active */
    /* the enclosing au-interval is unchanged, but its cost may change */
    aui_left = KheDrsAUIntervalFindLeft(eisc, 0 /* not open_index */);
    aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
    aui_before = KheDrsAUIntervalMerge(aui_left, aui_right, false);
    aui_after = KheDrsAUIntervalMerge(aui_left, aui_right, true);
    drs->solve_start_cost += KheDrsAUIntervalCost(aui_after, eisc)
      - KheDrsAUIntervalCost(aui_before, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_close_cost, NULL,
      eisc->monitor, "close1a", child_index, "+-",
      KheDrsAUIntervalCost(aui_after, eisc),
      KheDrsAUIntervalCost(aui_before, eisc));

    /* the a-intervals on each side merge */
    ai_left = KheDrsAIntervalFindLeft(eisc, 0 /* not open_index */);
    ai_right = KheDrsAIntervalFindRight(eisc, open_index, true);
    ai_merged = KheDrsAIntervalMerge(ai_left, ai_right);
    drs->solve_start_cost += KheDrsAIntervalCost(ai_merged, eisc)
      - KheDrsAIntervalCost(ai_left, eisc)
      - KheDrsAIntervalCost(ai_right, eisc);
    KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_close_cost, NULL,
      eisc->monitor, "close1b", child_index, "+--",
      KheDrsAIntervalCost(ai_merged, eisc),
      KheDrsAIntervalCost(ai_left, eisc), KheDrsAIntervalCost(ai_right, eisc));
  }

  /* merge the two relevant closed seqs */
  dcs1 = HaArray(eisc->closed_seqs, 0);
  dcs2 = HaArray(eisc->closed_seqs, open_index + 1);
  KheDrsClosedSeqMerge(dcs1, dcs2, eisc, drs);
  HaArrayPut(eisc->closed_seqs, open_index + 1, NULL);  /* defensive */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostSetClosedValue(KHE_DRS_EXPR_INT_SEQ_COST eisc,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eisc, set its value suitably for its closed state.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostSetClosedValue(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* clear out eisc->first_open_child_index and eisc->closed_seqs */
  eisc->first_open_child_index = 0;
  while( HaArrayCount(eisc->closed_seqs) > 0 &&
      HaArrayLast(eisc->closed_seqs) == NULL )
    HaArrayDeleteLast(eisc->closed_seqs);
  HnAssert(HaArrayCount(eisc->closed_seqs) == 1,
    "KheDrsExprIntSeqCostSetClosedValue internal error 1");

  /* nothing else to do here, since eisc has a cost but no value */
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSeqCostChildDomTest(                       */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_child_index)                  */
/*                                                                           */
/*  Return a dom test suited to a child of eisc on its open_child_index'th   */
/*  open day.                                                                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprIntSeqCostChildDomTest(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_child_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int a, b;

  /* calculate a */
  a = (eisc->max_limit < INT_MAX ? -1 : INT_MAX);

  /* calculate b */
  b = (eisc->min_limit > 0 ? INT_MAX : 0);

  /* make and return the dom test */
  return KheDrsDomTestMake(false, 0, INT_MAX, a, b, false, 0, NULL, drs);
}



/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSeqCostDomTest(                            */
/*    KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_child_index)                  */
/*                                                                           */
/*  Return a dom test suited to eisc on its open_child_index'th open day.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprIntSeqCostDomTest(
  KHE_DRS_EXPR_INT_SEQ_COST eisc, int open_child_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int a, b, q;

  /* calculate a */
  if( eisc->max_limit < INT_MAX )
  {
    q = KheDrsExprOpenChildrenAtOrAfter((KHE_DRS_EXPR) eisc, open_child_index);
    a = eisc->max_limit - q;
  }
  else
    a = INT_MAX;

  /* calculate b */
  if( eisc->min_limit > 0 )
    b = eisc->min_limit;
  else
    b = 0;

  /* make and return the dom test */
  return KheDrsDomTestMake(eisc->allow_zero, eisc->min_limit, eisc->max_limit,
    a, b, eisc->cost_fn != KHE_QUADRATIC_COST_FUNCTION, eisc->combined_weight,
    eisc->monitor, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDebugActiveLen(KHE_DRS_EXPR_INT_SEQ_COST eisc, char *label,   */
/*    KHE_DRS_CLOSED_SEQ dcs, int active_len,KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Debug these values if wanted.                                            */
/*                                                                           */
/*****************************************************************************/
/* ***
static void KheDrsExprDebugSoln(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  int verbosity, int indent, FILE *fp);
*** */

static void KheDrsDebugActiveLen(KHE_DRS_EXPR_INT_SEQ_COST eisc, char *label,
  int open_index, KHE_DRS_EXPR child_e, KHE_DRS_SOLN prev_soln,
  KHE_DRS_CLOSED_SEQ dcs, int active_len, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( RERUN && drs->rerun != NULL &&
      strcmp(KheMonitorId(eisc->monitor), RERUN_MONITOR_ID) == 0 )
  {
    fprintf(stderr, "  [ active_len %s: ", label);
    KheDrsClosedSeqDebug(dcs, 1, -1, stderr);
    fprintf(stderr, " --> %d\n", active_len);
    fprintf(stderr, "    child at open_index %d:", open_index);
    /* KheDrsExprDebugSoln(child_e, prev_soln, 1, 2, stderr); */
    KheDrsExprDebug(child_e, drs, 1, 2, stderr);
    fprintf(stderr, "  ]\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostEvalDay(KHE_DRS_EXPR_INT_SEQ_COST eisc,         */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc in next_soln.                                   */
/*                                                                           */
/*  Implementation note.  Much of this code is the same as the code for      */
/*  closing each child_e, except that the a-interval and au-interval at      */
/*  left both have length active_len, and the cost is updated in             */
/*  next_soln->cost rather than in drs->solve_start_cost.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCostEvalDay(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int open_index, i1, i2, active_len, di, adjusted_len;  KHE_DRS_EXPR child_e;
  KHE_DRS_AU_INTERVAL aui_left, aui_right, aui_merged;
  KHE_DRS_AU_INTERVAL aui_before, aui_after;
  KHE_DRS_A_INTERVAL ai_right_before, ai_right_after;
  KHE_DRS_A_INTERVAL ai_left, ai_right, ai_merged;
  KHE_DRS_CLOSED_SEQ dcs;

  ** get di, the open day index of next_soln **
  di = KheDrsSolnOpenDayIndex(next_soln);

  ** initialize active_len **
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eisc, di) )
  {
    ** first day **
    dcs = HaArrayFirst(eisc->closed_seqs);
    active_len = dcs->active_at_right;
    if( KheDrsClosedSeqAllActive(dcs) )
      active_len += eisc->history;
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  _%d", active_len);
  }
  else
  {
    ** not first day, so retrieve previous active_len **
    active_len = KheDrsExprSigVal((KHE_DRS_EXPR) eisc, di - 1, prev_soln);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  $%d", active_len);
  }

  ** handle each child_e whose last open day is di **
  KheDrsExprForEachOpenDayChild(eisc, di, child_e, open_index)
  {
    if( child_e->u.int_val == 0 )
    {
      ** child_e moves from unassigned to inactive: update cost **
      ** the enclosing au-interval splits **
      aui_left = KheDrsAUIntervalMakeLeft(eisc, open_index, active_len);
      aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
      aui_merged = KheDrsAUIntervalMerge(aui_left, aui_right, false);
      next_soln->cost += KheDrsAUIntervalCost(aui_left, eisc)
	+ KheDrsAUIntervalCost(aui_right, eisc)
	- KheDrsAUIntervalCost(aui_merged, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search", -1, "++-",
	KheDrsAUIntervalCost(aui_left, eisc),
	KheDrsAUIntervalCost(aui_right, eisc),
	KheDrsAUIntervalCost(aui_merged, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 0aui +%.5f +%.5f -%.5f",
	  KheCostShow(KheDrsAUIntervalCost(aui_left, eisc)),
          KheCostShow(KheDrsAUIntervalCost(aui_right, eisc)),
          KheCostShow(KheDrsAUIntervalCost(aui_merged, eisc)));

      ** the a-interval to the right changes its unassigned_precedes **
      ai_right_before = KheDrsAIntervalFindRight(eisc, open_index, true);
      ai_right_after  = KheDrsAIntervalFindRight(eisc, open_index, false);
      next_soln->cost += KheDrsAIntervalCost(ai_right_after, eisc)
	- KheDrsAIntervalCost(ai_right_before, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search2", -1, "+-",
        KheDrsAIntervalCost(ai_right_after, eisc),
        KheDrsAIntervalCost(ai_right_before, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 0ai +%.5f -%.5f",
	  KheCostShow(KheDrsAIntervalCost(ai_right_after, eisc)),
          KheCostShow(KheDrsAIntervalCost(ai_right_before, eisc)));

      ** set active_len for next iteration (child_e is now inactive) **
      dcs = HaArray(eisc->closed_seqs, open_index + 1);
      active_len = dcs->active_at_right;
      KheDrsDebugActiveLen(eisc, "inactive", open_index, child_e, prev_soln,
	dcs, active_len, drs);
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " %d", active_len);
    }
    else
    {
      ** child_e moves from unassigned to active: update cost **
      ** the enclosing au-interval is unchanged, but its cost may change **
      aui_left = KheDrsAUIntervalMakeLeft(eisc, open_index, active_len);
      aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
      aui_before = KheDrsAUIntervalMerge(aui_left, aui_right, false);
      aui_after = KheDrsAUIntervalMerge(aui_left, aui_right, true);
      next_soln->cost += KheDrsAUIntervalCost(aui_after, eisc)
	- KheDrsAUIntervalCost(aui_before, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search", -1, "+-",
        KheDrsAUIntervalCost(aui_after, eisc),
	KheDrsAUIntervalCost(aui_before, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 1aui +%.5f -%.5f",
	  KheCostShow(KheDrsAUIntervalCost(aui_after, eisc)),
          KheCostShow(KheDrsAUIntervalCost(aui_before, eisc)));

      ** the a-intervals on each side merge **
      ai_left = KheDrsAIntervalMakeLeft(eisc, open_index, active_len);
      ai_right = KheDrsAIntervalFindRight(eisc, open_index, true);
      ai_merged = KheDrsAIntervalMerge(ai_left, ai_right);
      next_soln->cost += KheDrsAIntervalCost(ai_merged, eisc)
	- KheDrsAIntervalCost(ai_left, eisc)
	- KheDrsAIntervalCost(ai_right, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search2", -1, "+--",
        KheDrsAIntervalCost(ai_merged, eisc),
	KheDrsAIntervalCost(ai_left, eisc),
	KheDrsAIntervalCost(ai_right, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 1ai +%.5f -%.5f -%.5f",
	  KheCostShow(KheDrsAIntervalCost(ai_merged, eisc)),
          KheCostShow(KheDrsAIntervalCost(ai_left, eisc)),
          KheCostShow(KheDrsAIntervalCost(ai_right, eisc)));

      ** set active_len for next iteration (child_e is now active) **
      dcs = HaArray(eisc->closed_seqs, open_index + 1);
      if( KheDrsClosedSeqAllActive(dcs) )
	active_len += 1 + dcs->active_at_right;
      else
        active_len = dcs->active_at_right;
      KheDrsDebugActiveLen(eisc, "active", open_index, child_e, prev_soln,
	dcs, active_len, drs);
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " %d", active_len);
    }
  }

  if( !KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eisc, di) )
  {
    ** not last day; store active_len (adjusted) in next_soln **
    adjusted_len = KheDrsAdjustedSigVal(active_len,
      eisc->adjust_type, eisc->min_limit, eisc->max_limit, 0, 0);
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eisc, di, next_soln, adjusted_len);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, ":%d$ %s\n", adjusted_len, KheMonitorId(eisc->monitor));
  }
  else
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "_ %s\n", KheMonitorId(eisc->monitor));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostEvalSignature(KHE_DRS_EXPR_INT_SEQ_COST eisc,   */
/*    KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE sig,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc in sig.                                         */
/*                                                                           */
/*  Implementation note.  Much of this code is the same as the code for      */
/*  closing each child_e, except that the a-interval and au-interval at      */
/*  left both have length active_len, and the cost is updated in             */
/*  next_soln->cost rather than in drs->solve_start_cost.                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostEvalSignature(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DRS_SOLN prev_soln, int next_di, KHE_DRS_SIGNATURE sig,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int open_index, i1, i2, active_len, adjusted_len;  KHE_DRS_EXPR child_e;
  KHE_DRS_AU_INTERVAL aui_left, aui_right, aui_merged;
  KHE_DRS_AU_INTERVAL aui_before, aui_after;
  KHE_DRS_A_INTERVAL ai_right_before, ai_right_after;
  KHE_DRS_A_INTERVAL ai_left, ai_right, ai_merged;
  KHE_DRS_CLOSED_SEQ dcs;

  /* initialize active_len */
  if( KheDrsExprOpenDayIsFirst((KHE_DRS_EXPR) eisc, next_di) )
  {
    /* first day */
    dcs = HaArrayFirst(eisc->closed_seqs);
    active_len = dcs->active_at_right;
    if( KheDrsClosedSeqAllActive(dcs) )
      active_len += eisc->history;
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  _%d", active_len);
  }
  else
  {
    /* not first day, so retrieve previous active_len */
    active_len = KheDrsExprSigVal((KHE_DRS_EXPR) eisc, next_di - 1, prev_soln);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "  $%d", active_len);
  }

  /* handle each child_e whose last open day is next_di */
  KheDrsExprForEachOpenDayChild(eisc, next_di, child_e, open_index)
  {
    if( child_e->u.int_val == 0 )
    {
      /* child_e moves from unassigned to inactive: update cost */
      /* the enclosing au-interval splits */
      aui_left = KheDrsAUIntervalMakeLeft(eisc, open_index, active_len);
      aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
      aui_merged = KheDrsAUIntervalMerge(aui_left, aui_right, false);
      sig->cost += KheDrsAUIntervalCost(aui_left, eisc)
	+ KheDrsAUIntervalCost(aui_right, eisc)
	- KheDrsAUIntervalCost(aui_merged, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search", -1, "++-",
	KheDrsAUIntervalCost(aui_left, eisc),
	KheDrsAUIntervalCost(aui_right, eisc),
	KheDrsAUIntervalCost(aui_merged, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 0aui +%.5f +%.5f -%.5f",
	  KheCostShow(KheDrsAUIntervalCost(aui_left, eisc)),
          KheCostShow(KheDrsAUIntervalCost(aui_right, eisc)),
          KheCostShow(KheDrsAUIntervalCost(aui_merged, eisc)));

      /* the a-interval to the right changes its unassigned_precedes */
      ai_right_before = KheDrsAIntervalFindRight(eisc, open_index, true);
      ai_right_after  = KheDrsAIntervalFindRight(eisc, open_index, false);
      sig->cost += KheDrsAIntervalCost(ai_right_after, eisc)
	- KheDrsAIntervalCost(ai_right_before, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search2", -1, "+-",
        KheDrsAIntervalCost(ai_right_after, eisc),
        KheDrsAIntervalCost(ai_right_before, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 0ai +%.5f -%.5f",
	  KheCostShow(KheDrsAIntervalCost(ai_right_after, eisc)),
          KheCostShow(KheDrsAIntervalCost(ai_right_before, eisc)));

      /* set active_len for next iteration (child_e is now inactive) */
      dcs = HaArray(eisc->closed_seqs, open_index + 1);
      active_len = dcs->active_at_right;
      KheDrsDebugActiveLen(eisc, "inactive", open_index, child_e, prev_soln,
	dcs, active_len, drs);
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " %d", active_len);
    }
    else
    {
      /* child_e moves from unassigned to active: update cost */
      /* the enclosing au-interval is unchanged, but its cost may change */
      aui_left = KheDrsAUIntervalMakeLeft(eisc, open_index, active_len);
      aui_right = KheDrsAUIntervalFindRight(eisc, open_index, drs);
      aui_before = KheDrsAUIntervalMerge(aui_left, aui_right, false);
      aui_after = KheDrsAUIntervalMerge(aui_left, aui_right, true);
      sig->cost += KheDrsAUIntervalCost(aui_after, eisc)
	- KheDrsAUIntervalCost(aui_before, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search", -1, "+-",
        KheDrsAUIntervalCost(aui_after, eisc),
	KheDrsAUIntervalCost(aui_before, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 1aui +%.5f -%.5f",
	  KheCostShow(KheDrsAUIntervalCost(aui_after, eisc)),
          KheCostShow(KheDrsAUIntervalCost(aui_before, eisc)));

      /* the a-intervals on each side merge */
      ai_left = KheDrsAIntervalMakeLeft(eisc, open_index, active_len);
      ai_right = KheDrsAIntervalFindRight(eisc, open_index, true);
      ai_merged = KheDrsAIntervalMerge(ai_left, ai_right);
      sig->cost += KheDrsAIntervalCost(ai_merged, eisc)
	- KheDrsAIntervalCost(ai_left, eisc)
	- KheDrsAIntervalCost(ai_right, eisc);
      KheDrsUpdateRerunCost(drs, &eisc->rerun_open_and_search_cost, NULL,
	eisc->monitor, "search2", -1, "+--",
        KheDrsAIntervalCost(ai_merged, eisc),
	KheDrsAIntervalCost(ai_left, eisc),
	KheDrsAIntervalCost(ai_right, eisc));
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " 1ai +%.5f -%.5f -%.5f",
	  KheCostShow(KheDrsAIntervalCost(ai_merged, eisc)),
          KheCostShow(KheDrsAIntervalCost(ai_left, eisc)),
          KheCostShow(KheDrsAIntervalCost(ai_right, eisc)));

      /* set active_len for next iteration (child_e is now active) */
      dcs = HaArray(eisc->closed_seqs, open_index + 1);
      if( KheDrsClosedSeqAllActive(dcs) )
	active_len += 1 + dcs->active_at_right;
      else
        active_len = dcs->active_at_right;
      KheDrsDebugActiveLen(eisc, "active", open_index, child_e, prev_soln,
	dcs, active_len, drs);
      if( DEBUG_LAST_EDGE )
	fprintf(stderr, " %d", active_len);
    }
  }

  if( !KheDrsExprOpenDayIsLast((KHE_DRS_EXPR) eisc, next_di) )
  {
    /* not last day; store active_len (adjusted) in sig */
    adjusted_len = KheDrsAdjustedSigVal(active_len,
      eisc->adjust_type, eisc->min_limit, eisc->max_limit, 0);
    KheDrsExprPutSigVal((KHE_DRS_EXPR) eisc, next_di, sig, adjusted_len);
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, ":%d$ %s\n", adjusted_len, KheMonitorId(eisc->monitor));
  }
  else
    if( DEBUG_LAST_EDGE )
      fprintf(stderr, "_ %s\n", KheMonitorId(eisc->monitor));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostInitRerunCost(KHE_DRS_EXPR_INT_SEQ_COST eisc)   */
/*                                                                           */
/*  Initialize the rerun cost fields of eisc.                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCostInitRerunCost(KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  eisc->rerun_open_and_search_cost = eisc->rerun_open_and_close_cost =
    KheMonitorCost(eisc->monitor);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostCheckRerunCost(KHE_DRS_EXPR_INT_SEQ_COST eisc)  */
/*                                                                           */
/*  Check the rerun cost fields of eisc.                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCostCheckRerunCost(KHE_DRS_EXPR_INT_SEQ_COST eisc)
{
  if( eisc->rerun_open_and_search_cost != KheMonitorCost(eisc->monitor) )
    fprintf(stderr, "  open_and_search_cost %.5f, monitor cost %.5f (%s)\n",
      KheCostShow(eisc->rerun_open_and_search_cost),
      KheCostShow(KheMonitorCost(eisc->monitor)), KheMonitorId(eisc->monitor));
  if( eisc->rerun_open_and_close_cost != KheMonitorCost(eisc->monitor) )
    fprintf(stderr, "  open_and_close_cost  %.5f, monitor cost %.5f (%s)\n",
      KheCostShow(eisc->rerun_open_and_close_cost),
      KheCostShow(KheMonitorCost(eisc->monitor)), KheMonitorId(eisc->monitor));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCostDoDebug(KHE_DRS_EXPR_INT_SEQ_COST eisc,         */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eisc which is specific to it.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCostDoDebug(KHE_DRS_EXPR_INT_SEQ_COST eisc,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_SEQ_COST(%s, %s)", KheMonitorId(eisc->monitor),
    KheDrsAdjustTypeShow(eisc->adjust_type));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eisc, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - deferred functions"                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make child a child of parent.                                            */
/*                                                                           */
/*  Implementation note.  During expression construction, all expressions    */
/*  are closed by the time they are made children of their parents (by       */
/*  calls on this function).  Hence KheDrsExprChildHasClosed below.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;

  /* link parent and child */
  HnAssert(HaArrayCount(parent->parents) == 0,
    "KheDrsExprAddChild internal error:  too late to add child");
  prnt.expr = parent;
  prnt.index = HaArrayCount(parent->children);
  HaArrayAddLast(child->parents, prnt);
  HaArrayAddLast(parent->children, child);

  /* update state in each parent */
  switch( parent->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrAddChild((KHE_DRS_EXPR_OR) parent, child, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndAddChild((KHE_DRS_EXPR_AND) parent, child, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumAddChild((KHE_DRS_EXPR_INT_SUM) parent, child, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumAddChild((KHE_DRS_EXPR_FLOAT_SUM) parent, child, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevAddChild((KHE_DRS_EXPR_INT_DEV) parent, child, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevAddChild((KHE_DRS_EXPR_FLOAT_DEV) parent, child, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostAddChild((KHE_DRS_EXPR_INT_SUM_COST) parent,
	child, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostAddChild((KHE_DRS_EXPR_INT_SEQ_COST) parent,
	child, drs);
      break;

    default:

      HnAbort("KheDrsExprAddChild internal error (tag %d)", parent->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,      */
/*    int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  As part of opening e for solving, inform e that its child_index'th       */
/*  child, child_e, has opened.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,
  int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_EXPR e2;

  /* add child_e to e->open_children */
  if( e->tag == KHE_DRS_EXPR_INT_SEQ_COST_TAG )
  {
    /* increase child_e's last open day if required */
    if( HaArrayCount(e->open_children) > 0 )
    {
      e2 = HaArrayLast(e->open_children);
      if( e2->open_day_range.last > child_e->open_day_range.last )
        child_e->open_day_range.last = e2->open_day_range.last;
    }

    /* add child_e to the end of e->open_children */
    HaArrayAddLast(e->open_children, child_e);
  }
  else
  {
    /* add child_e to e->open_children in last day order */
    HaArrayAddLast(e->open_children, NULL);
    for( i = HaArrayCount(e->open_children) - 2;  i >= 0;  i-- )
    {
      e2 = HaArray(e->open_children, i);
      if( e2->open_day_range.last <= child_e->open_day_range.last )
	break;
      HaArrayPut(e->open_children, i + 1, e2);
    }
    HaArrayPut(e->open_children, i + 1, child_e);
  }

  /* add child_e's last open day to e's open_day_range */
  KheDrsDayRangeAdd(&e->open_day_range, child_e->open_day_range.last);

  /* alter the state of e to take account of child_e opening */
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasOpened((KHE_DRS_EXPR_OR) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasOpened((KHE_DRS_EXPR_AND) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasOpened((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasOpened((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasOpened((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasOpened((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostChildHasOpened((KHE_DRS_EXPR_INT_SUM_COST) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostChildHasOpened((KHE_DRS_EXPR_INT_SEQ_COST) e,
	child_e, child_index, drs);
      break;

    default:

      HnAbort("KheDrsExprChildHasOpened internal error (tag %d)", e->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,                            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Inform internal expression e that its child_index'th child, child_e,     */
/*  has closed.  The child will have its correct closed_state and value      */
/*  fields when this is called.                                              */
/*                                                                           */
/*  Although KheDrsExprChildHasOpened adds child_e to e->open_children,      */
/*  KheDrsExprChildHasClosed does not remove child_e from e->open_children.  */
/*  This is because children are closed in the same order they were opened,  */
/*  but there is no efficient way to remove them in this order, short of     */
/*  something too complicated such as a circular list.                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* remove child_e from e->open_children */
  /* *** no longer doing this
  int pos;
  if( !HaArrayContains(e->open_children, child_e, &pos) )
    HnAbort("KheDrsExprChildHasClosed internal error");
  HaArrayDeleteAndPlug(e->open_children, pos);
  *** */

  /* do the type-specific part of this operation */
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasClosed((KHE_DRS_EXPR_OR) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasClosed((KHE_DRS_EXPR_AND) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasClosed((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasClosed((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasClosed((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasClosed((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostChildHasClosed((KHE_DRS_EXPR_INT_SUM_COST) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostChildHasClosed((KHE_DRS_EXPR_INT_SEQ_COST) e,
	child_e, child_index, drs);
      break;

    default:

      HnAbort("KheDrsExprChildHasClosed internal error (tag %d)", e->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing e, set its value suitably for its closed state.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskSetClosedValue((KHE_DRS_EXPR_ASSIGNED_TASK) e, drs);
      break;

    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeSetClosedValue((KHE_DRS_EXPR_BUSY_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeSetClosedValue((KHE_DRS_EXPR_FREE_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeSetClosedValue((KHE_DRS_EXPR_WORK_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDaySetClosedValue((KHE_DRS_EXPR_BUSY_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDaySetClosedValue((KHE_DRS_EXPR_FREE_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDaySetClosedValue((KHE_DRS_EXPR_WORK_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrSetClosedValue((KHE_DRS_EXPR_OR) e, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndSetClosedValue((KHE_DRS_EXPR_AND) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumSetClosedValue((KHE_DRS_EXPR_INT_SUM) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumSetClosedValue((KHE_DRS_EXPR_FLOAT_SUM) e,drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevSetClosedValue((KHE_DRS_EXPR_INT_DEV) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevSetClosedValue((KHE_DRS_EXPR_FLOAT_DEV) e,drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostSetClosedValue((KHE_DRS_EXPR_INT_SUM_COST) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostSetClosedValue((KHE_DRS_EXPR_INT_SEQ_COST) e, drs);
      break;

    default:

      HnAbort("KheDrsExprSetClosedValue internal error (tag %d)", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprChildDomTest(KHE_DRS_EXPR e,                  */
/*    int open_child_index)                                                  */
/*                                                                           */
/*  Return a dom test suited to a child of e on its open_child_index'th      */
/*  open day.                                                                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprChildDomTest(KHE_DRS_EXPR e,
  int open_child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT parent;
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:

      HnAbort("KheDrsExprChildDomTest: unexpected tag %d", e->tag);
      /* no break, to keep compiler happy */

    case KHE_DRS_EXPR_OR_TAG:
    case KHE_DRS_EXPR_AND_TAG:
    case KHE_DRS_EXPR_INT_SUM_TAG:
    case KHE_DRS_EXPR_FLOAT_SUM_TAG:
    case KHE_DRS_EXPR_INT_DEV_TAG:
    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      HnAssert(HaArrayCount(e->parents) > 0,
	"KheDrsExprChildDomTest internal error");
      parent = HaArrayFirst(e->parents);
      return KheDrsExprChildDomTest(parent.expr, open_child_index, drs);

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      return KheDrsExprIntSumCostChildDomTest((KHE_DRS_EXPR_INT_SUM_COST) e,
	open_child_index, drs);

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      return KheDrsExprIntSeqCostChildDomTest((KHE_DRS_EXPR_INT_SEQ_COST) e,
	open_child_index, drs);

    default:

      HnAbort("KheDrsExprChildDomTest internal error (tag %d)", e->tag);
      return KheDrsExprIntSeqCostChildDomTest((KHE_DRS_EXPR_INT_SEQ_COST) e,
	open_child_index, drs);  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprDomTest(KHE_DRS_EXPR e, int open_child_index) */
/*                                                                           */
/*  Return a dom test suited to e on its open_child_index'th open day.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsExprDomTest(KHE_DRS_EXPR e,
  int open_child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT parent;
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:

      HnAbort("KheDrsExprDomTest: unexpected tag %d", e->tag);
      /* no break, to keep compiler happy */

    case KHE_DRS_EXPR_OR_TAG:
    case KHE_DRS_EXPR_AND_TAG:
    case KHE_DRS_EXPR_INT_SUM_TAG:
    case KHE_DRS_EXPR_FLOAT_SUM_TAG:
    case KHE_DRS_EXPR_INT_DEV_TAG:
    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      HnAssert(HaArrayCount(e->parents) > 0,
	"KheDrsExprDomTest internal error");
      parent = HaArrayFirst(e->parents);
      return KheDrsExprChildDomTest(parent.expr, open_child_index, drs);

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      return KheDrsExprIntSumCostDomTest((KHE_DRS_EXPR_INT_SUM_COST) e,
	open_child_index, drs);

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      return KheDrsExprIntSeqCostDomTest((KHE_DRS_EXPR_INT_SEQ_COST) e,
	open_child_index, drs);

    default:

      HnAbort("KheDrsExprDomTest internal error (tag %d)", e->tag);
      return KheDrsExprIntSeqCostDomTest((KHE_DRS_EXPR_INT_SEQ_COST) e,
	open_child_index, drs);  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,          */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Inform leaf expression e that dtd has been assigned drd.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskLeafSet((KHE_DRS_EXPR_ASSIGNED_TASK) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeLeafSet((KHE_DRS_EXPR_BUSY_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeLeafSet((KHE_DRS_EXPR_FREE_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeLeafSet((KHE_DRS_EXPR_WORK_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayLeafSet((KHE_DRS_EXPR_BUSY_DAY) e, dtd, drd, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayLeafSet((KHE_DRS_EXPR_FREE_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayLeafSet((KHE_DRS_EXPR_WORK_DAY) e, dtd, drd);
      break;

    default:

      HnAbort("KheDrsExprLeafSet internal error: tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafClear(KHE_DRS_EXPR e)                                 */
/*                                                                           */
/*  Inform leaf expression e that the assignment of drd to dtd carried out   */
/*  by KheDrsExprLeafSet above has been removed.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafClear(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskLeafClear((KHE_DRS_EXPR_ASSIGNED_TASK) e);
      break;

    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeLeafClear((KHE_DRS_EXPR_BUSY_TIME) e);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeLeafClear((KHE_DRS_EXPR_FREE_TIME) e);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeLeafClear((KHE_DRS_EXPR_WORK_TIME) e);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayLeafClear((KHE_DRS_EXPR_BUSY_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayLeafClear((KHE_DRS_EXPR_FREE_DAY) e);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayLeafClear((KHE_DRS_EXPR_WORK_DAY) e);
      break;

    default:

      HnAbort("KheDrsExprLeafClear internal error: tag %d");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEvalDay(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,           */
/*    KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Evaluate expression e in next_soln, whose day is one of e's open days.   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprEvalDay(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,
  KHE_DRS_SOLN next_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:

      HnAbort("KheDrsExprEvalDay: external tag (%d)", e->tag);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrEvalDay((KHE_DRS_EXPR_OR) e, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndEvalDay((KHE_DRS_EXPR_AND) e, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumEvalDay((KHE_DRS_EXPR_INT_SUM) e,
	prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumEvalDay((KHE_DRS_EXPR_FLOAT_SUM) e,
	prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevEvalDay((KHE_DRS_EXPR_INT_DEV) e,
	prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevEvalDay((KHE_DRS_EXPR_FLOAT_DEV) e,
	prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostEvalDay((KHE_DRS_EXPR_INT_SUM_COST) e,
	prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostEvalDay((KHE_DRS_EXPR_INT_SEQ_COST) e,
	prev_soln, next_soln, drs);
      break;

    default:

      HnAbort("KheDrsExprEvalDay internal error");
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEvalSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,     */
/*    int next_di, KHE_DRS_SIGNATURE ss, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Like KheDrsExprEvalDay except that the results go into ss rather         */
/*  than into next_soln.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprEvalSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN prev_soln,
  int next_di, KHE_DRS_SIGNATURE sig, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:

      HnAbort("KheDrsExprEvalSignature: external tag (%d)", e->tag);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrEvalSignature((KHE_DRS_EXPR_OR) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndEvalSignature((KHE_DRS_EXPR_AND) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumEvalSignature((KHE_DRS_EXPR_INT_SUM) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumEvalSignature((KHE_DRS_EXPR_FLOAT_SUM) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevEvalSignature((KHE_DRS_EXPR_INT_DEV) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevEvalSignature((KHE_DRS_EXPR_FLOAT_DEV) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostEvalSignature((KHE_DRS_EXPR_INT_SUM_COST) e,
	prev_soln, next_di, sig, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostEvalSignature((KHE_DRS_EXPR_INT_SEQ_COST) e,
	prev_soln, next_di, sig, drs);
      break;

    default:

      HnAbort("KheDrsExprEvalSignature internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostInitRerunCost(KHE_DRS_EXPR_COST ec)                   */
/*                                                                           */
/*  Initialize the rerun cost of e, if it has a rerun cost field.            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostInitRerunCost(KHE_DRS_EXPR_COST ec)
{
  ec->rerun_open_and_search_cost = ec->rerun_open_and_close_cost =
    KheMonitorCost(ec->monitor);
  /* ***
  switch( e->tag )
  {
    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostInitRerunCost((KHE_DRS_EXPR_INT_SUM_COST) e);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostInitRerunCost((KHE_DRS_EXPR_INT_SEQ_COST) e);
      break;

    default:

      ** do nothing **
      break;
  }
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp)      */
/*                                                                           */
/*  Unindented debug print of that part of e which is specific to it.        */
/*  If soln != NULL, also print e's signature (if any) in soln.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDoDebug((KHE_DRS_EXPR_ASSIGNED_TASK) e, soln, fp);
      break;

    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDoDebug((KHE_DRS_EXPR_BUSY_TIME) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDoDebug((KHE_DRS_EXPR_FREE_TIME) e, soln, fp);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDoDebug((KHE_DRS_EXPR_WORK_TIME) e, soln, fp);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDoDebug((KHE_DRS_EXPR_BUSY_DAY) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDoDebug((KHE_DRS_EXPR_FREE_DAY) e, soln, fp);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDoDebug((KHE_DRS_EXPR_WORK_DAY) e, soln, fp);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDoDebug((KHE_DRS_EXPR_OR) e, soln, fp);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDoDebug((KHE_DRS_EXPR_AND) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDoDebug((KHE_DRS_EXPR_INT_SUM) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDoDebug((KHE_DRS_EXPR_FLOAT_SUM) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDoDebug((KHE_DRS_EXPR_INT_DEV) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDoDebug((KHE_DRS_EXPR_FLOAT_DEV) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_COST_TAG:

      KheDrsExprIntSumCostDoDebug((KHE_DRS_EXPR_INT_SUM_COST) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COST_TAG:

      KheDrsExprIntSeqCostDoDebug((KHE_DRS_EXPR_INT_SEQ_COST) e, soln, fp);
      break;

    default:

      HnAbort("KheDrsExprDoDebug internal error: unknown tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "conversion of event resource monitors into expression trees"  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorCmp(const void *t1, const void *t2)                        */
/*                                                                           */
/*  Comparison function for sorting (and uniqueifying) an array of monitors. */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorCmp(const void *t1, const void *t2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) t1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) t2;
  return KheMonitorSolnIndex(m1) - KheMonitorSolnIndex(m2);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheTaskToDrsTask(KHE_TASK task,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Convert task into the drs task corresponding to its proper root.         */
/*  The value is guaranteed to be non-NULL.                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheTaskToDrsTask(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int task_index;  KHE_DRS_TASK res;
  task_index = KheTaskSolnIndex(KheTaskProperRoot(task));
  if( task_index >= HaArrayCount(drs->all_root_tasks) )
    res = NULL;
  else
    res = HaArray(drs->all_root_tasks, task_index);
  HnAssert(res != NULL, "KheTaskToDrsTask internal error");
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddAssignResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_ASSIGN_RESOURCE_MONITOR arm)                                       */
/*                                                                           */
/*  Make and add in one or more expression trees representing arm.           */
/*  The context of the call ensures that arm has the correct resource type.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddAssignResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_ASSIGN_RESOURCE_MONITOR arm)
{
  KHE_DRS_EXPR_INT_SUM_COST eisc;  KHE_CONSTRAINT c;  KHE_EVENT_RESOURCE er;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  /* KHE_DRS_DOM_TEST dom_test; */
  int min_limit, i, j;  KHE_DRS_TASK dt;  KHE_TASK task;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  er = KheAssignResourceMonitorEventResource(arm);
  HnAssert(KheEventResourceResourceType(er) == drs->resource_type,
    "KheDrsAddAssignResourceMonitor internal error");
  c = KheMonitorConstraint((KHE_MONITOR) arm);
  combined_wt = KheConstraintCombinedWeight(c);
  if( combined_wt > 0 )
  {
    cost_fn = KheConstraintCostFunction(c);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    {
      /* one tree per task on day */
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	HaArrayForEach(dt->days, dtd, j)
	  if( dtd->task == task )
	  {
	    eisc = KheDrsExprIntSumCostMake((KHE_MONITOR) arm, cost_fn,
	      combined_wt, 1, INT_MAX, false, 0, 0, 0,
	      /* KheDrsDomTestMakeUnused(), */ NULL, drs);
	    eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
	  }
      }
    }
    else
    {
      /* a single tree, whose root has one child per task on day */
      min_limit = KheEventDuration(KheEventResourceEvent(er));
      /* ***
      dom_test = KheDrsDomTestMake(KHE_DRS_DOM_GE,
	cost_fn != KHE_QUADRATIC_COST_FUNCTION, combined_wt);
      *** */
      eisc = KheDrsExprIntSumCostMake((KHE_MONITOR) arm, cost_fn, combined_wt,
	min_limit, INT_MAX, false, 0, 0, 0, /* dom_test, */ NULL, drs);
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	HaArrayForEach(dt->days, dtd, j)
	  if( dtd->task == task )
	  {
	    eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	  }
      }
      KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddPreferResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    KHE_PREFER_RESOURCES_MONITOR prm)                                      */
/*                                                                           */
/*  Make and add in one or more expression trees representing prm.           */
/*  The context of the call ensures that prm has the correct resource type.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddPreferResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_PREFER_RESOURCES_MONITOR prm)
{
  KHE_CONSTRAINT c;  KHE_RESOURCE_GROUP rg;  KHE_EVENT_RESOURCE er;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  KHE_PREFER_RESOURCES_CONSTRAINT prc;  int i, j;  KHE_DRS_TASK_ON_DAY dtd;
  KHE_DRS_EXPR_ASSIGNED_TASK eat;  KHE_DRS_EXPR_INT_SUM_COST eisc;
  KHE_TASK task;  KHE_DRS_TASK dt;
  er = KhePreferResourcesMonitorEventResource(prm);
  HnAssert(KheEventResourceResourceType(er) == drs->resource_type,
    "KheDrsAddPreferResourcesMonitor internal error");
  c = KheMonitorConstraint((KHE_MONITOR) prm);
  combined_wt = KheConstraintCombinedWeight(c);
  if( combined_wt > 0 )
  {
    cost_fn = KheConstraintCostFunction(c);
    prc = (KHE_PREFER_RESOURCES_CONSTRAINT) c;
    rg = KhePreferResourcesConstraintDomainComplement(prc);
    if( cost_fn != KHE_LINEAR_COST_FUNCTION )
    {
      /* a single tree, whose root has one child per task on day */
      eisc = KheDrsExprIntSumCostZeroMaxMake((KHE_MONITOR) prm,
	cost_fn, combined_wt, NULL, drs);
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	HaArrayForEach(dt->days, dtd, j)
	  if( dtd->task == task )
	  {
	    eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	  }
      }
      KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
    }
    else if( KheResourceGroupResourceCount(rg) > 0 )
    {
      /* one tree per task on day */
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	HaArrayForEach(dt->days, dtd, j)
	  if( dtd->task == task )
	  {
	    eisc = KheDrsExprIntSumCostZeroMaxMake((KHE_MONITOR) prm,
	      cost_fn, combined_wt, NULL, drs);
	    eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
	  }
      }
    }
    else
    {
      /* we omit the tree for this, because its cost is already */
      /* included in the *asst_cost attribute of each task */
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,   */
/*    bool allow_zero, KHE_COST combined_wt, KHE_COST_FUNCTION cost_fn,      */
/*    KHE_DRS_DOM_TEST *alpha_dom_test, KHE_DRS_DOM_TEST *beta_dom_test)     */
/*                                                                           */
/*  Convert limits (min_limit, max_limit, allow_zero) into dominance tests.  */
/*  if min_limit <= 0 there is no min limit; if max_limit >= max_max_limit   */
/*  there is no maximum limit.                                               */
/*                                                                           */
/*  Return true if there is at least one limit, otherwise return false.      */
/*                                                                           */
/*  The values returned in *alpha_dom_test and *beta_dom_test are referred   */
/*  to as alpha and beta in the User's Guide.                                */
/*                                                                           */
/*  Note.  If there is a trivial allow_zero flag (if min_limit <= 1), an     */
/*  adjustment will have already been made by now to remove it.  So this     */
/*  call may assume that allow_zero implies min_limit > 1.                   */
/*                                                                           */
/*****************************************************************************/

/* *** equivalent function sti ll to do
static bool KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,
  bool allow_zero, KHE_COST combined_wt, KHE_COST_FUNCTION cost_fn,
  KHE_DRS_DOM_TEST *alpha_dom_test, KHE_DRS_DOM_TEST *beta_dom_test)
{
  if( allow_zero ** && min_limit > 1 ** )
  {
    ** fall back on equality when there is a non-trivial allow_zero flag **
    *alpha_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ,
      cost_fn != KHE_QUADRATIC_COST_FUNCTION, combined_wt);
    *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ, false, 0);
    return true;
  }
  else if( max_limit >= max_max_limit )
  {
    if( min_limit <= 0 )
    {
      ** no max_limit, and no min_limit (should never happen) **
      HnAbort("KheLimitsToTests: no max limit, no min limit");
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMakeUnused();
      return false;
    }
    else
    {
      ** no max_limit, and min_limit **
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_GE_LOOSE,
	min_limit, cost_fn != KHE_QUADRATIC_COST_FUNCTION, combined_wt);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_GE, false, 0);
      return true;
    }
  }
  else
  {
    if( min_limit <= 0 )
    {
      ** max_limit, and no min_limit **
      *alpha_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE,
	cost_fn != KHE_QUADRATIC_COST_FUNCTION, combined_wt);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE, false, 0);
      return true;
    }
    else
    {
      ** max_limit, and min_limit **
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_EQ_LOOSE,
	min_limit, cost_fn != KHE_QUADRATIC_COST_FUNCTION, combined_wt);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ, false, 0);
      return true;
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsConstraintCouldHaveCost(KHE_COST combined_wt,                 */
/*    int min_limit, int max_limit, int max_max_limit)                       */
/*                                                                           */
/*  Return true if a constraint with these attributes could generate a       */
/*  non-zero cost.                                                           */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsConstraintCouldHaveCost(KHE_COST combined_wt,
  int min_limit, int max_limit, int max_max_limit)
{
  return combined_wt > 0 && (min_limit > 0 || max_limit < max_max_limit);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_LIMIT_RESOURCES_MONITOR lrm)                                       */
/*                                                                           */
/*  Make and add in one or more expression trees representing lrm.           */
/*                                                                           */
/*  The context of the call ensures that at least one of lrm's event         */
/*  resources has the correct resource type.  We abort if not all do.        */
/*                                                                           */
/*  The documentation mentions an unlikely case where the expression can     */
/*  be split up.  We haven't bothered testing for that here.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_RESOURCES_MONITOR lrm)
{
  KHE_DRS_EXPR_INT_SUM_COST eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int min_limit, max_limit, i, egi, j, k;  KHE_RESOURCE_GROUP rg;
  KHE_LIMIT_RESOURCES_CONSTRAINT lrc;  KHE_EVENT_RESOURCE er;
  /* KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test; */
  KHE_DRS_TASK_ON_DAY dtd;  KHE_TASK task;  KHE_DRS_TASK dt;
  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  c = KheMonitorConstraint((KHE_MONITOR) lrm);
  combined_wt = KheConstraintCombinedWeight(c);
  cost_fn = KheConstraintCostFunction(c);
  lrc = (KHE_LIMIT_RESOURCES_CONSTRAINT) c;
  rg = KheLimitResourcesConstraintDomain(lrc);
  min_limit = KheLimitResourcesConstraintMinimum(lrc);
  max_limit = KheLimitResourcesConstraintMaximum(lrc);
  /* ***
  if( combined_wt > 0
      && KheLimitsToTests(min_limit, max_limit, INT_MAX,
	false, combined_wt, cost_fn, &alpha_dom_test, &beta_dom_test) )
  *** */
  if( KheDrsConstraintCouldHaveCost(combined_wt, min_limit, max_limit, INT_MAX))
  {
    eisc = KheDrsExprIntSumCostMake((KHE_MONITOR) lrm, cost_fn, combined_wt,
      min_limit, max_limit, false, 0, 0, 0, /* alpha_dom_test, */ NULL, drs);
    egi = KheLimitResourcesMonitorEventGroupIndex(lrm);
    for( i=0; i < KheLimitResourcesConstraintEventResourceCount(lrc, egi); i++ )
    {
      er = KheLimitResourcesConstraintEventResource(lrc, egi, i);
      HnAssert(KheEventResourceResourceType(er) == drs->resource_type,
	"KheDrsAddLimitResourcesMonitor: not all event resources have type %s",
	KheResourceTypeId(drs->resource_type));
      for( j = 0;  j < KheEventResourceTaskCount(drs->soln, er);  j++ )
      {
	task = KheEventResourceTask(drs->soln, er, j);
	dt = KheTaskToDrsTask(task, drs);
	HaArrayForEach(dt->days, dtd, k)
	  if( dtd->task == task )
	  {
	    eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	  }
      }
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddEventResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_MONITOR m)                                                         */
/*                                                                           */
/*  Add an expression tree to drs for event resource monitor m, if relevant. */
/*  The context of the call ensures that m has the correct resource type.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddEventResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_ASSIGN_RESOURCE_MONITOR_TAG:

      KheDrsAddAssignResourceMonitor(drs, (KHE_ASSIGN_RESOURCE_MONITOR) m);
      break;

    case KHE_PREFER_RESOURCES_MONITOR_TAG:

      KheDrsAddPreferResourcesMonitor(drs, (KHE_PREFER_RESOURCES_MONITOR) m);
      break;

    case KHE_AVOID_SPLIT_ASSIGNMENTS_MONITOR_TAG:

      /* event resource monitor, but we are currently not handling it */
      break;

    case KHE_LIMIT_RESOURCES_MONITOR_TAG:

      KheDrsAddLimitResourcesMonitor(drs, (KHE_LIMIT_RESOURCES_MONITOR) m);
      break;

    default:

      HnAbort("KheDrsAddEventResourceMonitor internal error (tag %d)",
	KheMonitorTag(m));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "conversion of resource monitors into expression trees"        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddAvoidUnavailableTimesMonitor(                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs,                                       */
/*    KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm, KHE_DRS_RESOURCE dr)         */
/*                                                                           */
/*  Make and add in one or more expression trees for dr representing autm.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddAvoidUnavailableTimesMonitor(
  KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm, KHE_DRS_RESOURCE dr)
{
  KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT c;  KHE_DRS_EXPR_BUSY_TIME eb;
  int i;  KHE_TIME t;  KHE_DRS_EXPR_INT_SUM_COST eisc;
  KHE_COST_FUNCTION cost_fn;  KHE_TIME_GROUP tg;  KHE_COST combined_wt;
  /* KHE_DRS_DOM_TEST unused_dom_test; */  KHE_DRS_RESOURCE_ON_DAY drd;

  c = KheAvoidUnavailableTimesMonitorConstraint(autm);
  tg = KheAvoidUnavailableTimesConstraintUnavailableTimes(c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  if( combined_wt > 0 && KheTimeGroupTimeCount(tg) > 0 )
  {
    /* unused_dom_test = KheDrsDomTestMakeUnused(); */
    cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION || KheTimeGroupTimeCount(tg) == 1 )
    {
      /* a separate expression tree for each time */
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	eisc = KheDrsExprIntSumCostZeroMaxMake((KHE_MONITOR) autm, cost_fn,
	  combined_wt, dr, drs);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eb = KheDrsExprBusyTimeMake(drd, t, /* unused_dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eb, drs);
	KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
      }
    }
    else
    {
      /* a single expression tree (max limit of 0, in effect) */
      eisc = KheDrsExprIntSumCostZeroMaxMake((KHE_MONITOR) autm, cost_fn,
	combined_wt, dr, drs);
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eb = KheDrsExprBusyTimeMake(drd, t, /* unused_dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eb, drs);
      }
      KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If tg contains a complete day's worth of times, starting from            */
/*  start_index, then return true and set *day_tg to the day time group,     */
/*  otherwise return false.                                                  */
/*                                                                           */
/*****************************************************************************/

static bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)
{
  KHE_TIME tg_t, day_t;  int i, day_index;  KHE_TIME_GROUP d_tg;
  tg_t = KheTimeGroupTime(tg, start_index);
  day_index = KheFrameTimeIndex(drs->days_frame, tg_t);
  d_tg = KheFrameTimeTimeGroup(drs->days_frame, tg_t);
  for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
  {
    day_t = KheTimeGroupTime(d_tg, i);
    if( start_index + i >= KheTimeGroupTimeCount(tg) )
      return *day = NULL, false;
    tg_t = KheTimeGroupTime(tg, start_index + i);
    if( day_t != tg_t )
      return *day = NULL, false;
  }
  return *day = HaArray(drs->all_days, day_index), true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsBuildOrExpr(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,           */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, bool force,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an OR expression with one BUSY child for each time of tg, and add   */
/*  the OR expression as a child of prnt.  If force is false and tg has      */
/*  exactly one time, optimize by adding just the BUSY expression for that   */
/*  time as a child of prnt.                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsBuildOrExpr(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, /* KHE_DRS_DOM_TEST dom_test, */ bool force,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR eo;  KHE_DRS_EXPR_BUSY_TIME ebt;  int i;  KHE_TIME t;
  /* KHE_DRS_DOM_TEST unused_dom_test; */  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;
  if( !force && KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single BUSY_TIME expression */
    t = KheTimeGroupTime(tg, 0);
    drd = KheDrsResourceOnDayAtTime(dr, t, drs);
    ebt = KheDrsExprBusyTimeMake(drd, t, /* dom_test, */ drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebt, drs);
  }
  else if( !force && KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
  {
    /* just a single BUSY_DAY expression */
    drd = KheDrsResourceOnDay(dr, day);
    ebd = KheDrsExprBusyDayMake(drd, day->time_group, /* dom_test, */ drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebd, drs);
  }
  else
  {
    /* an OR expression with BUSY_DAY and BUSY_TIME children */
    eo = KheDrsExprOrMake(/* dom_test, */ dr, drs);
    /* unused_dom_test = KheDrsDomTestMakeUnused(); */
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day) )
      {
	/* add a BUSY_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	ebd = KheDrsExprBusyDayMake(drd, day->time_group, /* dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebd, drs);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	/* add a BUSY_TIME expression */
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	ebt = KheDrsExprBusyTimeMake(drd, t, /* unused_dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebt, drs);
      }
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) eo, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eo, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsBuildAndExpr(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,          */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, bool force,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an AND expression with one FREE child for each time of tg, and      */
/*  add the AND expression as a child of prnt.  If force is false and tg     */
/*  has exactly one time, optimize by adding just the FREE expression for    */
/*  that time as a child of prnt.                                            */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsBuildAndExpr(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, /* KHE_DRS_DOM_TEST dom_test, */ bool force,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND ea;  KHE_DRS_EXPR_FREE_TIME eft;  int i;  KHE_TIME t;
  /* KHE_DRS_DOM_TEST unused_dom_test; */  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;
  if( !force && KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single FREE expression */
    t = KheTimeGroupTime(tg, 0);
    drd = KheDrsResourceOnDayAtTime(dr, t, drs);
    eft = KheDrsExprFreeTimeMake(drd, t, /* dom_test, */ drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eft, drs);
  }
  else if( !force && KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
  {
    /* just a single FREE_DAY expression */
    drd = KheDrsResourceOnDay(dr, day);
    efd = KheDrsExprFreeDayMake(drd, day->time_group, /* dom_test, */ drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) efd, drs);
  }
  else
  {
    /* an AND expression with FREE_DAY and FREE_TIME children */
    ea = KheDrsExprAndMake(/* dom_test, */ dr, drs);
    /* unused_dom_test = KheDrsDomTestMakeUnused(); */
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day) )
      {
	/* add a FREE_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	efd = KheDrsExprFreeDayMake(drd, day->time_group, /* dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) efd, drs);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	/* add a FREE_TIME expression */
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eft = KheDrsExprFreeTimeMake(drd, t, /* unused_dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) eft, drs);
      }
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) ea, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ea, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsClusterMonitorContainsWholePositiveDay(                       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int start_index,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)                     */
/*                                                                           */
/*  If cbtm contains a sequence of singleton time groups starting at         */
/*  start_index that have positive polarity and cover a complete day,        */
/*  return true and set *day to the day;  otherwise return false.            */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsClusterMonitorContainsWholePositiveDay(
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)
{
  KHE_TIME day_t;  int i, tg_count, day_index;  KHE_TIME_GROUP m_tg, d_tg;
  KHE_POLARITY po;  KHE_TIME t;
  tg_count = KheClusterBusyTimesMonitorTimeGroupCount(cbtm);
  m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, start_index, &po);
  if( po == KHE_POSITIVE && KheTimeGroupTimeCount(m_tg) == 1 )
  {
    t = KheTimeGroupTime(m_tg, 0);
    d_tg = KheFrameTimeTimeGroup(drs->days_frame, t);
    day_index = KheFrameTimeIndex(drs->days_frame, t);
    for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
    {
      day_t = KheTimeGroupTime(d_tg, i);
      if( start_index + i >= tg_count )
	return *day = NULL, false;
      m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, start_index + i, &po);
      if( po != KHE_POSITIVE || KheTimeGroupTimeCount(m_tg) != 1 ||
	  KheTimeGroupTime(m_tg, 0) != day_t )
	return *day = NULL, false;
    }
    return *day = HaArray(drs->all_days, day_index), true;
  }
  else
    return *day = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddClusterBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, KHE_DRS_RESOURCE dr)              */
/*                                                                           */
/*  Make and add in one or more expression trees representing cbtm.          */
/*                                                                           */
/*  Implementation note.  If cbtm contains a sequence of consecutive         */
/*  singleton time groups with positive polarity that add up to a complete   */
/*  day, then one BUSY_DAY expression is generated for them instead of one   */
/*  BUSY_TIME expression for each.  This is correct:  the sum reported       */
/*  in either case is 1 if the day is busy and 0 otherwise.  It cannot       */
/*  be more than 1 because the resource takes at most one task each day.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddClusterBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, KHE_DRS_RESOURCE dr)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;  bool allow_zero;
  int i, min_limit, max_limit, history_before, history, history_after, count;
  KHE_DRS_EXPR_INT_SUM_COST eisc;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_POLARITY po;  /* KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test; */
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;

  if( DEBUG_MONITOR(cbtm) )
    fprintf(stderr, "[ KheDrsAddClusterBusyTimesMonitor(drs, %s, %s)\n",
      KheMonitorId((KHE_MONITOR) cbtm), KheResourceId(dr->resource));

  c = KheClusterBusyTimesMonitorConstraint(cbtm);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheClusterBusyTimesConstraintMinimum(c);
  max_limit = KheClusterBusyTimesConstraintMaximum(c);
  allow_zero = KheClusterBusyTimesConstraintAllowZero(c);
  history_before = KheClusterBusyTimesMonitorHistoryBefore(cbtm);
  history_after = KheClusterBusyTimesMonitorHistoryAfter(cbtm);
  history = KheClusterBusyTimesMonitorHistory(cbtm);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  count = KheClusterBusyTimesConstraintTimeGroupCount(c);
  /* ***
  if( combined_wt > 0 && this test for non-trivial st ill to do
      KheLimitsToTests(min_limit, max_limit, count,
	allow_zero, combined_wt, cost_fn, &alpha_dom_test, &beta_dom_test) )
  *** */
  if( DEBUG_MONITOR(cbtm) )
    fprintf(stderr, "  KheDrsConstraintCouldHaveCost(%.5f, %d, %d, %d) = %s\n",
      KheCostShow(combined_wt), min_limit, max_limit, count,
      bool_show(KheDrsConstraintCouldHaveCost(combined_wt, min_limit,
	max_limit, count)));
  if( KheDrsConstraintCouldHaveCost(combined_wt, min_limit, max_limit,count))
  {
    /* one int_sum_cost expression at top of tree */
    if( max_limit >= count )
      max_limit = INT_MAX;
    eisc = KheDrsExprIntSumCostMake((KHE_MONITOR) cbtm, cost_fn, combined_wt,
      min_limit, max_limit, allow_zero, history_before, history_after,
      history, /* alpha_dom_test, */ dr, drs);

    /* one OR or AND expression for each time group, or a BUSY_DAY */
    /* expression if all singles of one day present and positive */
    for( i = 0;  i < count;  i++ )
    {
      if( KheDrsClusterMonitorContainsWholePositiveDay(cbtm, i, drs, &day) )
      {
	/* add a BUSY_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	ebd = KheDrsExprBusyDayMake(drd, day->time_group, /* beta_dom_test, */
	  drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebd, drs);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
	if( po == KHE_POSITIVE )
	{
	  /* an OR expression with one BUSY expression for each time of tg */
	  KheDrsBuildOrExpr(dr,(KHE_DRS_EXPR) eisc, tg, /* beta_dom_test, */
	    false, drs);
	}
	else
	{
	  /* an AND expression with one FREE expression for each time of tg */
	  KheDrsBuildAndExpr(dr, (KHE_DRS_EXPR) eisc, tg, /* beta_dom_test, */
	    false, drs);
	}
      }
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
    if( DEBUG_MONITOR(cbtm) )
      KheDrsExprDebug((KHE_DRS_EXPR) eisc, drs, 2, 2, stderr);
  }
  if( DEBUG_MONITOR(cbtm) )
    fprintf(stderr, "] KheDrsAddClusterBusyTimesMonitor returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, KHE_DRS_RESOURCE dr)                */
/*                                                                           */
/*  Make and add in one or more expression trees representing lbtm.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, j, junk, min_limit, max_limit;  bool allow_zero;
  KHE_DRS_EXPR_INT_SUM_COST eisc;  KHE_TIME t;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_EXPR_INT_DEV eid;
  KHE_DRS_EXPR_BUSY_TIME ebt;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;
  /* KHE_DRS_DOM_TEST unused_dom_test, alpha_dom_test, beta_dom_test; */

  /* boilerplate */
  /* r = KheLimitBusyTimesMonitorResource(lbtm); */
  c = KheLimitBusyTimesMonitorConstraint(lbtm);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitBusyTimesConstraintMinimum(c);
  max_limit = KheLimitBusyTimesConstraintMaximum(c);
  allow_zero = KheLimitBusyTimesConstraintAllowZero(c);
  /* unused_dom_test = KheDrsDomTestMakeUnused(); */
  /* le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE, false, 0); */
  /* KheDrsResource not safe to call here; but we don't need it anyway
  HnAssert(KheDrsResource(r, drs) == dr,
    "KheDrsAddLimitBusyTimesMonitor internal error 1");
  dr = KheDrsResource(r, drs); */

  /* adjust allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  if( combined_wt == 0 )
    return;

  if( cost_fn == KHE_LINEAR_COST_FUNCTION ||
      KheLimitBusyTimesMonitorTimeGroupCount(lbtm) == 1 )
  {
    /* one INT_SUM_COST for each time group, unless the time group is */
    /* one day and min_limit = 0 and max_limit = 1 */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &junk);
      if( min_limit == 0 && max_limit == 1 &&
	  KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
	  KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
      {
	/* constraint specifies at most one task per day; trivial, so omit */
      }
      /* ***
      else if( KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	  allow_zero, combined_wt, cost_fn, &alpha_dom_test, &beta_dom_test) )
      *** */
      else if( KheDrsConstraintCouldHaveCost(combined_wt, min_limit, max_limit,
	    KheTimeGroupTimeCount(tg)) )
      {
	eisc = KheDrsExprIntSumCostMake((KHE_MONITOR) lbtm, cost_fn,
	  combined_wt, min_limit, max_limit, allow_zero, 0, 0, 0,
	  /* alpha_dom_test, */ dr, drs);
	for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
	{
	  t = KheTimeGroupTime(tg, j);
	  drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	  ebt = KheDrsExprBusyTimeMake(drd, t, /* unused_dom_test, */ drs);
	  KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebt, drs);
	}
	KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
      }
    }
  }
  else
  {
    /* one INT_SUM_COST at root of tree */
    eisc = KheDrsExprIntSumCostZeroMaxMake((KHE_MONITOR) lbtm, cost_fn,
      combined_wt, dr, drs);

    /* one INT_DEV and INT_SUM for each time group */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &junk);
      /* ***
      if( KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	allow_zero, combined_wt, cost_fn, &alpha_dom_test, &beta_dom_test) )
      *** */
      if( KheDrsConstraintCouldHaveCost(combined_wt, min_limit, max_limit,
	    KheTimeGroupTimeCount(tg)) )
      {
	eid = KheDrsExprIntDevMake(min_limit, max_limit, allow_zero, dr, drs);
	eis = KheDrsExprIntSumMake(/* alpha_dom_test, */ dr, drs);
	for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
	{
	  t = KheTimeGroupTime(tg, j);
	  drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	  ebt = KheDrsExprBusyTimeMake(drd, t, /* unused_dom_test, */ drs);
	  KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) ebt, drs);
	}
	KheDrsExprInitEnd((KHE_DRS_EXPR) eis, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eis, drs);
	KheDrsExprInitEnd((KHE_DRS_EXPR) eid, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eid, drs);
      }
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitWorkloadMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_LIMIT_WORKLOAD_MONITOR lwm, KHE_DRS_RESOURCE dr)                   */
/*                                                                           */
/*  Make and add in one or more expression trees representing lwm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitWorkloadMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_WORKLOAD_MONITOR lwm, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_WORKLOAD_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, j, min_limit, max_limit;  bool allow_zero;  float junk;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_DRS_EXPR_WORK_TIME ew;
  /* KHE_DRS_EXPR_INT_SUM eis; */  KHE_DRS_EXPR_INT_SUM_COST eisc;
  KHE_DRS_EXPR_FLOAT_SUM efs;  KHE_DRS_EXPR_FLOAT_DEV efd;
  /* KHE_DRS_DOM_TEST unused_dom_test, alpha_dom_test, beta_dom_test; */
  KHE_DRS_RESOURCE_ON_DAY drd;

  /* boilerplate */
  c = KheLimitWorkloadMonitorConstraint(lwm);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitWorkloadConstraintMinimum(c);
  max_limit = KheLimitWorkloadConstraintMaximum(c);
  allow_zero = KheLimitWorkloadConstraintAllowZero(c);
  /* unused_dom_test = KheDrsDomTestMakeUnused(); */
  /* le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE, false, 0); */

  /* adjust allow_zero */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;

  /* ***
  if( KheLimitsToTests(min_limit, max_limit, INT_MAX, allow_zero,
      combined_wt, cost_fn, &alpha_dom_test, &beta_dom_test) )
  *** */
  if( KheDrsConstraintCouldHaveCost(combined_wt, min_limit, max_limit, INT_MAX))
  {
    /* one INT_SUM_COST at the root */
    eisc = KheDrsExprIntSumCostZeroMaxMake((KHE_MONITOR) lwm, cost_fn,
      combined_wt, dr, drs);

    /* one FLOAT_DEV and FLOAT_SUM for each time group */
    for( i = 0;  i < KheLimitWorkloadMonitorTimeGroupCount(lwm);  i++ )
    {
      tg = KheLimitWorkloadMonitorTimeGroup(lwm, i, &junk);
      efd = KheDrsExprFloatDevMake(min_limit, max_limit, allow_zero, dr, drs);
      efs = KheDrsExprFloatSumMake(/* alpha_dom_test, */ dr, drs);
      for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
      {
	t = KheTimeGroupTime(tg, j);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	ew = KheDrsExprWorkTimeMake(drd, t, /* unused_dom_test, */ drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) efs, (KHE_DRS_EXPR) ew, drs);
      }
      KheDrsExprInitEnd((KHE_DRS_EXPR) efs, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) efd, (KHE_DRS_EXPR) efs, drs);
      KheDrsExprInitEnd((KHE_DRS_EXPR) efd, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) efd, drs);
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitActiveIntervalsMonitor(                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs,                                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, KHE_DRS_RESOURCE dr)          */
/*                                                                           */
/*  Make and add in one or more expression trees representing laim.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitActiveIntervalsMonitor(
  KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_POLARITY po;  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, min_limit, max_limit, history_before, history_after, history, tg_count;
  KHE_DRS_EXPR_INT_SEQ_COST eisc;
  /* KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test; */

  /* boilerplate */
  if( DEBUG8 )
    fprintf(stderr, "[ KheDrsAddLimitActiveIntervalsMonitor(drs, %s, %s)\n",
      KheMonitorId((KHE_MONITOR) laim), KheResourceId(dr->resource));
  c = KheLimitActiveIntervalsMonitorConstraint(laim);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheLimitActiveIntervalsConstraintMinimum(c);
  max_limit = KheLimitActiveIntervalsConstraintMaximum(c);
  history_before = KheLimitActiveIntervalsMonitorHistoryBefore(laim);
  history_after = KheLimitActiveIntervalsMonitorHistoryAfter(laim);
  history = KheLimitActiveIntervalsMonitorHistory(laim);
  tg_count = KheLimitActiveIntervalsConstraintTimeGroupCount(c);

  /* return if constraint does nothing */
  /* *** KheDrsConstraintCouldHaveCost does this now
  if( (min_limit <= 1 && max_limit >= tg_count) || combined_wt == 0 )
  {
    if( DEBUG8 )
      fprintf(stderr,
	"] KheDrsAddLimitActiveIntervalsMonitor returning (no effect)\n");
    return;
  }
  *** */

  /* work out the various dom_tests */
  /* adjusted_min_limit = max(0, min_limit - history - history_after); */
  /* ***
  if( KheLimitsToTests(min_limit, max_limit, tg_count, false,
	combined_wt, cost_fn, &alpha_dom_test, &beta_dom_test) )
  *** */
  if( KheDrsConstraintCouldHaveCost(combined_wt, min_limit, max_limit,tg_count))
  {
    /* one INT_SEQ_COST expression at top of tree */
    if( max_limit >= tg_count )
      max_limit = INT_MAX;
    eisc = KheDrsExprIntSeqCostMake((KHE_MONITOR) laim, cost_fn, combined_wt,
      min_limit, max_limit, history_before, history_after, history,
      /* alpha_dom_test, */ dr, drs);

    /* one OR or AND expression for each time group */
    for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
    {
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
      if( po == KHE_POSITIVE )
      {
	/* an OR expression with one BUSY expression for each time of tg */
	KheDrsBuildOrExpr(dr, (KHE_DRS_EXPR) eisc, tg, /* beta_dom_test, */
	  true, drs);
      }
      else
      {
	/* an AND expression with one FREE expression for each time of tg */
	KheDrsBuildAndExpr(dr, (KHE_DRS_EXPR) eisc, tg, /* beta_dom_test, */
	  true, drs);
      }
    }
    KheDrsExprInitEnd((KHE_DRS_EXPR) eisc, drs);
    if( DEBUG8 )
      KheDrsExprDebug((KHE_DRS_EXPR) eisc, drs, 1, 2, stderr);
  }
  if( DEBUG8 )
    fprintf(stderr, "] KheDrsAddLimitActiveIntervalsMonitor returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,           */
/*    KHE_MONITOR m, KHE_DRS_RESOURCE dr)                                    */
/*                                                                           */
/*  Add an expression tree to drs for resource monitor m, if relevant;       */
/*  dr is the drs resource corresponding to the resource monitored by m.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m, KHE_DRS_RESOURCE dr)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_AVOID_CLASHES_MONITOR_TAG:

      /* resource monitor, but no need to handle it */
      break;

    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:

      KheDrsAddAvoidUnavailableTimesMonitor(drs,
	(KHE_AVOID_UNAVAILABLE_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:

      /* resource monitor, but we are currently not handling it */
      break;

    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      KheDrsAddClusterBusyTimesMonitor(drs,
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      KheDrsAddLimitBusyTimesMonitor(drs, (KHE_LIMIT_BUSY_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      KheDrsAddLimitWorkloadMonitor(drs, (KHE_LIMIT_WORKLOAD_MONITOR) m, dr);
      break;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      KheDrsAddLimitActiveIntervalsMonitor(drs,
	(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m, dr);
      break;

    default:

      HnAbort("KheDrsAddResourceMonitor internal error (tag %d)",
	KheMonitorTag(m));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public functions"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,  */
/*    KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)                             */
/*                                                                           */
/*  Make and return a new dynamic resource solver with these attributes.     */
/*                                                                           */
/*****************************************************************************/
typedef HA_ARRAY(KHE_MONITOR) ARRAY_KHE_MONITOR;

KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)
{
  KHE_DYNAMIC_RESOURCE_SOLVER res;  HA_ARENA a;  int i, j;  KHE_INSTANCE ins;
  KHE_FRAME days_frame;  KHE_DRS_DAY day;  KHE_EVENT_TIMETABLE_MONITOR etm;
  KHE_EVENT_RESOURCE er;  KHE_RESOURCE r;  KHE_DRS_RESOURCE dr;
  KHE_TASK_CLASS tc;  KHE_MONITOR m;  ARRAY_KHE_MONITOR er_monitors;

  /*************************************************************************/
  /*                                                                       */
  /*  Make the basic object and initialize its fields                      */
  /*                                                                       */
  /*************************************************************************/

  /* get the common frame and event timetable monitor; return NULL if can't */
  ins = KheSolnInstance(soln);
  if( DEBUG4 )
    fprintf(stderr, "[ KheDynamicResourceSolverMake(%s, %s, options)\n",
      KheInstanceId(ins), KheResourceTypeId(rt));
  days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  if( days_frame == NULL || etm == NULL )
  {
    if( DEBUG4 )
      fprintf(stderr, "] KheDynamicResourceSolverMake returning NULL\n");
    return NULL;
  }

  /* make the basic object */
  a = KheSolnArenaBegin(soln, false);
  HaMake(res, a);

  /* fields which are constant throughout the lifetime of the solver */
  res->arena = a;
  res->soln = soln;
  res->resource_type = rt;
  res->options = options;
  res->days_frame = days_frame;
  res->etm = etm;
  res->task_class_solver = NULL;  /* initialized below */
  HaArrayInit(res->all_resources, a);
  HaArrayInit(res->all_days, a);
  HaArrayInit(res->all_root_tasks, a);
  HaArrayInit(res->all_cost_exprs, a);
  res->postorder_count = 0;
  res->increment = 0;

  /* free list fields */
  HaArrayInit(res->free_closed_seqs, a);
  HaArrayInit(res->free_dom_tests, a);
  HaArrayInit(res->free_signatures, a);
  HaArrayInit(res->free_solns, a);
  HaArrayInit(res->free_soln_lists, a);
  HaArrayInit(res->free_soln_tries, a);
  HaArrayInit(res->free_soln_set_parts, a);
  HaArrayInit(res->free_indexed_soln_sets, a);
  HaArrayInit(res->free_soln_sets, a);
  HaArrayInit(res->free_packed_soln_days, a);
  HaArrayInit(res->free_packed_solns, a);

  /* fields that vary with the solve; their values here don't really matter */
  res->selected_resource_set = KheResourceSetMake(rt, a);
  HaArrayInit(res->selected_day_ranges, a);
  res->solve_soln_limit = 0;
  res->solve_init_cost = 0;
  res->solve_start_cost = 0;
  HaArrayInit(res->open_resources, a);
  HaArrayInit(res->open_days, a);
  HaArrayInit(res->open_exprs, a);
  HaArrayInit(res->extend_prev_tasks, a);
  HaArrayInit(res->extend_prev_signatures, a);
  res->extend_cost = 0;
  res->extend_must_assign_count = 0;
  res->rerun = NULL;

  /* fields for recording running time */
#if TESTING
  res->timer = KheTimerMake("dynamic", KHE_NO_TIME, a);
  HaArrayInit(res->table_size_per_day, a);
  HaArrayInit(res->running_time_per_day, a);
  HaArrayInit(res->ancestor_freq, a);
#endif


  /*************************************************************************/
  /*                                                                       */
  /*  Add days, resources, tasks, and expression trees                     */
  /*                                                                       */
  /*  When adding each drs resource, its resource monitors' expression     */
  /*  trees are not added.  They get added later.                          */
  /*                                                                       */
  /*  When adding each drs task class, its drs tasks are added as well,    */
  /*  both to the task class and to drs->all_root_tasks.  If a drs task's  */
  /*  corresponding KHE task is currently assigned to resource r, the      */
  /*  drs task is also assigned to the corresponding drs resource.         */
  /*                                                                       */
  /*************************************************************************/

  /* drs day objects for the days of the frame */
  for( i = 0;  i < KheFrameTimeGroupCount(days_frame);  i++ )
  {
    day = KheDrsDayMake(i, res);
    HaArrayAddLast(res->all_days, day);
  }

  /* resources of rt but not their monitors (assumes days done) */
  HaArrayInit(res->all_resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    dr = KheDrsResourceMake(r, res);
    HaArrayAddLast(res->all_resources, dr);
  }

  /* task classes and tasks, exiting if overlaps (assumes days and resources) */
  res->task_class_solver = KheTaskClassSolverMake(soln, rt, days_frame, a);
  KheTaskClassSolverAddAllTasks(res->task_class_solver);
  if( DEBUG3 )
    KheTaskClassSolverDebug(res->task_class_solver, 2, 2, stderr);
  for( i = 0;  i < KheTaskClassSolverClassCount(res->task_class_solver);  i++ )
  {
    tc = KheTaskClassSolverClass(res->task_class_solver, i);
    if( KheTaskClassNoOverlap(tc) )
      KheDrsTaskClassMake(tc, res);
    else
    {
      KheDynamicResourceSolverDelete(res);
      if( DEBUG4 )
	fprintf(stderr,
	  "] KheDynamicResourceSolverMake returning NULL (overlap in tc)\n");
      return NULL;
    }
  }

  /* resource monitors */
  HaArrayForEach(res->all_resources, dr, i)
    KheDrsResourceAddMonitors(dr, res);

  /* event resource monitors of type rt (assumes tasks done) */
  HaArrayInit(er_monitors, a);
  for( i = 0;  i < KheInstanceEventResourceCount(ins);  i++ )
  {
    er = KheInstanceEventResource(ins, i);
    if( KheEventResourceResourceType(er) == rt )
      for( j = 0;  j < KheSolnEventResourceMonitorCount(soln, er);  j++ )
      {
	m = KheSolnEventResourceMonitor(soln, er, j);
	HaArrayAddLast(er_monitors, m);
      }
  }
  HaArraySortUnique(er_monitors, &KheMonitorCmp);
  HaArrayForEach(er_monitors, m, i)
    KheDrsAddEventResourceMonitor(res, m);
  HaArrayFree(er_monitors);

  /* all done */
  if( DEBUG4 )
  {
    KheDynamicResourceSolverDebug(res, 2, 2, stderr);
    fprintf(stderr, "] KheDynamicResourceSolverMake returning\n");
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST gcd(KHE_COST a, KHE_COST b)                                     */
/*                                                                           */
/*  Here we must have a > 0 and b >= 0.  If b == 0, return a.  If b > 0,     */
/*  return the greatest common divisor of a and b.                           */
/*                                                                           */
/*****************************************************************************/

#define swap(a, b, tmp) ((tmp) = (a), (a) = (b), (b) = (tmp))

static KHE_COST gcd(KHE_COST a, KHE_COST b)
{
  KHE_COST tmp;
  HnAssert(a > 0 && b >= 0, "gcd internal error");
  while( b > 0 )
  {
    if( a > b )
      swap(a, b, tmp);
    b = b % a;
  }
  return a;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddWeight(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_COST weight)   */
/*                                                                           */
/*  Add weight to drs.  This accumulates the gcd of the weights.             */
/*                                                                           */
/*  Implementation note.  The test causes weights which are larger than      */
/*  or equal to the initial solution cost to be ignored.  Any solutions      */
/*  that incorporate this weight will be deleted before being inserted       */
/*  into the indexed list data structure that uses drs->increment.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddWeight(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_COST weight)
{
  if( weight < KheSolnCost(drs->soln) )
    drs->increment = gcd(weight, drs->increment);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Delete drs, reclaiming its memory.                                       */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KheSolnArenaEnd(drs->soln, drs->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Add resource r to the solve.                                             */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r)
{
  HnAssert(KheResourceResourceType(r) == drs->resource_type,
    "KheDynamicResourceSolverAddResource: r (%s) has resource type %s; expect"
    "ing %s", KheResourceId(r), KheResourceTypeId(KheResourceResourceType(r)),
    KheResourceTypeId(drs->resource_type));
  HnAssert(!KheResourceSetContainsResource(drs->selected_resource_set, r),
    "KheDynamicResourceSolverAddResource: r (%s) already added",
    KheResourceId(r));
  KheResourceSetAddResource(drs->selected_resource_set, r);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int first_day_index, int last_day_index)                               */
/*                                                                           */
/*  Add day range first_day_index .. last_day_index to the solve.            */
/*                                                                           */
/*****************************************************************************/

static bool KheDayIndexInRange(int di, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  return 0 <= di && di < KheFrameTimeGroupCount(drs->days_frame);
}

void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int first_day_index, int last_day_index)
{
  KHE_DRS_DAY_RANGE si;
  HnAssert(first_day_index <= last_day_index,
    "KheDynamicResourceSolverAddResource: empty day range (%d .. %d)",
    first_day_index, last_day_index);
  HnAssert(KheDayIndexInRange(first_day_index, drs),
    "KheDynamicResourceSolverAddResource: first_day_index (%d) out of range "
    "(0 .. %d)", first_day_index, KheFrameTimeGroupCount(drs->days_frame));
  HnAssert(KheDayIndexInRange(last_day_index, drs),
    "KheDynamicResourceSolverAddResource: last_day_index (%d) out of range "
    "(0 .. %d)", last_day_index, KheFrameTimeGroupCount(drs->days_frame));
  si = KheDrsDayRangeMake(first_day_index, last_day_index);
  KheDrsDayRangeMerge(&drs->selected_day_ranges, si);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebugBrief(KHE_DRS_SOLN soln, int verbosity,              */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of soln onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

/* *** perfectly good, but currently unused
static void KheDrsSolnDebugBrief(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  int val, i;  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE dr;
  if( indent >= 0 )
  {
    ** header line showing day, cost, and signature **
    fprintf(fp, "%*s", indent, "");
    fprintf(stderr, "[ Soln %s %.5f ", KheDrsDayId(soln->day),
      KheCostShow(soln->cost));
    HaArrayForEach(soln->sig, val, i)
      fprintf(stderr, "%s%d", i == 0 ? "" : ".", val);
    fprintf(stderr, "\n");

    ** second line showing prev_tasks **
    fprintf(fp, "%*s", indent + 2, "");
    HaArrayForEach(soln->prev_tasks, dtd, i)
    {
      dr = HaArray(drs->open_resources, i);
      fprintf(stderr, "%s%s: %s", i == 0 ? "" : ", ",
	KheResourceId(dr->resource),
	dtd == NULL || dtd->task == NULL ? "-" : KheTaskId(dtd->task));
    }
    fprintf(stderr, "\n");

    ** signature key and finish off **
    KheDrsDayDebugSignatureKey(soln->day, drs, 1, indent + 2, fp);
    fprintf(stderr, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "Soln(%s %.5f)", KheDrsDayId(soln->day),
      KheCostShow(soln->cost));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsRerun(KHE_DYNAMIC_RESOURCE_SOLVER drs,                        */
/*    KHE_DRS_PACKED_SOLN soln)                                              */
/*                                                                           */
/*  Do a rerun.  Upon return, the solver will be closed on soln.             */
/*                                                                           */
/*  A rerun is considered to be happening when RERUN && drs->rerun != NULL.  */
/*  This function is only called when RERUN is true; it sets drs->rerun to   */
/*  a non-NULL value initially, and clears it to NULL on exit.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsRerun(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,
  KHE_DRS_DOM_KIND cache_dom_kind, KHE_DRS_PACKED_SOLN soln)
{
  KHE_DRS_PACKED_SOLN init_soln2, new_best_soln2;
  KHE_DRS_EXPR_COST e;  int i;

  /* initialize the rerun cost fields in all cost expressions */
  HaArrayForEach(drs->all_cost_exprs, e, i)
    KheDrsExprCostInitRerunCost(e);

  /* carry out the open, search, and close of the rerun */
  drs->rerun = soln;
  KheDrsSolveOpen(drs, false, soln_limit, main_dom_kind, use_cache,
    cache_dom_kind, &init_soln2);
  if( DEBUG1 )
    KheDrsPackedSolnDebug(init_soln2, "init_soln2", drs, 2, 2, stderr);
  if( !KheDrsSolveSearch(drs, false, &new_best_soln2) )
    HnAbort("KheDrsRerun internal error (rerun failed to find a new best)");
  HnAssert(soln->cost == new_best_soln2->cost,
    "KheDrsRerun internal error (rerun new best has different cost)");
  if( DEBUG1 )
    KheDrsPackedSolnDebug(new_best_soln2, "new_best_soln2", drs, 2, 2, stderr);
  KheDrsSolveClose(drs, new_best_soln2, true);
  drs->rerun = NULL;

  /* check the rerun cost fields in all cost expressions */
  /* *** moved to KheDrsSolveClose
  fprintf(stderr, "[ KheDrsRerun checking costs:\n");
  for( i = 0;  i < HaArrayCount(drs->all_cost_exprs);  i = j )
  {
    e = HaArray(drs->all_cost_exprs, i);
    for( j = i + 1;  j < HaArrayCount(drs->all_cost_exprs);  j++ )
    {
      e2 = HaArray(drs->all_cost_exprs, j);
      if( e->monitor != e2->monitor )
	break;
    }
    KheDrsExprCostCheckRerunCost(drs, i, j - 1);
  }
  fprintf(stderr, "]\n");
  *** */

  /* delete the packed solutions made by this function */
  KheDrsPackedSolnDelete(init_soln2, drs);
  KheDrsPackedSolnDelete(new_best_soln2, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceSolverDoSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,        */
/*    KHE_DRS_DOM_KIND cache_dom_kind, bool test_only)                       */
/*                                                                           */
/*  Solve drs.  If test_only is false, adopt the best solution found and     */
/*  return true if it is an improvement on the original.                     */
/*                                                                           */
/*****************************************************************************/

bool KheDynamicResourceSolverDoSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,
  KHE_DRS_DOM_KIND cache_dom_kind, bool test_only)
{
  int rcount, i;  KHE_RESOURCE r;  KHE_DRS_DAY_RANGE ddr;
  KHE_DRS_PACKED_SOLN init_soln, new_best_soln, junk;  KHE_COST init_cost;

  /* debug print */
  init_cost = KheSolnCost(drs->soln);
  if( DEBUG1 )
  {
    fprintf(stderr,"[ KheDynamicResourceSolverDoSolve(drs, %d, %s, %s, %s) "
      "cost %.5f\n", soln_limit == INT_MAX ? -1 : soln_limit,
      KheDomKindShow(main_dom_kind), bool_show(use_cache),
      KheDomKindShow(cache_dom_kind), KheCostShow(init_cost));
    fprintf(stderr, "  resources:  ");
    rcount = KheResourceSetResourceCount(drs->selected_resource_set);
    for( i = 0;  i < rcount;  i++ )
    {
      r = KheResourceSetResource(drs->selected_resource_set, i);
      fprintf(stderr, "%s%s", i == 0 ? "" : ", ", KheResourceId(r));
    }
    fprintf(stderr, "\n");
    fprintf(stderr, "  day ranges: ");
    HaArrayForEach(drs->selected_day_ranges, ddr, i)
      fprintf(stderr, "%s%d-%d", i == 0 ? "" : ", ", ddr.first, ddr.last);
    fprintf(stderr, "\n");
  }

  /* open, search, close, and possibly rerun */
  KheDrsSolveOpen(drs, true, soln_limit, main_dom_kind, use_cache,
    cache_dom_kind, &init_soln);
  if( DEBUG1 )
    KheDrsPackedSolnDebug(init_soln, "init_soln", drs, 2, 2, stderr);
  if( !KheDrsSolveSearch(drs, true, &new_best_soln) )
  {
    /* no new best; close using init_soln */
    KheDrsSolveClose(drs, init_soln, false);
  }
  else if( RERUN )
  {
    /* close using init_soln */
    if( DEBUG1 )
      KheDrsPackedSolnDebug(new_best_soln, "new_best_soln", drs, 2, 2, stderr);
    KheDrsSolveClose(drs, init_soln, false);

    /* rerun new_best_soln (drs is closed on new_best_soln after this) */
    KheDrsRerun(drs, soln_limit, main_dom_kind, use_cache, cache_dom_kind,
      new_best_soln);

    /* if test only, return to init_soln */
    if( test_only )
    {
      KheDrsSolveOpen(drs, false, soln_limit, main_dom_kind, use_cache,
	cache_dom_kind, &junk);
      KheDrsPackedSolnDelete(junk, drs);
      KheDrsSolveClose(drs, init_soln, false);
    }
  }
  else
  {
    /* have new best solution, close using that */
    if( DEBUG1 )
      KheDrsPackedSolnDebug(new_best_soln, "new_best_soln", drs, 2, 2, stderr);
    KheDrsSolveClose(drs, new_best_soln, false);
  }

  /* delete init_soln and (if present) new_best_soln */
  KheDrsPackedSolnDelete(init_soln, drs);
  if( new_best_soln != NULL )
    KheDrsPackedSolnDelete(new_best_soln, drs);

  /* clear out selections ready for a fresh set of resources and days */
  KheResourceSetClear(drs->selected_resource_set);
  HaArrayClear(drs->selected_day_ranges);

  /* return true if the solution has been improved */
  HnAssert(KheSolnCost(drs->soln) <= init_cost,
    "KheDynamicResourceSolverDoSolve internal error: new %.5f > old %.5f",
    KheCostShow(KheSolnCost(drs->soln)), KheCostShow(init_cost));
  if( KheSolnCost(drs->soln) < init_cost )
  {
    if( DEBUG1 )
      fprintf(stderr, "] KheDynamicResourceSolverDoSolve returning true "
	"(new %.5f < old %.5f)\n", KheCostShow(KheSolnCost(drs->soln)),
	KheCostShow(init_cost));
    return true;
  }
  else
  {
    if( DEBUG1 )
      fprintf(stderr, "] KheDynamicResourceSolverDoSolve returning false\n");
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,        */
/*    KHE_DRS_DOM_KIND cache_dom_kind)                                       */
/*                                                                           */
/*  Perform one solve, changing the solution and returning true if a new     */
/*  best solution is found.                                                  */
/*                                                                           */
/*****************************************************************************/

bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,
  KHE_DRS_DOM_KIND cache_dom_kind)
{
  return KheDynamicResourceSolverDoSolve(drs, soln_limit, main_dom_kind,
    use_cache, cache_dom_kind, false);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,       */
/*    int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,        */
/*    KHE_DRS_DOM_KIND cache_dom_kind)                                       */
/*                                                                           */
/*  Perform a test solve.  This is identical to a regular solve except       */
/*  that the solution is never changed, not even if a new best is found.     */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND main_dom_kind, bool use_cache,
  KHE_DRS_DOM_KIND cache_dom_kind)
{
  KheDynamicResourceSolverDoSolve(drs, soln_limit, main_dom_kind,
    use_cache, cache_dom_kind, true);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of drs onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE dr;  int i;
  HnAssert(indent >= 0, "KheDynamicResourceSolverDebug: indent < 0");
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( verbosity > 1 && indent >= 0 )
  {
    /* full print */
    fprintf(fp, "[ DynamicResourceSolver(%s, %s)\n",
      KheInstanceId(KheSolnInstance(drs->soln)),
      KheResourceTypeId(drs->resource_type));
    HaArrayForEach(drs->all_resources, dr, i)
      KheDrsResourceDebug(dr, 2, indent + 2, fp);
    HaArrayForEach(drs->all_days, day, i)
      KheDrsDayDebug(day, drs, 2, indent + 2, fp);
    fprintf(fp, "%*s]", indent, "");
  }
  else
  {
    /* brief print */
    fprintf(fp, "DynamicResourceSolver(%s, %s, %d resources, %d days)",
      KheInstanceId(KheSolnInstance(drs->soln)),
      KheResourceTypeId(drs->resource_type),
      HaArrayCount(drs->all_resources), HaArrayCount(drs->all_days));
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "testing"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheDynamicResourceSolverSolveStatsCount(                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return the number of stats associated with this solve.                   */
/*                                                                           */
/*****************************************************************************/

int KheDynamicResourceSolverSolveStatsCount(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
#if TESTING
  return HaArrayCount(drs->running_time_per_day);
#else
  return 0;
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverSolveStats(KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    int i, int *table_size, float *running_time)                           */
/*                                                                           */
/*  Return the i'th stats.                                                   */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverSolveStats(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int i, int *table_size, float *running_time)
{
#if TESTING
  *table_size = HaArray(drs->table_size_per_day, i);
  *running_time = HaArray(drs->running_time_per_day, i);
#else
  *table_size = 0;
  *running_time = 0.0;
#endif
}
