
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_dynamic_resource.c                                  */
/*  DESCRIPTION:  Dynamic resource solving                                   */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_p.h"
#include <limits.h>
#include <float.h>
#include <math.h>

#define TESTING 1

#define bool_show(x) ((x) ? "true" : "false")
#define min(x, y) ((x) < (y) ? (x) : (y))
#define max(x, y) ((x) > (y) ? (x) : (y))

#define UNDEF_INT -1
#define UNDEF_FLOAT -1.0

#define DEBUG1 1		/* main functions		*/
#define DEBUG2 0		/* final solutions		*/
#define DEBUG3 0		/* checking adopted solutions	*/
#define DEBUG4 0		/* KheDrsSolnMake		*/
#define DEBUG5 0		/* KheDynamicResourceSolverDebug*/
#define DEBUG6 0		/* KheDrsFullSolnAdopt		*/
#define DEBUG7 0		/* running time (testing only)	*/
#define DEBUG8 0		/* manual tests               	*/
#define DEBUG9 0		/* hash table efficiency      	*/
#define DEBUG10 1		/* task classes                 */
#define DEBUG11 1		/* brief print of costs		*/
#define DEBUG12 0		/* trie dominance tests 	*/
#define DEBUG13 0		/* monitor sorting              */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations"                                            */
/*                                                                           */
/*  This submodule contains forward type declarations, when needed.          */
/*                                                                           */
/*  There are many types and no obvious best ordering, so here and           */
/*  thoughout this file we've followed the traditional ordering, except      */
/*  that we handle resources before days, because the implementation         */
/*  opens resources before days.  Here is the order:                         */
/*                                                                           */
/*    KHE entities        Corresponding DRS entities                         */
/*    -----------------------------------------------------------------      */
/*    Resources           DRS resources                                      */
/*    Times               DRS days                                           */
/*    Events              DRS tasks                                          */
/*    Constraints         DRS dominance tests and expressions                */
/*    Solutions           DRS solutions and the solver                       */
/*    -----------------------------------------------------------------      */
/*                                                                           */
/*****************************************************************************/

/* type declarations (resources) */
typedef struct khe_drs_resource_on_day_rec *KHE_DRS_RESOURCE_ON_DAY;
typedef HA_ARRAY(KHE_DRS_RESOURCE_ON_DAY) ARRAY_KHE_DRS_RESOURCE_ON_DAY;

typedef struct khe_drs_resource_rec *KHE_DRS_RESOURCE;
typedef HA_ARRAY(KHE_DRS_RESOURCE) ARRAY_KHE_DRS_RESOURCE;

/* type declarations (times) */
typedef struct khe_drs_day_rec *KHE_DRS_DAY;
typedef HA_ARRAY(KHE_DRS_DAY) ARRAY_KHE_DRS_DAY;

/* type declarations (events) */
typedef struct khe_drs_task_on_day_rec *KHE_DRS_TASK_ON_DAY;
typedef HA_ARRAY(KHE_DRS_TASK_ON_DAY) ARRAY_KHE_DRS_TASK_ON_DAY;

typedef struct khe_drs_task_rec *KHE_DRS_TASK;
typedef HA_ARRAY(KHE_DRS_TASK) ARRAY_KHE_DRS_TASK;

typedef struct khe_drs_task_class_rec *KHE_DRS_TASK_CLASS;
typedef HA_ARRAY(KHE_DRS_TASK_CLASS) ARRAY_KHE_DRS_TASK_CLASS;

/* type declarations (constraints) */
typedef struct khe_drs_dom_test_rec KHE_DRS_DOM_TEST;
typedef HA_ARRAY(KHE_DRS_DOM_TEST) ARRAY_KHE_DRS_DOM_TEST;

typedef struct khe_drs_expr_rec *KHE_DRS_EXPR;
typedef HA_ARRAY(KHE_DRS_EXPR) ARRAY_KHE_DRS_EXPR;

/* type declarations (solutions) */
typedef struct khe_drs_soln_set_rec *KHE_DRS_SOLN_SET;
typedef HA_ARRAY(KHE_DRS_SOLN_SET) ARRAY_KHE_DRS_SOLN_SET;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - resources"                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_RESOURCE_ON_DAY - what one resource is doing on one day     */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_resource_on_day_rec {
  KHE_DRS_RESOURCE		encl_dr;		/* the resource      */
  KHE_DRS_DAY			day;			/* the day           */
  KHE_DRS_TASK_ON_DAY		closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_EXPR		external_today;		/* leaf today        */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_RESOURCE - one resource                                     */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_resource_rec {
  KHE_RESOURCE			resource;		/* the resource      */
  ARRAY_KHE_DRS_RESOURCE_ON_DAY	days;			/* doing each day    */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - times"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DAY_RANGE - a range of days                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_day_range_rec {
  int			first;
  int			last;
} KHE_DRS_DAY_RANGE;

typedef HA_ARRAY(KHE_DRS_DAY_RANGE) ARRAY_KHE_DRS_DAY_RANGE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DAY - one day                                               */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_day_rec {
  int				frame_index;		/* index in frame    */
  int				open_day_index;		/* index when open   */
  KHE_TIME_GROUP		time_group;		/* day's time group  */
  ARRAY_KHE_DRS_TASK_CLASS	task_classes;		/* today's tasks     */
  ARRAY_KHE_DRS_EXPR		internal_today;		/* active today      */
  ARRAY_KHE_DRS_DOM_TEST	dom_tests;		/* for each sig pos  */
  HA_ARRAY_INT			eq_dom_test_indexes;	/* where = tests are */
  KHE_DRS_SOLN_SET		soln_set;		/* today's solns     */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - events"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK_ON_DAY - what one proper root task is doing on one day */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_on_day_rec {
  KHE_DRS_TASK			encl_dt;		/* enclosing task    */
  KHE_DRS_DAY			day;			/* the day           */
  KHE_TASK			task;			/* atomic task today */
  KHE_TIME			time;			/* time on day       */
  KHE_DRS_RESOURCE_ON_DAY	closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_EXPR		external_today;		/* leaf today        */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK - one proper root task                                 */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_rec {
  KHE_DRS_TASK_CLASS		encl_dtc;
  int				index_in_encl_dtc;
  bool				open;			/* true when open    */
  KHE_TASK			task;
  KHE_DRS_RESOURCE		closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_TASK_ON_DAY	days;
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK_CLASS - a set of tasks, taken from one KHE_TASK_CLASS  */
/*                                                                           */
/*  For efficiency, open_tasks is managed lazily.  Every unassigned task is  */
/*  in it, but not in order, and there may also be assigned tasks in it.     */
/*  These defects are corrected when the task class is opened for solving.   */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_class_rec {
  KHE_TASK_CLASS		orig_task_class;	/* origin of this    */
  KHE_DRS_DAY_RANGE		day_range;		/* days running      */
  ARRAY_KHE_DRS_TASK		all_tasks;		/* all tasks in class*/
  ARRAY_KHE_DRS_TASK		open_tasks;		/* unassigned tasks  */
  int				tmp_used;		/* used by curr edge */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - constraints"                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_KIND - the kind of dominance testing to use             */
/*                                                                           */
/*****************************************************************************/

/* *** defined in khe_solvers.h
typedef enum {
  KHE_DRS_DOM_NONE,
  KHE_DRS_DOM_WEAK,
  KHE_DRS_DOM_MEDIUM,
  KHE_DRS_DOM_STRONG,
  KHE_DRS_DOM_TRIE,
} KHE_DRS_DOM_KIND;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_TYPE - the type of a dominance test                     */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_DOM_UNUSED,		/* "-"      should never be used      */
  KHE_DRS_DOM_GE,		/* ">="     min limit only            */
  KHE_DRS_DOM_LE,		/* "<="     max limit only            */
  KHE_DRS_DOM_EQ,		/* "==",    for uncertain cases       */
  KHE_DRS_DOM_GE_LOOSE,		/* ">=*"    min limit with loosening  */
  KHE_DRS_DOM_EQ_LOOSE		/* "==*"    max + min with loosening  */
} KHE_DRS_DOM_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_TEST - a dominance test                                 */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_dom_test_rec {
  KHE_DRS_DOM_TYPE	type;			/* type of test              */
  int			min_limit;		/* GE_LOOSE and EQ_LOOSE     */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_TAG                                                    */
/*                                                                           */
/*  Tags for expression types.                                               */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_EXPR_BUSY_TIME_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_FREE_TIME_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_WORK_TIME_TAG,		/* float-valued      */
  KHE_DRS_EXPR_BUSY_DAY_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_FREE_DAY_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_WORK_DAY_TAG,		/* float-valued      */
  KHE_DRS_EXPR_ASSIGNED_TASK_TAG,	/* int-valued        */
  KHE_DRS_EXPR_OR_TAG,			/* int-valued	     */
  KHE_DRS_EXPR_AND_TAG,			/* int-valued	     */
  KHE_DRS_EXPR_INT_SUM_TAG,		/* int-valued	     */
  /* KHE_DRS_EXPR_INT_SEQ_TAG, */
  KHE_DRS_EXPR_FLOAT_SUM_TAG,		/* float-valued      */
  KHE_DRS_EXPR_INT_DEV_TAG,
  KHE_DRS_EXPR_FLOAT_DEV_TAG,
  KHE_DRS_EXPR_COST_TAG,
  /* KHE_DRS_EXPR_COST_SUM_TAG, */
  KHE_DRS_EXPR_INT_SUM_COMB_TAG,	/* int-valued	     */
  KHE_DRS_EXPR_INT_SEQ_COMB_TAG		/* int-valued	     */
} KHE_DRS_EXPR_TAG;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_COMB_TYPE - types of comb expressions                       */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_COMB_ORDINARY,
  KHE_DRS_COMB_NO_MAX,
  KHE_DRS_COMB_LINEAR,
  KHE_DRS_COMB_STEP
} KHE_DRS_COMB_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_VALUE - the value of an expression; could be int or float   */
/*                                                                           */
/*****************************************************************************/

typedef union {
  int			int_val;		/* value, if int   */
  float			float_val;		/* value, if float */
  KHE_COST		cost_val;		/* value, if cost  */
} KHE_DRS_VALUE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_PARENT - a expression's parent                              */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_parent_rec {
  KHE_DRS_EXPR		expr;			/* the parent expression */
  int			index;			/* index in parent       */
} KHE_DRS_PARENT;

typedef HA_ARRAY(KHE_DRS_PARENT) ARRAY_KHE_DRS_PARENT;


/*****************************************************************************/
/*                                                                           */
/*  INHERIT_KHE_DRS_EXPR - fields inherited by all expression types          */
/*                                                                           */
/*  For each active day di of e, except its very last active day,            */
/*  sig_index[di - e->first_day_index] is the position of e's value in       */
/*  the signatures of solutions ending on day di.                            */
/*                                                                           */
/*  For each active day di of e, child_index[di - e->first_day_index] is     */
/*  the index of the first child whose last_day_index is di or more.  There  */
/*  is also a sentinel value, so for each active day di, the iterator        */
/*                                                                           */
/*    start_child_index = child_index[di - e->first_day_index];              */
/*    stop_child_index  = child_index[di + 1 - e->first_day_index];          */
/*    for( i = start_child_index;  i < stop_child_index;  i++ )              */
/*      child_e = HaArray(e->children, i);                                   */
/*                                                                           */
/*  visits the children of e whose last_day_index is di.                     */
/*                                                                           */
/*****************************************************************************/

#define INHERIT_KHE_DRS_EXPR						\
  KHE_DRS_EXPR_TAG	tag;			/* tag field      */	\
  bool			gathered;		/* for opening    */	\
  KHE_DRS_DOM_TEST	dom_test;		/* dom test       */	\
  int			postorder_index;	/* expr index     */	\
  KHE_DRS_VALUE		u;			/* value          */	\
  ARRAY_KHE_DRS_PARENT	parents;		/* parent exprs   */	\
  ARRAY_KHE_DRS_EXPR	children;		/* child exprs    */	\
  KHE_DRS_DAY_RANGE	open_day_range;		/* open day range */	\
  ARRAY_KHE_DRS_EXPR	open_children;		/* open children  */	\
  HA_ARRAY_INT		open_child_indexes;	/* for each day   */	\
  HA_ARRAY_INT		sig_indexes;		/* for each day   */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR - abstract supertype of expressions                    */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_expr_rec {
  INHERIT_KHE_DRS_EXPR
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_BUSY_TIME                                              */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy at a given      */
/*  time, and 0 when the resource is free then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time we're busy at    */
} *KHE_DRS_EXPR_BUSY_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_TIME) ARRAY_KHE_DRS_EXPR_BUSY_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FREE_TIME                                              */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free at a given      */
/*  time, and 0 when the resource is busy then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time we're free at    */
} *KHE_DRS_EXPR_FREE_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_FREE_TIME) ARRAY_KHE_DRS_EXPR_FREE_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_WORK_TIME                                              */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource at    */
/*  a given time.  This will be 0.0 if the resource is free then.  The       */
/*  expression stores the time, but only for debugging.                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time of this workload */
} *KHE_DRS_EXPR_WORK_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_WORK_TIME) ARRAY_KHE_DRS_EXPR_WORK_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_BUSY_DAY                                               */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy on a given      */
/*  day, and 0 when the resource is free then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're busy at     */
} *KHE_DRS_EXPR_BUSY_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_DAY) ARRAY_KHE_DRS_EXPR_BUSY_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FREE_DAY                                               */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free on a given      */
/*  day, and 0 when the resource is busy then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_FREE_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_FREE_DAY) ARRAY_KHE_DRS_EXPR_FREE_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_WORK_DAY                                               */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource on    */
/*  a given day.  This will be 0.0 if the resource is free then.  The        */
/*  expression stores the day (as a time group), but only for debugging.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_WORK_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_WORK_DAY) ARRAY_KHE_DRS_EXPR_WORK_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_ASSIGNED_TASK                                          */
/*                                                                           */
/*  An expression derived from an atomic task, whose value is 1 if the task  */
/*  is assigned a resource from the given resource group, and 0 otherwise.   */
/*                                                                           */
/*  We include the task because we are going to merge equal expressions of   */
/*  this kind, and that is only possible when the task is in the expression. */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_assigned_task_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_TASK_ON_DAY		task_on_day;
  KHE_RESOURCE_GROUP		resource_group;
} *KHE_DRS_EXPR_ASSIGNED_TASK;

typedef HA_ARRAY(KHE_DRS_EXPR_ASSIGNED_TASK) ARRAY_KHE_DRS_EXPR_ASSIGNED_TASK;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_OR                                                     */
/*                                                                           */
/*  An expression whose value is the logical `or' of its children's values.  */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*  The closed state here is the number of closed children with value 1.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_or_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_OR;

typedef HA_ARRAY(KHE_DRS_EXPR_OR) ARRAY_KHE_DRS_EXPR_OR;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_AND                                                    */
/*                                                                           */
/*  An expression whose value is the logical `and' of its children's values. */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*  The closed state here is the number of closed children with value 0.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_and_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_AND;

typedef HA_ARRAY(KHE_DRS_EXPR_AND) ARRAY_KHE_DRS_EXPR_AND;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SUM                                                */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.           */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*  We may be able to use just the value field to hold this, but if so that  */
/*  is still to do.                                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_INT_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM) ARRAY_KHE_DRS_EXPR_INT_SUM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT_SUM                                              */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.  The      */
/*  expression's value, and its children's values, have type float.          */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*  We may be able to use just the value field to hold this, but if so that  */
/*  is still to do.                                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_sum_rec {
  INHERIT_KHE_DRS_EXPR
  float				closed_state;
} *KHE_DRS_EXPR_FLOAT_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_SUM) ARRAY_KHE_DRS_EXPR_FLOAT_SUM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_DEV                                                */
/*                                                                           */
/*  An expression whose value is the deviation of its integer-valued child.  */
/*                                                                           */
/*  The closed state here is the value of the closed child.  There is no     */
/*  need to store it explicitly.                                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_dev_rec {
  INHERIT_KHE_DRS_EXPR
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
} *KHE_DRS_EXPR_INT_DEV;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_DEV) ARRAY_KHE_DRS_EXPR_INT_DEV;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT_DEV                                              */
/*                                                                           */
/*  An expression whose value is the integer-valued deviation of its         */
/*  float-valued child.                                                      */
/*                                                                           */
/*  The closed state here is the value of the closed child.  There is no     */
/*  need to store it explicitly.                                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_dev_rec {
  INHERIT_KHE_DRS_EXPR
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
} *KHE_DRS_EXPR_FLOAT_DEV;

typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_DEV) ARRAY_KHE_DRS_EXPR_FLOAT_DEV;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_COST                                                   */
/*                                                                           */
/*  An expression whose value is a cost based on its sole child's deviation. */
/*                                                                           */
/*  The closed state here is the value of the closed child.                  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_cost_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
} *KHE_DRS_EXPR_COST;

typedef HA_ARRAY(KHE_DRS_EXPR_COST) ARRAY_KHE_DRS_EXPR_COST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_COST_SUM                                               */
/*                                                                           */
/*  An expression whose value is the sum of the values of its children.      */
/*  All these values are costs.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_drs_expr_cost_sum_rec {
  INHERIT_KHE_DRS_EXPR
} *KHE_DRS_EXPR_COST_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_COST_SUM) ARRAY_KHE_DRS_EXPR_COST_SUM;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SUM_COMB                                           */
/*                                                                           */
/*  Combines INT_SUM, INT_DEV, and COST into a single expression.  This      */
/*  allows important optimizations in some cases, depending on comb_type.    */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*  We may be able to use just the value field to hold this, but if so that  */
/*  is still to do.                                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_comb_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
  KHE_DRS_COMB_TYPE		comb_type;	/* type of combined value    */
  int				history_before;	/* history value before      */
  int				history_after;	/* history value after       */
  int				history;	/* history value             */
  int				closed_state;
  /* int			open_children_count; */
  KHE_COST			closed_cost;
} *KHE_DRS_EXPR_INT_SUM_COMB;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM_COMB) ARRAY_KHE_DRS_EXPR_INT_SUM_COMB;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SEQ_COMB                                           */
/*                                                                           */
/*  Combines INT_SEQ, INT_DEV (without allow_zero), and COST into a single   */
/*  expression, for the same optimizations as KHE_DRS_EXPR_INT_SUM_COMB.     */
/*                                                                           */
/*  The closed state here will be a complex record of the state of the       */
/*  closed intervals.  This is very much still to do.                        */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_seq_comb_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  KHE_DRS_COMB_TYPE		comb_type;	/* type of combined value    */
  int				history_before;	/* history value before      */
  int				history_after;	/* history value after       */
  int				history;	/* history value             */
} *KHE_DRS_EXPR_INT_SEQ_COMB;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SEQ_COMB) ARRAY_KHE_DRS_EXPR_INT_SEQ_COMB;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type definitions - solutions and solver"                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN - a partial solution, part of the search tree          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_rec *KHE_DRS_SOLN;

struct khe_drs_soln_rec {
  KHE_DRS_SOLN		prev_soln;		/* edge from prev_soln       */
  KHE_DRS_DAY		day;			/* day containing this soln  */
  ARRAY_KHE_DRS_TASK_ON_DAY prev_tasks;		/* indexed by resources      */
  int			best_of;		/* how many this is best of  */
  KHE_COST		cost;			/* soln cost so far          */
  HA_ARRAY_INT		signature;		/* signature of this soln    */
};

typedef HA_ARRAY(KHE_DRS_SOLN) ARRAY_KHE_DRS_SOLN;
typedef HP_TABLE(KHE_DRS_SOLN) TABLE_KHE_DRS_SOLN;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_LIST - a list of solns; has several uses               */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_list_rec {
  ARRAY_KHE_DRS_SOLN	solns;			/* the solns                 */
} *KHE_DRS_SOLN_LIST;

typedef HA_ARRAY(KHE_DRS_SOLN_LIST) ARRAY_KHE_DRS_SOLN_LIST;
typedef HP_TABLE(KHE_DRS_SOLN_LIST) TABLE_KHE_DRS_SOLN_LIST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_TRIE - a trie of solns                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_trie_rec *KHE_DRS_SOLN_TRIE;
typedef HA_ARRAY(KHE_DRS_SOLN_TRIE) ARRAY_KHE_DRS_SOLN_TRIE;

struct khe_drs_soln_trie_rec {
  KHE_DRS_SOLN			soln;
  ARRAY_KHE_DRS_SOLN_TRIE	children;
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_SET - a set of solns, with a choice of implementations */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_soln_set_rec {
  int				count;			/* number of solns   */
  KHE_DRS_DOM_KIND		dom_kind;		/* kind of dominance */
  union {
    KHE_DRS_SOLN_LIST		none_dom_set;		/* none dominance    */
    TABLE_KHE_DRS_SOLN		weak_dom_set;		/* weak dominance    */
    TABLE_KHE_DRS_SOLN_LIST	medium_dom_set;		/* medium dominance  */
    KHE_DRS_SOLN_LIST		strong_dom_set;		/* strong dominance  */
    KHE_DRS_SOLN_TRIE		trie_dom_set;		/* trie dominance    */
  } u;
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DYNAMIC_RESOURCE_SOLVER - a dynamic resource solver             */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_MONITOR) ARRAY_KHE_MONITOR;

struct khe_dynamic_resource_solver_rec {

  /* fields which are constant throughout the lifetime of the solver */
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_RESOURCE_TYPE		resource_type;
  KHE_OPTIONS			options;
  KHE_FRAME			days_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_TASK_CLASS_SOLVER		task_class_solver;
  ARRAY_KHE_DRS_RESOURCE	all_resources;
  ARRAY_KHE_DRS_DAY		all_days;
  ARRAY_KHE_DRS_TASK		all_root_tasks;
  ARRAY_KHE_DRS_EXPR		all_root_exprs;	/* roots of expr trees */
  int				postorder_count;

  /* free list fields */
  ARRAY_KHE_DRS_SOLN		free_solns;
  ARRAY_KHE_DRS_SOLN_LIST	free_soln_lists;
  ARRAY_KHE_DRS_SOLN_TRIE	free_soln_tries;
  ARRAY_KHE_DRS_SOLN_SET	free_soln_sets;

  /* fields that vary with the solve */
  KHE_RESOURCE_SET		selected_resource_set; /* selected for solve */
  ARRAY_KHE_DRS_DAY_RANGE	selected_day_ranges;   /* selected for solve */
  ARRAY_KHE_DRS_RESOURCE	open_resources;
  ARRAY_KHE_DRS_DAY		open_days;
  ARRAY_KHE_DRS_EXPR		open_exprs;
  KHE_COST			total_closed_cost;
  KHE_COST			solve_init_cost;
  int				solve_soln_limit;
  int				solve_diversifier;
  bool				solve_rerun;

  /* fields for recording running time */
#if TESTING
  KHE_TIMER			timer;
  HA_ARRAY_FLOAT		running_time_per_day;
#endif
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE_ON_DAY"                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(KHE_DRS_RESOURCE dr,     */
/*    KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  Make a new resource on day object with these attributes.                 */
/*  This function assumes that dr is not assigned any task on day.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY res;
  HaMake(res, drs->arena);
  HaArrayInit(res->external_today, drs->arena);
  res->encl_dr = dr;
  res->day = day;
  res->closed_asst = NULL;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,         */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If dr has a BUSY_TIME, FREE_TIME, or WORK_TIME expression with the       */
/*  given tag, dom_test, and time, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2);

static bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_EXPR_FREE_TIME eft;
  KHE_DRS_EXPR_WORK_TIME ewt;
  HaArrayForEach(drd->external_today, e, i)
    if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_TIME_TAG:

	  ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
	  if( ebt->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_TIME_TAG:

	  eft = (KHE_DRS_EXPR_FREE_TIME) e;
	  if( eft->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_TIME_TAG:

	  ewt = (KHE_DRS_EXPR_WORK_TIME) e;
	  if( ewt->time == time )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsResourceOnDayHasTimeExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,          */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If dr has a BUSY_DAY, FREE_DAY, or WORK_DAY expression with the given    */
/*  tag, dom_test, and time group, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_EXPR_WORK_DAY ewd;
  if( KheTimeGroupTimeCount(tg) > 0 )
  {
    HaArrayForEach(drd->external_today, e, i)
      if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
      {
	switch( e->tag )
	{
	  case KHE_DRS_EXPR_BUSY_DAY_TAG:

	    ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
	    if( KheTimeGroupEqual(ebd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_FREE_DAY_TAG:

	    efd = (KHE_DRS_EXPR_FREE_DAY) e;
	    if( KheTimeGroupEqual(efd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_WORK_DAY_TAG:

	    ewd = (KHE_DRS_EXPR_WORK_DAY) e;
	    if( KheTimeGroupEqual(ewd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  default:

	    HnAbort("KheDrsResourceHasDayExpr: invalid tag (%d)", e->tag);
	}
      }
  }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,             */
/*    KHE_DRS_EXPR e)                                                        */
/*                                                                           */
/*  Add e, known to be new, to drd.                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR e)
{
  HaArrayAddLast(drd->external_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayIsFixed(KHE_DRS_RESOURCE_ON_DAY drd,             */
/*    int dr_index, KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY *dtd)             */
/*                                                                           */
/*  If the assignment of drd (whose resource's open index is dr_index)       */
/*  is fixed, return in *dtd the task on day that it is fixed to.            */
/*                                                                           */
/*  There are two reasons why drd may be fixed:                              */
/*                                                                           */
/*    * drd may have a closed assignment;                                    */
/*                                                                           */
/*    * drd's resource may have been assigned to some task on the previous   */
/*      day, and that task may still be running                              */
/*                                                                           */
/*  We can find out whether drd's resource has been assigned to some task    */
/*  on the previous day by consulting the edge leading into soln.            */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,
  KHE_DRS_RESOURCE dr, int dr_solve_index, KHE_DRS_TASK_ON_DAY *dtd);
static bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,
  KHE_DRS_TASK_ON_DAY *dtd);

static bool KheDrsResourceOnDayIsFixed(KHE_DRS_RESOURCE_ON_DAY drd,
  int dr_index, KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY *dtd)
{
  KHE_DRS_TASK_ON_DAY dtd1, dtd2;

  /* if drd has a closed assignment, it's fixed to that */
  if( drd->closed_asst != NULL )
    return *dtd = drd->closed_asst, true;

  /* if drd's resource is assigned to a task in soln which is still running, */
  /* then drd is fixed to that */
  if( KheDrsSolnResourceIsAssigned(soln, drd->encl_dr, dr_index, &dtd1) &&
	KheDrsTaskBusyOnDay(dtd1->encl_dt, drd->day, &dtd2) )
    return *dtd = dtd2, true;

  /* otherwise drd has no fixed assignment */
  return *dtd = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,               */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of drd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,
  int verbosity, int indent, FILE *fp);

static void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "%s:", KheTimeGroupId(drd->day->time_group));
  if( drd->closed_asst != NULL )
    KheDrsTaskOnDayDebug(drd->closed_asst, 1, -1, fp);
  else
    fprintf(fp, "-");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOpen(KHE_DRS_RESOURCE dr,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Open dr.  This involves unassigning all the tasks assigned dr which      */
/*  lie in the selected day ranges of drs, and gathering for opening the     */
/*  expressions that depend on what dr is doing on those days, including     */
/*  on days where dr remains assigned.                                       */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task);
static bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE si1, KHE_DRS_DAY_RANGE si2);
static KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last);
static void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsResourceOpen(KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY_RANGE ddr, open_day_range;  int i, j, k, open_day_index;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_TASK dt;  KHE_DRS_EXPR e;
  open_day_index = 0;
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      /* unassign any task assigned on drd, if it lies entirely in ddr */
      drd = HaArray(dr->days, j);
      if( drd->closed_asst != NULL )
      {
	dt = drd->closed_asst->encl_dt;
	if( KheDrsDayRangeSubset(dt->encl_dtc->day_range, ddr) )
	  KheDrsTaskUnAssign(dt, true);
      }

      /* gather for opening expressions depending on drd (unassigned or not) */
      open_day_range = KheDrsDayRangeMake(open_day_index, open_day_index);
      HaArrayForEach(drd->external_today, e, k)
      {
	e->open_day_range = open_day_range;
	KheDrsExprGatherForOpening(e, drs);
      }

      /* increase open_day_index */
      open_day_index++;
    }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLimit(KHE_MONITOR m)                                       */
/*                                                                           */
/*  Return the maximum limit of m if there is one, or else the minimum       */
/*  limit if there is one, or else 0.                                        */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLimit(KHE_MONITOR m)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;  int val;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      cbtc = KheClusterBusyTimesMonitorConstraint(
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m);
      val = KheClusterBusyTimesConstraintMaximum(cbtc);
      if( val < KheClusterBusyTimesConstraintTimeGroupCount(cbtc) )
	return val;
      val = KheClusterBusyTimesConstraintMinimum(cbtc);
      return val;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      lbtc = KheLimitBusyTimesMonitorConstraint(
	(KHE_LIMIT_BUSY_TIMES_MONITOR) m);
      val = KheLimitBusyTimesConstraintMaximum(lbtc);
      if( val < INT_MAX )
	return val;
      val = KheLimitBusyTimesConstraintMinimum(lbtc);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      /* could do better here */
      return 0;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      val = KheLimitActiveIntervalsMonitorMaximum(laim);
      if( val < KheLimitActiveIntervalsMonitorTimeGroupCount(laim) )
	return val;
      val = KheLimitActiveIntervalsMonitorMinimum(laim);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_AVOID_CLASHES_MONITOR_TAG:
    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:
    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:
    default:

      return 0;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)       */
/*                                                                           */
/*  Comparison function for sorting monitors so that those with the          */
/*  largest limits come first.                                               */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) p1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) p2;
  return KheMonitorLimit(m2) - KheMonitorLimit(m1);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new resource object with these attributes.                        */
/*                                                                           */
/*****************************************************************************/
static void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m, KHE_DRS_RESOURCE dr);

static KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE res;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_DAY day;  KHE_MONITOR m;

  /* initialize res, including one resource on day object for each day */
  HaMake(res, drs->arena);
  HaArrayInit(res->days, drs->arena);
  res->resource = r;
  HaArrayForEach(drs->all_days, day, i)
  {
    drd = KheDrsResourceOnDayMake(res, day, drs);
    HaArrayAddLast(res->days, drd);
  }

  /* add expressions to dr based on the resource's monitors */
  KheSolnResourceMonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    KheDrsAddResourceMonitor(drs, m, res);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDay(KHE_DRS_RESOURCE dr,         */
/*    KHE_DRS_DAY day)                                                       */
/*                                                                           */
/*  Return the resource on day object representing dr on day.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDay(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY day)
{
  return HaArray(dr->days, day->frame_index);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayAtTime(KHE_DRS_RESOURCE dr,   */
/*    KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Return the resource on day object whose resource is dr and whose day     */
/*  is time's day.                                                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayAtTime(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int day_index;
  day_index = KheFrameTimeIndex(drs->days_frame, time);
  return HaArray(dr->days, day_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,             */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dr onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,
  int indent, FILE *fp)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ %s", KheResourceId(dr->resource));
  if( verbosity > 1 && indent >= 0 )
  {
    fprintf(fp, "\n");
    HaArrayForEach(dr->days, drd, i)
      KheDrsResourceOnDayDebug(drd, verbosity, indent + 2, fp);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY_RANGE"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last)                */
/*                                                                           */
/*  Return a new interval with these endpoints.                              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last)
{
  KHE_DRS_DAY_RANGE res;
  res.first = first;
  res.last = last;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeAdd(KHE_DRS_DAY_RANGE *ddr, int index)                */
/*                                                                           */
/*  Update *ddr to include index.                                            */
/*                                                                           */
/*  Implementation note.  Both cases are needed here, as proved by a trace   */
/*  of Add([1, 0], 4).  The second case wrongly produces [1, 4].             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeAdd(KHE_DRS_DAY_RANGE *ddr, int index)
{
  if( ddr->first > ddr->last )
  {
    /* ddr is empty, so make it cover just index */
    ddr->first = ddr->last = index;
  }
  else
  {
    /* ddr is non-empty, so expand it to include index */
    if( index < ddr->first )
      ddr->first = index;
    if( index > ddr->last )
      ddr->last = index;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,            */
/*    KHE_DRS_DAY_RANGE ddr)                                                 */
/*                                                                           */
/*  Merge ddr into *day_ranges, a sorted array of day ranges.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  KHE_DRS_DAY_RANGE ddr)
{
  KHE_DRS_DAY_RANGE ddr2;  int i;

  /* if ddr is empty, it changes nothing (this can't happen, actually) */
  if( ddr.first > ddr.last )
    return;

  HaArrayForEach(*day_ranges, ddr2, i)
  {
    /***********************************************************************/
    /*                                                                     */
    /*  Invariant: day_ranges and ddr may have changed, but it remains     */
    /*  true that merging ddr with day_ranges gives the desired result.    */
    /*  Also, day_ranges[0..i-1] precede and do not interact with ddr.     */
    /*                                                                     */
    /***********************************************************************/

    if( ddr2.last + 1 < ddr.first )
    {
      /* ddr2 precedes ddr and does not interact with it; continue */
      continue;
    }

    if( ddr.last + 1 < ddr2.first )
    {
      /* ddr precedes ddr2 and does not interact with it, or with day_ranges */
      HaArrayAdd(*day_ranges, i, ddr);
      return;
    }

    /* ddr2 interacts with ddr, so expand ddr to include ddr2, then */
    /* delete ddr2 and continue */
    if( ddr2.first < ddr.first )
      ddr.first = ddr2.first;
    if( ddr2.last > ddr.last )
      ddr.last = ddr2.last;
    HaArrayDeleteAndShift(*day_ranges, i);
    i--;
  }

  /* ddr does not interact with any day_ranges, and it follows them all */
  HaArrayAddLast(*day_ranges, ddr);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeContains(KHE_DRS_DAY_RANGE ddr, int i)                 */
/*                                                                           */
/*  Return true if ddr contains i.                                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsDayRangeContains(KHE_DRS_DAY_RANGE ddr, int i)
{
  return ddr.first <= i && i <= ddr.last;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE ddr1, KHE_DRS_DAY_RANGE ddr2)*/
/*                                                                           */
/*  Return true if ddr1 is a subset of ddr2.                                 */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE ddr1, KHE_DRS_DAY_RANGE ddr2)
{
  return ddr1.first >= ddr2.first && ddr1.last <= ddr2.last;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeDebug(KHE_DRS_DAY_RANGE ddr,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of ddr onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeDebug(KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_TIME_GROUP tg1, tg2;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  tg1 = KheFrameTimeGroup(drs->days_frame, ddr.first);
  tg2 = KheFrameTimeGroup(drs->days_frame, ddr.last);
  if( tg1 == tg2 )
    fprintf(fp, "%s", KheTimeGroupId(tg1));
  else
    fprintf(fp, "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,           */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of *day_ranges onto fp with the given verbosity and indent.  */
/*                                                                           */
/*****************************************************************************/

/* *** good but currently unused
static void KheDrsDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY_RANGE ddr;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[");
  HaArrayForEach(*day_ranges, ddr, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    KheDrsDayRangeDebug(ddr, drs, 1, -1, fp);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayMake(int frame_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Make a new day object and add it to drs->all_days.  Just make the basic  */
/*  object; the task classes, expressions, and dominance tests will be       */
/*  added later.                                                             */
/*                                                                           */
/*  Here frame_index is the index of the new day in the common frame.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayMake(int frame_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY res;
  HaMake(res, drs->arena);
  HaArrayInit(res->task_classes, drs->arena);
  HaArrayInit(res->internal_today, drs->arena);
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  res->frame_index = frame_index;
  res->open_day_index = -1;  /* signals not currently open for solving */
  res->time_group = KheFrameTimeGroup(drs->days_frame, frame_index);
  res->soln_set = NULL;  /* will be initialized at the start of the solve */

  /* add the new day to drs->all_days */
  HaArrayAddLast(drs->all_days, res);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayOpen(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int open_day_index)                   */
/*                                                                           */
/*  Open day for solving.  It lies within day range ddr, and its index in    */
/*  the sequence of all open days is open_day_index.                         */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
KHE_DRS_SOLN_SET KheDrsSolnSetMake(KHE_DRS_DOM_KIND dom_kind,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsDayOpen(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,
  KHE_DRS_DOM_KIND dom_kind, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int open_day_index)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  day->open_day_index = open_day_index;
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassOpen(dtc, ddr, drs);
  HaArrayClear(day->internal_today);
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  day->soln_set = KheDrsSolnSetMake(dom_kind, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e)               */
/*                                                                           */
/*  Add open expression e to day.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e)
{
  HaArrayAddLast(day->internal_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)      */
/*                                                                           */
/*  Add dom_test to day and return its index in day->dom_tests.              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)
{
  int res;

  /* add dom_test to day */
  res = HaArrayCount(day->dom_tests);
  HnAssert(dom_test.type != KHE_DRS_DOM_UNUSED,
    "KheDrsDayAddDomTest internal error");
  HaArrayAddLast(day->dom_tests, dom_test);

  /* if dom_test is equality, record it in eq_dom_test_indexes */
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(day->eq_dom_test_indexes, res);

  /* return dom_test's index */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClose(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Close day.                                                               */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc);
static void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsDayClose(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  day->open_day_index = -1;  /* signals no longer open */
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassClose(dtc);
  HaArrayClear(day->internal_today);
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  KheDrsSolnSetFree(day->soln_set, drs);
  day->soln_set = NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddSoln(KHE_DRS_DAY day, KHE_DRS_SOLN soln,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add soln to day, or do nothing if soln has a hard constraint violation   */
/*  or contains too few or too many assignments.  Otherwise what to do       */
/*  depends on whether we are using strong dominance or weak dominance.      */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET dns, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsDayAddSoln(KHE_DRS_DAY day, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KheDrsSolnSetAddSoln(day->soln_set, soln, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,              */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of the dom tests of day.                                     */
/*                                                                           */
/*****************************************************************************/
static char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100]);
static KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type);

static void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,
  int indent, FILE *fp)
{
  int dom_test_freq[KHE_DRS_DOM_EQ_LOOSE + 1];  int i;
  KHE_DRS_DOM_TEST dom_test;  char buff[100];

  /* initialize dom_test_freq */
  for( i = KHE_DRS_DOM_UNUSED;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
    dom_test_freq[i] = 0;

  /* accumulate dom_test_freq */
  HaArrayForEach(day->dom_tests, dom_test, i)
    dom_test_freq[dom_test.type]++;

  /* display frequencies */
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ Day %s dom tests: ", KheTimeGroupId(day->time_group));
  for( i = 0;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
  {
    if( i > 0 )
      fprintf(fp, ", ");
    fprintf(fp, "%d %s", dom_test_freq[i],
      KheDrsDomTestShow(KheDrsDomTestMake(i), buff));
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of day onto fp.                                              */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);

static void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Day %s (frame %d, ", indent, "",
      KheTimeGroupId(day->time_group), day->frame_index);
    if( day->open_day_index >= 0 )
      fprintf(fp, "open %d)\n", day->open_day_index);
    else
      fprintf(fp, "closed)\n");
    if( verbosity > 1 )
    {
      HaArrayForEach(day->task_classes, dtc, i)
	KheDrsTaskClassDebug(dtc, drs, verbosity, indent + 2, fp);
      KheDrsDayDomTestsDebug(day, verbosity, indent + 2, fp);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "%s", KheTimeGroupId(day->time_group));
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_ON_DAY"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_ON_DAY KheDrsTaskOnDayMake(KHE_DRS_TASK encl_dt,            */
/*    KHE_DRS_DAY day, KHE_TASK task, KHE_TIME time,                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new task on day object with these attributes.                     */
/*  This function assumes that the task is not assigned anything.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK_ON_DAY KheDrsTaskOnDayMake(KHE_DRS_TASK encl_dt,
  KHE_DRS_DAY day, KHE_TASK task, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY res;
  HaMake(res, drs->arena);
  HaArrayInit(res->external_today, drs->arena);
  res->encl_dt = encl_dt;
  res->day = day;
  res->task = task;
  res->time = time;
  res->closed_asst = NULL;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskOnDayCmp(const void *t1, const void *t2)                   */
/*                                                                           */
/*  Comparison function for sorting an array of task on day objects by       */
/*  increasing day index.                                                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskOnDayCmp(const void *t1, const void *t2)
{
  KHE_DRS_TASK_ON_DAY dtd1, dtd2;
  dtd1 = * (KHE_DRS_TASK_ON_DAY *) t1;
  dtd2 = * (KHE_DRS_TASK_ON_DAY *) t2;
  return dtd1->day->frame_index - dtd2->day->frame_index;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsOptionalResourceGroupEqual(KHE_RESOURCE_GROUP rg1,            */
/*    KHE_RESOURCE_GROUP rg2)                                                */
/*                                                                           */
/*  Return true if rg1 and rg2 are both NULL, or both non-NULL and equal.    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsOptionalResourceGroupEqual(KHE_RESOURCE_GROUP rg1,
  KHE_RESOURCE_GROUP rg2)
{
  if( rg1 == NULL )
    return rg2 == NULL;
  else
    return rg2 != NULL && KheResourceGroupEqual(rg1, rg2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,         */
/*    KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res)                */
/*                                                                           */
/*  If dtd has an assigned task expression with resource group attribute     */
/*  rg, return true with *res set to that expression; else return false.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res)
{
  KHE_DRS_EXPR e;  int i;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  HaArrayForEach(dtd->external_today, e, i)
  {
    eat = (KHE_DRS_EXPR_ASSIGNED_TASK) e;
    if( KheDrsOptionalResourceGroupEqual(rg, eat->resource_group) )
      return *res = eat, true;
  }
  return *res = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,         */
/*    KHE_DRS_EXPR_ASSIGNED_TASK eat)                                        */
/*                                                                           */
/*  Add eat to dtd.                                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_EXPR_ASSIGNED_TASK eat)
{
  HaArrayAddLast(dtd->external_today, (KHE_DRS_EXPR) eat);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,                       */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of dtd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( dtd->task != NULL )
  {
    KheTaskDebug(dtd->task, 1, -1, fp);
    fprintf(fp, "(%s:%s:%s)", KheTimeId(dtd->time),
      KheTimeGroupId(dtd->day->time_group), dtd->closed_asst == NULL ? "-" :
      KheResourceId(dtd->closed_asst->encl_dr->resource));
  }
  else
  {
    fprintf(fp, "-(%s)", KheTimeGroupId(dtd->day->time_group));
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskAssign(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr, bool task)   */
/*                                                                           */
/*  Assign dr (which must be non-NULL) to dt, by setting the close_asst      */
/*  fields in dt, dt->days, and dr.  If task is true, also assign dt's KHE   */
/*  task to dr's KHE resource, aborting if that does not succeed.            */
/*                                                                           */
/*  The enclosing task class has a list of unassigned tasks; but do not      */
/*  remove dt from it, because we are being lazy:  the unassigned tasks      */
/*  will be sorted out later, by KheDrsTaskClassOrganizeOpenTasks.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskAssign(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr, bool task)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  HnAssert(dt->closed_asst == NULL, "KheDrsTaskAssign internal error 1");
  HnAssert(dr != NULL, "KheDrsTaskAssign internal error 2");
  if( task && !KheTaskAssignResource(dt->task, dr->resource) )
    HnAbort("KheDrsTaskAssign internal error 3");
  dt->closed_asst = dr;
  HaArrayForEach(dt->days, dtd, i)
  {
    drd = KheDrsResourceOnDay(dr, dtd->day);
    HnAssert(dtd->closed_asst == NULL, "KheDrsTaskAssign internal error 4");
    HnAssert(drd->closed_asst == NULL, "KheDrsTaskAssign internal error 5");
    dtd->closed_asst = drd;
    drd->closed_asst = dtd;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task)                      */
/*                                                                           */
/*  Unassign dt (which must be assigned some drs resource dr), by clearing   */
/*  the closed_asst fields in dt, dt->days, and dr.  If task is true, also   */
/*  unassign dt's KHE task, or do nothing if the task will not unassign.     */
/*                                                                           */
/*  The enclosing task class has a list of unassigned tasks; add dt to it.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;
  dr = dt->closed_asst;
  HnAssert(dr != NULL, "KheDrsTaskUnAssign internal error 1");
  if( task )
  {
    HnAssert(KheTaskAsstResource(dt->task) == dr->resource,
      "KheDrsTaskUnAssign internal error 2");
    if( !KheTaskUnAssignResource(dt->task) )
      return;
  }
  HaArrayForEach(dt->days, dtd, i)
  {
    drd = KheDrsResourceOnDay(dr, dtd->day);
    HnAssert(dtd->closed_asst == drd, "KheDrsTaskUnAssign internal error 3");
    HnAssert(drd->closed_asst == dtd, "KheDrsTaskUnAssign internal error 4");
    dtd->closed_asst = NULL;
    drd->closed_asst = NULL;
  }
  dt->closed_asst = NULL;
  HaArrayAddLast(dt->encl_dtc->open_tasks, dt);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskAddDays(KHE_TASK task, KHE_DRS_TASK dt,                      */
/*    int first_solve_index, int last_solve_index,                           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add the days covered by task and its descendants to dt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheTaskAddDays(KHE_TASK task, KHE_DRS_TASK dt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_TASK child_task;  int i, durn, day_index;  KHE_DRS_DAY day;
  KHE_MEET meet;  KHE_TIME start_time, time;  KHE_DRS_TASK_ON_DAY dtd;

  /* do it for task */
  meet = KheTaskMeet(task);
  HnAssert(meet != NULL, "KheTaskAddDays internal error (no meet)");
  start_time = KheMeetAsstTime(meet);
  HnAssert(start_time != NULL,  "KheTaskAddDays internal error (no time)");
  durn = KheMeetDuration(meet);
  for( i = 0;  i < durn;  i++ )
  {
    time = KheTimeNeighbour(start_time, i);
    day_index = KheFrameTimeIndex(drs->days_frame, time);
    day = HaArray(drs->all_days, day_index);
    dtd = KheDrsTaskOnDayMake(dt, day, task, time, drs);
    HaArrayAddLast(dt->days, dtd);
  }

  /* do it for task's descendants */
  for( i = 0;  i < KheTaskAssignedToCount(task);  i++ )
  {
    child_task = KheTaskAssignedTo(task, i);
    KheTaskAddDays(child_task, dt, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOrganizeDays(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Assuming that dt->days are sorted by increasing day index, organize      */
/*  them so that there is one per day, by inserting NULL entries where       */
/*  there are gaps.  There can be no cases of two on one day, thanks to an   */
/*  earlier call to KheTaskClassNoOverlap.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOrganizeDays(KHE_DRS_TASK dt)
{
  int init_len, reqd_len, i, final_pos;
  KHE_DRS_TASK_ON_DAY first_dtd, last_dtd, dtd;

  /* quit early if the number of days is fine as is */
  first_dtd = HaArrayFirst(dt->days);
  last_dtd = HaArrayLast(dt->days);
  reqd_len = last_dtd->day->frame_index - first_dtd->day->frame_index + 1;
  init_len = HaArrayCount(dt->days);
  if( reqd_len == init_len )
    return;

  /* add the required number of NULL entries, initially at the array's end */
  HaArrayFill(dt->days, reqd_len, NULL);

  /* move the entries to their final positions, working right to left */
  for( i = init_len - 1;  i >= 0;  i-- )
  {
    dtd = HaArray(dt->days, i);
    final_pos = dtd->day->frame_index - first_dtd->day->frame_index;
    if( final_pos != i )
    {
      HaArrayPut(dt->days, final_pos, dtd);
      HaArrayPut(dt->days, i, NULL);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assuming that r has the correct resource type, return its corresponding  */
/*  drs resource.                                                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  return HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskMake(KHE_DRS_TASK_CLASS encl_dtc,                 */
/*    int index_in_encl_dtc, KHE_TASK task, KHE_DYNAMIC_RESOURCE_SOLVER drs) */
/*                                                                           */
/*  Make a new drs task with these attributes, add it to drs->all_root_tasks */
/*  at the same index as it has in the solution, and return it.              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheDrsTaskMake(KHE_DRS_TASK_CLASS encl_dtc,
  int index_in_encl_dtc, KHE_TASK task, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;  int task_index;  KHE_RESOURCE r;

  /* get the basic object and initialize its fields */
  HnAssert(KheTaskIsProperRoot(task), "KheDrsTaskMake internal error");
  HaMake(res, drs->arena);
  HaArrayInit(res->days, drs->arena);
  res->encl_dtc = encl_dtc;
  res->index_in_encl_dtc = index_in_encl_dtc;
  res->open = false;
  res->task = task;
  res->closed_asst = NULL;

  /* add, sort, and organize the task on day objects */
  KheTaskAddDays(task, res, drs);
  HaArraySort(res->days, &KheDrsTaskOnDayCmp);
  KheDrsTaskOrganizeDays(res);

  /* add res to drs->all_root_tasks */
  task_index = KheTaskSolnIndex(task);
  HaArrayFill(drs->all_root_tasks, task_index + 1, NULL);
  HnAssert(HaArray(drs->all_root_tasks, task_index) == NULL,
    "KheDrsTaskMake internal error");
  HaArrayPut(drs->all_root_tasks, task_index, res);

  /* if task is assigned, make that assignment in res */
  r = KheTaskAsstResource(task);
  if( r != NULL )
    KheDrsTaskAssign(res, KheDrsResource(r, drs), false);

  /* and also return res */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskIndexInEnclCmp(const void *t1, const void *t2)             */
/*                                                                           */
/*  Comparison function for sorting an array of drs tasks into increasing    */
/*  order of their index in their enclosing task class.                      */
/*                                                                           */
/*  In addition, it moves any assigned tasks to the end.                     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskIndexInEnclCmp(const void *t1, const void *t2)
{
  KHE_DRS_TASK dt1 = * (KHE_DRS_TASK *) t1;
  KHE_DRS_TASK dt2 = * (KHE_DRS_TASK *) t2;
  if( dt1->closed_asst != NULL )
  {
    if( dt2->closed_asst != NULL )
    {
      /* dt1 is assigned, dt2 is assigned */
      return 0;
    }
    else
    {
      /* dt1 is assigned, dt2 is not assigned */
      return 1;
    }
  }
  else
  {
    if( dt2->closed_asst != NULL )
    {
      /* dt1 is not assigned, dt2 is assigned */
      return -1;
    }
    else
    {
      /* dt1 is not assigned, dt2 is not assigned */
      return dt1->index_in_encl_dtc - dt2->index_in_encl_dtc;
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOpen(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Open task for solving.  We'll have checked previously that its days      */
/*  all lie in the current open interval, and it must currently be closed    */
/*  but unassigned; so opening dt basically just gathers for opening all     */
/*  expressions that depend on dt's assignment.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOpen(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR e;  int i, j, di;
  KHE_DRS_DAY_RANGE open_day_range;

  /* open dt */
  HnAssert(!dt->open, "KheDrsTaskOpen internal error 1");
  HnAssert(dt->closed_asst == NULL, "KheDrsTaskOpen internal error 2");
  dt->open = true;

  /* gather external expressions for opening */
  HaArrayForEach(dt->days, dtd, i)
  {
    di = dtd->day->open_day_index;
    open_day_range = KheDrsDayRangeMake(di, di);
    HaArrayForEach(dtd->external_today, e, j)
    {
      e->open_day_range = open_day_range;
      KheDrsExprGatherForOpening(e, drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)               */
/*                                                                           */
/*  If dt is open, close it, with a call to KheDrsTaskAssign if dr != NULL.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)
{
  if( dt->open )
  {
    dt->open = false;
    if( dr != NULL )
      KheDrsTaskAssign(dt, dr, true);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,               */
/*    KHE_DRS_TASK_ON_DAY *dtd)                                              */
/*                                                                           */
/*  If dt is running on day, return true with *dtd set to its task on day    */
/*  object for that day.  Otherwise return false with *dtd set to NULL.      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,
  KHE_DRS_TASK_ON_DAY *dtd)
{
  KHE_DRS_TASK_ON_DAY dtd2;  int i;
  HaArrayForEach(dt->days, dtd2, i)
    if( dtd2->day == day )
      return *dtd = dtd2, true;
  return *dtd = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,                     */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dt onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  KheTaskDebug(dt->task, 1, -1, fp);
  fprintf(fp, "(%d, %s, %s; ", dt->index_in_encl_dtc,
    dt->open ? "open" : "closed",
    dt->closed_asst == NULL ? "-" : KheResourceId(dt->closed_asst->resource));
  HaArrayForEach(dt->days, dtd, i)
    KheDrsTaskOnDayDebug(dtd, verbosity, -1, fp);
  fprintf(fp, ")");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_CLASS"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassMake(KHE_TASK_CLASS tc,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task class based on tc and add it to its first day.       */
/*  Also make drs tasks corresponding to tc's tasks and add them to the      */
/*  drs class task and to drs->all_root_tasks.                               */
/*                                                                           */
/*  The following comment is obsolete; it relates to selecting tasks         */
/*  for opening during a solve, whereas here we select all tasks.            */
/*                                                                           */
/*  Here first_solve_index is the index in the solver (not in the days       */
/*  frame) of the first day that the tasks of tc are running, and similarly  */
/*  last_solve_index is the index in the solve of the last day that the      */
/*  tasks of tc are running.  Every day that the tasks of tc are running is  */
/*  a solve day, as we know thanks to an earlier call to KheTaskClassIsOpen. */
/*                                                                           */
/*  Which tasks are added to the result?  We have to add all tasks which     */
/*  are assigned selected resources, to allow them to be unassigned.  We     */
/*  also have to add at least as many tasks as there are selected resources  */
/*  (if there are that many in tc), to allow for the possibility of every    */
/*  selected resource being assigned a task of this class simultaneously.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassMake(KHE_TASK_CLASS tc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS res;  int i;  KHE_TASK task;  KHE_DRS_TASK dt;
  KHE_DRS_DAY day;

  /* get or make a drs task class object */
  HaMake(res, drs->arena);
  HaArrayInit(res->open_tasks, drs->arena);
  res->orig_task_class = tc;
  res->day_range = KheDrsDayRangeMake(KheTaskClassFirstDayIndex(tc),
    KheTaskClassLastDayIndex(tc));
  res->tmp_used = 0;

  /* make drs tasks for its khe tasks and add them */
  for( i = 0;  i < KheTaskClassTaskCount(tc);  i++ )
  {
    task = KheTaskClassTask(tc, i);
    dt = KheDrsTaskMake(res, i, task, drs);
    HaArrayAddLast(res->all_tasks, dt);
    if( dt->closed_asst == NULL )
      HaArrayAddLast(res->open_tasks, dt);
  }

  /* make res one of the task classes of its first day */
  day = HaArray(drs->all_days, res->day_range.first);
  HaArrayAddLast(day->task_classes, res);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassOrganizeOpenTasks(KHE_DRS_TASK_CLASS dtc)            */
/*                                                                           */
/*  Organize the open tasks of dtc for solving.                              */
/*                                                                           */
/*  When this function is called, open_tasks may be out of order, and,       */
/*  because open_tasks is handled lazily, some of them may in fact be        */
/*  assigned, so not open.  This function fixes both of those problems.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassOrganizeOpenTasks(KHE_DRS_TASK_CLASS dtc)
{
  KHE_DRS_TASK dt;

  /* sort open_tasks into order with assigned tasks at the end */
  HaArraySortUnique(dtc->open_tasks, &KheDrsTaskIndexInEnclCmp);

  /* remove assigned tasks from the end */
  while( HaArrayCount(dtc->open_tasks) > 0 )
  {
    dt = HaArrayLast(dtc->open_tasks);
    if( dt->closed_asst == NULL )
      break;
    HaArrayDeleteLast(dtc->open_tasks);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  If dtc is suitable (if its day range lies within ddr, and its resource   */
/*  domain includes at least one open resource), then open it.  Opening dtc  */
/*  basically means organizing its unassigned tasks into decreasing          */
/*  desirability order and opening them.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  int i;
  if( KheDrsDayRangeSubset(dtc->day_range, ddr) &&
      !KheResourceSetDisjointGroup(drs->selected_resource_set,
	KheTaskClassDomain(dtc->orig_task_class)) )
  {
    /* dtc can open; organize and open open_tasks; say none are used */
    KheDrsTaskClassOrganizeOpenTasks(dtc);
    HaArrayForEach(dtc->open_tasks, dt, i)
      KheDrsTaskOpen(dt, drs);
    dtc->tmp_used = 0;
  }
  else
  {
    /* dtc can't open; set dtc->tmp_used to make that clear */
    dtc->tmp_used = INT_MAX;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Close dtc.  Some of its tasks may be already closed; that's OK.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc)
{
  KHE_DRS_TASK dt;  int i;
  HaArrayForEach(dtc->open_tasks, dt, i)
    KheDrsTaskClose(dt, NULL);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,          */
/*    KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt)                                 */
/*                                                                           */
/*  If dtc accepts dr (because dr lies in its tasks' domains, and not all    */
/*  the tasks are assigned), then return true and set *dt to one of dtc's    */
/*  tasks that dr can be assigned to.  Otherwise return false.               */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,
  KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt)
{
  KHE_RESOURCE_GROUP domain;
  domain = KheTaskClassDomain(dtc->orig_task_class);
  if( dtc->tmp_used < HaArrayCount(dtc->open_tasks) &&
	KheResourceGroupContains(domain, dr->resource) )
  {
    *dt = HaArray(dtc->open_tasks, dtc->tmp_used);
    dtc->tmp_used++;
    return true;
  }
  else
    return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc)            */
/*                                                                           */
/*  Undo the effect of a corresponding KheDrsTaskClassAcceptResourceBegin    */
/*  which returned true.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc)
{
  dtc->tmp_used--;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc, int verbosity,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of dtc onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK dt;  int i, pos;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Task class ", indent, "");
    KheDrsDayRangeDebug(dtc->day_range, drs, 1, 0, fp);
    HaArrayForEach(dtc->all_tasks, dt, i)
    {
      fprintf(fp, "%*s  %c ", indent, "",
	HaArrayContains(dtc->open_tasks, dt, &pos) ? 'o' : 'c');
      KheDrsTaskDebug(dt, 1, 0, fp);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    fprintf(fp, "TaskClass(");
    KheDrsDayRangeDebug(dtc->day_range, drs, 1, -1, fp);
    fprintf(fp, ", %d tasks)", HaArrayCount(dtc->all_tasks));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DOM_TEST"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type)                */
/*                                                                           */
/*  Make and return a new dominance test object with no min limit attribute. */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.min_limit = -1;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val)  */
/*                                                                           */
/*  Make and return a new dominance test object with a min limit attribute.  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.min_limit = val;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)      */
/*                                                                           */
/*  Return true if dt1 and dt2 are equal.  This function assumes, as it      */
/*  may, that when the min_limit field is unused its value is always -1.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)
{
  return dt1.type == dt2.type && dt1.min_limit == dt2.min_limit;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)     */
/*                                                                           */
/*  Return true if val1 dominates val2, according to dt.                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominates internal error (KHE_DRS_DOM_UNUSED)");
      return false;  /* keep compiler happy */

    case KHE_DRS_DOM_GE:

      /* tight min limit */
      return val1 >= val2;

    case KHE_DRS_DOM_LE:

      /* max limit */
      return val1 <= val2;

    case KHE_DRS_DOM_EQ:

      /* values must be equal */
      return val1 == val2;

    case KHE_DRS_DOM_GE_LOOSE:

      /* loose min limit */
      return val1 >= val2 || val1 >= dt.min_limit;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* max and loose min; both conditions must hold (basically equality) */
      return (val1 <= val2) && (val1 >= val2 || val1 >= dt.min_limit);

    default:

      HnAbort("KheDrsDomTestDominates internal error (%d)", dt.type);
      return false;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,          */
/*    int *from_val, int *to_val)                                            */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that dominate val        */
/*  under dt.  That is, set it to the largest range of values x such that    */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, x, val)                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominatingRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;

    case KHE_DRS_DOM_GE:

      /* largest range of values x such that x >= val */
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_LE:

      /* largest range of values x such that x <= val */
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ:

      /* largest range of values x such that x == val */
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      /* largest range of values x such that x >= val || x >= dt.min_limit */
      /* i.e. such that x >= min(val, dt.min_limit) */
      *from_val = min(val, dt.min_limit), *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* largest range of values x st x <= val && x >= min(val, dt.min_limit) */
      *from_val = min(val, dt.min_limit), *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatingRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,           */
/*    int *from_val, int *to_val)                                            */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that are dominated by    */
/*  val under dt.  That is, set it to the largest range of values x s.t.     */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, val, x)                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominatedRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;

    case KHE_DRS_DOM_GE:

      /* largest range of values x such that val >= x */
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_LE:

      /* largest range of values x such that val <= x */
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ:

      /* largest range of values x such that val == x */
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      /* largest range of values x such that val >= x || val >= dt.min_limit */
      if( val >= dt.min_limit )
        *from_val = 0, *to_val = INT_MAX;
      else
        *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* largest range of x st val <= x && (val >= x || val >= dt.min_limit) */
      if( val >= dt.min_limit )
	*from_val = val, *to_val = INT_MAX;
      else
        *from_val = val, *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatedRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])             */
/*                                                                           */
/*  Show dt, using buff for scratch storage.                                 */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      sprintf(buff, "-");
      break;

    case KHE_DRS_DOM_GE:

      sprintf(buff, ">=");
      break;

    case KHE_DRS_DOM_LE:

      sprintf(buff, "<=");
      break;

    case KHE_DRS_DOM_EQ:

      sprintf(buff, "==");
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      sprintf(buff, ">=(%d)", dt.min_limit);
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      sprintf(buff, "=(%d)", dt.min_limit);
      break;

    default:

      sprintf(buff, "?");
      break;
  }
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "helper code for expressions, including deviations and costs"  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KheDrsExprInitBegin(res, tg, dt, drs, free_list)                         */
/*                                                                           */
/*  Shared generic code for creating an expression object and                */
/*  initializing those fields common to all expressions.                     */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprInitBegin(res, tg, dt, drs)				\
  HaMake(res, drs->arena);						\
  HaArrayInit(res->parents, drs->arena);				\
  HaArrayInit(res->children, drs->arena);				\
  HaArrayInit(res->open_children, drs->arena);				\
  HaArrayInit(res->open_child_indexes, drs->arena);			\
  HaArrayInit(res->sig_indexes, drs->arena);				\
  res->tag = tg;							\
  res->gathered = false;						\
  res->postorder_index = -1;						\
  res->open_day_range = KheDrsDayRangeMake(1, 0);			\
  res->dom_test = dt

#define OldKheDrsExprInit(res, tg, dt, drs, free_list)			\
  if( HaArrayCount(drs->free_list) > 0 )				\
  {									\
    res = HaArrayLastAndDelete(drs->free_list);				\
    HaArrayClear(res->parents);						\
    HaArrayClear(res->children);					\
    HaArrayClear(res->open_children);					\
    HaArrayClear(res->open_child_indexes);				\
    HaArrayClear(res->sig_indexes);					\
  }									\
  else									\
  {									\
    HaMake(res, drs->arena);						\
    HaArrayInit(res->parents, drs->arena);				\
    HaArrayInit(res->children, drs->arena);				\
    HaArrayInit(res->open_children, drs->arena);			\
    HaArrayInit(res->open_child_indexes, drs->arena);			\
    HaArrayInit(res->sig_indexes, drs->arena);				\
  }									\
  res->tag = tg;							\
  res->postorder_index = -1;						\
  res->open_day_range = KheDrsDayRangeMake(1, 0);			\
  res->dom_test = dt


/*****************************************************************************/
/*                                                                           */
/*  KheDrsExprInitEnd(res, drs)                                              */
/*                                                                           */
/*  Shared generic code for ending the intialization of res by setting       */
/*  its postorder index and (closed) value.                                  */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprInitEnd(res, drs)					\
  (res)->postorder_index = (drs)->postorder_count++,			\
  KheDrsExprSetClosedValue((KHE_DRS_EXPR) (res), (drs))


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprOpenDayIsFirst(KHE_DRS_EXPR e, int open_day_index)        */
/*                                                                           */
/*  Return true if open_day_index is the index of the first open day of e.   */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprOpenDayIsFirst(e, open_day_index)			\
  ((open_day_index) == (e)->open_day_range.first)


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprSigVal(KHE_DRS_EXPR e, int open_day_index,                 */
/*    KHE_DRS_SOLN soln)                                                     */
/*                                                                           */
/*  Retrieve from soln the signature of e on open day open_day_index.        */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprSigVal(e, open_day_index, soln)			\
  HaArray((soln)->signature,						\
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first))


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprPutSigVal(KHE_DRS_EXPR e, int open_day_index,             */
/*    KHE_DRS_SOLN soln, int sig_val)                                        */
/*                                                                           */
/*  Add sig_val, the signature of e on open_day_index, to soln.              */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprPutSigVal(e, open_day_index, soln, val)		     \
  HnAssert(HaArrayCount((soln)->signature) ==				     \
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first), \
    "KheDrsExprPutSigVal internal error"),				     \
  HaArrayAddLast((soln)->signature, (val));


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprForEachOpenDayChild(KHE_DRS_EXPR e, int open_day_index,   */
/*    KHE_DRS_EXPR x, int i)                                                 */
/*                                                                           */
/*  Iterator which visits every open child of e whose last open day is       */
/*  open_day_index, setting x to each child in turn and i to its index       */
/*  in e->open_children.                                                     */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprForEachOpenDayChild(e, odi, x, i)			      \
  i1 = HaArray((e)->open_child_indexes, (odi)   - (e)->open_day_range.first); \
  i2 = HaArray((e)->open_child_indexes, (odi)+1 - (e)->open_day_range.first); \
  for( (i) = i1;							      \
       (i) < i2 ? ((x) = HaArray((e)->open_children, i), true) : false;	      \
       (i)++ )


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprOpenDayIsLast(KHE_DRS_EXPR e, int open_day_index)         */
/*                                                                           */
/*  Return true if open_day_index is the index of the last open day of e.    */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprOpenDayIsLast(e, open_day_index)			      \
  ((open_day_index) == (e)->open_day_range.last)


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSumDev(int min_limit, int max_limit, bool allow_zero,          */
/*    int history_after, int val)                                            */
/*                                                                           */
/*  Deviation function to use with sum expressions; it has allow_zero.       */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSumDev(int min_limit, int max_limit, bool allow_zero,
  int history_after, int val)
{
  if( val == 0 && allow_zero )
    return 0;
  else if( val > max_limit )
    return val - max_limit;
  else if( val + history_after < min_limit )
    return min_limit - (val + history_after);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSeqDev(int min_limit, int max_limit,                           */
/*    int history_after, int val)                                            */
/*                                                                           */
/*  Deviation function to use with sequence expressions; no allow_zero.      */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSeqDev(int min_limit, int max_limit,
  int history_after, int val)
{
  if( val > max_limit )
    return val - max_limit;
  else if( val + history_after < min_limit )
    return min_limit - (val + history_after);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, */
/*    int dev)                                                               */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int dev)
{
  switch( cost_fn )
  {
    case KHE_STEP_COST_FUNCTION:

        return dev > 0 ? combined_weight : 0;

    case KHE_LINEAR_COST_FUNCTION:

        return dev * combined_weight;

    case KHE_QUADRATIC_COST_FUNCTION:

	return dev * dev * combined_weight;

    default:

	HnAbort("KheDrsCost internal error");
	return KheCost(0, 0);  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si)     */
/*                                                                           */
/*  If e is in the signature on day, return true and set *si to its          */
/*  position in the signature.  Otherwise return false.                      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si)
{
  int i = day->open_day_index;
  if( e->open_day_range.first <= i && i < e->open_day_range.last )
    return *si = HaArray(e->sig_indexes, i - e->open_day_range.first), true;
  else
    return *si = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,         */
/*    bool is_float, FILE *fp)                                               */
/*                                                                           */
/*  If soln != NULL and e has a signature entry in soln, debug that entry.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebugSignature(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  bool is_float, FILE *fp)
{
  int si;
  if( soln != NULL && KheDrsExprInSignature(e, soln->day, &si) )
  {
    if( is_float )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(soln->signature, si)/100.0);
    else
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeSetClosedValue(KHE_DRS_EXPR_BUSY_TIME ebt,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ebt, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeSetClosedValue(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ebt->resource_on_day->closed_asst;
  ebt->u.int_val = (dtd != NULL && dtd->time == ebt->time ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(                           */
/*    KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,                            */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_BUSY_TIME_TAG,
	dom_test, time, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_TIME) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_TIME_TAG, dom_test, drs);
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeCopy(KHE_DRS_EXPR_BUSY_TIME ebt,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ebt.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeCopy(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_TIME_TAG, ebt->dom_test, drs);
  res->u.int_val = ebt->u.int_val;
  res->time = ebt->time;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeLeafSet(KHE_DRS_EXPR_BUSY_TIME ebt,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ebt, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ebt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeLeafSet(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( dtd->time == ebt->time )
    ebt->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeLeafClear(KHE_DRS_EXPR_BUSY_TIME ebt)             */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprBusyTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeLeafClear(KHE_DRS_EXPR_BUSY_TIME ebt)
{
  ebt->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDoDebug(KHE_DRS_EXPR_BUSY_TIME ebt,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ebt which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeDoDebug(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "BUSY_TIME(%s, %s)",
    KheResourceId(ebt->resource_on_day->encl_dr->resource),
    KheTimeId(ebt->time));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ebt, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDebug(KHE_DRS_EXPR_BUSY_TIME ebt,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy time expression ebt.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyTimeDebug(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "BUSY_TIME(%s)", KheTimeId(ebt->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ebt, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    ** indented version without a soln value **
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ BUSY_TIME(%s) %s ]", KheTimeId(ebt->time),
      KheDrsDomTestShow(ebt->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_FREE_TIME_TAG,
      dom_test, time, drs, &e) )
    return (KHE_DRS_EXPR_FREE_TIME) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_TIME_TAG, dom_test, drs);
  res->u.int_val = 0;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeCopy(KHE_DRS_EXPR_FREE_TIME eft,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of eft.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeCopy(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_TIME_TAG, eft->dom_test, drs);
  res->u.int_val = eft->u.int_val;
  res->time = eft->time;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeSetClosedValue(KHE_DRS_EXPR_FREE_TIME eft,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eft, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeSetClosedValue(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = eft->resource_on_day->closed_asst;
  eft->u.int_val = (dtd != NULL && dtd->time == eft->time ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeLeafSet(KHE_DRS_EXPR_FREE_TIME eft,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of eft, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as eft is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeLeafSet(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( dtd->time == eft->time )
    eft->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeLeafClear(KHE_DRS_EXPR_FREE_TIME eft)             */
/*                                                                           */
/*  Inform leaf expression eft that the assignment of drd to dtd carried     */
/*  out by KheDrsExprFreeTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeLeafClear(KHE_DRS_EXPR_FREE_TIME eft)
{
  eft->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDoDebug(KHE_DRS_EXPR_FREE_TIME eft,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eft which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeDoDebug(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FREE_TIME(%s, %s)",
    KheResourceId(eft->resource_on_day->encl_dr->resource),
    KheTimeId(eft->time));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eft, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDebug(KHE_DRS_EXPR_FREE_TIME eft,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for free time expression eft.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeTimeDebug(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "FREE_TIME(%s)", KheTimeId(eft->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eft, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    ** indented version without a soln value **
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ FREE_TIME(%s) %s ]", KheTimeId(eft->time),
      KheDrsDomTestShow(eft->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_WORK_TIME_TAG,
	dom_test, time, drs, &e) )
    return (KHE_DRS_EXPR_WORK_TIME) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_TIME_TAG, dom_test, drs);
  res->u.float_val = 0.0;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeCopy(KHE_DRS_EXPR_WORK_TIME ewt,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ewt.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeCopy(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_TIME_TAG, ewt->dom_test, drs);
  res->u.float_val = ewt->u.float_val;
  res->time = ewt->time;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeSetClosedValue(KHE_DRS_EXPR_WORK_TIME ewt,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ewt, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeSetClosedValue(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ewt->resource_on_day->closed_asst;
  ewt->u.float_val = (dtd != NULL && dtd->time == ewt->time ?
    KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeLeafSet(KHE_DRS_EXPR_WORK_TIME ewt,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ewt, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ewt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeLeafSet(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( dtd->time == ewt->time )
    ewt->u.float_val = KheTaskWorkloadPerTime(dtd->task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeLeafClear(KHE_DRS_EXPR_WORK_TIME ewt)             */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprWorkTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeLeafClear(KHE_DRS_EXPR_WORK_TIME ewt)
{
  ewt->u.float_val = 0.0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDoDebug(KHE_DRS_EXPR_WORK_TIME ewt,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ewt which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeDoDebug(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "WORK_TIME(%s, %s)",
    KheResourceId(ewt->resource_on_day->encl_dr->resource),
    KheTimeId(ewt->time));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ewt, soln, true, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDebug(KHE_DRS_EXPR_WORK_TIME ewt,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for work time expression ewt.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkTimeDebug(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "WORK_TIME(%s)", KheTimeId(ewt->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ewt, soln->day, &si) )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(soln->signature, si) / 100.0);
  }
  else
  {
    ** indented version without a soln value **
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ WORK_TIME(%s) %s ]", KheTimeId(ewt->time),
      KheDrsDomTestShow(ewt->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasDayExpr(drd, KHE_DRS_EXPR_BUSY_DAY_TAG, dom_test,
      tg, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_DAY) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_DAY_TAG, dom_test, drs);
  res->u.int_val = 0;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time_group = tg;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayCopy(KHE_DRS_EXPR_BUSY_DAY ebd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ebd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayCopy(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_DAY_TAG, ebd->dom_test, drs);
  res->u.int_val = ebd->u.int_val;
  res->time_group = ebd->time_group;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDaySetClosedValue(KHE_DRS_EXPR_BUSY_DAY ebd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ebd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDaySetClosedValue(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ebd->resource_on_day->closed_asst;
  ebd->u.int_val = (dtd != NULL ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayLeafSet(KHE_DRS_EXPR_BUSY_DAY ebd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ebd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ebt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayLeafSet(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ebd->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayLeafClear(KHE_DRS_EXPR_BUSY_DAY ebd)               */
/*                                                                           */
/*  Inform leaf soln ebd that the assignment of drd to dtd carried out by    */
/*  KheDrsExprBusyDayLeafSet above has been removed.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayLeafClear(KHE_DRS_EXPR_BUSY_DAY ebd)
{
  ebd->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDoDebug(KHE_DRS_EXPR_BUSY_DAY ebd,                 */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ebd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayDoDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "BUSY_DAY(%s, %s)",
    KheResourceId(ebd->resource_on_day->encl_dr->resource),
    KheTimeGroupId(ebd->time_group));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ebd, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,                   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy day expression ebd.                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "BUSY_DAY(%s)", KheTimeGroupId(ebd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ebd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    ** indented version without a soln value **
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ BUSY_DAY(%s) %s ]", KheTimeGroupId(ebd->time_group),
      KheDrsDomTestShow(ebd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasDayExpr(drd, KHE_DRS_EXPR_FREE_DAY_TAG,
	dom_test, tg, drs, &e) )
    return (KHE_DRS_EXPR_FREE_DAY) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_DAY_TAG, dom_test, drs);
  res->u.int_val = 1;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time_group = tg;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayCopy(KHE_DRS_EXPR_FREE_DAY efd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of efd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayCopy(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_DAY_TAG, efd->dom_test, drs);
  res->u.int_val = efd->u.int_val;
  res->time_group = efd->time_group;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDaySetClosedValue(KHE_DRS_EXPR_FREE_DAY efd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDaySetClosedValue(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = efd->resource_on_day->closed_asst;
  efd->u.int_val = (dtd != NULL ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayLeafSet(KHE_DRS_EXPR_FREE_DAY efd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of efd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as efd is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayLeafSet(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  efd->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayLeafClear(KHE_DRS_EXPR_FREE_DAY efd)               */
/*                                                                           */
/*  Inform leaf expression eft that the assignment of drd to dtd carried     */
/*  out by KheDrsExprFreeDayLeafSet above has been removed.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayLeafClear(KHE_DRS_EXPR_FREE_DAY efd)
{
  efd->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDoDebug(KHE_DRS_EXPR_FREE_DAY efd,                 */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of efd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayDoDebug(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FREE_DAY(%s, %s)",
    KheResourceId(efd->resource_on_day->encl_dr->resource),
    KheTimeGroupId(efd->time_group));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) efd, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDebug(KHE_DRS_EXPR_FREE_DAY efd,                   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for free time expression efd.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeDayDebug(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "FREE_DAY(%s)", KheTimeGroupId(efd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    ** indented version without a soln value **
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ FREE_DAY(%s) %s ]", KheTimeGroupId(efd->time_group),
      KheDrsDomTestShow(efd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused, although I really should be using it
static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsHasDayExpr(drs, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, tg, &e) )
    return (KHE_DRS_EXPR_WORK_DAY) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, drs);
  res->u.float_val = 0.0;  ** default value, used sometimes **
  res->time_group = tg;
  KheDrsAddDayExpr(drs, tg, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayCopy(KHE_DRS_EXPR_WORK_DAY ewd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ewd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayCopy(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_DAY_TAG, ewd->dom_test, drs);
  res->u.float_val = ewd->u.float_val;
  res->time_group = ewd->time_group;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDaySetClosedValue(KHE_DRS_EXPR_WORK_DAY ewd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ewd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDaySetClosedValue(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ewd->resource_on_day->closed_asst;
  ewd->u.int_val = (dtd != NULL ? KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayLeafSet(KHE_DRS_EXPR_WORK_DAY ewd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ewd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ewd is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayLeafSet(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ewd->u.float_val = KheTaskWorkloadPerTime(dtd->task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayLeafClear(KHE_DRS_EXPR_WORK_DAY ewd)               */
/*                                                                           */
/*  Inform leaf expression ewt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprWorkDayLeafSet above has been removed.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayLeafClear(KHE_DRS_EXPR_WORK_DAY ewd)
{
  ewd->u.float_val = 0.0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDoDebug(KHE_DRS_EXPR_WORK_DAY ewd,                 */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ewd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayDoDebug(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "WORK_DAY(%s, %s)",
    KheResourceId(ewd->resource_on_day->encl_dr->resource),
    KheTimeGroupId(ewd->time_group));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ewd, soln, true, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDebug(KHE_DRS_EXPR_WORK_DAY ewd,                   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for work day expression ewd.                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkDayDebug(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "WORK_DAY(%s)", KheTimeGroupId(ewd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ewd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    ** indented version without a soln value **
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ WORK_DAY(%s) %s ]", KheTimeGroupId(ewd->time_group),
      KheDrsDomTestShow(ewd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_ASSIGNED_TASK"                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(                   */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_RESOURCE_GROUP rg,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an assigned task expression object with these attributes.           */
/*  Or if there is already such an expression in dtd's list, use that.       */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res);
static void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_EXPR_ASSIGNED_TASK eat);

static KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(
  KHE_DRS_TASK_ON_DAY dtd, KHE_RESOURCE_GROUP rg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_ASSIGNED_TASK res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  HnAssert(dtd->task != NULL, "KheDrsExprAssignedTaskMake internal error");
  if( KheDrsTaskOnDayHasAssignedTaskExpr(dtd, rg, &res) )
    return res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_ASSIGNED_TASK_TAG, dom_test, drs);
  res->u.int_val = 0;    /* default value, will be used sometimes */
  res->task_on_day = dtd;
  res->resource_group = rg;
  KheDrsTaskOnDayAddAssignedTaskExpr(dtd, res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprAssignedTaskValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,          */
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Return the value of eat when its task is assigned r (possibly NULL).     */
/*  This will be 1 when r is non-NULL and lies in eat's resource group.      */
/*  For efficienty, eat's resource group may be NULL, meaning all resources. */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprAssignedTaskValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_RESOURCE r)
{
  KHE_RESOURCE_GROUP rg;
  if( r == NULL )
    return 0;
  rg = eat->resource_group;
  return (rg == NULL || KheResourceGroupContains(rg, r)) ? 1 : 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskSetClosedValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eat, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskSetClosedValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_RESOURCE r;
  drd = eat->task_on_day->closed_asst;
  r = (drd == NULL ? NULL : drd->encl_dr->resource);
  eat->u.int_val = KheDrsExprAssignedTaskValue(eat, r);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskLeafSet(KHE_DRS_EXPR_ASSIGNED_TASK eat,       */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Update the value of eat to reflect the assignment of drd to dtd.         */
/*  Here dtd is known to be the relevant task, and drd is known to           */
/*  represent a non-NULL resource, not a non-assignment.  Also the           */
/*  value of eat must currently be 0.                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskLeafSet(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( KheDrsExprAssignedTaskValue(eat, drd->encl_dr->resource) == 1 )
    eat->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskLeafClear(KHE_DRS_EXPR_ASSIGNED_TASK eat)     */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprAssignedTaskLeafSet above has been removed.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskLeafClear(KHE_DRS_EXPR_ASSIGNED_TASK eat)
{
  eat->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDoDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,       */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eat which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskDoDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "ASSIGNED_TASK(");
  HnAssert(eat->task_on_day->task != NULL,
    "KheDrsExprAssignedTaskDoDebug internal error");
  KheTaskDebug(eat->task_on_day->task, 1, -1, fp);
  fprintf(fp, ", %s)", eat->resource_group == NULL ? "-" :
    KheResourceGroupId(eat->resource_group));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eat, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,         */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for assigned task expression eat.              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAssignedTaskDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "ASSIGNED_TASK");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eat, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ASSIGNED_TASK %s %s ]\n", indent, "",
      KheDrsDomTestShow(eat->dom_test, buff2),
      KheResourceGroupId(eat->resource_group));
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ASSIGNED_TASK %s %s ]",
      KheDrsDomTestShow(eat->dom_test, buff2),
      KheResourceGroupId(eat->resource_group));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_OR"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_OR object with these attributes.                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_OR_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eo for solving, eo's child_index'th child, child_e,   */
/*  has been opened.  Update eo's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of eo is the number of eo's closed children whose       */
/*  value is 1, so this update subtracts 1 from the closed state if          */
/*  child_e's value (which preserves its value before opening) is 1.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state -= 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eo, eo's child_index'th child, child_e, has been      */
/*  closed.  Update eo's closed state to reflect this.                       */
/*                                                                           */
/*  The closed state of eo is the number of eo's closed children whose       */
/*  value is 1, so this update adds 1 to the closed state if child_e's       */
/*  value (which is up to date because it is closed first) is 1.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrSetClosedValue(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eo, set its value suitably for its closed state.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrSetClosedValue(KHE_DRS_EXPR_OR eo,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eo->u.int_val = (eo->closed_state > 0 ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrEvalDay(KHE_DRS_EXPR_OR eo,                             */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eo on day open_day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrEvalDay(KHE_DRS_EXPR_OR eo, int open_day_index,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, val;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(eo, open_day_index) )
  {
    /* first day, so we have a 0 (false) value here */
    val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = KheDrsExprSigVal(eo, open_day_index - 1, prev_soln);
  }

  /* accumulate the values of the children of eo that finalized today */
  KheDrsExprForEachOpenDayChild(eo, open_day_index, child_e, i)
    if( child_e->u.int_val == 1 )
      val = 1;

  if( KheDrsExprOpenDayIsLast(eo, open_day_index) )
  {
    /* last day; add closed_state to the value */
    eo->u.int_val = (eo->closed_state + val > 0 ? 1 : 0);
  }
  else
  {
    /* this not eo's last day; store value in next_soln */
    KheDrsExprPutSigVal(eo, open_day_index, next_soln, val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDoDebug(KHE_DRS_EXPR_OR eo,                             */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eo which is specific to it.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrDoDebug(KHE_DRS_EXPR_OR eo,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "OR(%d)", eo->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eo, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprHeaderDebug(KHE_DRS_EXPR e, char *type_tag, FILE *fp)     */
/*                                                                           */
/*  Print the header of e (with the given type tag) onto fp.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprHeaderDebug(KHE_DRS_EXPR e, char *type_tag,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, FILE *fp)
{
  char buff2[100];
  fprintf(fp, "%s p%d o%d r%s", type_tag, e->postorder_index,
    HaArrayCount(e->open_children), KheDrsDomTestShow(e->dom_test, buff2));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDebug(KHE_DRS_EXPR_OR eo,                               */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for or expression eo.                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrDebug(KHE_DRS_EXPR_OR eo,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eo, "OR", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eo, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eo, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eo, "OR", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eo, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eo, "OR", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eo->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_AND"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_AND object with these attributes.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_AND_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening ea for solving, ea's child_index'th child, child_e,   */
/*  has been opened.  Update ea's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ea is the number of ea's closed children whose       */
/*  value is 0, so this update subtracts 1 from the closed state if          */
/*  child_e's value (which preserves its value before opening) is 0.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state -= 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing ea, ea's child_index'th child, child_e, has been      */
/*  closed.  Update ea's closed state to reflect this.                       */
/*                                                                           */
/*  The closed state of ea is the number of ea's closed children whose       */
/*  value is 0, so this update adds 1 to the closed state if child_e's       */
/*  value (which is up to date because it is closed first) is 0.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndSetClosedValue(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ea, set its value suitably for its closed state.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndSetClosedValue(KHE_DRS_EXPR_AND ea,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ea->u.int_val = (ea->closed_state > 0 ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndEvalDay(KHE_DRS_EXPR_AND ea,                           */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression ea on day open_day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndEvalDay(KHE_DRS_EXPR_AND ea,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(ea, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 1 (true) value here */
    ea->u.int_val = 1;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    ea->u.int_val = KheDrsExprSigVal(ea, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of ea that finalized today */
  KheDrsExprForEachOpenDayChild(ea, open_day_index, child_e, i)
    if( child_e->u.int_val == 0 )
      ea->u.int_val = 0;

  if( KheDrsExprOpenDayIsLast(ea, open_day_index) )
  {
    /* last day; add closed_state to the value */
    if( ea->closed_state > 0 )
      ea->u.int_val = 0;
  }
  else
  {
    /* this not ea's last day; store value in next_soln */
    KheDrsExprPutSigVal(ea, open_day_index, next_soln, ea->u.int_val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDoDebug(KHE_DRS_EXPR_AND ea,                           */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ea which is specific to it.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndDoDebug(KHE_DRS_EXPR_AND ea,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "AND(%d)", ea->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) ea, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDebug(KHE_DRS_EXPR_AND ea,                             */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for and expression ea.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndDebug(KHE_DRS_EXPR_AND ea,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ea, "AND", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ea, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ea, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ea, "AND", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ea, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ea, "AND", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(ea->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SUM_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eis for solving, eis's child_index'th child, child_e, */
/*  has been opened.  Update eis's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eis is the sum of the values of eis's s closed       */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->closed_state -= child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eis, eis's child_index'th child, child_e, has been    */
/*  closed.  Update eis's closed state to reflect this.                      */
/*                                                                           */
/*  The closed state of eis is the sum of the values of eis's s closed       */
/*  children, so this update adds child_e's value (which preserves           */
/*  its value before opening) to the closed state.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->closed_state += child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumSetClosedValue(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eis, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumSetClosedValue(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->u.int_val = eis->closed_state;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumEvalDay(KHE_DRS_EXPR_INT_SUM eis,                   */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eis on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumEvalDay(KHE_DRS_EXPR_INT_SUM eis,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(eis, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 0 value here */
    eis->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    eis->u.int_val = KheDrsExprSigVal(eis, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of eis that finalized today */
  KheDrsExprForEachOpenDayChild(eis, open_day_index, child_e, i)
    eis->u.int_val += child_e->u.int_val;

  if( KheDrsExprOpenDayIsLast(eis, open_day_index) )
  {
    /* last day; add closed_state to the value */
    eis->u.int_val += eis->closed_state;
  }
  else
  {
    /* this not eis's last day; store value in next_soln */
    KheDrsExprPutSigVal(eis, open_day_index, next_soln, eis->u.int_val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDoDebug(KHE_DRS_EXPR_INT_SUM eis,                   */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eis which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumDoDebug(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_SUM(%d)", eis->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eis, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDebug(KHE_DRS_EXPR_INT_SUM eis,                     */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy time expression eis.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumDebug(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eis, "INT_SUM", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eis, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eis, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eis, "INT_SUM", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eis, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eis, "INT_SUM", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eis->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ KheDrsExprIntSeqMake(int history_before,            */
/*    int history_after, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static KHE_DRS_EXPR_INT_SEQ KheDrsExprIntSeqMake(int history_before,
  int history_after, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SEQ_TAG, dom_test, drs);
  res->history_before = history_before;
  res->history_after = history_after;
  res->tmp_value = UNDEF_INT;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqFree(KHE_DRS_EXPR_INT_SEQ eis,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eis and its descendants.                                            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqFree(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eis->children, e, i)
    KheDrsExprFree(e, drs);
  HaArrayAddLast(drs->free_expr_int_seq, eis);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSeqDomTest(KHE_DRS_EXPR_INT_SEQ eis)       */
/*                                                                           */
/*  Return a suitable dom test for eis.                                      */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static KHE_DRS_DOM_TEST KheDrsExprIntSeqDomTest(KHE_DRS_EXPR_INT_SEQ eis)
{
  return KheExprMonotoneDomTest((KHE_DRS_EXPR) eis);  ** is this right? **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqBeginDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,   */
/*    KHE_DRS_SOLN prev_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Inform eis that day day_index is beginning.                              */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqBeginDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,
  KHE_DRS_SOLN prev_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->tmp_value = KheDrsExprIntBeginDay((KHE_DRS_EXPR) eis, day_index,
    eis->history_before, prev_soln, drs, "IntSeq");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqReceiveValueInt(KHE_DRS_EXPR_INT_SEQ eis,           */
/*    int value, KHE_DYNAMIC_RESOURCE_SOLVER drs)                            */
/*                                                                           */
/*  Inform eis that one of its children has just finalized its value.        */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqReceiveValueInt(KHE_DRS_EXPR_INT_SEQ eis,
  int value, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(value == 0 || value == 1,
    "KheDrsExprIntSeqReceiveValueInt internal error");
  if( value == 1 )
  {
    ** a sequence is beginning or continuing **
    eis->tmp_value++;
  }
  else if( eis->tmp_value > 0 )
  {
    ** a non-empty sequence has just ended **
    KheDrsExprReceiveValueInt(eis->parent, eis->tmp_value, drs);
    eis->tmp_value = 0;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqEndDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,     */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)                    */
/*                                                                           */
/*  Inform eis that day day_index is ending.                                 */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqEndDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->tmp_value = KheDrsExprIntEndDay((KHE_DRS_EXPR) eis, day_index,
    ** eis->tmp_value > 0, ** eis->tmp_value, soln, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqDoDebug(KHE_DRS_EXPR_INT_SEQ eis,                   */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eis which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqDoDebug(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DRS_SOLN soln, FILE *fp)
{
  ** who knows **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqDebug(KHE_DRS_EXPR_INT_SEQ eis,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of eis onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqDebug(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KheDrsExprIntDebug((KHE_DRS_EXPR) eis, "INT_SEQ", eis->tmp_value,
    drs, verbosity, indent, fp);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_SUM"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(                           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_SUM object with these attributes.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_SUM res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FLOAT_SUM_TAG, dom_test, drs);
  res->u.float_val = UNDEF_FLOAT;
  res->closed_state = 0.0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening efs for solving, efs's child_index'th child, child_e, */
/*  has been opened.  Update efs's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efs is the sum of the values of efs's closed         */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->closed_state -= child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening efs for solving, efs's child_index'th child, child_e, */
/*  has been closed.  Update efs's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efs is the sum of the values of efs's closed         */
/*  children, so this update adds child_e's value (which preserves           */
/*  its value before opening) to the closed state.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->closed_state += child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumSetClosedValue(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efs, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumSetClosedValue(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->u.float_val = efs->closed_state;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumEvalDay(KHE_DRS_EXPR_FLOAT_SUM efs,               */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efs on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumEvalDay(KHE_DRS_EXPR_FLOAT_SUM efs,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(efs, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 0 value here */
    efs->u.float_val = 0.0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    efs->u.float_val =
      (float) KheDrsExprSigVal(efs, open_day_index - 1, prev_soln) / 100.0;
  }

  /* accumulate the open values of the children of efs that finalized today */
  KheDrsExprForEachOpenDayChild(efs, open_day_index, child_e, i)
    efs->u.float_val += child_e->u.float_val;

  if( KheDrsExprOpenDayIsLast(efs, open_day_index) )
  {
    /* last day; add closed_state to the value */
    efs->u.float_val += efs->closed_state;
  }
  else
  {
    /* this not efs's last day; store value in next_soln */
    KheDrsExprPutSigVal(efs, open_day_index, next_soln,
      (int) (efs->u.float_val * 100.0));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDoDebug(KHE_DRS_EXPR_FLOAT_SUM efs,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of efs which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumDoDebug(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FLOAT_SUM(%.2f)", efs->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) efs, soln, true, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDebug(KHE_DRS_EXPR_FLOAT_SUM efs,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of efs in soln onto fp.                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumDebug(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efs, "FLOAT_SUM", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efs, soln->day, &si) )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(soln->signature, si)/100.0);
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efs, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efs, "FLOAT_SUM", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efs, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efs, "FLOAT_SUM", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(efs->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_DEV"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit,                 */
/*    int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_DEV object with these attributes.            */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_DEV res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_DEV_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eid for solving, eid's child_index'th child, child_e, */
/*  has been opened.  Update eid's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eid is the value of eid's single closed child.       */
/*  This is not stored explicitly, so this update does nothing.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eid for solving, eid's child_index'th child, child_e, */
/*  has been closed.  Update eid's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eid is the value of eid's single closed child.       */
/*  This is not stored explicitly, so this update does nothing.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevSetClosedValue(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eid, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevSetClosedValue(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(eid->children);
  eid->u.int_val = KheDrsSumDev(eid->min_limit, eid->max_limit,
    eid->allow_zero, 0, child_e->u.int_val);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevEvalDay(KHE_DRS_EXPR_INT_DEV eid,                   */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eid on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevEvalDay(KHE_DRS_EXPR_INT_DEV eid,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(eid, open_day_index) )
  {
    /* first day */
    eid->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprIntDevEvalDay internal error (non-first day)");
  }

  /* accumulate the open values of the children of eid that finalized today */
  KheDrsExprForEachOpenDayChild(eid, open_day_index, child_e, i)
    eid->u.int_val += KheDrsSumDev(eid->min_limit, eid->max_limit,
      eid->allow_zero, 0, child_e->u.int_val);

  if( KheDrsExprOpenDayIsLast(eid, open_day_index) )
  {
    /* last day; add closed_state to the value */
    /* don't add anything because the sole child must be open so there */
    /* is no closed state to add in */
  }
  else
  {
    /* this not eid's last day; store value in next_soln */
    HnAbort("KheDrsExprIntDevEvalDay internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDoDebug(KHE_DRS_EXPR_INT_DEV eid,                   */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eid which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevDoDebug(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_DEV(%d, %d, %s)", eid->min_limit, eid->max_limit,
    bool_show(eid->allow_zero));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eid, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDebug(KHE_DRS_EXPR_INT_DEV eid,                     */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of eid in soln onto fp.                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevDebug(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value (actually no soln value here) **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eid, "INT_DEV", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eid, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eid, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eid, "INT_DEV", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eid, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eid, "INT_DEV", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eid->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_DEV"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,             */
/*    int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_DEV object with these attributes.          */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_DEV res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FLOAT_DEV_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    float workload)                                                        */
/*                                                                           */
/*  Return the deviation of workload wrt efd.                                */
/*                                                                           */
/*****************************************************************************/

static int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,
  float workload)
{
  if( efd->allow_zero && workload < 0.001 )
    return 0;
  else if( workload < efd->min_limit - 0.001 )
    return (int) ceil(efd->min_limit - 0.001 - workload);
  else if( workload > efd->max_limit + 0.001 )
    return (int) ceil(workload - efd->max_limit - 0.001);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening efd for solving, efd's child_index'th child, child_e, */
/*  has been opened.  Update efd's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efd is the value of efd's single closed child,       */
/*  so this update needs to do nothing.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing efd, efd's child_index'th child, child_e,             */
/*  has been closed.  Update efd's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efd is the value of efd's single closed child,       */
/*  so this update needs to do nothing.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevSetClosedValue(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevSetClosedValue(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(efd->children);
  efd->u.int_val = KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevEvalDay(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efd on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevEvalDay(KHE_DRS_EXPR_FLOAT_DEV efd,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(efd, open_day_index) )
  {
    /* first day */
    efd->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprFloatDevEvalDay internal error (non-first day)");
  }

  /* accumulate the open values of the children of efd that finalized today */
  KheDrsExprForEachOpenDayChild(efd, open_day_index, child_e, i)
    efd->u.int_val += KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);

  if( KheDrsExprOpenDayIsLast(efd, open_day_index) )
  {
    /* last day; add closed_state to the value */
    /* don't add anything because the sole child must be open so there */
    /* is no closed state to add in */
  }
  else
  {
    /* this not efd's last day; store value in next_soln */
    HnAbort("KheDrsExprFloatDevEvalDay internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDoDebug(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of efd which is specific to it.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevDoDebug(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "FLOAT_DEV(%d, %d, %s)", efd->min_limit, efd->max_limit,
    bool_show(efd->allow_zero));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) efd, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDebug(KHE_DRS_EXPR_FLOAT_DEV efd,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for float dev expression efd.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevDebug(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value (actually no soln value here) **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efd, "FLOAT_DEV", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efd, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efd, "FLOAT_DEV", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efd, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efd, "FLOAT_DEV", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(efd->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_COST"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_COST KheDrsExprCostMake(KHE_MONITOR m,                      */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_COST object with these attributes.               */
/*                                                                           */
/*  This function has no dom_test parameter, because there is never any      */
/*  dominance testing at this expression.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_COST KheDrsExprCostMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_COST res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_COST_TAG, dom_test, drs);
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->u.int_val = -1;  /* no value at this expression */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostChildHasOpened(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening ec for solving, ec's child_index'th child, child_e,   */
/*  has been opened.  Update ec's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ec is the value of ec's single closed child, so      */
/*  this update does nothing.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostChildHasOpened(KHE_DRS_EXPR_COST ec,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* update drs->total_closed_cost still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostChildHasClosed(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing ec, ec's child_index'th child, child_e,               */
/*  has been closed.  Update ec's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ec is the value of ec's single closed child, so      */
/*  this update does nothing.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostChildHasClosed(KHE_DRS_EXPR_COST ec,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* update drs->total_closed_cost still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostSetClosedValue(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ec, set its value suitably for its closed state.      */
/*                                                                           */
/*  The cost here needs to be reported directly to somewhere; I don't        */
/*  know where yet, that's still to do.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostSetClosedValue(KHE_DRS_EXPR_COST ec,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* *** lhs here still to do; rhs is correct
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(ec->children);
  somewhere += KheDrsCost(ec->cost_fn, ec->combined_weight, child_e->u.int_val);
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostEvalDay(KHE_DRS_EXPR_COST ec,                         */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression ec on day open_day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostEvalDay(KHE_DRS_EXPR_COST ec,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;  KHE_COST cost;

  if( KheDrsExprOpenDayIsFirst(ec, open_day_index) )
  {
    /* first day */
    cost = KheCost(0, 0);
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprCostEvalDay internal error (non-first day)");
    cost =  KheCost(0, 0);  /* keep compiler happy */
  }

  /* accumulate the open values of the children of ec that finalized today */
  KheDrsExprForEachOpenDayChild(ec, open_day_index, child_e, i)
    cost += KheDrsCost(ec->cost_fn, ec->combined_weight, child_e->u.int_val);

  if( KheDrsExprOpenDayIsLast(ec, open_day_index) )
  {
    /* last day; add closed_state to the value */
    /* don't add anything because the sole child must be open so there */
    /* is no closed state to add in */
    if( cost > 0 )
    {
      if( drs->solve_rerun )
	fprintf(stderr, "  %s reporting cost %.5f\n",
	  KheMonitorId(ec->monitor), KheCostShow(cost));
      next_soln->cost += cost;
    }
  }
  else
  {
    /* this not ec's last day; store value in next_soln */
    HnAbort("KheDrsExprCostEvalDay internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDoDebug(KHE_DRS_EXPR_COST ec,                         */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of ec which is specific to it.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostDoDebug(KHE_DRS_EXPR_COST ec,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "COST(%s, %s, %.5f)", KheMonitorId(ec->monitor),
    KheCostFunctionShow(ec->cost_fn), KheCostShow(ec->combined_weight));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDebug(KHE_DRS_EXPR_COST ec,                           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for cost expression ec.                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostDebug(KHE_DRS_EXPR_COST ec,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ec, "COST", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ec, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ec, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ec, "COST", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ec, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ec, "COST", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(ec->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM_COMB"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_COMB_TYPE KheDrsCombType(KHE_COST_FUNCTION cost_fn,              */
/*    int max_limit)                                                         */
/*                                                                           */
/*  Return the comb type, as determined by the cost function and max limit.  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_COMB_TYPE KheDrsCombType(KHE_COST_FUNCTION cost_fn,
  int max_limit)
{
  if( max_limit == INT_MAX )
    return KHE_DRS_COMB_NO_MAX;
  else if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    return KHE_DRS_COMB_LINEAR;
  else if( cost_fn == KHE_STEP_COST_FUNCTION )
    return KHE_DRS_COMB_STEP;
  else
    return KHE_DRS_COMB_ORDINARY;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsCombTypeShow(KHE_DRS_COMB_TYPE dct)                          */
/*                                                                           */
/*  Return a string representation of dct.                                   */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsCombTypeShow(KHE_DRS_COMB_TYPE dct)
{
  switch( dct )
  {
    case KHE_DRS_COMB_NO_MAX:	return "NO_MAX";
    case KHE_DRS_COMB_LINEAR:	return "LINEAR";
    case KHE_DRS_COMB_STEP:	return "STEP";
    case KHE_DRS_COMB_ORDINARY:	return "ORDINARY";
    default:			return "??";
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprIntSumCombDev(KHE_DRS_EXPR_INT_SUM_COMB eisc,              */
/*    int children_with_value_1, int children_with_unknown_value)            */
/*                                                                           */
/*  Return the deviation of eisc, assuming that eisc has the given number    */
/*  of children with value 1 (not counting history), and the given number    */
/*  of children with unknown value (not counting history after).             */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprIntSumCombDev(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  int children_with_value_1, int children_with_unknown_value)
{
  int val;
  val = children_with_value_1 + eisc->history;
  if( val == 0 && eisc->allow_zero )
    return 0;
  if( val > eisc->max_limit )
    return val - eisc->max_limit;
  val += children_with_unknown_value + eisc->history_after;
  return (val < eisc->min_limit ? eisc->min_limit - val : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsExprIntSumCombClosedCost(KHE_DRS_EXPR_INT_SUM_COMB eisc)  */
/*                                                                           */
/*  Return the cost of eisc when only its closed children have known values. */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsExprIntSumCombClosedCost(KHE_DRS_EXPR_INT_SUM_COMB eisc)
{
  int dev;
  dev = KheDrsExprIntSumCombDev(eisc, eisc->closed_state,
    HaArrayCount(eisc->open_children));
    /* eisc->open_children_count); */
  return KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombUpdateClosedCost(KHE_DRS_EXPR_INT_SUM_COMB eisc,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Recalculate the closed cost of eisc (because something relevant to it    */
/*  has changed), and update eisc->closed_cost and drs->total_closed_cost.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombUpdateClosedCost(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  drs->total_closed_cost -= eisc->closed_cost;
  eisc->closed_cost = KheDrsExprIntSumCombClosedCost(eisc);
  drs->total_closed_cost += eisc->closed_cost;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM_COMB KheDrsExprIntSumCombMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    int min_limit, int max_limit, bool allow_zero,                         */
/*    int history_before, int history_after, int history,                    */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM_COMB object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM_COMB KheDrsExprIntSumCombMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int min_limit, int max_limit, bool allow_zero,
  int history_before, int history_after, int history,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM_COMB res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SUM_COMB_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  res->comb_type = KheDrsCombType(cost_fn, max_limit);
  res->history_before = history_before;
  res->history_after = history_after;
  res->history = history;
  res->closed_state = 0;
  /* res->open_children_count = 0; */
  res->closed_cost = 0;
  KheDrsExprIntSumCombUpdateClosedCost(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombChildHasOpened(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eisc for solving, eisc's child_index'th child,        */
/*  child_e, has been opened.  Update eisc's closed state to reflect this.   */
/*                                                                           */
/*  This function is called after child_e is added to eisc->open_children.   */
/*  So when KheDrsExprIntSumCombUpdateClosedCost retrieves the number of     */
/*  open children, it gets the correct (that is, the changed) value.         */
/*                                                                           */
/*  The closed state of eisc is the sum of the values of eisc's closed       */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*  Because eisc's value is a cost, drs->total_closed_cost is also updated.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombChildHasOpened(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eisc->closed_state -= child_e->u.int_val;
  /* eisc->open_children_count++; */
  KheDrsExprIntSumCombUpdateClosedCost(eisc, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombChildHasClosed(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eisc, eisc's child_index'th child, child_e, has       */
/*  been closed.  Update eisc's closed state to reflect this.                */
/*                                                                           */
/*  This is called after child_e is removed from eisc->open_children.        */
/*  So when KheDrsExprIntSumCombUpdateClosedCost retrieves the number of     */
/*  open children, it gets the correct (that is, the changed) value.         */
/*                                                                           */
/*  The closed state of eisc is the sum of the values of eisc's closed       */
/*  children, so this update adds child_e's value to the closed state.       */
/*                                                                           */
/*  Because eisc's value is a cost, drs->total_closed_cost is also updated.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombChildHasClosed(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eisc->closed_state += child_e->u.int_val;
  /* eisc->open_children_count--; */
  KheDrsExprIntSumCombUpdateClosedCost(eisc, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombSetClosedValue(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eisc, set its value suitably for its closed state.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombSetClosedValue(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* nothing to do here, since eisc has closed_cost instead of a value */
  /* ***
  int dev;  KHE_COST cost;
  dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
    eisc->history_after, eisc->closed_state);
  if( dev > 0 )
  {
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    if( cost > 0 )
    {
      ** reporting cost here still to do **
    }
  }
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombEvalDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,         */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc on day open_day_index.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombEvalDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, dev, val;  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ai = HaArray(eisc->open_child_indexes, open_day_index - e->first_day_index);
  bi = HaArray(eisc->open_child_indexes, open_day_index+1 - e->first_day_index)
    - ai;
  ci = HaArrayCount(eisc->open_child_indexes) - ai - bi;
  if( KheDrsExprOpenDayIsFirst(eisc, open_day_index) )
  {
    xi = eisc->history;
  }
  else
  {
    xi = KheDrsExprSigVal(eisc, open_day_index - 1, prev_soln) + eisc->history;
  }

  if( KheDrsExprOpenDayIsFirst(eisc, open_day_index) )
  {
    /* first day */
    val = eisc->history_before + eisc->closed_state;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = KheDrsExprSigVal(eisc, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of eisc that finalized today */
  KheDrsExprForEachOpenDayChild(eisc, open_day_index, child_e, i)
    val += child_e->u.int_val;

  /* adjust in various special cases */
  switch( eisc->comb_type )
  {
    case KHE_DRS_COMB_ORDINARY:

      /* no adjustment here */
      break;

    case KHE_DRS_COMB_NO_MAX:

      /* no maximum limit; store minimum limit at most */
      if( val > eisc->min_limit )
	val = eisc->min_limit;
      break;

    case KHE_DRS_COMB_LINEAR:

      /* linear cost; if over maximum, report excess immediately */
      if( val > eisc->max_limit )
      {
	cost = (val - eisc->max_limit) * eisc->combined_weight;
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting linear cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
	val = eisc->max_limit;
      }
      break;

    case KHE_DRS_COMB_STEP:

      /* step cost; store maximum plus one at most */
      if( val > eisc->max_limit + 1 )
	val = eisc->max_limit + 1;
      break;

    default:

      HnAbort("KheDrsExprIntSumCombEvalDay internal error");
      break;
  }

  if( KheDrsExprOpenDayIsLast(eisc, open_day_index) )
  {
    /* this is eisc's last day; report any cost immediately to next_soln */
    dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
      eisc->history_after, val);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->solve_rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_soln->cost += cost;
      }
    }
  }
  else
  {
    /* this not eisc's last day; store value in next_soln */
    KheDrsExprPutSigVal(eisc, open_day_index, next_soln, val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDoDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,         */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eisc which is specific to it.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombDoDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_SUM_COMB(%s, %s, %d)", KheMonitorId(eisc->monitor),
    KheDrsCombTypeShow(eisc->comb_type), eisc->closed_state);
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eisc, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for int sum comb expression eisc.              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
     KheDrsExprHeaderDebug((KHE_DRS_EXPR) eisc, "INT_SUM_COMB", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eisc, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eisc, "INT_SUM_COMB", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eisc, "INT_SUM_COMB", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eisc->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ_COMB"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ_COMB KheDrsExprIntSeqCombMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,    */
/*    int max_limit, int history_before, int history_after, int history,     */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ_COMB object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SEQ_COMB KheDrsExprIntSeqCombMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,
  int max_limit, int history_before, int history_after, int history,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ_COMB res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SEQ_COMB_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->comb_type = KheDrsCombType(cost_fn, max_limit);
  res->history_before = history_before;
  res->history_after = history_after;
  res->history = history;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of opening eisc for solving, eisc's child_index'th child,        */
/*  child_e, has been opened.  Update eisc's closed state to reflect this.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  As part of closing eisc, eisc's child_index'th child, child_e, has       */
/*  been opened.  Update eisc's closed state to reflect this.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombSetClosedValue(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eisc, set its value suitably for its closed state.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombSetClosedValue(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR KheDrsExprOptionalCopy(KHE_DRS_EXPR e,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  If e is a BUSY_TIME, FREE_TIME, WORK_TIME, BUSY_DAY, FREE_DAY, or        */
/*  WORK_DAY expression, return a fresh copy of it.  Otherwise return e.     */
/*                                                                           */
/*****************************************************************************/

/* *** might need this later
static KHE_DRS_EXPR KheDrsExprOptionalCopy(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprBusyTimeCopy((KHE_DRS_EXPR_BUSY_TIME) e,
	drs);

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprFreeTimeCopy((KHE_DRS_EXPR_FREE_TIME) e,
	drs);

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprWorkTimeCopy((KHE_DRS_EXPR_WORK_TIME) e,
	drs);

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprBusyDayCopy((KHE_DRS_EXPR_BUSY_DAY) e,
	drs);

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprFreeDayCopy((KHE_DRS_EXPR_FREE_DAY) e,
	drs);

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprWorkDayCopy((KHE_DRS_EXPR_WORK_DAY) e,
	drs);

    default:

      ** other expression types are not shared, just return e **
      return e;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombEvalDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,         */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc on day open_day_index.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombEvalDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombInternalDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,   */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process active expression eisc on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombInternalDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, dev, start_index, stop_index;
  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ** initialize eisc->value depending on whether day_index is its first day **
  if( day_index == eisc->day_range.first )
    eisc->u.int_val = eisc->history_before;
  else
  {
    ** this is not eisc's first day, so we need to retrieve a previous value **
    si = HaArray(eisc->sig_indexes, day_index-1 - eisc->day_range.first);
    eisc->u.int_val = HaArray(prev_soln->signature, si);
  }

  ** accumulate the values of the children of eisc that finalized today **
  start_index = HaArray(eisc->open_child_indexes,
    day_index - eisc->day_range.first);
  stop_index =  HaArray(eisc->open_child_indexes,
    day_index+1-eisc->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    if( child_e->u.int_val == 1 )
    {
      ** a sequence is beginning or continuing **
      eisc->u.int_val++;

      ** adjust in various special cases **
      switch( eisc->comb_type )
      {
	case KHE_DRS_COMB_ORDINARY:

	  ** no adjustment here **
	  break;

	case KHE_DRS_COMB_NO_MAX:

	  ** no maximum limit; store minimum limit at most **
	  if( eisc->u.int_val > eisc->min_limit )
	    eisc->u.int_val = eisc->min_limit;
	  break;

	case KHE_DRS_COMB_LINEAR:

	  ** linear cost; if over maximum, report excess immediately **
	  if( eisc->u.int_val > eisc->max_limit )
	  {
	    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0,
	      eisc->u.int_val);
	    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	    if( cost > 0 )
	    {
	      if( drs->solve_rerun )
		fprintf(stderr, "  %s reporting linear cost %.5f\n",
		  KheMonitorId(eisc->monitor), KheCostShow(cost));
	      next_soln->cost += cost;
	    }
	    eisc->u.int_val = eisc->max_limit;
	  }
	  break;

	case KHE_DRS_COMB_STEP:

	  ** step cost; store maximum plus one at most **
	  if( eisc->u.int_val > eisc->max_limit + 1 )
	    eisc->u.int_val = eisc->max_limit + 1;
	  break;

	default:

	  HnAbort("KheDrsExprIntSeqCombDoDay internal error");
	  break;
      }
    }
    else if( eisc->u.int_val > 0 )
    {
      ** a non-empty sequence not at the end of the cycle has just ended **
      dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0, eisc->u.int_val);
      if( dev > 0 )
      {
	cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
      }
      eisc->u.int_val = 0;
    }
  }

  if( day_index < eisc->day_range.last )
  {
    ** this not eisc's last day; store value in next_soln **
    si = HaArray(eisc->sig_indexes, day_index - eisc->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprIntSeqInternalDoDay internal error 3");
    HaArrayAddLast(next_soln->signature, eisc->u.int_val);
  }
  else if( eisc->u.int_val > 0 )
  {
    ** this is the last day and we have a sequence to report **
    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit,
      eisc->history_after, eisc->u.int_val);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->solve_rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_soln->cost += cost;
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,           */
/*    int day_index,                                                         */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eisc on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, dev, start_index, stop_index;
  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ** initialize eisc->value depending on whether day_index is its first day **
  if( day_index == eisc->day_range.first )
    eisc->u.int_val = eisc->history_before;
  else
  {
    ** this is not eisc's first day, so we need to retrieve a previous value **
    si = HaArray(eisc->sig_indexes, day_index-1 - eisc->day_range.first);
    eisc->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  ** accumulate the values of the children of eisc that finalized today **
  start_index = HaArray(eisc->open_child_indexes,
    day_index - eisc->day_range.first);
  stop_index =  HaArray(eisc->open_child_indexes,
    day_index+1-eisc->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    if( child_e->u.int_val == 1 )
    {
      ** a sequence is beginning or continuing **
      eisc->u.int_val++;

      ** adjust in various special cases **
      switch( eisc->comb_type )
      {
	case KHE_DRS_COMB_ORDINARY:

	  ** no adjustment here **
	  break;

	case KHE_DRS_COMB_NO_MAX:

	  ** no maximum limit; store minimum limit at most **
	  if( eisc->u.int_val > eisc->min_limit )
	    eisc->u.int_val = eisc->min_limit;
	  break;

	case KHE_DRS_COMB_LINEAR:

	  ** linear cost; if over maximum, report excess immediately **
	  if( eisc->u.int_val > eisc->max_limit )
	  {
	    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0,
	      eisc->u.int_val);
	    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	    if( cost > 0 )
	    {
	      if( drs->solve_rerun )
		fprintf(stderr, "  %s reporting linear cost %.5f\n",
		  KheMonitorId(eisc->monitor), KheCostShow(cost));
	      next_soln->cost += cost;
	    }
	    eisc->u.int_val = eisc->max_limit;
	  }
	  break;

	case KHE_DRS_COMB_STEP:

	  ** step cost; store maximum plus one at most **
	  if( eisc->u.int_val > eisc->max_limit + 1 )
	    eisc->u.int_val = eisc->max_limit + 1;
	  break;

	default:

	  HnAbort("KheDrsExprIntSeqCombDoDay internal error");
	  break;
      }
    }
    else if( eisc->u.int_val > 0 )
    {
      ** a non-empty sequence not at the end of the cycle has just ended **
      dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0, eisc->u.int_val);
      if( dev > 0 )
      {
	cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
      }
      eisc->u.int_val = 0;
    }
  }

  if( day_index < eisc->day_range.last )
  {
    ** this not eisc's last day; store value in next_soln **
    si = HaArray(eisc->sig_indexes, day_index - eisc->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprIntSeqDoDay internal error 3");
    HaArrayAddLast(next_soln->signature, eisc->u.int_val);
  }
  else if( eisc->u.int_val > 0 )
  {
    ** this is the last day and we have a sequence to report **
    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit,
      eisc->history_after, eisc->u.int_val);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->solve_rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_soln->cost += cost;
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDoDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,         */
/*    KHE_DRS_SOLN soln, FILE *fp)                                           */
/*                                                                           */
/*  Unindented debug print of that part of eisc which is specific to it.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombDoDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "INT_SEQ_COMB(%s, %s)", KheMonitorId(eisc->monitor),
    KheDrsCombTypeShow(eisc->comb_type));
  KheDrsExprDebugSignature((KHE_DRS_EXPR) eisc, soln, false, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of eisc in soln onto fp.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff2[100];  int si;
  if( soln != NULL )
  {
    ** unindented version with a soln value **
    fprintf(fp, "INT_SEQ_COMB %s", KheMonitorId(eisc->monitor));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eisc, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    ** indented version without a soln value **
    fprintf(fp, "%*s[ INT_SEQ_COMB %s %s\n", indent, "",
      KheMonitorId(eisc->monitor), KheDrsDomTestShow(eisc->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    ** unindented version without a soln value **
    fprintf(fp, "[ INT_SEQ_COMB %s %s (%d children) ]",
      KheMonitorId(eisc->monitor), KheDrsDomTestShow(eisc->dom_test, buff2),
      HaArrayCount(eisc->children));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make child a child of parent.                                            */
/*                                                                           */
/*  Implementation note.  During expression construction, all expressions    */
/*  are closed by the time they are made children of their parents (by       */
/*  calls on this function).  Hence KheDrsExprChildHasClosed below.          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;
  HnAssert(HaArrayCount(parent->parents) == 0,
    "KheDrsExprAddChild internal error:  too late to add child to parent");
  prnt.expr = parent;
  prnt.index = HaArrayCount(parent->children);
  HaArrayAddLast(parent->children, child);
  HaArrayAddLast(child->parents, prnt);
  KheDrsExprChildHasClosed(parent, child,
    HaArrayCount(parent->children) - 1, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprCmp(const void *t1, const void *t2)                        */
/*                                                                           */
/*  Comparison function for sorting an array of expressions by increasing    */
/*  postorder index.                                                         */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprCmp(const void *t1, const void *t2)
{
  KHE_DRS_EXPR e1 = * (KHE_DRS_EXPR *) t1;
  KHE_DRS_EXPR e2 = * (KHE_DRS_EXPR *) t2;
  return e1->postorder_index - e2->postorder_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Gather e into drs->open_exprs, the list of expressions that are to be    */
/*  opened.  This does not actually open e; that comes later.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  if( !e->gathered )
  {
    e->gathered = true;
    HaArrayAddLast(drs->open_exprs, e);
    HaArrayForEach(e->parents, prnt, i)
      KheDrsExprGatherForOpening(prnt.expr, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,      */
/*    int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  As part of opening e for solving, inform e that its child_index'th       */
/*  child, child_e, has opened.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,
  int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_EXPR e2;

  /* add child_e's last solve day to open_day_range */
  KheDrsDayRangeAdd(&e->open_day_range, child_e->open_day_range.last);

  /* add child_e to e->open_children in last day order */
  HaArrayAddLast(e->open_children, NULL);
  for( i = HaArrayCount(e->open_children) - 2;  i >= 0;  i-- )
  {
    e2 = HaArray(e->open_children, i);
    if( e2->open_day_range.last <= child_e->open_day_range.last )
      break;
    HaArrayPut(e->open_children, i + 1, e2);
  }
  HaArrayPut(e->open_children, i + 1, child_e);

  /* delete child_e's value from the closed state of e */
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasOpened((KHE_DRS_EXPR_OR) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasOpened((KHE_DRS_EXPR_AND) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasOpened((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasOpened((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasOpened((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasOpened((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostChildHasOpened((KHE_DRS_EXPR_COST) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombChildHasOpened((KHE_DRS_EXPR_INT_SUM_COMB) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombChildHasOpened((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	child_e, child_index, drs);
      break;

    default:

      HnAbort("KheDrsExprChildHasOpened internal error (tag %d)", e->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Open e for solving.                                                      */
/*                                                                           */
/*  By the time this is called, all of e's open children will have           */
/*  called KheDrsExprChildHasOpened.                                         */
/*                                                                           */
/*  The value of e->open_child_indexes[i - e->open_day_range.first] is the   */
/*  index in e->open_children of the first child whose last open day is i:   */
/*                                                                           */
/*    start = HaArray(e->open_child_indexes, i - e->open_day_range.first);   */
/*    stop =  HaArray(e->open_child_indexes, i+1 - e->open_day_range.first); */
/*    for( j = start;  j < stop;  j++ )                                      */
/*      child_e = HaArray(e->open_children, j);                              */
/*                                                                           */
/*  visits all children whose last open day is i.                            */
/*                                                                           */
/*  Implementation note - how e->open_day_range is set                       */
/*  --------------------------------------------------                       */
/*  In external expressions, e->open_day_range receives its final value      */
/*  when KheDrsExprGatherForOpening is called.  In internal expressions,     */
/*  its value is initially empty and is updated by each call to              */
/*  KheDrsExprChildHasOpened until it reaches its final value.               */
/*                                                                           */
/*****************************************************************************/
static int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test);
static void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e);

static void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i, curr_last_day_index;
  KHE_DRS_DAY day;  KHE_DRS_EXPR child_e;

  /* inform e's parents that e is open; or drs->init_open_cost if no parent */
  e->gathered = false;
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasOpened(prnt.expr, e, prnt.index, drs);
  /* *** not sure about this, it is c' but not c(e, x0)
  if( HaArrayCount(e->parents) > 0 )
    HaArrayForEach(e->parents, prnt, i)
      KheDrsExprChildHasOpened(prnt.expr, e, prnt.index);
  else
    drs->total_closed_cost -= e->u.cost_val;
  *** */

  /* nothing more to do if e is an external expression */
  if( e->tag <= KHE_DRS_EXPR_ASSIGNED_TASK_TAG )
    return;

  /* build open_child_indexes array into e's open children (already sorted) */
  HaArrayClear(e->open_child_indexes);
  curr_last_day_index = e->open_day_range.first - 1;
  HaArrayForEach(e->open_children, child_e, i)
  {
    while( child_e->open_day_range.last > curr_last_day_index )
    {
      HaArrayAddLast(e->open_child_indexes, i);
      curr_last_day_index++;
    }
  }
  HaArrayAddLast(e->open_child_indexes, i);

  /* inform e's days that e is open */
  HaArrayClear(e->sig_indexes);
  for( i = e->open_day_range.first;  i <= e->open_day_range.last;  i++ )
  {
    /* add e to day's list of open internal expressions */
    day = HaArray(drs->open_days, i);
    KheDrsDayAddOpenExpr(day, e);

    /* signatures and dom tests (needed on each day except the last) */
    if( i < e->open_day_range.last )
      HaArrayAddLast(e->sig_indexes, KheDrsDayAddDomTest(day, e->dom_test));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,                            */
/*    KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Inform internal expression e that its child_index's child, child_e,      */
/*  has closed.  The child will have its correct closed_state and value      */
/*  fields when this is called.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int pos;

  /* remove child_e from e->open_children */
  if( !HaArrayContains(e->open_children, child_e, &pos) )
    HnAbort("KheDrsExprChildHasClosed internal error");
  HaArrayDeleteAndPlug(e->open_children, pos);

  /* do the type-specific part of this operation */
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasClosed((KHE_DRS_EXPR_OR) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasClosed((KHE_DRS_EXPR_AND) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasClosed((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasClosed((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasClosed((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasClosed((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostChildHasClosed((KHE_DRS_EXPR_COST) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombChildHasClosed((KHE_DRS_EXPR_INT_SUM_COMB) e,
	child_e, child_index, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombChildHasClosed((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	child_e, child_index, drs);
      break;

    default:

      HnAbort("KheDrsExprChildHasClosed internal error (tag %d)", e->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing e, set its value suitably for its closed state.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeSetClosedValue((KHE_DRS_EXPR_BUSY_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeSetClosedValue((KHE_DRS_EXPR_FREE_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeSetClosedValue((KHE_DRS_EXPR_WORK_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDaySetClosedValue((KHE_DRS_EXPR_BUSY_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDaySetClosedValue((KHE_DRS_EXPR_FREE_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDaySetClosedValue((KHE_DRS_EXPR_WORK_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskSetClosedValue((KHE_DRS_EXPR_ASSIGNED_TASK) e, drs);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrSetClosedValue((KHE_DRS_EXPR_OR) e, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndSetClosedValue((KHE_DRS_EXPR_AND) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumSetClosedValue((KHE_DRS_EXPR_INT_SUM) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumSetClosedValue((KHE_DRS_EXPR_FLOAT_SUM) e,drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevSetClosedValue((KHE_DRS_EXPR_INT_DEV) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevSetClosedValue((KHE_DRS_EXPR_FLOAT_DEV) e,drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostSetClosedValue((KHE_DRS_EXPR_COST) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombSetClosedValue((KHE_DRS_EXPR_INT_SUM_COMB) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombSetClosedValue((KHE_DRS_EXPR_INT_SEQ_COMB) e, drs);
      break;

    default:

      HnAbort("KheDrsExprSetClosedValue internal error (tag %d)", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Close e, which may be internal or external.  Whatever it needs to work   */
/*  out its closed value (an assignment, closed children, or closed state)   */
/*  is up to date.  At the end, inform e's parents that e has closed.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;

  /* set e's closed value */
  KheDrsExprSetClosedValue(e, drs);

  /* clear fields that are meaningful only when e is open */
  e->open_day_range = KheDrsDayRangeMake(1, 0);
  HaArrayClear(e->open_children);
  HaArrayClear(e->open_child_indexes);
  HaArrayClear(e->sig_indexes);

  /* inform e's parents that e has closed */
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasClosed(prnt.expr, e, prnt.index, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,          */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Inform leaf expression e that dtd has been assigned drd.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeLeafSet((KHE_DRS_EXPR_BUSY_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeLeafSet((KHE_DRS_EXPR_FREE_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeLeafSet((KHE_DRS_EXPR_WORK_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayLeafSet((KHE_DRS_EXPR_BUSY_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayLeafSet((KHE_DRS_EXPR_FREE_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayLeafSet((KHE_DRS_EXPR_WORK_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskLeafSet((KHE_DRS_EXPR_ASSIGNED_TASK) e, dtd, drd);
      break;

    default:

      HnAbort("KheDrsExprLeafSet internal error: tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafClear(KHE_DRS_EXPR e)                                 */
/*                                                                           */
/*  Inform leaf expression e that the assignment of drd to dtd carried out   */
/*  by KheDrsExprLeafSet above has been removed.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafClear(KHE_DRS_EXPR e)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeLeafClear((KHE_DRS_EXPR_BUSY_TIME) e);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeLeafClear((KHE_DRS_EXPR_FREE_TIME) e);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeLeafClear((KHE_DRS_EXPR_WORK_TIME) e);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayLeafClear((KHE_DRS_EXPR_BUSY_DAY) e);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayLeafClear((KHE_DRS_EXPR_FREE_DAY) e);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayLeafClear((KHE_DRS_EXPR_WORK_DAY) e);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskLeafClear((KHE_DRS_EXPR_ASSIGNED_TASK) e);
      break;

    default:

      HnAbort("KheDrsExprLeafClear internal error: tag %d");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEvalDay(KHE_DRS_EXPR e,                                   */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression e on day open_day_index.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprEvalDay(KHE_DRS_EXPR e, int open_day_index,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      HnAbort("KheDrsExprEvalDay: external tag (%d)", e->tag);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrEvalDay((KHE_DRS_EXPR_OR) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndEvalDay((KHE_DRS_EXPR_AND) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumEvalDay((KHE_DRS_EXPR_INT_SUM) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    /* *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqEvalDay((KHE_DRS_EXPR_INT_SEQ) e, open_day_index,
	prev_soln, next_soln, drs);
      break;
    *** */

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumEvalDay((KHE_DRS_EXPR_FLOAT_SUM) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevEvalDay((KHE_DRS_EXPR_INT_DEV) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevEvalDay((KHE_DRS_EXPR_FLOAT_DEV) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostEvalDay((KHE_DRS_EXPR_COST) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombEvalDay((KHE_DRS_EXPR_INT_SUM_COMB) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombEvalDay((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    default:

      HnAbort("KheDrsExprEvalDay internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprFirstDebugDayIndex(KHE_DRS_EXPR e)                         */
/*                                                                           */
/*  The minimum, over e and its descendants, of the first_day_index field    */
/*  of e.                                                                    */
/*                                                                           */
/*  Implementation note.  We only search down the first children.  If that   */
/*  is wrong in some obscure case, it doesn't matter, because this function  */
/*  is called only when debugging.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheDrsExprFirstDebugDayIndex(KHE_DRS_EXPR e)
{
  int res;
  res = e->day_range.first;
  while( HaArrayCount(e->children) > 0 )
  {
    e = HaArrayFirst(e->children);
    if( e->day_range.first < res )
      res = e->day_range.first;
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Carry out the part of KheDrsExprDebug which prints the children.         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  ** KHE_DRS_EXPR child_e;  int i, di, first_day_index; **
  ** bool child_printed; **
  if( soln == NULL )
  {
    ** print all children **
    ** *** needs revision here still to do
    di = e->day_range.first;
    HaArrayForEach(e->children, child_e, i)
    {
      for(; HaArray(e->open_child_indexes, di - e->day_range.first) <= i; di++ )
	fprintf(fp, "%*sending on %s:\n", indent + 2, "",
	  KheTimeGroupId(KheFrameTimeGroup(drs->days_frame, di)));
      KheDrsExprDebug(child_e, soln, drs, verbosity, indent + 2, fp);
    }
    *** **
  }
  else
  {
    ** print the children with a descendant with an entry in the signature **
    if( HaArrayCount(e->children) > 0 )
    {
      fprintf(fp, "(");
      ** *** needs revision here stil l to do
      child_printed = false;
      HaArrayForEach(e->children, child_e, i)
      {
	first_day_index = KheDrsExprFirstDebugDayIndex(child_e);
	if( first_day_index <= soln->day->solve_index &&
	    soln->day->solve_index <= child_e->last_day_index )
	{
	  if( child_printed )
	    fprintf(fp, ", ");
	  KheDrsExprDebug(child_e, soln, drs, verbosity, -1, fp);
	  child_printed = true;
	}
      }
      *** **
      fprintf(fp, ")");
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp)      */
/*                                                                           */
/*  Unindented debug print of that part of e which is specific to it.        */
/*  If soln != NULL, also print e's signature (if any) in soln.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDoDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln, FILE *fp)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDoDebug((KHE_DRS_EXPR_BUSY_TIME) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDoDebug((KHE_DRS_EXPR_FREE_TIME) e, soln, fp);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDoDebug((KHE_DRS_EXPR_WORK_TIME) e, soln, fp);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDoDebug((KHE_DRS_EXPR_BUSY_DAY) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDoDebug((KHE_DRS_EXPR_FREE_DAY) e, soln, fp);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDoDebug((KHE_DRS_EXPR_WORK_DAY) e, soln, fp);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDoDebug((KHE_DRS_EXPR_ASSIGNED_TASK) e, soln, fp);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDoDebug((KHE_DRS_EXPR_OR) e, soln, fp);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDoDebug((KHE_DRS_EXPR_AND) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDoDebug((KHE_DRS_EXPR_INT_SUM) e, soln, fp);
      break;

    /* *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqDoDebug((KHE_DRS_EXPR_INT_SEQ) e, soln, fp);
      break;
    *** */

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDoDebug((KHE_DRS_EXPR_FLOAT_SUM) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDoDebug((KHE_DRS_EXPR_INT_DEV) e, soln, fp);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDoDebug((KHE_DRS_EXPR_FLOAT_DEV) e, soln, fp);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostDoDebug((KHE_DRS_EXPR_COST) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombDoDebug((KHE_DRS_EXPR_INT_SUM_COMB) e, soln, fp);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombDoDebug((KHE_DRS_EXPR_INT_SEQ_COMB) e, soln, fp);
      break;

    default:

      HnAbort("KheDrsExprDoDebug internal error: unknown tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebug(KHE_DRS_EXPR e,                                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebug(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  char buff2[100];  KHE_DRS_EXPR child_e;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprDoDebug(e, NULL, fp);
    fprintf(fp, " pi %d oc %d dt %s\n", e->postorder_index,
    HaArrayCount(e->open_children), KheDrsDomTestShow(e->dom_test, buff2));
    HaArrayForEach(e->children, child_e, i)
      KheDrsExprDebug(child_e, drs, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    KheDrsExprDoDebug(e, NULL, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugSoln(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.  Here      */
/*  e is known to be active in soln.                                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebugSoln(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  KheDrsExprDoDebug(e, soln, fp);
  if( indent > 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.            */
/*                                                                           */
/*  If soln != NULL, only those parts of e that have a value stored in       */
/*  soln's signature are printed, along with those values.  In that case     */
/*  indent is ignored and the whole print is on one line.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void OldKheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDebug((KHE_DRS_EXPR_BUSY_TIME) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDebug((KHE_DRS_EXPR_FREE_TIME) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDebug((KHE_DRS_EXPR_WORK_TIME) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDebug((KHE_DRS_EXPR_BUSY_DAY) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDebug((KHE_DRS_EXPR_FREE_DAY) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDebug((KHE_DRS_EXPR_WORK_DAY) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDebug((KHE_DRS_EXPR_ASSIGNED_TASK) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDebug((KHE_DRS_EXPR_OR) e, soln, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDebug((KHE_DRS_EXPR_AND) e, soln, drs, verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDebug((KHE_DRS_EXPR_INT_SUM) e, soln, drs,
	verbosity, indent, fp);
      break;

    ** *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqDebug((KHE_DRS_EXPR_INT_SEQ) e, soln, drs, verbosity,
	indent, fp);
      break;
    *** **

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDebug((KHE_DRS_EXPR_FLOAT_SUM) e, soln, drs,
	verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDebug((KHE_DRS_EXPR_INT_DEV) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDebug((KHE_DRS_EXPR_FLOAT_DEV) e, soln, drs,
	verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostDebug((KHE_DRS_EXPR_COST) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombDebug((KHE_DRS_EXPR_INT_SUM_COMB) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombDebug((KHE_DRS_EXPR_INT_SEQ_COMB) e, soln, drs,
	verbosity, indent, fp);
      break;

    default:

      HnAbort("KheDrsExprDebug internal error: unknown tag %d", e->tag);
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnMake(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,     */
/*    KHE_COST cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Make a new soln with these attributes.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnMake(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,
  KHE_COST cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;

  /* get the basic object */
  if( HaArrayCount(drs->free_solns) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_solns);
    HaArrayClear(res->prev_tasks);
    HaArrayClear(res->signature);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->prev_tasks, drs->arena);
    HaArrayInit(res->signature, drs->arena);
  }
  
  /* initialize the fields */
  res->prev_soln = prev_soln;
  res->day = day;
  res->best_of = 1;
  res->cost = cost;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnFree(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Free soln.                                                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnFree(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_solns, soln);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnEvaluate(KHE_DRS_SOLN soln,                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Copy soln and evaluate the copy.  Here soln has a non-NULL prev_soln.    */
/*                                                                           */
/*****************************************************************************/
static void KheDrsDayAddSoln(KHE_DRS_DAY day, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnEvaluate(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;  KHE_DRS_EXPR e;  int i;  KHE_DRS_DAY day;

  /* quit early if over solve_soln_limit */
  if( --drs->solve_soln_limit < 0 )
    return;

  /* make res, a copy of soln */
  HnAssert(soln->prev_soln != NULL, "KheDrsSolnEvaluate internal error");
  if( DEBUG4 )
    fprintf(stderr, "[ KheDrsSolnMake(soln, drs)\n");
  res = KheDrsSolnMake(soln->prev_soln, soln->day, soln->prev_soln->cost, drs);
  HaArrayAppend(res->prev_tasks, soln->prev_tasks, i);

  /* set signature */
  day = soln->day;
  HaArrayForEach(day->internal_today, e, i)
    KheDrsExprEvalDay(e, day->open_day_index, soln->prev_soln, res, drs);

  /* check cost limit and either add to day or free soln */
  if( res->cost < drs->solve_init_cost )
    KheDrsDayAddSoln(res->day, res, drs);
  else
    KheDrsSolnFree(res, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnPushAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,      */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Push the assignment of drd to dtd onto soln, and update leaf exprs.      */
/*  Here drd must be non-NULL, but dtd could be NULL, in which case no leaf  */
/*  expr updates are required, because the default state is correct then.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnPushAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  /* add the task to the edge */
  HaArrayAddLast(soln->prev_tasks, dtd);

  if( dtd != NULL )
  {
    /* update leaf expressions of the resource on day */
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd);

    /* update leaf expressions of the task on day */
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnPopAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,       */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Undo the effect of the corresponding KheDrsSolnPushAsst.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnPopAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  /* remove the task from the edge */
  HaArrayDeleteLast(soln->prev_tasks);

  if( dtd != NULL )
  {
    /* update leaf expressions of the resource */
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafClear(e);

    /* update leaf expressions of the task */
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafClear(e);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignatureFullHash(KHE_DRS_SOLN soln)                       */
/*                                                                           */
/*  Hash function for hashing the full signature of soln.                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignatureFullHash(KHE_DRS_SOLN soln)
{
  int res, val, i;
  res = 0;
  HaArrayForEach(soln->signature, val, i)
    res = (res + val) * 3;
  return (res < 0 ? - res : res);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignatureFullHashUntyped(void *p)                          */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignatureHash.                              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignatureFullHashUntyped(void *p)
{
  return KheDrsSolnSignatureFullHash((KHE_DRS_SOLN) p);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignaturePartialHash(KHE_DRS_SOLN soln)                    */
/*                                                                           */
/*  Hash function for hashing the partial signature of soln, that is, only   */
/*  the parts of it whose dominance tests are equality (or loose equality).  */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignaturePartialHash(KHE_DRS_SOLN soln)
{
  int res, index, val, i;  KHE_DRS_DAY day;
  res = 0;
  day = soln->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
  {
    val = HaArray(soln->signature, index);
    res = (res + val) * 3;
  }
  if( res < 0 )
    res = - res;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignatureFullEqual(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)*/
/*                                                                           */
/*  Return true if the signatures of soln1 and soln2 are equal.  The lengths */
/*  of their signatures must be equal.                                       */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignatureFullEqual(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)
{
  int i;
  HnAssert(HaArrayCount(soln1->signature) == HaArrayCount(soln2->signature),
    "KheDrsSolnSignaturesEqual: signatures have different lengths");
  for( i = 0;  i < HaArrayCount(soln1->signature);  i++ )
    if( HaArray(soln1->signature, i) != HaArray(soln2->signature, i) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignatureFullEqualUntyped(void *p1, void *p2)             */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignaturesEqual.                            */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignatureFullEqualUntyped(void *p1, void *p2)
{
  return KheDrsSolnSignatureFullEqual((KHE_DRS_SOLN) p1,
    (KHE_DRS_SOLN) p2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignaturePartialEqual(KHE_DRS_SOLN soln1,                 */
/*    KHE_DRS_SOLN soln2)                                                    */
/*                                                                           */
/*  Return true if the signatures of soln1 and soln2 are partially equal,    */
/*  that is, the entries whose dominance tests are equality (or loose        */
/*  equality) are equal.                                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignaturePartialEqual(KHE_DRS_SOLN soln1,
  KHE_DRS_SOLN soln2)
{
  int i, index;  KHE_DRS_DAY day;
  HnAssert(HaArrayCount(soln1->signature) == HaArrayCount(soln2->signature),
   "KheDrsSolnSignaturePartialEqual: signatures have different lengths");
  day = soln1->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
    if( HaArray(soln1->signature, index) != HaArray(soln2->signature, index) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSignatureDebug(KHE_DRS_SOLN soln, FILE *fp)               */
/*                                                                           */
/*  Debug print of the signature of soln onto fp (no verbosity or indent).   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSignatureDebug(KHE_DRS_SOLN soln, FILE *fp)
{
  int i, val;
  fprintf(fp, "[");
  HaArrayForEach(soln->signature, val, i)
  {
    if( i > 0 )
      fprintf(fp, " ");
    fprintf(fp, "%d", val);
  }
  fprintf(fp, "]");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSignatureDebugUntyped(void *p, FILE *fp)                  */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignatureDebug.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSignatureDebugUntyped(void *p, FILE *fp)
{
  KheDrsSolnSignatureDebug((KHE_DRS_SOLN) p, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2,         */
/*    int start_depth)                                                       */
/*                                                                           */
/*  Return true if soln1 dominates soln2.  They come from the same day.      */
/*  Parameter start_depth says how far along the signature to start.         */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2,
  int start_depth)
{
  KHE_DRS_DOM_TEST dt;  int i, val1, val2, sig_len;  KHE_DRS_DAY day;
  HnAssert(soln1->day == soln2->day,
    "KheDrsSolnDominates internal error: soln1->day = %s, soln2->day = %s",
    soln1->day != NULL ? KheTimeGroupId(soln1->day->time_group) : "-",
    soln2->day != NULL ? KheTimeGroupId(soln2->day->time_group) : "-");
  if( soln2->cost < soln1->cost )
    return false;
  day = soln1->day;
  sig_len = HaArrayCount(soln1->signature);
  for( i = start_depth;  i < sig_len;  i++ )
  {
    dt = HaArray(day->dom_tests, i);
    val1 = HaArray(soln1->signature, i);
    val2 = HaArray(soln2->signature, i);
    if( !KheDrsDomTestDominates(dt, val1, val2) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnOverWrite(KHE_DRS_SOLN dest_soln, KHE_DRS_SOLN src_soln)  */
/*                                                                           */
/*  Overwrite src_soln onto dest_soln.  Their signatures are known to be     */
/*  equal, so those don't need to be touched.  Also the best_of field        */
/*  is not touched here; it gets updated separately.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnOverWrite(KHE_DRS_SOLN dest_soln, KHE_DRS_SOLN src_soln)
{
  int i;
  HaArrayClear(dest_soln->prev_tasks);
  HaArrayAppend(dest_soln->prev_tasks, src_soln->prev_tasks, i);
  dest_soln->prev_soln = src_soln->prev_soln;
  dest_soln->cost = src_soln->cost;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,                     */
/*    KHE_DRS_RESOURCE dr, int dr_solve_index, KHE_DRS_TASK_ON_DAY *dtd)     */
/*                                                                           */
/*  If dr is assigned in soln, return true and set *dtd to what it is        */
/*  assigned to; otherwise return false and set *dtd to NULL.                */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,
  KHE_DRS_RESOURCE dr, int dr_solve_index, KHE_DRS_TASK_ON_DAY *dtd)
{
  if( soln == NULL )
    return *dtd = NULL, false;
  *dtd = HaArray(soln->prev_tasks, dr_solve_index);
  return *dtd != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDoExtend(KHE_DRS_SOLN soln,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend soln from soln->prev_soln to soln->day; build soln->prev_tasks    */
/*  up to full length first.                                                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDoExtend(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE dr;  int i, dr_index;  KHE_DRS_TASK dt;
  KHE_DRS_TASK_CLASS dtc;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_TASK_ON_DAY dtd;  /* KHE_DRS_SOLN soln2; */
  if( HaArrayCount(soln->prev_tasks) >= HaArrayCount(drs->open_resources) )
  {
    /* soln is ready to use; add it to soln->day */
    KheDrsSolnEvaluate(soln, drs);
    drs->solve_diversifier++;
  }
  else
  {
    dr_index = HaArrayCount(soln->prev_tasks);
    dr = HaArray(drs->open_resources, dr_index);
    drd = KheDrsResourceOnDay(dr, soln->day);
    if( KheDrsResourceOnDayIsFixed(drd, dr_index, soln->prev_soln, &dtd) )
    {
      /* try dtd only */
      KheDrsSolnPushAsst(soln, dtd, drd);
      KheDrsSolnDoExtend(soln, drs);
      KheDrsSolnPopAsst(soln, dtd, drd);
    }
    else
    {
      /* try a task from any class that will accept dtd */
      HaArrayForEach(soln->day->task_classes, dtc, i)
	if( KheDrsTaskClassAcceptResourceBegin(dtc, dr, &dt) )
	{
	  if( !KheDrsTaskBusyOnDay(dt, soln->day, &dtd) )
	    HnAbort("KheDrsSolnDoExtend internal error");
	  KheDrsSolnPushAsst(soln, dtd, drd);
	  KheDrsSolnDoExtend(soln, drs);
	  KheDrsSolnPopAsst(soln, dtd, drd);
	  KheDrsTaskClassAcceptResourceEnd(dtc);
	}

      /* try a free day */
      KheDrsSolnPushAsst(soln, NULL, drd);
      KheDrsSolnDoExtend(soln, drs);
      KheDrsSolnPopAsst(soln, NULL, drd);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnExtend(KHE_DRS_SOLN soln, KHE_DRS_DAY new_day,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend the dynamic programming search by searching out of soln.  Each    */
/*  new task starts on new_day.                                              */
/*                                                                           */
/*  Implementation note.  The call to KheDrsEdgeFree is safe here, because   */
/*  edge de does not go into the graph structure; only copies of it do.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;
  soln = KheDrsSolnMake(prev_soln, new_day, 0 /* cost not used here */, drs);
  KheDrsSolnDoExtend(soln, drs);
  KheDrsSolnFree(soln, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebugBasics(KHE_DRS_SOLN soln, FILE *fp)                  */
/*                                                                           */
/*  Produce an unindented one-line debug print of the basics of soln.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebugBasics(KHE_DRS_SOLN soln, FILE *fp)
{
  fprintf(fp, "SOLN(%s, %d, %.5f)",
    soln->day == NULL ? "-" : KheTimeGroupId(soln->day->time_group),
    soln->best_of, KheCostShow(soln->cost));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of soln onto fp with the given verbosity and indent.         */
/*                                                                           */
/*  NB If verbosity == 1 or indent < 0, it is safe to pass NULL for drs.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_RESOURCE dr;  int i, count;  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR e;
  if( indent < 0 )
    KheDrsSolnDebugBasics(soln, fp);
  else
  {
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsSolnDebugBasics(soln, fp);
    fprintf(fp, "\n");
    count = HaArrayCount(soln->prev_tasks);
    if( verbosity >= 2 && count > 0 )
    {
      HnAssert(count == HaArrayCount(drs->open_resources),
	"KheDrsSolnDebug internal error");
      HaArrayForEach(drs->open_resources, dr, i)
      {
	fprintf(fp, "%*s  %s: ", indent, "", KheResourceId(dr->resource));
	dtd = HaArray(soln->prev_tasks, i);
	if( dtd != NULL )
	  KheDrsTaskOnDayDebug(dtd, 1, 0, fp);
	else
	  fprintf(fp, "-\n");
      }
      if( verbosity >= 3 )
      {
	HaArrayForEach(soln->day->internal_today, e, i)
	  KheDrsExprDebugSoln(e, soln, drs, verbosity, indent + 2, fp);
      }
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_LIST"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_LIST KheDrsSolnListMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Return a new, empty KHE_DRS_SOLN_LIST object.                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_LIST KheDrsSolnListMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST res;
  if( HaArrayCount(drs->free_soln_lists) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_lists);
    HaArrayClear(res->solns);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->solns, drs->arena);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListFree(KHE_DRS_SOLN_LIST ns,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ns and the solns in it.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListFree(KHE_DRS_SOLN_LIST ns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(ns->solns, soln, i)
    KheDrsSolnFree(soln, drs);
  HaArrayAddLast(drs->free_soln_lists, ns);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnListSignaturePartialHashUntyped(void *p)                   */
/*                                                                           */
/*  Hash function for solnss, being the partial hash of the set's first      */
/*  element.                                                                 */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnListSignaturePartialHashUntyped(void *p)
{
  KHE_DRS_SOLN_LIST ns;  KHE_DRS_SOLN soln;
  ns = (KHE_DRS_SOLN_LIST) p;
  soln = HaArrayFirst(ns->solns);
  return KheDrsSolnSignaturePartialHash(soln);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnListSignaturePartialEqualUntyped(void *p1, void *p2)      */
/*                                                                           */
/*  Equality function for solns, being the partial equality of the sets'     */
/*  first elements.                                                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnListSignaturePartialEqualUntyped(void *p1, void *p2)
{
  KHE_DRS_SOLN_LIST ns1, ns2;  KHE_DRS_SOLN soln1, soln2;
  ns1 = (KHE_DRS_SOLN_LIST) p1;
  ns2 = (KHE_DRS_SOLN_LIST) p2;
  soln1 = HaArrayFirst(ns1->solns);
  soln2 = HaArrayFirst(ns2->solns);
  return KheDrsSolnSignaturePartialEqual(soln1, soln2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListDoStrongDominanceTest(KHE_DRS_SOLN_LIST ns,           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)  */
/*                                                                           */
/*  Carry out a strong dominance test between ns and soln.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListDoStrongDominanceTest(KHE_DRS_SOLN_LIST ns,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)
{
  KHE_DRS_SOLN other_soln;  int i;

  /* if soln is dominated by anything else, delete soln and return */
  HaArrayForEach(ns->solns, other_soln, i)
    if( KheDrsSolnDominates(other_soln, soln, 0) )
    {
      KheDrsSolnFree(soln, drs);
      return;
    }

  /* remove other solns that soln dominates */
  HaArrayForEach(ns->solns, other_soln, i)
    if( KheDrsSolnDominates(soln, other_soln, 0) )
    {
      KheDrsSolnFree(other_soln, drs);
      HaArrayDeleteAndPlug(ns->solns, i);
      (*total_count)--;
      i--;
    }

  /* add soln to ns */
  HaArrayAddLast(ns->solns, soln);
  (*total_count)++;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListSignatureDebugUntyped(void *p, FILE *fp)              */
/*                                                                           */
/*  Debug print of the signature of soln set p.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListSignatureDebugUntyped(void *p, FILE *fp)
{
  KHE_DRS_SOLN_LIST ns;  KHE_DRS_SOLN soln;
  ns = (KHE_DRS_SOLN_LIST) p;
  soln = HaArrayFirst(ns->solns);
  KheDrsSolnSignatureDebug(soln, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListDebug(KHE_DRS_SOLN_LIST dsl,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of dsl onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListDebug(KHE_DRS_SOLN_LIST dsl,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN soln;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ SolnList(%d solns)\n", indent, "",
      HaArrayCount(dsl->solns));
    HaArrayForEach(dsl->solns, soln, i)
    {
      KheDrsSolnDebug(soln, drs, verbosity, indent + 2, fp);
      if( i > 50 )
      {
	fprintf(fp, "%*s... omitting the rest\n", indent + 2, "");
	break;
      }
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "SolnList(%d solns)", HaArrayCount(dsl->solns));

}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_TRIE"                                            */
/*                                                                           */
/*  Abstraction function                                                     */
/*  --------------------                                                     */
/*                                                                           */
/*  A soln_trie represents a set of solns.  If soln_trie == NULL, the set    */
/*  is empty.  Otherwise the set consists of soln_trie->soln (if non-NULL)   */
/*  plus the sets stored in each of the children.                            */
/*                                                                           */
/*  Invariant                                                                */
/*  ---------                                                                */
/*                                                                           */
/*  Exactly one of these three cases applies:                                */
/*                                                                           */
/*    (1)  soln_trie == NULL                                                 */
/*    (2)  soln_trie->soln != NULL                                           */
/*    (3)  HaArrayCount(soln_trie->children) > 0                             */
/*                                                                           */
/*  and in the last case, the last child of soln_trie->children is non-NULL. */
/*  Each soln appears in the trie in a place determined by the sequence      */
/*  of values in its signature, in the usual trie manner.                    */
/*                                                                           */
/*  There is no requirement that two or more solns be present in case (3).   */
/*  If there are no solns, case (1) always applies; if there are two or      */
/*  more solns, case (3) always applies; but if there is exactly one soln,   */
/*  then either of cases (2) or (3) applies.  This is because it would be    */
/*  too slow and tedious to detect situations where a case (3) soln could    */
/*  be simplified to a case (2) soln.                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_TRIE KheDrsSolnTrieMake(KHE_DRS_SOLN soln,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new trie object containing soln.  It satisfies case (2) above.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_TRIE KheDrsSolnTrieMake(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE res;
  HnAssert(soln != NULL, "KheDrsSolnTrieMake internal error (soln == NULL)");
  if( HaArrayCount(drs->free_soln_tries) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_tries);
    HaArrayClear(res->children);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->children, drs->arena);
  }
  res->soln = soln;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieFree(KHE_DRS_SOLN_TRIE *soln_trie,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free *soln_trie and all its descendants and solns.                       */
/*  Set *soln_trie to NULL.                                                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieFree(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;
  if( *soln_trie == NULL )
  {
    /* case (1), do nothing */
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2), delete (*soln_trie)->soln and *soln_trie */
    KheDrsSolnFree((*soln_trie)->soln, drs);
    HaArrayAddLast(drs->free_soln_tries, *soln_trie);
    *soln_trie = NULL;
  }
  else
  {
    /* case (3), delete children and *soln_trie */
    for( i = 0;  i < HaArrayCount((*soln_trie)->children);  i++ )
      KheDrsSolnTrieFree(&HaArray((*soln_trie)->children, i), drs);
    HaArrayAddLast(drs->free_soln_tries, *soln_trie);
    *soln_trie = NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieInsertChild(KHE_DRS_SOLN_TRIE soln_trie,              */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert soln in its appropriate place among the children of soln_trie.    */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnTrieInsert(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnTrieInsertChild(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int index;
  index = HaArray(soln->signature, depth);
  HaArrayFill(soln_trie->children, index + 1, NULL);
  KheDrsSolnTrieInsert(&HaArray(soln_trie->children, index), soln,
    depth + 1, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieInsert(KHE_DRS_SOLN_TRIE *soln_trie,                  */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert soln into *soln_trie, possibly changing *soln_trie as a result.   */
/*  The depth parameter indicates how far along the signature we are.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieInsert(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( *soln_trie == NULL )
  {
    /* case (1), make a new soln containing just soln, -> case (2) */
    *soln_trie = KheDrsSolnTrieMake(soln, drs);
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2), add soln and (*soln_trie)->soln to children, -> case (3) */
    KheDrsSolnTrieInsertChild(*soln_trie, soln, depth, drs);
    soln = (*soln_trie)->soln;
    (*soln_trie)->soln = NULL;
    KheDrsSolnTrieInsertChild(*soln_trie, soln, depth, drs);
  }
  else
  {
    /* case (3), add soln to children, -> case (3) */
    KheDrsSolnTrieInsertChild(*soln_trie, soln, depth, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,                */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Return true if there is a soln in soln_trie that dominates soln.         */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  if( soln_trie == NULL )
  {
    /* case (1); no solns in soln_trie, so none dominate */
    return false;
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); the only possible dominator is soln_trie->soln */
    return KheDrsSolnDominates(soln_trie->soln, soln, depth);
  }
  else
  {
    /* case (3); find range of dominators and try each of them */
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->signature, depth);
    KheDrsDomTestDominatingRange(dom_test, val, &from_val, &to_val);
    if( to_val >= HaArrayCount(soln_trie->children) )
      to_val = HaArrayCount(soln_trie->children) - 1;
    for( i = from_val;  i <= to_val;  i++ )
      if( KheDrsSolnTrieDominates(HaArray(soln_trie->children,i), soln, depth+1, drs))
	return true;
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,         */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,         */
/*    int *total_count)                                                      */
/*                                                                           */
/*  Remove those elements of *soln_trie that are dominated by soln.  This    */
/*  can change *soln_trie.                                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *total_count)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  if( *soln_trie == NULL )
  {
    /* case (1); there is nothing to remove */
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2); if (*soln_trie)->soln is dominated, remove whole trie */
    if( KheDrsSolnDominates(soln, (*soln_trie)->soln, depth) )
    {
      KheDrsSolnTrieFree(soln_trie, drs);
      (*total_count)--;
    }
  }
  else
  {
    /* case (3); removed dominated solns among the children */
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->signature, depth);
    KheDrsDomTestDominatedRange(dom_test, val, &from_val, &to_val);
    if( to_val >= HaArrayCount((*soln_trie)->children) )
      to_val = HaArrayCount((*soln_trie)->children) - 1;
    for( i = from_val;  i <= to_val;  i++ )
      KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, i), soln,
	depth + 1, drs, total_count);

    /* remove trailing NULL children */
    while( HaArrayCount((*soln_trie)->children) > 0 &&
	HaArrayLast((*soln_trie)->children) == NULL )
      HaArrayDeleteLast((*soln_trie)->children);

    /* if children array is empty then the whole trie is empty, so delete it */
    if( HaArrayCount((*soln_trie)->children) == 0 )
      KheDrsSolnTrieFree(soln_trie, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDoStrongDominanceTest(KHE_DRS_SOLN_TRIE *soln_trie,   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int *total_count)                                                      */
/*                                                                           */
/*  Carry out a strong dominance test between *soln_trie and soln.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDoStrongDominanceTest(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)
{
  if( DEBUG12 )
    fprintf(stderr, "[ KheDrsSolnTrieDoStrongDominanceTest(soln for %s)\n",
      KheTimeGroupId(soln->day->time_group));
  if( KheDrsSolnTrieDominates(*soln_trie, soln, 0, drs) )
  {
    /* soln is dominated by some other soln, so delete soln */
    if( DEBUG12 )
      fprintf(stderr, "  trie deleting soln for %s\n",
	KheTimeGroupId(soln->day->time_group));
    KheDrsSolnFree(soln, drs);
  }
  else
  {
    /* remove solns that soln dominates, and insert soln */
    if( DEBUG12 )
      fprintf(stderr, "  trie inserting soln for %s\n",
	KheTimeGroupId(soln->day->time_group));
    KheDrsSolnTrieRemoveDominated(soln_trie, soln, 0, drs, total_count);
    KheDrsSolnTrieInsert(soln_trie, soln, 0, drs);
    (*total_count)++;
  }
  if( DEBUG12 )
    fprintf(stderr, "] KheDrsSolnTrieDoStrongDominanceTest returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieExtend(KHE_DRS_SOLN_TRIE soln_trie,                   */
/*    int day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Extend the solutions in soln_trie to next_day.                           */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnExtend(KHE_DRS_SOLN soln, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnTrieExtend(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_DAY new_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); extend soln_trie->soln */
    KheDrsSolnExtend(soln_trie->soln, new_day, drs);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieExtend(child_trie, new_day, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieExtractFinalSoln(KHE_DRS_SOLN_TRIE soln_trie,         */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Extract the unique solution (if any) from soln_trie.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieExtractFinalSoln(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); extract soln_trie->soln */
    HnAssert(*final_soln == NULL,
      "KheDrsSolnTrieExtractFinalSoln internal error 1");
    *final_soln = soln_trie->soln;
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieExtractFinalSoln(child_trie, final_soln);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDoDebug(KHE_DRS_SOLN_TRIE soln_trie, int index,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Carry out the recursive part of KheDrsSolnTrieDebug.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDoDebug(KHE_DRS_SOLN_TRIE soln_trie, int index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); soln_trie->soln */
    fprintf(fp, "%*s%d: ", indent, "", index);
    KheDrsSolnDebug(soln_trie->soln, drs, verbosity, indent, fp);
  }
  else
  {
    /* case (3); handle children */
    fprintf(fp, "%*s[ %d:\n", indent, "", index);
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieDoDebug(child_trie, i, drs, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDebug(KHE_DRS_SOLN_TRIE soln_trie,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of soln_trie onto fp with the given verbosity and indent.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDebug(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ SolnTrie\n", indent, "");
    if( soln_trie == NULL )
    {
      /* case (1); do nothing */
    }
    else if( soln_trie->soln != NULL )
    {
      /* case (2); soln_trie->soln */
      KheDrsSolnDebug(soln_trie->soln, drs, verbosity, indent + 2, fp);
    }
    else
    {
      /* case (3); handle children */
      HaArrayForEach(soln_trie->children, child_trie, i)
	KheDrsSolnTrieDoDebug(child_trie, i, drs, verbosity, indent + 2, fp);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_SET"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_SET KheDrsSolnSetMake(KHE_TIME_GROUP tg,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a new KHE_DRS_SOLN_SET object.                                    */
/*                                                                           */
/*****************************************************************************/

KHE_DRS_SOLN_SET KheDrsSolnSetMake(/* KHE_TIME_GROUP tg, */
  KHE_DRS_DOM_KIND dom_kind, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_SET res;
  if( HaArrayCount(drs->free_soln_sets) > 0 )
    res = HaArrayLastAndDelete(drs->free_soln_sets);
  else
    HaMake(res, drs->arena);
  res->count = 0;
  res->dom_kind = dom_kind;
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      res->u.none_dom_set = KheDrsSolnListMake(drs);
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableInit(res->u.weak_dom_set,
	&KheDrsSolnSignatureFullHashUntyped,
	&KheDrsSolnSignatureFullEqualUntyped,
	&KheDrsSolnSignatureDebugUntyped, drs->arena);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableInit(res->u.medium_dom_set,
	&KheDrsSolnListSignaturePartialHashUntyped,
	&KheDrsSolnListSignaturePartialEqualUntyped,
	&KheDrsSolnListSignatureDebugUntyped, drs->arena);
      break;

    case KHE_DRS_DOM_STRONG:

      res->u.strong_dom_set = KheDrsSolnListMake(drs);
      break;

    case KHE_DRS_DOM_TRIE:

      res->u.trie_dom_set = NULL;  /* represents the empty set */
      break;

    default:

      HnAbort("KheDrsSolnSetMake internal error: unknown dom_kind (%d)\n",
	dom_kind);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free dns and everything in it.                                           */
/*                                                                           */
/*  Most of the objects go onto free lists in drs for re-use later.  The     */
/*  exceptions are HpTable objects; they return to the arena they came from. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int pos;  KHE_DRS_SOLN soln;  KHE_DRS_SOLN_LIST ns;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListFree(dns->u.none_dom_set, drs);
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(dns->u.weak_dom_set, soln, pos)
	KheDrsSolnFree(soln, drs);
      HpTableFree(dns->u.weak_dom_set);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, ns, pos)
	KheDrsSolnListFree(ns, drs);
      HpTableFree(dns->u.medium_dom_set);
      break;

    case KHE_DRS_DOM_STRONG:

      KheDrsSolnListFree(dns->u.strong_dom_set, drs);
      break;

    case KHE_DRS_DOM_TRIE:

      KheDrsSolnTrieFree(&dns->u.trie_dom_set, drs);
      HnAssert(dns->u.trie_dom_set == NULL,
	"KheDrsSolnSetClear internal error");
      break;

    default:

      HnAbort("KheDrsSolnSetClear internal error: unknown dom_kind (%d)\n",
	dns->dom_kind);
  }
  HaArrayAddLast(drs->free_soln_sets, dns);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET dns, KHE_DRS_SOLN soln,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add soln to dns.                                                         */
/*                                                                           */
/*  Strong dominance                                                         */
/*  ----------------                                                         */
/*                                                                           */
/*  For each soln other_soln in dns, if soln is dominated by other_soln,     */
/*  then delete soln and return.                                             */
/*                                                                           */
/*  Otherwise, delete each other_soln that soln dominates, then add soln.    */
/*                                                                           */
/*  Weak dominance                                                           */
/*  --------------                                                           */
/*                                                                           */
/*  If dns does not already contain a soln with the same signature as soln,  */
/*  just add soln to dns and return.                                         */
/*                                                                           */
/*  If dns already contains a soln (call it other_soln) with the same        */
/*  signature as soln, keep whichever one of soln and other_soln has the     */
/*  smallest cost, by overwriting other_soln's fields with soln's fields,    */
/*  if required.                                                             */
/*                                                                           */
/*  If soln and other_soln have equal cost, keep one chosen at random, with  */
/*  weights which ensure that if there are K solns with the same signature   */
/*  and the same cost, then each has probability 1/K of being the one kept.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET dns, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN other_soln;  /* int asst_count; */
  KHE_DRS_SOLN_LIST soln_set, other_soln_set;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      /* no dominance testing, just add soln */
      HaArrayAddLast(dns->u.none_dom_set->solns, soln);
      break;

    case KHE_DRS_DOM_WEAK:

      /* do equivalence testing */
      if( HpTableAddUnique(dns->u.weak_dom_set, (void *) soln, soln,other_soln))
	dns->count++;
      else
      {
	/* other_soln will be best of one more soln than before */
	other_soln->best_of++;

	/* ensure other_soln is the better of other_soln and soln */
	if( soln->cost < other_soln->cost || (soln->cost == other_soln->cost &&
	      drs->solve_diversifier % other_soln->best_of == 0) )
	{
	  /* soln is better than other_soln, so overwrite with soln */
	  KheDrsSolnOverWrite(other_soln, soln);
	}

	/* free soln */
	KheDrsSolnFree(soln, drs);
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      /* make soln_set holding just soln and try adding it to medium_dom_set */
      soln_set = KheDrsSolnListMake(drs);
      HaArrayAddLast(soln_set->solns, soln);
      if( HpTableAddUnique(dns->u.medium_dom_set, (void *) soln_set, soln_set,
	    other_soln_set) )
      {
	/* soln_set added, so that's all that's needed */
	dns->count++;
      }
      else
      {
	/* remove soln_set and do strong dom. of soln with other_soln_set */
	HaArrayClear(soln_set->solns);
	KheDrsSolnListFree(soln_set, drs);
        KheDrsSolnListDoStrongDominanceTest(other_soln_set, soln, drs,
	  &dns->count);
      }
      break;

    case KHE_DRS_DOM_STRONG:
    
      /* carry out strong dominance test between soln_set and soln */
      KheDrsSolnListDoStrongDominanceTest(dns->u.strong_dom_set, soln,
	drs, &dns->count);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieDoStrongDominanceTest(&dns->u.trie_dom_set, soln, drs,
	&dns->count);
      break;
    
    default:

      HnAbort("KheDrsSolnSetAddSoln internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetExtend(KHE_DRS_SOLN_SET dns, KHE_DRS_DAY new_day,      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend dns by adding tasks from new_day.                                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetExtend(KHE_DRS_SOLN_SET dns, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_set;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      HaArrayForEach(dns->u.none_dom_set->solns, soln, j)
	KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_WEAK:

      if( DEBUG9 )
      {
	float probes;
	probes = HpTableProbeLength(dns->u.weak_dom_set);
	fprintf(stderr, "  hash table: count %d, size %d, probes %.2f\n",
	  dns->count, HpTableSize(dns->u.weak_dom_set), probes);
	if( probes > 14.0 )
	  HpTableDebug(dns->u.weak_dom_set, 2, stderr);
      }
      HpTableForEachValue(dns->u.weak_dom_set, soln, i)
	KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	  KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_STRONG:

      HaArrayForEach(dns->u.strong_dom_set->solns, soln, j)
	KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieExtend(dns->u.trie_dom_set, new_day, drs);
      break;
    
    default:

      HnAbort("KheDrsSolnSetExtend internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetExtractFinalSoln(KHE_DRS_SOLN_SET dns,                 */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Extract the unique final solution (if any) from dns.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetExtractFinalSoln(KHE_DRS_SOLN_SET dns,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_set;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      HaArrayForEach(dns->u.none_dom_set->solns, soln, j)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetExtractFinalSoln internal error 3");
	*final_soln = soln;
      }
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(dns->u.weak_dom_set, soln, i)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetExtractFinalSoln internal error 1");
	*final_soln = soln;
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	{
	  HnAssert(*final_soln == NULL,
	    "KheDrsSolnSetExtractFinalSoln internal error 2");
	  *final_soln = soln;
	}
      break;

    case KHE_DRS_DOM_STRONG:

      HaArrayForEach(dns->u.strong_dom_set->solns, soln, j)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetExtractFinalSoln internal error 3");
	*final_soln = soln;
      }
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieExtractFinalSoln(dns->u.trie_dom_set, final_soln);
      break;
    
    default:

      HnAbort("KheDrsSolnSetExtractFinalSoln internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetDebug(KHE_DRS_SOLN_SET dns,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of dns onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetDebug(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_SOLN soln;  int i;  KHE_DRS_SOLN_LIST soln_list;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListDebug(dns->u.none_dom_set, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_DOM_WEAK:

      fprintf(fp, "%*s[ WeakTable\n", indent, "");
      HpTableForEachValue(dns->u.weak_dom_set, soln, i)
	KheDrsSolnDebug(soln, drs, verbosity, indent + 2, fp);
      fprintf(fp, "%*s]\n", indent, "");
      break;

    case KHE_DRS_DOM_MEDIUM:

      fprintf(fp, "%*s[ MediumTable\n", indent, "");
      HpTableForEachValue(dns->u.medium_dom_set, soln_list, i)
	KheDrsSolnListDebug(soln_list, drs, verbosity, indent, fp);
      fprintf(fp, "%*s]\n", indent, "");
      break;

    case KHE_DRS_DOM_STRONG:

      KheDrsSolnListDebug(dns->u.strong_dom_set, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieDebug(dns->u.trie_dom_set, drs, verbosity, indent, fp);
      break;
    
    default:

      HnAbort("KheDrsSolnSetDebug internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "conversion of event resource monitors into expression trees"  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorCmp(const void *t1, const void *t2)                        */
/*                                                                           */
/*  Comparison function for sorting (and uniqueifying) an array of monitors. */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorCmp(const void *t1, const void *t2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) t1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) t2;
  return KheMonitorSolnIndex(m1) - KheMonitorSolnIndex(m2);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheTaskToDrsTask(KHE_TASK task,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Convert task into the drs task corresponding to its proper root.         */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheTaskToDrsTask(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int task_index;
  task_index = KheTaskSolnIndex(KheTaskProperRoot(task));
  if( task_index >= HaArrayCount(drs->all_root_tasks) )
    return NULL;
  else
    return HaArray(drs->all_root_tasks, task_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddAssignResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_ASSIGN_RESOURCE_MONITOR arm)                                       */
/*                                                                           */
/*  Make and add in one or more expression trees representing arm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddAssignResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_ASSIGN_RESOURCE_MONITOR arm)
{
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int min_limit, i, j;  KHE_DRS_TASK dt;  KHE_TASK task;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_COST ec;  KHE_DRS_EXPR_INT_DEV eid;  KHE_EVENT_RESOURCE er;
  er = KheAssignResourceMonitorEventResource(arm);
  if( KheEventResourceResourceType(er) == drs->resource_type )
  {
    c = KheMonitorConstraint((KHE_MONITOR) arm);
    combined_wt = KheConstraintCombinedWeight(c);
    cost_fn = KheConstraintCostFunction(c);
    min_limit = KheEventDuration(KheEventResourceEvent(er));
    if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    {
      /* one tree per task on day */
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      ec = KheDrsExprCostMake((KHE_MONITOR)arm,cost_fn,combined_wt,drs);
	      eid = KheDrsExprIntDevMake(1, INT_MAX, false, drs);
	      eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eat, drs);
	      KheDrsExprInitEnd(eid, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eid, drs);
	      KheDrsExprInitEnd(ec, drs);
	      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
	    }
	}
      }
    }
    else
    {
      /* a single tree, whose root has one child per task on day */
      eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) arm, cost_fn, combined_wt,
	min_limit, INT_MAX, false, 0, 0, KheDrsDomTestMake(KHE_DRS_DOM_GE),drs);
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	    }
	}
      }
      KheDrsExprInitEnd(eisc, drs);
      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddPreferResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    KHE_PREFER_RESOURCES_MONITOR prm)                                      */
/*                                                                           */
/*  Make and add in one or more expression trees representing prm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddPreferResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_PREFER_RESOURCES_MONITOR prm)
{
  KHE_CONSTRAINT c;  KHE_RESOURCE_GROUP rg;  KHE_EVENT_RESOURCE er;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  KHE_PREFER_RESOURCES_CONSTRAINT prc;  int i, j;  KHE_DRS_TASK_ON_DAY dtd;
  KHE_DRS_EXPR_COST ec;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_DOM_TEST dom_test;
  KHE_TASK task;  KHE_DRS_TASK dt;
  er = KhePreferResourcesMonitorEventResource(prm);
  if( KheEventResourceResourceType(er) == drs->resource_type )
  {
    c = KheMonitorConstraint((KHE_MONITOR) prm);
    combined_wt = KheConstraintCombinedWeight(c);
    cost_fn = KheConstraintCostFunction(c);
    prc = (KHE_PREFER_RESOURCES_CONSTRAINT) c;
    rg = KhePreferResourcesConstraintDomainComplement(prc);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    {
      /* one tree per task on day */
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      ec = KheDrsExprCostMake((KHE_MONITOR)prm,cost_fn,combined_wt,drs);
	      eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eat, drs);
	      KheDrsExprInitEnd(ec, drs);
	      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
	    }
	}
      }
    }
    else
    {
      /* a single tree, whose root has one grandchild per task on day */
      ec = KheDrsExprCostMake((KHE_MONITOR) prm, cost_fn, combined_wt, drs);
      dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
      eis = KheDrsExprIntSumMake(dom_test, drs);
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eat, drs);
	    }
	}
      }
      KheDrsExprInitEnd(eis, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis, drs);
      KheDrsExprInitEnd(ec, drs);
      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,   */
/*    bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,                     */
/*    KHE_DRS_DOM_TEST *beta_dom_test)                                       */
/*                                                                           */
/*  Convert limits (min_limit, max_limit, allow_zero) into dominance tests.  */
/*  if min_limit <= 0 there is no min limit; if max_limit >= max_max_limit   */
/*  there is no maximum limit.                                               */
/*                                                                           */
/*  The values returned in *alpha_dom_test and *beta_dom_test are referred   */
/*  to as alpha and beta in the User's Guide.                                */
/*                                                                           */
/*  Note.  If there is a trivial allow_zero flag (if min_limit <= 1), an     */
/*  adjustment will have already been made by now to remove it.  So this     */
/*  call may assume that allow_zero implies min_limit > 1.                   */
/*                                                                           */
/*****************************************************************************/

static void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,
  bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,
  KHE_DRS_DOM_TEST *beta_dom_test)
{
  if( allow_zero /* && min_limit > 1 */ )
  {
    /* fall back on equality when there is a non-trivial allow_zero flag */
    *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
  }
  else if( max_limit >= max_max_limit )
  {
    if( min_limit <= 0 )
    {
      /* no max_limit, and no min_limit (should never happen) */
      HnAbort("KheLimitsToTests: no max limit, no min limit");
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    }
    else
    {
      /* no max_limit, and min_limit */
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_GE_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_GE);
    }
  }
  else
  {
    if( min_limit <= 0 )
    {
      /* max_limit, and no min_limit */
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
    }
    else
    {
      /* max_limit, and min_limit */
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_EQ_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_LIMIT_RESOURCES_MONITOR lrm)                                       */
/*                                                                           */
/*  Make and add in one or more expression trees representing lrm.           */
/*                                                                           */
/*  The documentation mentions an unlikely case where the expression can     */
/*  be split up.  We haven't bothered testing for that here.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_RESOURCES_MONITOR lrm)
{
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int min_limit, max_limit, i, egi, j;  KHE_RESOURCE_GROUP rg;
  KHE_LIMIT_RESOURCES_CONSTRAINT lrc;  KHE_EVENT_RESOURCE er;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_TASK task;  KHE_DRS_TASK dt;
  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  c = KheMonitorConstraint((KHE_MONITOR) lrm);
  combined_wt = KheConstraintCombinedWeight(c);
  cost_fn = KheConstraintCostFunction(c);
  lrc = (KHE_LIMIT_RESOURCES_CONSTRAINT) c;
  rg = KheLimitResourcesConstraintDomain(lrc);
  min_limit = KheLimitResourcesConstraintMinimum(lrc);
  max_limit = KheLimitResourcesConstraintMaximum(lrc);
  KheLimitsToTests(min_limit, max_limit, INT_MAX, false,
    &alpha_dom_test, &beta_dom_test);
  eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) lrm, cost_fn, combined_wt,
    min_limit, max_limit, false, 0, 0, alpha_dom_test, drs);
  egi = KheLimitResourcesMonitorEventGroupIndex(lrm);
  for( i = 0; i < KheLimitResourcesConstraintEventResourceCount(lrc, egi); i++ )
  {
    er = KheLimitResourcesConstraintEventResource(lrc, egi, i);
    for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
    {
      task = KheEventResourceTask(drs->soln, er, i);
      dt = KheTaskToDrsTask(task, drs);
      if( dt != NULL )
      {
	HaArrayForEach(dt->days, dtd, j)
	  if( dtd->task == task )
	  {
	    eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat, drs);
	  }
      }
    }
  }
  KheDrsExprInitEnd(eisc, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddEventResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_MONITOR m)                                                         */
/*                                                                           */
/*  Add an expression tree to drs for event resource monitor m, if relevant. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddEventResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_ASSIGN_RESOURCE_MONITOR_TAG:

      KheDrsAddAssignResourceMonitor(drs, (KHE_ASSIGN_RESOURCE_MONITOR) m);
      break;

    case KHE_PREFER_RESOURCES_MONITOR_TAG:

      KheDrsAddPreferResourcesMonitor(drs, (KHE_PREFER_RESOURCES_MONITOR) m);
      break;

    case KHE_AVOID_SPLIT_ASSIGNMENTS_MONITOR_TAG:

      /* event resource monitor, but we are currently not handling it */
      break;

    case KHE_LIMIT_RESOURCES_MONITOR_TAG:

      KheDrsAddLimitResourcesMonitor(drs, (KHE_LIMIT_RESOURCES_MONITOR) m);
      break;

    default:

      HnAbort("KheDrsAddEventResourceMonitor internal error (tag %d)",
	KheMonitorTag(m));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "conversion of resource monitors into expression trees"        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddAvoidUnavailableTimesMonitor(                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs,                                       */
/*    KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm)                              */
/*                                                                           */
/*  Make and add in one or more expression trees representing autm.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddAvoidUnavailableTimesMonitor(
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm,
  KHE_DRS_RESOURCE dr)
{
  KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT c;  KHE_DRS_EXPR_BUSY_TIME eb;
  int i;  KHE_TIME t;  KHE_DRS_EXPR_COST ec;  KHE_COST_FUNCTION cost_fn;
  KHE_TIME_GROUP tg;  KHE_DRS_EXPR_INT_SUM eis;  KHE_COST combined_wt;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test;  /* KHE_RESOURCE r; */
  KHE_DRS_RESOURCE_ON_DAY drd;

  /* boilerplate */
  c = KheAvoidUnavailableTimesMonitorConstraint(autm);
  tg = KheAvoidUnavailableTimesConstraintUnavailableTimes(c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  if( combined_wt > 0 && KheTimeGroupTimeCount(tg) > 0 )
  {
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION || KheTimeGroupTimeCount(tg) == 1 )
    {
      /* a separate expression tree for each time */
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	ec = KheDrsExprCostMake((KHE_MONITOR) autm, cost_fn, combined_wt, drs);
	HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eb = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eb, drs);
	KheDrsExprInitEnd(ec, drs);
      }
    }
    else
    {
      /* a single expression tree (max limit of 0, in effect) */
      ec = KheDrsExprCostMake((KHE_MONITOR) autm, cost_fn, combined_wt, drs);
      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
      le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
      eis = KheDrsExprIntSumMake(le_dom_test, drs);
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eb = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eb, drs);
      }
      KheDrsExprInitEnd(eis, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis, drs);
      KheDrsExprInitEnd(ec, drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If tg contains a complete day's worth of times, starting from            */
/*  start_index, then return true and set *day_tg to the day time group,     */
/*  otherwise return false.                                                  */
/*                                                                           */
/*****************************************************************************/

static bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)
{
  KHE_TIME tg_t, day_t;  int i, day_index;  KHE_TIME_GROUP d_tg;
  tg_t = KheTimeGroupTime(tg, start_index);
  day_index = KheFrameTimeIndex(drs->days_frame, tg_t);
  d_tg = KheFrameTimeTimeGroup(drs->days_frame, tg_t);
  for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
  {
    day_t = KheTimeGroupTime(d_tg, i);
    if( start_index + i >= KheTimeGroupTimeCount(tg) )
      return *day = NULL, false;
    tg_t = KheTimeGroupTime(tg, start_index + i);
    if( day_t != tg_t )
      return *day = NULL, false;
  }
  return *day = HaArray(drs->all_days, day_index), true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeOrExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,    */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an OR expression with one BUSY child for each time of tg, and add   */
/*  the OR expression as a child of prnt.  If tg has exactly one time,       */
/*  optimize by adding just the BUSY expression for that time as a child     */
/*  of prnt.                                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheMakeOrExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR eo;  KHE_DRS_EXPR_BUSY_TIME ebt;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single BUSY_TIME expression */
    t = KheTimeGroupTime(tg, 0);
    drd = KheDrsResourceOnDayAtTime(dr, t, drs);
    ebt = KheDrsExprBusyTimeMake(drd, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebt, drs);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
  {
    /* just a single BUSY_DAY expression */
    drd = KheDrsResourceOnDay(dr, day);
    ebd = KheDrsExprBusyDayMake(drd, day->time_group, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebd, drs);
  }
  else
  {
    /* an OR expression with BUSY_DAY and BUSY_TIME children */
    eo = KheDrsExprOrMake(dom_test, drs);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day) )
      {
	/* add a BUSY_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	ebd = KheDrsExprBusyDayMake(drd, day->time_group, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebd, drs);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	/* add a BUSY_TIME expression */
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	ebt = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebt, drs);
      }
    }
    KheDrsExprInitEnd(eo, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eo, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeAndExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,   */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an AND expression with one FREE child for each time of tg, and      */
/*  add the AND expression as a child of prnt.  If tg has exactly one time,  */
/*  optimize by adding just the FREE expression for that time as a child     */
/*  of prnt.                                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheMakeAndExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND ea;  KHE_DRS_EXPR_FREE_TIME eft;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single FREE expression */
    t = KheTimeGroupTime(tg, 0);
    drd = KheDrsResourceOnDayAtTime(dr, t, drs);
    eft = KheDrsExprFreeTimeMake(drd, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eft, drs);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
  {
    /* just a single FREE_DAY expression */
    drd = KheDrsResourceOnDay(dr, day);
    efd = KheDrsExprFreeDayMake(drd, day->time_group, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) efd, drs);
  }
  else
  {
    /* an AND expression with FREE_DAY and FREE_TIME children */
    ea = KheDrsExprAndMake(dom_test, drs);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day) )
      {
	/* add a FREE_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	efd = KheDrsExprFreeDayMake(drd, day->time_group, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) efd, drs);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	/* add a FREE_TIME expression */
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eft = KheDrsExprFreeTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) eft, drs);
      }
    }
    KheDrsExprInitEnd(ea, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ea, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsClusterMonitorContainsWholePositiveDay(                       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int start_index,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)                     */
/*                                                                           */
/*  If cbtm contains a sequence of singleton time groups starting at         */
/*  day_index that have positive polarity and cover a complete day, return   */
/*  true and set *day to the day;  otherwise return false.                   */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsClusterMonitorContainsWholePositiveDay(
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)
{
  KHE_TIME day_t;  int i, tg_count, day_index;  KHE_TIME_GROUP m_tg, d_tg;
  KHE_POLARITY po;  KHE_TIME t;
  tg_count = KheClusterBusyTimesMonitorTimeGroupCount(cbtm);
  m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, start_index, &po);
  if( po == KHE_POSITIVE && KheTimeGroupTimeCount(m_tg) == 1 )
  {
    t = KheTimeGroupTime(m_tg, 0);
    d_tg = KheFrameTimeTimeGroup(drs->days_frame, t);
    day_index = KheFrameTimeIndex(drs->days_frame, t);
    for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
    {
      day_t = KheTimeGroupTime(d_tg, i);
      if( start_index + i >= tg_count )
	return *day = NULL, false;
      m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, start_index + i, &po);
      if( po != KHE_POSITIVE || KheTimeGroupTimeCount(m_tg) != 1 ||
	  KheTimeGroupTime(m_tg, 0) != day_t )
	return *day = NULL, false;
    }
    return *day = HaArray(drs->all_days, day_index), true;
  }
  else
    return *day = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddClusterBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, KHE_DRS_RESOURCE dr)              */
/*                                                                           */
/*  Make and add in one or more expression trees representing cbtm.          */
/*                                                                           */
/*  Implementation note.  If cbtm contains a sequence of consecutive         */
/*  singleton time groups with positive polarity that add up to a complete   */
/*  day, then one BUSY_DAY expression is generated for them instead of one   */
/*  BUSY_TIME expression for each.  This is correct:  the sum reported       */
/*  in either case is 1 if the day is busy and 0 otherwise.  It cannot       */
/*  be more than 1 because the resource takes at most one task each day.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddClusterBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, KHE_DRS_RESOURCE dr)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;  bool allow_zero;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_POLARITY po;  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;
  /* KHE_RESOURCE r; */

  /* boilerplate */
  c = KheClusterBusyTimesMonitorConstraint(cbtm);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheClusterBusyTimesConstraintMinimum(c);
  max_limit = KheClusterBusyTimesConstraintMaximum(c);
  allow_zero = KheClusterBusyTimesConstraintAllowZero(c);
  history_before = KheClusterBusyTimesMonitorHistory(cbtm);
  history_after = KheClusterBusyTimesMonitorHistoryAfter(cbtm);

  /* adjust if allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  tg_count = KheClusterBusyTimesConstraintTimeGroupCount(c);
  if( (min_limit <= 0 && max_limit >= tg_count) || combined_wt == 0 )
    return;

  /* work out the various dom_tests */
  KheLimitsToTests(min_limit, max_limit, tg_count, allow_zero,
    &alpha_dom_test, &beta_dom_test);

  /* one int_cost_comb expression at top of tree */
  eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) cbtm, cost_fn, combined_wt,
    min_limit, max_limit, allow_zero, history_before, history_after,
    alpha_dom_test, drs);

  /* one OR or AND expression for each time group */
  /* or a BUSY_DAY or FREE_DAY expression if all singles of one day present */
  for( i = 0;  i < tg_count;  i++ )
  {
    if( KheDrsClusterMonitorContainsWholePositiveDay(cbtm, i, drs, &day) )
    {
      /* add a BUSY_DAY expression */
      drd = KheDrsResourceOnDay(dr, day);
      ebd = KheDrsExprBusyDayMake(drd, day->time_group, beta_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebd, drs);
      i += KheTimeGroupTimeCount(day->time_group) - 1;
    }
    else
    {
      tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
      if( po == KHE_POSITIVE )
      {
	/* make an OR expression with one BUSY expression for each time of tg */
	KheMakeOrExprAndChildren(dr,(KHE_DRS_EXPR) eisc, tg,beta_dom_test,drs);
      }
      else
      {
	/* make an AND expression with one FREE expression for each time of tg */
	KheMakeAndExprAndChildren(dr,(KHE_DRS_EXPR)eisc, tg,beta_dom_test,drs);
      }
    }
  }
  KheDrsExprInitEnd(eisc, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, KHE_DRS_RESOURCE dr)                */
/*                                                                           */
/*  Make and add in one or more expression trees representing lbtm.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, j, junk, min_limit, max_limit;  bool allow_zero;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_TIME t;
  KHE_DRS_EXPR_INT_SUM eis, eis2;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_INT_DEV eid;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;
  KHE_RESOURCE r;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;

  /* boilerplate */
  r = KheLimitBusyTimesMonitorResource(lbtm);
  c = KheLimitBusyTimesMonitorConstraint(lbtm);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitBusyTimesConstraintMinimum(c);
  max_limit = KheLimitBusyTimesConstraintMaximum(c);
  allow_zero = KheLimitBusyTimesConstraintAllowZero(c);
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
  dr = KheDrsResource(r, drs);

  /* adjust allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  if( combined_wt == 0 )
    return;

  if( cost_fn == KHE_LINEAR_COST_FUNCTION ||
      KheLimitBusyTimesMonitorTimeGroupCount(lbtm) == 1 )
  {
    /* one INT_SUM_COMB for each time group */
    /* unless the time group is one day, min_limit = 0, and max_limit = 1 */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &junk);
      if( min_limit == 0 && max_limit == 1 &&
	  KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
	  KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
      {
	/* constraint specifies at most one task per day; trivial, so omit */
      }
      else if( min_limit > 0 || max_limit < KheTimeGroupTimeCount(tg) )
      {
	KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	  allow_zero, &alpha_dom_test, &beta_dom_test);
	eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) lbtm, cost_fn,
	  combined_wt, min_limit, max_limit, allow_zero, 0, 0,
	  alpha_dom_test, drs);
	for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
	{
	  t = KheTimeGroupTime(tg, j);
	  drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	  ebt = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	  KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebt, drs);
	}
	KheDrsExprInitEnd(eisc, drs);
	HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
      }
    }
  }
  else
  {
    /* one COST and INT_SUM at root of tree */
    ec = KheDrsExprCostMake((KHE_MONITOR) lbtm, cost_fn, combined_wt, drs);
    HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
    eis = KheDrsExprIntSumMake(le_dom_test, drs);

    /* one INT_DEV and INT_SUM for each time group */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &junk);
      eid = KheDrsExprIntDevMake(min_limit, max_limit, allow_zero, drs);
      KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	allow_zero, &alpha_dom_test, &beta_dom_test);
      eis2 = KheDrsExprIntSumMake(alpha_dom_test, drs);
      for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
      {
	t = KheTimeGroupTime(tg, j);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	ebt = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis2, (KHE_DRS_EXPR) ebt, drs);
      }
      KheDrsExprInitEnd(eis2, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eis2, drs);
      KheDrsExprInitEnd(eid, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eid, drs);
    }
    KheDrsExprInitEnd(eis, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis, drs);
    KheDrsExprInitEnd(ec, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitWorkloadMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_LIMIT_WORKLOAD_MONITOR lwm, KHE_DRS_RESOURCE dr)                   */
/*                                                                           */
/*  Make and add in one or more expression trees representing lwm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitWorkloadMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_WORKLOAD_MONITOR lwm, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_WORKLOAD_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, j, min_limit, max_limit;  bool allow_zero;  float junk;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_DRS_EXPR_WORK_TIME ew;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_FLOAT_SUM efs;  KHE_DRS_EXPR_FLOAT_DEV efd;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;
  KHE_DRS_RESOURCE_ON_DAY drd;

  /* boilerplate */
  c = KheLimitWorkloadMonitorConstraint(lwm);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitWorkloadConstraintMinimum(c);
  max_limit = KheLimitWorkloadConstraintMaximum(c);
  allow_zero = KheLimitWorkloadConstraintAllowZero(c);
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);

  /* adjust allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  if( combined_wt == 0 )
    return;

  /* one COST and INT_SUM at the root */
  ec = KheDrsExprCostMake((KHE_MONITOR) lwm, cost_fn, combined_wt, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
  eis = KheDrsExprIntSumMake(le_dom_test, drs);

  /* one FLOAT_DEV and FLOAT_SUM for each time group */
  KheLimitsToTests(min_limit, max_limit, INT_MAX, allow_zero,
    &alpha_dom_test, &beta_dom_test);
  for( i = 0;  i < KheLimitWorkloadMonitorTimeGroupCount(lwm);  i++ )
  {
    tg = KheLimitWorkloadMonitorTimeGroup(lwm, i, &junk);
    efd = KheDrsExprFloatDevMake(min_limit, max_limit, allow_zero, drs);
    efs = KheDrsExprFloatSumMake(alpha_dom_test, drs);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      drd = KheDrsResourceOnDayAtTime(dr, t, drs);
      ew = KheDrsExprWorkTimeMake(drd, t, unused_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) efs, (KHE_DRS_EXPR) ew, drs);
    }
    KheDrsExprInitEnd(efs, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) efd, (KHE_DRS_EXPR) efs, drs);
    KheDrsExprInitEnd(efd, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) efd, drs);
  }
  KheDrsExprInitEnd(eis, drs);
  KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis, drs);
  KheDrsExprInitEnd(ec, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitActiveIntervalsMonitor(                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs,                                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, KHE_DRS_RESOURCE dr)          */
/*                                                                           */
/*  Make and add in one or more expression trees representing laim.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitActiveIntervalsMonitor(
  KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SEQ_COMB eisc;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;

  /* boilerplate */
  c = KheLimitActiveIntervalsMonitorConstraint(laim);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheLimitActiveIntervalsConstraintMinimum(c);
  max_limit = KheLimitActiveIntervalsConstraintMaximum(c);
  history_before = KheLimitActiveIntervalsMonitorHistoryBefore(laim);
  history_after = KheLimitActiveIntervalsMonitorHistoryAfter(laim);
  history = KheLimitActiveIntervalsMonitorHistory(laim);
  tg_count = KheLimitActiveIntervalsConstraintTimeGroupCount(c);

  /* return if constraint does nothing */
  if( (min_limit <= 1 && max_limit >= tg_count) || combined_wt == 0 )
    return;

  /* work out the various dom_tests */
  KheLimitsToTests(min_limit, max_limit, tg_count, false,
    &alpha_dom_test, &beta_dom_test);

  /* one int_seq_comb expression at top of tree */
  eisc = KheDrsExprIntSeqCombMake((KHE_MONITOR) laim, cost_fn, combined_wt,
    min_limit, max_limit, history_before, history_after, history,
    alpha_dom_test, drs);

  /* one OR or AND expression for each time group */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    if( po == KHE_POSITIVE )
    {
      /* make an OR expression with one BUSY expression for each time of tg */
      KheMakeOrExprAndChildren(dr, (KHE_DRS_EXPR) eisc, tg,beta_dom_test,drs);
    }
    else
    {
      /* make an AND expression with one FREE expression for each time of tg */
      KheMakeAndExprAndChildren(dr, (KHE_DRS_EXPR) eisc,tg,beta_dom_test,drs);
    }
  }
  KheDrsExprInitEnd(eisc, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,           */
/*    KHE_MONITOR m, KHE_DRS_RESOURCE dr)                                    */
/*                                                                           */
/*  Add an expression tree to drs for resource monitor m, if relevant;       */
/*  dr is the drs resource corresponding to the resource monitored by m.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m, KHE_DRS_RESOURCE dr)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_AVOID_CLASHES_MONITOR_TAG:

      /* resource monitor, but no need to handle it */
      break;

    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:

      KheDrsAddAvoidUnavailableTimesMonitor(drs,
	(KHE_AVOID_UNAVAILABLE_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:

      /* resource monitor, but we are currently not handling it */
      break;

    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      KheDrsAddClusterBusyTimesMonitor(drs,
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      KheDrsAddLimitBusyTimesMonitor(drs, (KHE_LIMIT_BUSY_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      KheDrsAddLimitWorkloadMonitor(drs, (KHE_LIMIT_WORKLOAD_MONITOR) m, dr);
      break;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      KheDrsAddLimitActiveIntervalsMonitor(drs,
	(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m, dr);
      break;

    default:

      HnAbort("KheDrsAddResourceMonitor internal error (tag %d)",
	KheMonitorTag(m));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public functions"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,  */
/*    KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)                             */
/*                                                                           */
/*  Make and return a new dynamic resource solver with these attributes.     */
/*                                                                           */
/*****************************************************************************/

KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)
{
  KHE_DYNAMIC_RESOURCE_SOLVER res;  HA_ARENA a;  int i, j;
  KHE_FRAME days_frame;  KHE_EVENT_TIMETABLE_MONITOR etm;
  KHE_TASK_CLASS tc;  KHE_INSTANCE ins;  KHE_EVENT_RESOURCE er;
  KHE_MONITOR m;  KHE_RESOURCE r;  KHE_DRS_RESOURCE dr;
  ARRAY_KHE_MONITOR er_monitors;

  /* get the common frame and event timetable monitor; return NULL if can't */
  days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  if( days_frame == NULL || etm == NULL )
    return NULL;

  /*************************************************************************/
  /*                                                                       */
  /*  Make the basic object and initialize its fields                      */
  /*                                                                       */
  /*************************************************************************/

  /* make the basic object */
  a = KheSolnArenaBegin(soln, false);
  HaMake(res, a);

  /* fields which are constant throughout the lifetime of the solver */
  res->arena = a;
  res->soln = soln;
  res->resource_type = rt;
  res->options = options;
  res->days_frame = days_frame;
  res->etm = etm;
  res->task_class_solver = NULL;  /* initialized below */
  HaArrayInit(res->all_resources, a);
  HaArrayInit(res->all_days, a);
  HaArrayInit(res->all_root_tasks, a);
  HaArrayInit(res->all_root_exprs, a);
  res->postorder_count = 0;

  /* free list fields - solns and days */
  /* HaArrayInit(res->free_edge, a); */
  HaArrayInit(res->free_solns, a);
  HaArrayInit(res->free_soln_lists, a);
  HaArrayInit(res->free_soln_tries, a);
  HaArrayInit(res->free_soln_sets, a);

  /* fields that vary with the solve */
  res->selected_resource_set = KheResourceSetMake(rt, a);
  HaArrayInit(res->open_resources, a);
  HaArrayInit(res->selected_day_ranges, a);
  HaArrayInit(res->open_days, a);
  HaArrayInit(res->open_exprs, a);
  res->solve_diversifier = 0;
  res->solve_rerun = false;

  /* fields for recording running time */
#if TESTING
  res->timer = KheTimerMake("dynamic", KHE_NO_TIME, a);
  HaArrayInit(res->running_time_per_day, a);
#endif


  /*************************************************************************/
  /*                                                                       */
  /*  Add days, tasks, resources, and expression trees                     */
  /*                                                                       */
  /*  When adding each drs resource, its resource monitors' expression     */
  /*  trees are also added.  This assumes that all drs days are made.      */
  /*                                                                       */
  /*  When adding each drs task class, its drs tasks are added as well,    */
  /*  both to the task class and to drs->all_root_tasks.  If a drs task's  */
  /*  corresponding KHE task is currently assigned to resource r, the      */
  /*  drs task is also assigned to the corresponding drs resource.  This   */
  /*  assumes that all drs days and all drs resources are already made.    */
  /*                                                                       */
  /*  When adding the expression tree for each event resource monitor,     */
  /*  it is assumed that drs->all_root_tasks is initialized.               */
  /*                                                                       */
  /*************************************************************************/

  /* make one drs day object for each day in the frame */
  for( i = 0;  i < KheFrameTimeGroupCount(days_frame);  i++ )
    KheDrsDayMake(i, res);

  /* make one drs resource for each resource in rt */
  HaArrayInit(res->all_resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    dr = KheDrsResourceMake(r, res);
    HaArrayAddLast(res->all_resources, dr);
  }

  /* make one drs task class for each no-overlap task class */
  /* this code assumes that drs resources are already made */
  res->task_class_solver = KheTaskClassSolverMake(soln, rt, days_frame, a);
  KheTaskClassSolverAddAllTasks(res->task_class_solver);
  if( DEBUG10 )
    KheTaskClassSolverDebug(res->task_class_solver, 2, 2, stderr);
  for( i = 0;  i < KheTaskClassSolverClassCount(res->task_class_solver);  i++ )
  {
    tc = KheTaskClassSolverClass(res->task_class_solver, i);
    if( KheTaskClassNoOverlap(tc) )
      KheDrsTaskClassMake(tc, res);
  }

  /* make one expression tree for each event resource monitor of type rt */
  ins = KheSolnInstance(soln);
  HaArrayInit(er_monitors, a);
  for( i = 0;  i < KheInstanceEventResourceCount(ins);  i++ )
  {
    er = KheInstanceEventResource(ins, i);
    if( KheEventResourceResourceType(er) == rt )
      for( j = 0;  j < KheSolnEventResourceMonitorCount(soln, er);  j++ )
      {
	m = KheSolnEventResourceMonitor(soln, er, j);
	HaArrayAddLast(er_monitors, m);
      }
  }
  HaArraySortUnique(er_monitors, &KheMonitorCmp);
  HaArrayForEach(er_monitors, m, i)
    KheDrsAddEventResourceMonitor(res, m);
  HaArrayFree(er_monitors);

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Delete drs, reclaiming its memory.                                       */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KheSolnArenaEnd(drs->soln, drs->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Add resource r to the solve.                                             */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r)
{
  HnAssert(KheResourceResourceType(r) == drs->resource_type,
    "KheDynamicResourceSolverAddResource: r has wrong resource type for drs");
  HnAssert(!KheResourceSetContainsResource(drs->selected_resource_set, r),
    "KheDynamicResourceSolverAddResource: r (%s) already added",
    KheResourceId(r));
  KheResourceSetAddResource(drs->selected_resource_set, r);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int first_day_index, int last_day_index)                               */
/*                                                                           */
/*  Add day range first_day_index .. last_day_index to the solve.            */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int first_day_index, int last_day_index)
{
  KHE_DRS_DAY_RANGE si;
  HnAssert(first_day_index <= last_day_index,
    "KheDynamicResourceSolverAddResource: empty day range (%d .. %d)",
    first_day_index, last_day_index);
  HnAssert(first_day_index >= 0,
    "KheDynamicResourceSolverAddResource: first_day_index (%d) out of range",
    first_day_index);
  HnAssert(last_day_index < KheFrameTimeGroupCount(drs->days_frame),
    "KheDynamicResourceSolverAddResource: last_day_index (%d) out of range",
    last_day_index);
  si = KheDrsDayRangeMake(first_day_index, last_day_index);
  KheDrsDayRangeMerge(&drs->selected_day_ranges, si);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    KHE_DRS_DOM_KIND dom_kind)                                             */
/*                                                                           */
/*  Open drs for solving.                                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolveOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs, int soln_limit,
  KHE_DRS_DOM_KIND dom_kind, KHE_MARK *mark)
{
  KHE_DRS_DAY_RANGE ddr;  KHE_DRS_DAY day;  int i, j;  KHE_DRS_EXPR e;
  KHE_DRS_RESOURCE dr;  KHE_RESOURCE r;

  /* boilerplate */
  drs->solve_init_cost = KheSolnCost(drs->soln);
  drs->solve_soln_limit = (soln_limit <= 0 ? INT_MAX : soln_limit);
  *mark = KheMarkBegin(drs->soln);

  /* clear the open resources, open days, and open expressions */
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);

  /* open selected resources (adds drs->open_resources and drs->open_exprs) */
  for( i = 0; i < KheResourceSetResourceCount(drs->selected_resource_set); i++ )
  {
    r = KheResourceSetResource(drs->selected_resource_set, i);
    dr = HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
    KheDrsResourceOpen(dr, drs);
    HaArrayAddLast(drs->open_resources, dr);
  }

  /* open the selected days (adds drs->open_days and drs->open_exprs) */
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      day = HaArray(drs->all_days, j);
      KheDrsDayOpen(day, ddr, dom_kind, drs, HaArrayCount(drs->open_days));
      HaArrayAddLast(drs->open_days, day);
    }

  /* sort drs->open_exprs by increasing postorder index, then open them */
  HaArraySort(drs->open_exprs, &KheDrsExprCmp);
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprOpen(e, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveDoSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,                 */
/*    KHE_DRS_DOM_KIND dom_kind)                                             */
/*                                                                           */
/*  Carry out the actual solve.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolveDoSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_DOM_KIND dom_kind)
{
  KHE_DRS_DAY new_day;  KHE_DRS_SOLN_SET dns;
  int i;  KHE_DRS_SOLN soln;

  dns = KheDrsSolnSetMake(dom_kind, drs);
  soln = KheDrsSolnMake(NULL, NULL, drs->total_closed_cost, drs);
  KheDrsSolnSetAddSoln(dns, soln, drs);
  HaArrayForEach(drs->open_days, new_day, i)
  {
    KheDrsSolnSetExtend(dns, new_day, drs);
    dns = new_day->soln_set;
#if TESTING
    HaArray(drs->running_time_per_day, i) = KheTimerElapsedTime(drs->timer);
#endif
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_MARK mark,    */
/*    bool testing_only)                                                     */
/*                                                                           */
/*  Close drs for solving.  If testing_only, this is for testing only, so    */
/*  do not adopt any solution that may have been found.                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_MARK mark,
  bool testing_only)
{
  KHE_DRS_DAY day;  int i;  KHE_DRS_EXPR e;  KHE_DRS_SOLN soln, final_soln;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;

  /* extract final_soln */
  day = HaArrayLast(drs->open_days);
  final_soln = NULL;
  KheDrsSolnSetExtractFinalSoln(day->soln_set, &final_soln);
  if( DEBUG2 )
  {
    if( final_soln != NULL )
    {
      fprintf(stderr, "  final solution:\n");
      KheDrsSolnDebug(final_soln, drs, 1, 2, stderr);
    }
    else
      fprintf(stderr, "  no final solution\n");
  }

  if( testing_only || final_soln == NULL ||
      final_soln->cost >= drs->solve_init_cost )
  {
    /* testing, or no final soln, or no improvement; return to initial soln */
    KheMarkEnd(mark, true);
    HnAssert(KheSolnCost(drs->soln) == drs->solve_init_cost,
      "KheDynamicResourceSolverSolve internal error 2");
  }
  else
  {
    /* improvement; traverse final solution, closing resources and tasks */
    for( soln = final_soln;  soln->prev_soln != NULL;  soln = soln->prev_soln )
    {
      HaArrayForEach(soln->prev_tasks, dtd, i)
      {
	if( dtd != NULL )
	{
	  dr = HaArray(drs->open_resources, i);
	  drd = KheDrsResourceOnDay(dr, soln->day);
	  KheDrsTaskClose(dtd->encl_dt, drd->encl_dr);
	}
      }
    }
    KheMarkEnd(mark, false);
    HnAssert(KheSolnCost(drs->soln) == final_soln->cost,
      "KheDynamicResourceSolverSolve internal error 3");
    HnAssert(KheSolnCost(drs->soln) < drs->solve_init_cost,
      "KheDynamicResourceSolverSolve internal error 4");
  }

  /* close the open days */
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayClose(day, drs);

  /* close the open expressions */
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprClose(e, drs);

  /* close drs */
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int soln_limit, KHE_DRS_DOM_KIND dom_kind)                             */
/*                                                                           */
/*  Solve drs.  Adopt the best solution found and return true if it is an    */
/*  improvement on the original.                                             */
/*                                                                           */
/*****************************************************************************/

bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND dom_kind)
{
  KHE_COST final_cost;  bool res;  KHE_MARK mark;
  if( DEBUG1 )
    fprintf(stderr, "[ KheDynamicResourceSolverSolve(drs)\n");

  /* open drs, solve, and close drs */
  KheDrsSolveOpen(drs, soln_limit, dom_kind, &mark);
  KheDrsSolveDoSolve(drs, dom_kind);
  KheDrsSolveClose(drs, mark, false);

  /* tidy up and return */
  drs->solve_diversifier++;
  KheResourceSetClear(drs->selected_resource_set);
  HaArrayClear(drs->selected_day_ranges);
  final_cost = KheSolnCost(drs->soln);
  HnAssert(final_cost <= drs->solve_init_cost,
    "KheDynamicResourceSolverSolve: final cost (%.5f) exceeds initial cost "
    "(%.5f)", KheCostShow(final_cost),KheCostShow(drs->solve_init_cost));
  res = (final_cost < drs->solve_init_cost);
  if( DEBUG1 )
    fprintf(stderr, "] KheDynamicResourceSolverSolve returning %s\n",
      bool_show(res));
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of drs onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE dr;  int i;
  HnAssert(indent >= 0, "KheDynamicResourceSolverDebug: indent < 0");
  fprintf(fp, "%*s[ DynamicResourceSolver(", indent, "");
  HaArrayForEach(drs->all_resources, dr, i)
    KheDrsResourceDebug(dr, 2, indent + 2, fp);
  HaArrayForEach(drs->all_days, day, i)
    KheDrsDayDebug(day, drs, 2, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "testing"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void TestName(KHE_INSTANCE ins, char *suffix, char buff[200])            */
/*                                                                           */
/*  Set buff to the name of a test with these attributes.                    */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static void TestName(KHE_INSTANCE ins, char *suffix, char buff[200])
{
  int len;

  /* add standard prefix to buff */
  len = snprintf(buff, 200, "%s", KheInstanceId(ins));

  /* add suffix */
  if( strcmp(suffix, "") != 0 )
    len += snprintf(&buff[len], 200 - len, "-%s", suffix);
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  KHE_DATASET_POINTS_TYPE KheDomKindToPointsType(KHE_DRS_DOM_KIND dom_kind)*/
/*                                                                           */
/*  Return the dataset points type corresponding to dom_kind.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DATASET_POINTS_TYPE KheDomKindToPointsType(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:	return KHE_DATASET_POINTS_TRIANGLE;
    case KHE_DRS_DOM_WEAK:	return KHE_DATASET_POINTS_SQUARE;
    case KHE_DRS_DOM_MEDIUM:	return KHE_DATASET_POINTS_CIRCLE;
    case KHE_DRS_DOM_STRONG:	return KHE_DATASET_POINTS_PLUS;
    case KHE_DRS_DOM_TRIE:	return KHE_DATASET_POINTS_CROSS;

    default:

      HnAbort("KheDomKindToPointsType: unknown dom_kind (%d)", dom_kind);
      return KHE_DATASET_POINTS_CROSS;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDomKindLabel(KHE_DRS_DOM_KIND dom_kind)                         */
/*                                                                           */
/*  Return a suitable label for dom_kind.                                    */
/*                                                                           */
/*****************************************************************************/

static char *KheDomKindLabel(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:	return "None";
    case KHE_DRS_DOM_WEAK:	return "Weak";
    case KHE_DRS_DOM_MEDIUM:	return "Medium";
    case KHE_DRS_DOM_STRONG:	return "Strong";
    case KHE_DRS_DOM_TRIE:	return "Trie";

    default:

      HnAbort("KheDomKindLabel: unknown dom_kind (%d)", dom_kind);
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsPrintPoint(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)            */
/*                                                                           */
/*  Return true if we want to print a point with index i.                    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsPrintPoint(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)
{
  return HaArrayCount(drs->open_days) <= 28 || i % 2 == 0 ||
    i == HaArrayCount(drs->open_days) - 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheBelowCaption(KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[200])    */
/*                                                                           */
/*  Print a suitable value for the below caption onto buff.                  */
/*                                                                           */
/*****************************************************************************/

static void KheBelowCaption(KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[200])
{
  int pos, i;  KHE_DRS_DAY_RANGE ddr;  KHE_TIME_GROUP tg1, tg2;

  /* print day ranges */
  pos = 0;
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
  {
    if( i > 0 )
    {
      sprintf(&buff[pos], ", ");
      pos = strlen(buff);
    }
    tg1 = KheFrameTimeGroup(drs->days_frame, ddr.first);
    tg2 = KheFrameTimeGroup(drs->days_frame, ddr.last);
    if( tg1 == tg2 )
      sprintf(&buff[pos], "%s", KheTimeGroupId(tg1));
    else
      sprintf(&buff[pos], "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
    pos = strlen(buff);
    if( pos > 150 )
    {
      sprintf(&buff[pos], "...");
      pos = strlen(buff);
      break;
    }
  }

  /* print resource count */
  sprintf(&buff[pos], " (%d resources)",
    KheResourceSetResourceCount(drs->selected_resource_set));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,       */
/*    int soln_limit)                                                        */
/*                                                                           */
/*  Test drs.                                                                */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit)
{
#if TESTING
  KHE_DRS_DOM_KIND dom_kind;  int di, i;  KHE_MARK mark;
  char time_file_name[200], size_file_name[200], below_caption[200];
  KHE_FILE time_file, size_file;  KHE_GRAPH time_graph, size_graph;
  KHE_DATASET_POINTS_TYPE points_type; KHE_DATASET_PAIRS_TYPE pairs_type;
  KHE_DATASET kd;  char *label;  float val;  KHE_DRS_DAY day;

  /* initialize the time file and graph */
  KheBelowCaption(drs, below_caption);
  TestName(KheSolnInstance(drs->soln), "time", time_file_name);
  time_file = KheFileBegin(time_file_name, KHE_FILE_LOUT_STANDALONE);
  time_graph = KheGraphBegin(time_file);
  KheGraphSetBelowCaption(time_graph, below_caption);
  KheGraphSetLeftCaptionAndGap(time_graph, "Secs", "1.0c");

  /* initialize the table size file and graph */
  TestName(KheSolnInstance(drs->soln), "size", size_file_name);
  size_file = KheFileBegin(size_file_name, KHE_FILE_LOUT_STANDALONE);
  size_graph = KheGraphBegin(size_file);
  KheGraphSetBelowCaption(size_graph, below_caption);
  KheGraphSetLeftCaptionAndGap(size_graph, "Size", "1.0c");

  /* one solve for each kind of dominance testing */
  for( di = KHE_DRS_DOM_NONE;  di <= KHE_DRS_DOM_TRIE;  di++ )
  {
    dom_kind = (KHE_DRS_DOM_KIND) di;

    /* clear out running time stuff */
    HaArrayClear(drs->running_time_per_day);
    HaArrayFill(drs->running_time_per_day, HaArrayCount(drs->open_days), 0.0);
    KheTimerResetStartTime(drs->timer);

    /* solve drs for dominance test i */
    KheDrsSolveOpen(drs, soln_limit, dom_kind, &mark);
    KheDrsSolveDoSolve(drs, dom_kind);
    KheDrsSolveClose(drs, mark, true);

    /* sort out the points type and pairs type */
    points_type = KheDomKindToPointsType(dom_kind);
    pairs_type = KHE_DATASET_PAIRS_DASHED;
    label = KheDomKindLabel(dom_kind);

    /* add one data set to the running time graph */
    kd = KheDataSetAdd(time_graph, points_type, pairs_type, label);
    HaArrayForEach(drs->running_time_per_day, val, i)
      if( KheDrsPrintPoint(drs, i) )
        KhePointAdd(kd, (float) i + 1, val);

    /* add one data set to the table size graph */
    kd = KheDataSetAdd(size_graph, points_type, pairs_type, label);
    KhePointAdd(kd, 0.0, 0.0);
    HaArrayForEach(drs->open_days, day, i)
      if( KheDrsPrintPoint(drs, i) )
        KhePointAdd(kd, (float) i + 1, day->soln_set->count);
  }

  /* tidy up graphs and stats files */
  KheGraphEnd(time_graph);
  KheFileEnd(time_file);
  KheGraphEnd(size_graph);
  KheFileEnd(size_file);
#else
  HnAbort("KheDynamicResourceSolverTest:  TESTING flag in "
    "khe_sr_dynamic_resource.c not set");
#endif
}
